Type: Package
Package: r5r
Title: Rapid Realistic Routing with 'R5'
Version: 2.4.0
Authors@R: c(
    person("Marcus", "Saraiva", , "marcus.saraiva@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0001-6218-2338")),
    person("Rafael H. M.", "Pereira", , "rafa.pereira.br@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-2125-7465")),
    person("Daniel", "Herszenhut", , "dhersz@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0001-8066-1105")),
    person("Alex", "Magnus", , "alex.magnus@mail.utoronto.ca", role = "aut"),
    person("Matthew Wigginton", "Bhagat-Conway", , "mwbc@unc.edu", role = "aut",
           comment = c(ORCID = "0000-0002-1210-2982")),
    person("Carlos Kaue Vieira", "Braga", , "kaue@kauebraga.dev", role = "ctb",
           comment = c(ORCID = "0000-0002-6104-7297")),
    person("Luyu", "Liu", , "luyuliu@ufl.edu", role = "ctb",
           comment = c(ORCID = "0000-0002-6684-5570")),
    person("Egor", "Kotov", , "kotov.egor@gmail.com", role = c("ctb"),
           comment = c(ORCID = "0000-0001-6690-5345")),
    person("Daniel", "Snow", , "dan@sno.ws", role = "ctb"),
    person("Ipea - Institute for Applied Economic Research", role = c("cph", "fnd")),
    person("Department of Geography & Planning, University of Toronto", role ="fnd")
    )
Description: Rapid realistic routing on multimodal transport networks
    (walk, bike, public transport and car) using 'R5', the Rapid Realistic
    Routing on Real-world and Reimagined networks engine
    <https://github.com/conveyal/r5>. The package allows users to generate
    detailed routing analysis or calculate travel time and monetary cost matrices 
    using seamless parallel computing on top of the R5 Java machine.  While R5
    is developed by Conveyal, the package r5r is independently developed
    by a team at the Institute for Applied Economic Research (Ipea) with
    contributions from collaborators. Apart from the documentation in this
    package, users will find additional information on R5 documentation at
    <https://docs.conveyal.com/>. Although we try to keep new releases of
    r5r in synchrony with R5, the development of R5 follows Conveyal's
    independent update process. Hence, users should confirm the R5 version
    implied by the Conveyal user manual (see
    <https://docs.conveyal.com/changelog>) corresponds with the R5 version
    that r5r depends on. This version of r5r depends on R5 v7.1.
License: MIT + file LICENSE
URL: https://github.com/ipeaGIT/r5r, https://ipeagit.github.io/r5r/
BugReports: https://github.com/ipeaGIT/r5r/issues
Depends: R (>= 3.6)
Imports: checkmate, cli, data.table, isoband, jsonlite, lifecycle,
        methods, purrr, rJava (>= 0.9-10), rlang, sf (>= 1.0-12),
        sfheaders, utils, zip
Suggests: accessibility, covr, dplyr, fs, ggplot2 (>= 3.3.1),
        gtfstools, h3jsr (>= 1.3.0), interp, knitr, mapview, parallel,
        patchwork, rJavaEnv, rmarkdown, testthat
VignetteBuilder: knitr
Encoding: UTF-8
SystemRequirements: Java JDK (>= 21.0)
Config/roxygen2/version: 8.0.0
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-19 20:33:44 UTC; rafap
Author: Marcus Saraiva [aut] (ORCID: <https://orcid.org/0000-0001-6218-2338>),
  Rafael H. M. Pereira [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2125-7465>),
  Daniel Herszenhut [aut] (ORCID:
    <https://orcid.org/0000-0001-8066-1105>),
  Alex Magnus [aut],
  Matthew Wigginton Bhagat-Conway [aut] (ORCID:
    <https://orcid.org/0000-0002-1210-2982>),
  Carlos Kaue Vieira Braga [ctb] (ORCID:
    <https://orcid.org/0000-0002-6104-7297>),
  Luyu Liu [ctb] (ORCID: <https://orcid.org/0000-0002-6684-5570>),
  Egor Kotov [ctb] (ORCID: <https://orcid.org/0000-0001-6690-5345>),
  Daniel Snow [ctb],
  Ipea - Institute for Applied Economic Research [cph, fnd],
  Department of Geography & Planning, University of Toronto [fnd]
Maintainer: Rafael H. M. Pereira <rafa.pereira.br@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-20 06:20:25 UTC
Built: R 4.5.2; ; 2026-05-21 23:32:14 UTC; unix

# Generated by roxygen2: do not edit by hand

export(accessibility)
export(arrival_travel_time_matrix)
export(build_network)
export(check_transit_availability)
export(detailed_itineraries)
export(download_r5)
export(expanded_travel_time_matrix)
export(find_snap)
export(get_gtfs_errors)
export(isochrone)
export(pareto_frontier)
export(r5r_cache)
export(r5r_sitrep)
export(read_fare_structure)
export(setup_fare_structure)
export(setup_r5)
export(stop_r5)
export(street_network_bbox)
export(street_network_to_sf)
export(transit_network_to_sf)
export(travel_time_matrix)
export(write_fare_structure)
importClassesFrom(rJava,jobjRef)
importFrom(data.table,"%between%")
importFrom(data.table,"%chin%")
importFrom(data.table,"%like%")
importFrom(data.table,":=")
importFrom(data.table,.N)
importFrom(lifecycle,deprecated)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)


## ----message = FALSE----------------------------------------------------------
# # increase Java memory
# options(java.parameters = "-Xmx2G")
# 
# # load libraries
# library(r5r)
# library(sf)
# library(data.table)
# library(ggplot2)
# library(dplyr)
# 
# # build a routable transport network with r5r
# data_path <- system.file("extdata/spo", package = "r5r")
# r5r_network <- build_network(data_path)
# 
# # routing inputs
# mode <- c('walk', 'transit')
# max_walk_time <- 30 # minutes
# max_trip_duration <- 90 # minutes
# 
# # load origin/destination points
# points <- fread(file.path(data_path, "spo_hexgrid.csv"))
# 
# # departure datetime
# departure_datetime = as.POSIXct("13-05-2019 14:00:00",
#                                 format = "%d-%m-%Y %H:%M:%S")

## ----message = FALSE----------------------------------------------------------
# # estimate accessibility
# acc <- r5r::accessibility(
#   r5r_network,
#   origins = points,
#   destinations = points,
#   opportunities_colnames = 'schools',
#   mode = mode,
#   max_walk_time = max_walk_time,
#   decay_function = "step",
#   cutoffs = 45,
#   departure_datetime = departure_datetime,
#   progress = FALSE,
#   time_window = 60,
#   percentiles = c(10, 20, 50, 70, 80)
#   )
# 
# head(acc, n = 10)
# 

## -----------------------------------------------------------------------------
# # summarize
# df <- acc[, .(min_acc = min(accessibility),
#               median = accessibility[which(percentile == 50)],
#               max_acc = max(accessibility)), by = id]
# 
# # plot
# ggplot(data=df) +
#   geom_linerange(color='gray', alpha=.5, aes(x = reorder(id, median) ,
#                       y=median, ymin=min_acc, ymax=max_acc)) +
#   geom_point(color='#0570b0', size=.5, aes(x = reorder(id, median), y=median)) +
#   labs(y='N. of schools accessible\nby public transport', x='Origins sorted by accessibility',
#        title="Accessibility uncertainty between 2pm and 3pm",
#        subtitle = 'Upper limit 10% and lower limit 80% of the times') +
#   theme_classic() +
#   theme(axis.text.x=element_blank(),
#         axis.ticks.x=element_blank())

## ----message = FALSE----------------------------------------------------------
# # estimate travel time matrix
# ttm <- travel_time_matrix(
#   r5r_network,
#   origins = points,
#   destinations = points,
#   mode = mode,
#   max_walk_time = max_walk_time,
#   max_trip_duration = max_trip_duration,
#   departure_datetime = departure_datetime,
#   progress = TRUE,
#   time_window = 60,
#   percentiles = c(10, 20, 50, 70, 80)
#   )
# 
# head(ttm, n = 10)
# 

## -----------------------------------------------------------------------------
# ettm <- r5r::expanded_travel_time_matrix(
#   r5r_network,
#   origins = points[1:30,],
#   destinations = points[31:61,],
#   mode = mode,
#   max_walk_time = max_walk_time,
#   max_trip_duration = max_trip_duration,
#   departure_datetime = departure_datetime,
#   progress = FALSE,
#   time_window = 20
#   )
# 
# head(ettm, n = 10)
# 

## ----message = FALSE----------------------------------------------------------
# r5r::stop_r5(r5r_network)
# rJava::.jgc(R.gc = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

## ----message = FALSE----------------------------------------------------------
# options(java.parameters = "-Xmx2G")
# 
# library(r5r)
# library(sf)
# library(data.table)
# library(ggplot2)

## ----message = FALSE----------------------------------------------------------
# # system.file returns the directory with example data inside the r5r package
# # set data path to directory containing your own data if not running this example
# data_path <- system.file("extdata/poa", package = "r5r")
# 
# r5r_network <- build_network(data_path)

## ----message = FALSE----------------------------------------------------------
# # read all points in the city
# points <- fread(file.path(data_path, "poa_hexgrid.csv"))
# 
# # subset point with the geolocation of the central bus station
# central_bus_stn <- points[291,]
# 
# # isochrone intervals
# time_intervals <- seq(0, 100, 10)
# 
# # routing inputs
# mode <- c("WALK", "TRANSIT")
# max_walk_time <- 30      # in minutes
# max_trip_duration <- 90  # in minutes
# time_window <- 60        # in minutes
# departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
#                                  format = "%d-%m-%Y %H:%M:%S")
# 
# # calculate travel time matrix
# iso1 <- r5r::isochrone(
#   r5r_network,
#   origins = central_bus_stn,
#   mode = mode,
#   polygon_output = TRUE,
#   cutoffs = time_intervals,
#   departure_datetime = departure_datetime,
#   max_walk_time = max_walk_time,
#   max_trip_duration = max_trip_duration,
#   time_window = time_window,
#   progress = FALSE,
#   zoom = 10
#   )
# 

## ----message = FALSE----------------------------------------------------------
# head(iso1)

## ----message = FALSE----------------------------------------------------------
# # extract OSM network
# street_net <- street_network_to_sf(r5r_network)
# main_roads <- subset(street_net$edges, street_class %like% 'PRIMARY|SECONDARY')
# 
# colors <- c('#ffe0a5','#ffcb69','#ffa600','#ff7c43','#f95d6a',
#             '#d45087','#a05195','#665191','#2f4b7c','#003f5c')
# 
# ggplot() +
#   geom_sf(data = iso1, aes(fill=factor(isochrone)), color = NA, alpha = .7) +
#   geom_sf(data = main_roads, color = "gray55", size=0.01, alpha = 0.2) +
#   geom_point(data = central_bus_stn, aes(x=lon, y=lat, color='Central bus\nstation')) +
#   scale_fill_manual(values = rev(colors) ) +
#   scale_color_manual(values=c('Central bus\nstation'='black')) +
#   labs(fill = "Travel time\n(in minutes)", color='') +
#   theme_minimal() +
#   theme(axis.title = element_blank())

## ----message = FALSE----------------------------------------------------------
# # calculate travel time matrix
# iso2 <- r5r::isochrone(
#   r5r_network,
#   origins = central_bus_stn,
#   mode = mode,
#   polygon_output = FALSE,
#   cutoffs = time_intervals,
#   departure_datetime = departure_datetime,
#   max_walk_time = max_walk_time,
#   max_trip_duration = max_trip_duration,
#   time_window = time_window,
#   progress = FALSE
#   )
# 
# head(iso2)
# 

## ----message = FALSE----------------------------------------------------------
# ggplot() +
#   geom_sf(data = iso2, aes(color=factor(isochrone)), alpha = .7) +
#   scale_color_manual(values = rev(colors) ) +
#   geom_point(data = central_bus_stn, aes(x=lon, y=lat), color='black') +
#   labs(color = "Travel time\n(in minutes)", color='sadasd') +
#   theme_minimal() +
#   theme(axis.title = element_blank())

## ----message = FALSE----------------------------------------------------------
# r5r::stop_r5(r5r_network)
# rJava::.jgc(R.gc = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

## removes files previously created by 'build_network()'
#data_path <- system.file("extdata/poa", package = "r5r")
#existing_files <- list.files(data_path)
#files_to_keep <- c(
#  "poa_hexgrid.csv",
#  "poa_osm.pbf",
#  "poa_points_of_interest.csv",
#  "poa_eptc.zip",
#  "poa_trensurb.zip",
#  'fares'
#  )
#files_to_remove <- existing_files[! existing_files %in% files_to_keep]
#invisible(file.remove(file.path(data_path, files_to_remove)))

## ----eval = FALSE-------------------------------------------------------------
# # from CRAN
# install.packages('r5r')
# 
# # dev version with latest features
# devtools::install_github("ipeaGIT/r5r", subdir = "r-package")

## ----eval = FALSE-------------------------------------------------------------
# # install {rJavaEnv} from CRAN
# install.packages("rJavaEnv")
# 
# # check version of Java currently installed (if any)
# rJavaEnv::java_check_version_rjava()
# 
# ## if this is the first time you use {rJavaEnv}, you might need to run this code
# ## below to consent the installation of Java.
# # rJavaEnv::rje_consent(provided = TRUE)
# 
# # install Java 21
# rJavaEnv::java_quick_install(version = 21)
# 
# # check if Java was successfully installed
# rJavaEnv::java_check_version_rjava()
# 

## ----message = FALSE----------------------------------------------------------
# options(java.parameters = "-Xmx2G")
# 
# # By default, {r5r} uses all CPU cores available. If you want to limit the
# # number of CPUs to 4, for example, you can run:
# options(java.parameters = c("-Xmx2G", "-XX:ActiveProcessorCount=4"))

## ----message = FALSE, warning = FALSE-----------------------------------------
# library(r5r)
# library(sf)
# library(data.table)
# library(ggplot2)

## -----------------------------------------------------------------------------
# data_path <- system.file("extdata/poa", package = "r5r")
# list.files(data_path)

## -----------------------------------------------------------------------------
# poi <- fread(file.path(data_path, "poa_points_of_interest.csv"))
# head(poi)

## -----------------------------------------------------------------------------
# points <- fread(file.path(data_path, "poa_hexgrid.csv"))
# 
# # sample points
# sampled_rows <- sample(1:nrow(points), 200, replace=TRUE)
# points <- points[ sampled_rows, ]
# head(points)

## ----message = FALSE----------------------------------------------------------
# # Indicate the path where OSM and GTFS data are stored
# r5r_network <- build_network(data_path = data_path)

## ----message = FALSE----------------------------------------------------------
# # set departure datetime input
# departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
#                                  format = "%d-%m-%Y %H:%M:%S")
# 
# # calculate accessibility
# access <- accessibility(
#   r5r_network,
#   origins = points,
#   destinations = points,
#   opportunities_colnames = c("schools", "healthcare"),
#   mode = c("WALK", "TRANSIT"),
#   departure_datetime = departure_datetime,
#   decay_function = "step",
#   cutoffs = 60
#   )
# 
# head(access)

## ----message = FALSE----------------------------------------------------------
# # set inputs
# mode <- c("WALK", "TRANSIT")
# max_walk_time <- 30 # minutes
# max_trip_duration <- 120 # minutes
# departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
#                                  format = "%d-%m-%Y %H:%M:%S")
# 
# # calculate a travel time matrix
# ttm <- travel_time_matrix(
#   r5r_network,
#   origins = poi,
#   destinations = poi,
#   mode = mode,
#   departure_datetime = departure_datetime,
#   max_walk_time = max_walk_time,
#   max_trip_duration = max_trip_duration
#   )
# 
# head(ttm)

## ----ttm head, echo=FALSE, message=FALSE, out.width='100%', eval = FALSE------
# knitr::include_graphics("https://github.com/ipeaGIT/r5r/blob/master/r-package/inst/img/vig_output_ttm.png?raw=true")

## ----message = FALSE----------------------------------------------------------
# # calculate a travel time matrix
# ettm <- expanded_travel_time_matrix(
#   r5r_network,
#   origins = poi,
#   destinations = poi,
#   mode = mode,
#   departure_datetime = departure_datetime,
#   breakdown = TRUE,
#   max_walk_time = max_walk_time,
#   max_trip_duration = max_trip_duration
#   )
# 
# head(ettm)

## ----message = FALSE----------------------------------------------------------
# # set inputs
# origins <- poi[10,]
# destinations <- poi[12,]
# mode <- c("WALK", "TRANSIT")
# max_walk_time <- 60 # minutes
# departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
#                                  format = "%d-%m-%Y %H:%M:%S")
# 
# # calculate detailed itineraries
# det <- detailed_itineraries(
#   r5r_network,
#   origins = origins,
#   destinations = destinations,
#   mode = mode,
#   departure_datetime = departure_datetime,
#   max_walk_time = max_walk_time,
#   shortest_path = FALSE
#   )
# 
# head(det)

## ----detailed head, echo = FALSE, out.width='100%', message = FALSE, eval = FALSE----
# knitr::include_graphics("https://github.com/ipeaGIT/r5r/blob/master/r-package/inst/img/vig_output_detailed.png?raw=true")

## ----message = FALSE----------------------------------------------------------
# # extract OSM network
# street_net <- r5r::street_network_to_sf(r5r_network)
# 
# # extract public transport network
# transit_net <- r5r::transit_network_to_sf(r5r_network)
# 
# # plot
# ggplot() +
#   geom_sf(data = street_net$edges, color='gray85') +
#   geom_sf(data = det, aes(color=mode)) +
#   facet_wrap(.~option) +
#   theme_void()
# 

## ----ggplot2 output, echo = FALSE, out.width='100%', message = FALSE, eval = FALSE----
# knitr::include_graphics("https://github.com/ipeaGIT/r5r/blob/master/r-package/inst/img/vig_detailed_ggplot.png?raw=true")

## ----message = FALSE----------------------------------------------------------
# r5r::stop_r5(r5r_network)
# rJava::.jgc(R.gc = TRUE)

## ----eval = TRUE, include = FALSE, message = FALSE----------------------------
# clean cache (CRAN policy)
r5r::r5r_cache(delete_file = 'all')

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

## ----echo = FALSE, fig.width=7, fig.height=4----------------------------------
# library(ggplot2)
# 
# # data.frame
# df <- structure(list(option = c(1, 2, 3, 4, 5),
#                      modes = c("Walk", "Bus","Bus + Bus", "Subway", "Bus + Subway"),
#                      time = c(50, 35, 29, 20, 15),
#                      cost = c(0, 3, 5, 6, 8)), class = "data.frame",
#                 row.names = c(NA, -5L))
# 
# # figure
# ggplot(data=df, aes(x=cost, y=time, label = modes)) +
#   geom_step(linetype = "dashed") +
#   geom_point() +
#   geom_text(color='gray30', hjust = -.2, nudge_x = 0.05, angle = 45) +
#   labs(title='Pareto frontier of alternative routes from A to B', subtitle = 'Hypotetical example') +
#   scale_x_continuous(name="Travel cost (BRL)", breaks=seq(0,12,3)) +
#   scale_y_continuous(name="Travel time (minutes)", breaks=seq(0,60,10)) +
#   coord_cartesian(xlim = c(0,14), ylim = c(0, 60)) +
#   theme_classic()

## ----message = FALSE----------------------------------------------------------
# # increase Java memory
# options(java.parameters = "-Xmx2G")
# 
# # load libraries
# library(r5r)
# library(data.table)
# library(ggplot2)
# library(dplyr)
# 
# # build a routable transport network with r5r
# data_path <- system.file("extdata/poa", package = "r5r")
# r5r_network <- build_network(data_path)
# 
# # routing inputs
# mode <- c('walk', 'transit')
# max_trip_duration <- 90 # minutes
# 
# # load origin/destination points of interest
# points <- fread(file.path(data_path, "poa_points_of_interest.csv"))
# 

## -----------------------------------------------------------------------------
# # create basic fare structure
# fare_structure <- setup_fare_structure(
#   r5r_network,
#   base_fare = 4.8,
#   by = "MODE"
#   )
# 
# # update the cost of bus and train fares
# fare_structure$fares_per_type[, fare := fcase(type == "BUS", 4.80,
#                                              type == "RAIL", 4.50)]
# 
# # update the cost of tranfers
# fare_structure$fares_per_transfer[, fare := fcase(first_leg == "BUS" & second_leg == "BUS", 7.2,
#                                                  first_leg != second_leg, 8.37)]
# 
# # update transfer_time_allowance to 60 minutes
# fare_structure$transfer_time_allowance <- 60
# 
# fare_structure$fares_per_type[type == "RAIL", unlimited_transfers := TRUE]
# fare_structure$fares_per_type[type == "RAIL", allow_same_route_transfer := TRUE]
# 

## -----------------------------------------------------------------------------
# # save fare rules to temp file
# temp_fares <- tempfile(pattern = "fares_poa", fileext = ".zip")
# r5r::write_fare_structure(fare_structure, file_path = temp_fares)
# 
# 
# fare_structure <- r5r::read_fare_structure(file.path(data_path, "fares/fares_poa.zip"))

## -----------------------------------------------------------------------------
# departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
#                                  format = "%d-%m-%Y %H:%M:%S")
# 
# prtf <- pareto_frontier(
#   r5r_network,
#   origins = points,
#   destinations = points,
#   mode = c("WALK", "TRANSIT"),
#   departure_datetime = departure_datetime,
#   fare_structure = fare_structure,
#   fare_cutoffs = c(1, 4.5, 4.8, 7.20, 8.37),
#   progress = TRUE
#   )
# 
# head(prtf)

## ----echo = FALSE, fig.width=7, fig.height=4, warning=FALSE-------------------
# # select origin and destinations
# pf2 <- dplyr::filter(prtf, to_id == 'farrapos_station'  &
#                        from_id %in% c('moinhos_de_vento_hospital',
#                                       'praia_de_belas_shopping_center'))
# 
# # recode modes
# pf2[, modes := fcase(monetary_cost == 1,   'Walk',
#                      monetary_cost == 4.5, 'Train',
#                      monetary_cost == 4.8, 'Bus',
#                      monetary_cost == 7.2, 'Bus + Bus',
#                      monetary_cost == 8.37, 'Bus + Train')]
# 
# # plot
# ggplot(data=pf2, aes(x=monetary_cost, y=travel_time, color=from_id, label = modes)) +
#   geom_step(linetype = "dashed") +
#   geom_point() +
#   geom_text(color='gray30', hjust = -.2, nudge_x = 0.05, angle = 45) +
#   labs(title='Pareto frontier of alternative routes from Farrapos station to:',
#        subtitle = 'Praia de Belas shopping mall and Moinhos hospital',
#        color='Destination') +
#   scale_x_continuous(name="Travel cost ($)", breaks=seq(0,12,2)) +
#   scale_y_continuous(name="Travel time (minutes)", breaks=seq(0,120,20)) +
#   coord_cartesian(xlim = c(0,12), ylim = c(0, 120)) +
#   theme_classic() + theme(legend.position=c(.2,0.8))

## ----message = FALSE----------------------------------------------------------
# r5r::stop_r5(r5r_network)
# rJava::.jgc(R.gc = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)


## ----message=FALSE------------------------------------------------------------
# # increase Java memory
# options(java.parameters = "-Xmx2G")
# 
# # load libraries
# library(r5r)
# library(dplyr)
# library(data.table)
# library(ggplot2)
# 
# # data path
# data_path <- system.file("extdata/poa", package = "r5r")
# 
# # build network
# r5r_network <- r5r::build_network(
#   data_path = data_path,
#   verbose = FALSE
#   )

## ----message=FALSE------------------------------------------------------------
# # read data.frame with new car speeds
# edge_speed_factors <- read.csv(
#   file.path(data_path, "poa_osm_congestion.csv")
#   )
# 
# head(edge_speed_factors)

## ----message=FALSE------------------------------------------------------------
# # origins and destination points
# points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))
# 
# # travel time matrix
# ttm_congestion <- r5r::travel_time_matrix(
#   r5r_network = r5r_network,
#   origins = points,
#   destinations = points,
#   mode = 'car',
#   departure_datetime = Sys.time(),
#   max_trip_duration = 30,
#   new_carspeeds = edge_speed_factors,
#   carspeed_scale = 0.8
# )

## ----message=FALSE------------------------------------------------------------
# # path to OSM pbf
# pbf_path <- paste0(data_path, "/poa_osm.pbf")
# 
# # read layer of lines from pbf
# roads <- sf::st_read(
#   pbf_path,
#   layer = 'lines',
#   quiet = TRUE
#   )
# 
# # Filter only road types of interest
# rt <- c("motorway", "primary", "secondary", "tertiary")
# 
# roads <- roads |>
#   select(osm_id, highway) |>
#   filter(highway %in% rt)
# 
# head(roads)

## -----------------------------------------------------------------------------
# # map
# plot(roads["highway"])
# 

## -----------------------------------------------------------------------------
# new_edge_speeds <- roads |>
#   mutate(
#     osm_id = as.numeric(osm_id),
#     max_speed = case_when(
#       highway == "motorway"  ~ 0.75,
#       highway == "primary"   ~ 0.8,
#       highway == "secondary" ~ 0.85,
#       highway == "tertiary"  ~ 0.9)) |>
#   sf::st_drop_geometry()
# 
# new_edge_speeds$speed_type <- "scale"
# 
# head(new_edge_speeds)
# 

## ----message = FALSE, warning=FALSE-------------------------------------------
# # travel time matrix
# ttm_congestion <- r5r::travel_time_matrix(
#   r5r_network = r5r_network,
#   origins = points,
#   destinations = points,
#   mode = 'car',
#   departure_datetime = Sys.time(),
#   max_trip_duration = 30,
#   new_carspeeds = new_edge_speeds
#   )
# 

## ----message = FALSE----------------------------------------------------------
# # edit table with custom speeds to 40 km/h
# new_edge_speeds40 <- new_edge_speeds |>
#   mutate(max_speed = 40,
#          speed_type = "km/h")
# 
# # travel time matrix
# ttm_congestion <- r5r::travel_time_matrix(
#   r5r_network = r5r_network,
#   origins = points,
#   destinations = points,
#   mode = 'car',
#   departure_datetime = Sys.time(),
#   max_trip_duration = 30,
#   new_carspeeds = new_edge_speeds40
#   )

## ----message = FALSE----------------------------------------------------------
# # read sf with congestion polygons
# congestion_poly <- readRDS(file.path(data_path, "poa_poly_congestion.rds"))
# 
# # preview
# mapview::mapview(congestion_poly, zcol="scale")

## ----message = FALSE----------------------------------------------------------
# head(congestion_poly)

## ----message = FALSE----------------------------------------------------------
# ttm_congestion <- r5r::travel_time_matrix(
#   r5r_network = r5r_network,
#   origins = points,
#   destinations = points,
#   mode = 'car',
#   departure_datetime = Sys.time(),
#   max_trip_duration = 30,
#   new_carspeeds = congestion_poly,
#   carspeed_scale = 0.95
#   )

## ----message=FALSE------------------------------------------------------------
# # read data.frame with new lts
# edge_lts <- read.csv(
#   file.path(data_path, "poa_osm_lts.csv")
#   )
# 
# head(edge_lts)

## ----message = FALSE----------------------------------------------------------
# ttm_new_lts <- r5r::travel_time_matrix(
#   r5r_network = r5r_network,
#   origins = points,
#   destinations = points,
#   mode = 'bicycle',
#   departure_datetime = Sys.time(),
#   max_trip_duration = 30,
#   new_lts = edge_lts
#   )

## ----message = FALSE----------------------------------------------------------
# # read sf with congestion polygons
# lts_lines <- readRDS(file.path(data_path, "poa_ls_lts.rds"))
# 
# # preview
# mapview::mapview(lts_lines, zcol="lts")

## ----message = FALSE----------------------------------------------------------
# ttm_new_lts <- r5r::travel_time_matrix(
#   r5r_network = r5r_network,
#   origins = points,
#   destinations = points,
#   mode = 'bicycle',
#   departure_datetime = Sys.time(),
#   max_trip_duration = 30,
#   new_lts = lts_lines
#   )

## ----message = FALSE----------------------------------------------------------
# # stop an specific r5r network
# r5r::stop_r5(r5r_network)
# 
# # or stop all r5r networks at once
# r5r::stop_r5()
# rJava::.jgc(R.gc = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

## ----message = FALSE----------------------------------------------------------
# options(java.parameters = "-Xmx2G")
# 
# library(r5r)
# library(sf)
# library(data.table)
# library(ggplot2)
# library(patchwork)
# library(dplyr)
# library(h3jsr)

## -----------------------------------------------------------------------------
# # setup and load Porto Alegre multimodal network into memory
# 
# # system.file returns the directory with example data inside the r5r package
# # set data path to directory containing your own data if not using the examples
# data_path <- system.file("extdata/poa", package = "r5r")
# 
# r5r_network <- build_network(data_path)
# 
# # load transit network as an SF
# transit_network <- transit_network_to_sf(r5r_network)
# 
# # map
# ggplot() +
#   geom_sf(data=transit_network$routes, aes(color=mode)) +
#   theme_void()

## -----------------------------------------------------------------------------
# fare_structure <- setup_fare_structure(r5r_network,
#                                        base_fare = 4.8,
#                                        by = "MODE")

## -----------------------------------------------------------------------------
# head(fare_structure, n=7)
# 

## -----------------------------------------------------------------------------
# fare_structure$max_discounted_transfers
# fare_structure$transfer_time_allowance <- 60 # update transfer_time_allowance
# fare_structure$fare_cap

## -----------------------------------------------------------------------------
# fare_structure$fares_per_type

## -----------------------------------------------------------------------------
# fare_structure$fares_per_type[type == "RAIL", unlimited_transfers := TRUE]
# fare_structure$fares_per_type[type == "RAIL", fare := 4.50]
# fare_structure$fares_per_type[type == "RAIL", allow_same_route_transfer := TRUE]

## -----------------------------------------------------------------------------
# fare_structure$fares_per_type

## -----------------------------------------------------------------------------
# fare_structure$fares_per_transfer

## -----------------------------------------------------------------------------
# # conditional update fare value
# fare_structure$fares_per_transfer[first_leg == "BUS" & second_leg == "BUS", fare := 7.2]

## -----------------------------------------------------------------------------
# # conditional update fare value
# fare_structure$fares_per_transfer[first_leg != second_leg, fare := 8.37]
# 
# # use fcase instead ?
# fare_structure$fares_per_transfer[, fare := fcase(first_leg == "BUS" & second_leg == "BUS", 7.2,
#                                                  first_leg != second_leg, 8.37)]
# 

## -----------------------------------------------------------------------------
# # remove row
# fare_structure$fares_per_transfer <- fare_structure$fares_per_transfer[!(first_leg == "RAIL" & second_leg == "RAIL")]
# 

## -----------------------------------------------------------------------------
# fare_structure$fares_per_transfer

## -----------------------------------------------------------------------------
# tail(fare_structure$fares_per_route)

## -----------------------------------------------------------------------------
# ## load input data
# points <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))
# 
# # calculate travel times function
# calculate_travel_times <- function(fare) {
#   ttm_df <- travel_time_matrix(
#     r5r_network,
#     origins = points,
#     destinations = points,
#     mode = c("WALK", "TRANSIT"),
#     departure_datetime = as.POSIXct(
#       "13-05-2019 14:00:00",
#       format = "%d-%m-%Y %H:%M:%S"
#     ),
#     time_window = 1,
#     fare_structure = fare_structure,
#     max_fare = fare,
#     max_trip_duration = 40,
#     max_walk_time = 20
#   )
# 
#   return(ttm_df)
# }
# 
# 
# # calculate travel times, and combine results
# ttm <- calculate_travel_times(fare = Inf)
# ttm_500 <- calculate_travel_times(fare = 5)
# 
# # merge results
# ttm[ttm_500, on = .(from_id, to_id), travel_time_500 := i.travel_time_p50]
# ttm[, travel_time_unl := travel_time_p50]
# ttm[, travel_time_p50 := NULL]

## -----------------------------------------------------------------------------
# tail(ttm, 10)

## -----------------------------------------------------------------------------
# # plot of overall travel time differences between limited and unlimited cost travel time matrices
# time_difference = ttm[!is.na(travel_time_500), .(count = .N),
#                       by = .(travel_time_unl, travel_time_500)]
# 
# p1 <- ggplot(time_difference, aes(y = travel_time_unl, x = travel_time_500)) +
#   geom_point(size = 0.7) +
#   coord_fixed() +
#   scale_x_continuous(breaks = seq(0, 45, 5)) +
#   scale_y_continuous(breaks = seq(0, 45, 5)) +
#   theme_light() +
#   theme(legend.position = "none") +
#   labs(y = "travel time (minutes)\nunestricted monetary cost",
#        x = "travel time (minutes)\nmonetary cost restricted to BRL 5.00"
#        )
# 
# # plot of unreachable destinations when the monetary cost limit is too low
# unreachable <- ttm[, .(count = .N), by = .(travel_time_unl, is.na(travel_time_500))]
# unreachable[, perc := count / sum(count, na.rm = T), by = .(travel_time_unl)]
# unreachable <- unreachable[is.na == TRUE]
# unreachable <- na.omit(unreachable)
# 
# p2 <- ggplot(unreachable, aes(x=travel_time_unl, y=perc)) +
#   geom_col() +
#   coord_flip() +
#   scale_x_continuous(breaks = seq(0, 45, 5)) +
#   scale_y_continuous(limits = c(0, 1), breaks = seq(0, 1, 0.2),
#                      labels = paste0(seq(0, 100, 20), "%")) +
#   theme_light() +
#   labs(x = "travel time (minutes)\nwithout monetary cost restriction",
#        y = "% of unreachable destinations\nconsidering a R$ 5.00 monetary cost limit")
# 
# # combine both plots using patchwork
# p1 + p2 + plot_annotation(subtitle = "Comparing travel times with and without monetary cost restriction")
# 

## -----------------------------------------------------------------------------
# # calculate accessibility function
# calculate_accessibility <- function(fare, fare_string) {
#   access_df <- accessibility(
#     r5r_network,
#     origins = points,
#     destinations = points,
#     mode = c("WALK", "TRANSIT"),
#     departure_datetime = as.POSIXct(
#       "13-05-2019 14:00:00",
#       format = "%d-%m-%Y %H:%M:%S"),
#     time_window = 1,
#     opportunities_colname = "healthcare",
#     cutoffs = 40,
#     fare_structure = fare_structure,
#     max_fare = fare,
#     max_trip_duration = 40,
#     max_walk_time = 20,
#     progress = FALSE)
# 
#   access_df$max_fare <- fare_string
# 
#   return(access_df)
# }
# 
# # calculate accessibility, combine results, and convert to SF
# access_500 <- calculate_accessibility(fare=5, fare_string="R$ 5.00 budget")
# access_unl <- calculate_accessibility(fare=Inf, fare_string="Unlimited budget")
# 
# access <- rbind(access_500, access_unl)
# 
# # bring geometry
# access$geometry <- h3jsr::cell_to_polygon(access$id)
# access <- st_as_sf(access)
# 

## -----------------------------------------------------------------------------
# # plot accessibility maps
# ggplot(data = access) +
#   geom_sf(aes(fill = accessibility), color=NA, size = 0.2) +
#   scale_fill_distiller(palette = "Spectral") +
#   facet_wrap(~max_fare) +
#   labs(subtitle = "Effect of monetary cost on accessibility") +
#   theme_minimal() +
#   theme(legend.position = "bottom",
#         axis.text = element_blank())
# 

## ----message = FALSE----------------------------------------------------------
# r5r::stop_r5(r5r_network)
# rJava::.jgc(R.gc = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

## ----message = FALSE----------------------------------------------------------
# # increase Java memory
# options(java.parameters = "-Xmx2G")
# 
# # load libraries
# library(r5r)
# library(sf)
# library(ggplot2)
# library(data.table)
# 
# # build a routable transport network with r5r
# data_path <- system.file("extdata/poa", package = "r5r")
# r5r_network <- build_network(data_path)
# 
# # routing inputs
# mode <- c('walk', 'transit')
# max_trip_duration <- 60 # minutes
# 
# # departure time
# departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
#                                  format = "%d-%m-%Y %H:%M:%S")
# 
# # load origin/destination points
# poi <- fread(file.path(data_path, "poa_points_of_interest.csv"))
# 

## ----message = FALSE----------------------------------------------------------
# # set inputs
# origins <- poi[10,]
# destinations <- poi[12,]
# mode <- c("WALK", "TRANSIT")
# max_walk_time <- 60
# departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
#                                  format = "%d-%m-%Y %H:%M:%S")
# 
# # calculate detailed itineraries
# det <- detailed_itineraries(
#   r5r_network,
#   origins = origins,
#   destinations = destinations,
#   mode = mode,
#   departure_datetime = departure_datetime,
#   max_walk_time = max_walk_time,
#   suboptimal_minutes = 8,
#   shortest_path = FALSE
#   )
# 
# head(det)

## ----detailed head, echo = FALSE, out.width='100%', message = FALSE, eval = FALSE----
# knitr::include_graphics("https://github.com/ipeaGIT/r5r/blob/master/r-package/inst/img/vig_output_detailed.png?raw=true")

## ----message = FALSE----------------------------------------------------------
# # extract OSM network
# street_net <- r5r::street_network_to_sf(r5r_network)
# 
# # extract public transport network
# transit_net <- r5r::transit_network_to_sf(r5r_network)
# 
# # plot
# fig <- ggplot() +
#         geom_sf(data = street_net$edges, color='gray85') +
#         geom_sf(data = subset(det, option <4), aes(color=mode)) +
#         facet_wrap(.~option) +
#         theme_void()
# 
# fig

## ----message = FALSE, eval = FALSE--------------------------------------------
# # SAVE image
# ggsave(plot = fig, filename = 'inst/img/vig_detailed_ggplot.png',
#        height = 5, width = 15, units='cm', dpi=200)

## ----ggplot2 output, echo = FALSE, out.width='100%', message = FALSE, eval = FALSE----
# knitr::include_graphics("https://github.com/ipeaGIT/r5r/blob/master/r-package/inst/img/vig_detailed_ggplot.png?raw=true")

## ----message = FALSE, eval = FALSE--------------------------------------------
# library(gtfstools)
# 
# # location of your frequency-based GTFS
# freq_gtfs_file <- system.file("extdata/spo/spo.zip", package = "r5r")
# 
# # read GTFS data
# freq_gtfs <- gtfstools::read_gtfs(freq_gtfs_file)
# 
# # convert from frequencies to time tables
# stop_times_gtfs <- gtfstools::frequencies_to_stop_times(freq_gtfs)
# 
# # save it as a new GTFS.zip file
# gtfstools::write_gtfs(gtfs = stop_times_gtfs,
#                       path = tempfile(pattern = 'stop_times_gtfs', fileext = '.zip'))
# 
# 

## ----message = FALSE----------------------------------------------------------
# r5r::stop_r5(r5r_network)
# rJava::.jgc(R.gc = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

## ----message = FALSE----------------------------------------------------------
# options(java.parameters = "-Xmx2G")
# 
# library(r5r)
# library(accessibility)
# library(sf)
# library(data.table)
# library(ggplot2)
# library(interp)
# library(h3jsr)
# library(dplyr)

## ----message = FALSE----------------------------------------------------------
# # system.file returns the directory with example data inside the r5r package
# # set data path to directory containing your own data if not running this example
# data_path <- system.file("extdata/poa", package = "r5r")
# 
# r5r_network <- build_network(data_path)

## ----message = FALSE----------------------------------------------------------
# # read all points in the city
# points <- fread(file.path(data_path, "poa_hexgrid.csv"))
# 
# # routing inputs
# mode <- c("WALK", "TRANSIT")
# max_walk_time <- 30      # in minutes
# travel_time_cutoff <- 20 # in minutes
# time_window <- 60        # in minutes
# departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
#                                  format = "%d-%m-%Y %H:%M:%S")
# 
# # calculate accessibility
# access1 <- r5r::accessibility(
#   r5r_network,
#   origins = points,
#   destinations = points,
#   mode = mode,
#   opportunities_colnames = c("schools", "healthcare"),
#   decay_function = "step",
#   cutoffs = travel_time_cutoff,
#   departure_datetime = departure_datetime,
#   max_walk_time = max_walk_time,
#   time_window = time_window,
#   progress = FALSE
#   )
# 
# head(access1)

## ----message = FALSE----------------------------------------------------------
# # calculate travel time matrix
# ttm <- r5r::travel_time_matrix(
#   r5r_network,
#   origins = points,
#   destinations = points,
#   mode = mode,
#   departure_datetime = departure_datetime,
#   max_walk_time = max_walk_time,
#   time_window = time_window,
#   progress = FALSE
#   )
# 
# head(ttm)

## ----message = FALSE----------------------------------------------------------
# # calculate accessibility
# access_edu <- accessibility::cumulative_cutoff(
#   travel_matrix = ttm,
#   land_use_data = points,
#   opportunity = 'schools',
#   travel_cost = 'travel_time_p50',
#   cutoff = 20
#   )
# 
# access_health <- accessibility::cumulative_cutoff(
#   travel_matrix = ttm,
#   land_use_data = points,
#   opportunity = 'healthcare',
#   travel_cost = 'travel_time_p50',
#   cutoff = 20
#   )
# 
# head(access_edu)
# head(access_health)

## ----message = FALSE, out.width='100%'----------------------------------------
# # retrieve polygons of H3 spatial grid
# grid <- h3jsr::cell_to_polygon(points$id, simple = FALSE)
# 
# # merge accessibility estimates
# access_sf <- left_join(grid, access1, by = c('h3_address'='id'))
# 
# # plot
# ggplot() +
#   geom_sf(data = access_sf, aes(fill = accessibility), color= NA) +
#   scale_fill_viridis_c(direction = -1, option = 'B') +
#   labs(fill = "Number of\nfacilities within\n20 minutes") +
#   theme_minimal() +
#   theme(axis.title = element_blank()) +
#   facet_wrap(~opportunity) +
#   theme_void()
# 

## ----message = FALSE, out.width='100%'----------------------------------------
# # interpolate estimates to get spatially smooth result
# access_schools <- access1 %>%
#   filter(opportunity == "schools") %>%
#   inner_join(points, by='id') %>%
#   with(interp::interp(lon, lat, accessibility)) %>%
#   with(cbind(acc=as.vector(z),  # Column-major order
#              x=rep(x, times=length(y)),
#              y=rep(y, each=length(x)))) %>% as.data.frame() %>% na.omit() %>%
#   mutate(opportunity = "schools")
# 
# access_health <- access1 %>%
#   filter(opportunity == "healthcare") %>%
#   inner_join(points, by='id') %>%
#   with(interp::interp(lon, lat, accessibility)) %>%
#   with(cbind(acc=as.vector(z),  # Column-major order
#              x=rep(x, times=length(y)),
#              y=rep(y, each=length(x)))) %>% as.data.frame() %>% na.omit() %>%
#   mutate(opportunity = "healthcare")
# 
# access.interp <- rbind(access_schools, access_health)
# 
# # find results' bounding box to crop the map
# bb_x <- c(min(access.interp$x), max(access.interp$x))
# bb_y <- c(min(access.interp$y), max(access.interp$y))
# 
# # extract OSM network, to plot over map
# street_net <- street_network_to_sf(r5r_network)
# 
# # plot
# ggplot(na.omit(access.interp)) +
#   geom_sf(data = street_net$edges, color = "gray55", size=0.01, alpha = 0.7) +
#   geom_contour_filled(aes(x=x, y=y, z=acc), alpha=.7) +
#   scale_fill_viridis_d(direction = -1, option = 'B') +
#   scale_x_continuous(expand=c(0,0)) +
#   scale_y_continuous(expand=c(0,0)) +
#   coord_sf(xlim = bb_x, ylim = bb_y, datum = NA) +
#   labs(fill = "Number of\nfacilities within\n20 minutes") +
#   theme_void() +
#   facet_wrap(~opportunity)

## ----message = FALSE----------------------------------------------------------
# r5r::stop_r5(r5r_network)
# rJava::.jgc(R.gc = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

## removes files previously created by 'build_network()'
#data_path <- system.file("extdata/poa", package = "r5r")
#existing_files <- list.files(data_path)
#files_to_keep <- c(
#  "poa_hexgrid.csv",
#  "poa_osm.pbf",
#  "poa_points_of_interest.csv",
#  "poa_eptc.zip",
#  "poa_trensurb.zip",
#  'fares'
#  )
#files_to_remove <- existing_files[! existing_files %in% files_to_keep]
#invisible(file.remove(file.path(data_path, files_to_remove)))

## ----message = FALSE----------------------------------------------------------
# # increase Java memory
# options(java.parameters = "-Xmx2G")
# 
# # load libraries
# library(r5r)
# library(data.table)
# library(ggplot2)
# 
# # build a routable transport network with r5r
# data_path <- system.file("extdata/poa", package = "r5r")
# r5r_network <- build_network(data_path)
# 
# # routing inputs
# mode <- c('walk', 'transit')
# max_trip_duration <- 60 # minutes
# 
# # departure time
# departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
#                                  format = "%d-%m-%Y %H:%M:%S")
# 
# # load origin/destination points
# points <- fread(file.path(data_path, "poa_points_of_interest.csv"))
# 

## ----message = FALSE----------------------------------------------------------
# # estimate travel time matrix
# ttm <- travel_time_matrix(
#   r5r_network,
#   origins = points,
#   destinations = points,
#   mode = mode,
#   max_trip_duration = max_trip_duration,
#   departure_datetime = departure_datetime
#   )
# 
# head(ttm, n = 10)
# 

## ----message = FALSE----------------------------------------------------------
# ettm <- expanded_travel_time_matrix(
#   r5r_network,
#   origins = points,
#   destinations = points,
#   mode = mode,
#   max_trip_duration = max_trip_duration,
#   departure_datetime = departure_datetime
#   )
# 
# head(ettm, n = 10)

## ----message = FALSE----------------------------------------------------------
# ettm2 <- expanded_travel_time_matrix(
#   r5r_network,
#   origins = points,
#   destinations = points,
#   mode = mode,
#   max_trip_duration = max_trip_duration,
#   departure_datetime = departure_datetime,
#   breakdown = TRUE
#   )
# 
# head(ettm2, n = 10)

## ----message = FALSE----------------------------------------------------------
# ettm_window <- expanded_travel_time_matrix(
#   r5r_network,
#   origins = points,
#   destinations = points,
#   mode = mode,
#   max_trip_duration = max_trip_duration,
#   departure_datetime = departure_datetime,
#   breakdown = TRUE,
#   time_window = 10
#   )
# 
# ettm_window[15:25,]

## ----message = FALSE----------------------------------------------------------
# 
# arrival_datetime <- as.POSIXct(
#  "13-05-2019 14:00:00",
#  format = "%d-%m-%Y %H:%M:%S"
# )
# 
# arrival_ttm <- arrival_travel_time_matrix(
#   r5r_network,
#   origins = points,
#   destinations = points,
#   mode = c("WALK", "TRANSIT"),
#   arrival_datetime = arrival_datetime,
#   max_trip_duration = 60
# )
# 
# head(arrival_ttm, n = 10)

## ----message = FALSE----------------------------------------------------------
# r5r::stop_r5(r5r_network)
# rJava::.jgc(R.gc = TRUE)

## ----eval = TRUE, include = FALSE, message = FALSE----------------------------
# clean cache (CRAN policy)
r5r::r5r_cache(delete_file = 'all')

#' @param elevation A string. The name of the impedance function to be used to
#'        calculate impedance for walking and cycling based on street slopes.
#'        Available options include `TOBLER` (Default) and `MINETTI`, or `NONE`
#'        to ignore elevation. R5 loads elevation data from `.tif` files saved
#'        inside the `data_path` directory. Elevation raster must be in WGS 84
#'        (EPSG:4326) coordinate reference system. See more info in the Details
#'        section below.

#' @param percentiles An integer vector (max length of 5). Specifies the
#'   percentile to use when returning accessibility estimates within the given
#'   time window. Please note that this parameter is applied to the travel time
#'   estimates that generate the accessibility results, and not to the
#'   accessibility distribution itself (i.e. if the 25th percentile is
#'   specified, the accessibility is calculated from the 25th percentile travel
#'   time, which may or may not be equal to the 25th percentile of the
#'   accessibility distribution itself). Defaults to 50, returning the
#'   accessibility calculated from the median travel time. If a vector with
#'   length bigger than 1 is passed, the output contains an additional column
#'   that specifies the percentile of each accessibility estimate. Due to
#'   upstream restrictions, only 5 percentiles can be specified at a time. For
#'   more details, please see `R5` documentation at <https://docs.conveyal.com/analysis/methodology#accounting-for-variability>.

# initial message about ram memory
.onAttach <- function(lib, pkg) { # nocov start
  msg <- cli::format_inline(
    "Please make sure you have already allocated some memory to Java by running
    {.code options(java.parameters = '-Xmx2G')} *before* loading library(r5r)

    You should replace {.val 2G} by the amount of memory you'll need.
    Currently, Java memory is set to {.val {getOption('java.parameters')}}."
  )

  packageStartupMessage(msg)
  } # nocov end

# package global variables
r5r_env <- new.env(parent = emptyenv())

.onLoad <- function(lib, pkg) { # nocov start

  # JAR version
  r5r_env$r5_jar_version <- "7.5.1" # "7.4.0"
  r5r_env$r5_jar_size <- 64437969 #  63866633

  # create dir to store R5 Jar
  cache_d <- paste0('r5r/r5_jar_v', r5r_env$r5_jar_version)
  r5r_env$cache_dir <- tools::R_user_dir(cache_d, which = 'cache')
  if (!dir.exists(r5r_env$cache_dir)) dir.create(r5r_env$cache_dir, recursive = TRUE)
  # gsub("\\\\", "/", r5r_env$cache_dir)

  ## delete any JAR files from old releases
  dir_above <- dirname(r5r_env$cache_dir)
  all_cache <- list.files(dir_above, pattern = 'r5',full.names = TRUE)
  old_cache <- all_cache[!grepl(r5r_env$r5_jar_version, all_cache)]
  if(length(old_cache)>0){ unlink(old_cache, recursive = TRUE) }

} # nocov end

#' @section Elevation:
#'
#' More information about the `TOBLER` and `MINETTI` options to calculate the
#' effects of elevation on travel times can be found in the references below:
#'
#' - Campbell, M. J., et al (2019). Using crowdsourced fitness tracker data to
#' model the relationship between slope and travel rates. Applied geography, 106,
#' 93-107. \doi{10.1016/j.apgeog.2019.03.008}.
#' - Minetti, A. E., et al (2002). Energy cost of walking and running at extreme
#' uphill and downhill slopes. Journal of applied physiology. \doi{10.1152/japplphysiol.01177.2001}.
#' - Tobler, W. (1993). Three presentations on geographical analysis and modeling:
#' Non-isotropic geographic modeling speculations on the geometry of geography
#' global spatial analysis. Technical Report. National center for geographic
#' information and analysis. 93 (1). \url{https://escholarship.org/uc/item/05r820mz}.

# options(java.parameters = '-Xmx16384m')
# options(java.parameters = c("-XX:+UseConcMarkSweepGC", "-Xmx16384m"))

# library(r5r)
devtools::load_all(".")
library(tidyverse)
library(sf)
library(h3jsr)
library(data.table)
library(jsonlite)

# POA --------------
data_path <- "~/Repos/r5r_fares/poa"
core_poa <- setup_r5(data_path = data_path, verbose = FALSE)

# RIO --------------
data_path <- "~/Repos/r5r_fares/rio"
core_rio <- setup_r5(data_path = data_path, verbose = FALSE)


# SPO --------------
data_path <- system.file("extdata/spo", package = "r5r")
core_spo <- setup_r5(data_path = data_path, verbose = FALSE)


# get fare settings -------------
f_struct <- setup_fare_calculator(core_poa, base_fare = 470, by = "MODE")
f_struct$base_fare <- 500

f_struct <- setup_fare_calculator(core_poa,
                                  base_fare = 470,
                                  by = "MODE")

write_fare_calculator(f_struct, "c:\fares_poa.zip")

f_struct <- read_fare_calculator("c:\fares_poa.zip")


f_struct <- fare_calculator_settings(core_poa, 470, "AGENCY")
f_struct <- fare_calculator_settings(core_poa, 470, "AGENCY_ID")
f_struct <- fare_calculator_settings(core_poa, 470, "AGENCY_NAME")
f_struct <- fare_calculator_settings(core_poa, 470, "GENERIC")

f_struct <- fare_calculator_settings(core_rio, 405, "MODE")
f_struct <- fare_calculator_settings(core_spo, 440, "MODE")

f_struct$base_fare
f_struct$max_discounted_transfers

f_struct$fare_per_mode

View(f_struct$fare_per_mode)
View(f_struct$fare_per_transfer)
View(f_struct$routes_info)

zip::zip_list()


f_struct_json <- jsonlite::toJSON(f_struct)
clipr::write_clip(f_struct_json)


###
t_net <- transit_network_to_sf(core_poa)
t_net$routes %>%
  filter(str_starts(short_name, "R")) %>%
  # filter(str_starts(short_name, "T")) %>%
  mapview::mapview()

#' Extract the geographic bounding box of the transport network
#'
#' Extracts the geographic bounding box of the street network layer from a
#' routable transport network built with [build_network()]). It is a fast and
#' memory-efficient alternative to `sf::st_bbox(street_network_to_sf(r5r_net))`.
#'
#' @template r5r_network
#' @template r5r_core
#' @param output A character string specifying the desired output format. One of
#'   `"polygon"` (the default), `"bbox"`, or `"vector"`.
#'
#' @return By default (`output = "polygon"`), an `sf` object with a single `POLYGON`
#'   geometry. If `output = "bbox"`, an `sf` `bbox` object. If `output = "vector"`,
#'   a named numeric vector with `xmin`, `ymin`, `xmax`, `ymax` coordinates.
#'   All outputs use the WGS84 coordinate reference system (EPSG: 4326).
#'
#' @family network functions
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#' library(sf)
#'
#' data_path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(data_path)
#'
#' # Get the network's bounding box as an sf polygon (default)
#' poly <- street_network_bbox(r5r_network, output = "polygon")
#' plot(poly)
#'
#' # Get an sf bbox object (order is xmin, ymin, xmax, ymax)
#' box <- street_network_bbox(r5r_network , output = "bbox")
#' box
#'
#' # Get a simple named vector (order now also xmin, ymin, xmax, ymax)
#' vec <- street_network_bbox(r5r_network , output = "vector")
#' vec
#'
#' stop_r5(r5r_network)
#' @export
street_network_bbox <- function(
  r5r_network,
  output = c("polygon", "bbox", "vector"),
  r5r_core = deprecated()
) {
  # deprecating r5r_core --------------------------------------
  if (lifecycle::is_present(r5r_core)) {
    cli::cli_warn(c(
      "!" = "The `r5r_core` argument is deprecated as of r5r v2.3.0.",
      "i" = "Please use the `r5r_network` argument instead."
    ))
    r5r_network <- r5r_core
  }

  # --- Input validation ---
  checkmate::assert_class(r5r_network, "r5r_network")
  output <- match.arg(output)

  # --- Call the optimized Java method ---
  coords_vec <- rJava::.jcall(
    r5r_network@jcore,
    "[D", # Signature for "returns an array of doubles"
    "getNetworkEnvelopeAsArray"
  )

  # --- Assign names based on the documented order from the Java method ---
  names(coords_vec) <- c("xmin", "xmax", "ymin", "ymax")

  # --- Reorder to match sf::st_bbox standard ---
  coords_vec <- coords_vec[c("xmin", "ymin", "xmax", "ymax")]

  # --- Format the output as requested ---
  if (output == "vector") {
    return(coords_vec)
  }

  bbox <- sf::st_bbox(coords_vec, crs = 4326) # WGS 84

  if (output == "bbox") {
    return(bbox)
  }

  if (output == "polygon") {
    return(sf::st_as_sfc(bbox))
  }
}

context("Testing street_network_bbox")

testthat::skip_on_cran()

test_that("street_network_bbox returns an sf polygon by default", {
  bbox_poly <- street_network_bbox(r5r_network)

  # Check class and type
  expect_s3_class(bbox_poly, "sfc")
  expect_s3_class(bbox_poly, "sfc_POLYGON")
  expect_equal(
    as.character(sf::st_geometry_type(bbox_poly, by_geometry = FALSE)),
    "POLYGON"
  )

  # Check properties
  expect_equal(length(bbox_poly), 1)
  expect_equal(sf::st_crs(bbox_poly), sf::st_crs(4326))
})

test_that("street_network_bbox returns a bbox object when output = 'bbox'", {
  bbox_obj <- street_network_bbox(r5r_network, output = "bbox")

  # Check class and properties
  expect_s3_class(bbox_obj, "bbox")
  expect_true(is.numeric(bbox_obj))
  expect_length(bbox_obj, 4)
  expect_named(bbox_obj, c("xmin", "ymin", "xmax", "ymax"))
})

test_that("street_network_bbox returns a named vector when output = 'vector'", {
  bbox_vec <- street_network_bbox(r5r_network, output = "vector")

  # Check class and properties
  expect_true(is.numeric(bbox_vec))
  expect_false(inherits(bbox_vec, "bbox"))
  expect_length(bbox_vec, 4)
  expect_named(bbox_vec, c("xmin", "ymin", "xmax", "ymax"))
})

test_that("bbox and vector outputs are consistent in value and order", {
  bbox_obj <- street_network_bbox(r5r_network, output = "bbox")
  bbox_vec <- street_network_bbox(r5r_network, output = "vector")

  # Check that the numeric values are equal, ignoring names and attributes.
  expect_equal(as.numeric(bbox_obj), unclass(unname(bbox_vec)))

  # Check that the names are identical.
  expect_identical(names(bbox_obj), names(bbox_vec))
})

test_that("street_network_bbox gives deprecation warning for r5r_core", {
  # Check that the warning is thrown
  bbox_poly <- expect_warning(
    street_network_bbox(r5r_core = r5r_network),
    regexp = "The `r5r_core` argument is deprecated"
  )

  # Check that the function still returns the correct output
  expect_s3_class(bbox_poly, "sfc")
  expect_s3_class(bbox_poly, "sfc_POLYGON")
})

test_that("street_network_bbox throws an error for invalid output argument", {
  expect_error(
    street_network_bbox(r5r_network, output = "invalid_option"),
    regexp = "'arg' should be one of"
  )
})

# options(java.parameters = '-Xmx16384m')
# options(java.parameters = c("-XX:+UseConcMarkSweepGC", "-Xmx16384m"))

# library(r5r)
devtools::load_all(".")
# library(ggplot2)
library(data.table)
library(tidyverse)
# build transport network
# data_path <- system.file("extdata/spo", package = "r5r")
data_path <- system.file("extdata/poa", package = "r5r")
data_path <- paste0(data_path, "/../../extdata/poa")
r5r_core <- setup_r5(data_path = data_path, verbose = FALSE, overwrite = FALSE)

# load origin/destination points

departure_datetime <- as.POSIXct("13-05-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S")

# points <- fread(file.path(data_path, "spo_hexgrid.csv"))
points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))
# points <- fread(file.path(data_path, "poa_hexgrid.csv"))
# dest <- points

# dir.create(here::here("csv"))

# r5r_core$setCsvOutput(here::here("csv"))

  normal_ttm <- travel_time_matrix(r5r_core, origins = points, #[id == "89a9012a3cfffff",],
                                   destinations = points, #[id == "89a901284a3ffff",],
                            mode = c("CAR", "TRANSIT"),
                            departure_datetime = departure_datetime,
                            max_trip_duration = 60,
                            max_walk_time = 15,
                            time_window = 30,
                            percentiles = c(1, 25, 50, 75, 99),
                            verbose = FALSE,
                            progress = TRUE)

  ttm2 <- travel_time_matrix(r5r_core, origins = points, #[id == "89a9012a3cfffff",],
                                   destinations = points, #[id == "89a901284a3ffff",],
                                   mode = c("CAR", "TRANSIT"),
                                   departure_datetime = departure_datetime,
                                   max_trip_duration = 60,
                                   max_walk_time = 15,
                                   time_window = 30,
                                   percentiles = c(1, 25, 50, 75, 99),
                                   verbose = FALSE,
                                   progress = TRUE)

  ttm3 <- travel_time_matrix(r5r_core, origins = points, #[id == "89a9012a3cfffff",],
                                   destinations = points, #[id == "89a901284a3ffff",],
                                   mode = c("CAR", "TRANSIT"),
                                   departure_datetime = departure_datetime,
                                   max_trip_duration = 60,
                                   max_walk_time = 15,
                             max_car_time = 15,
                                   time_window = 30,
                                   percentiles = c(1, 25, 50, 75, 99),
                                   verbose = FALSE,
                                   progress = TRUE)

  dit <- detailed_itineraries(r5r_core, origins = points[5,], #[id == "89a9012a3cfffff",],
                              destinations = points[11,], #[id == "89a901284a3ffff",],
                              mode = c("CAR", "TRANSIT"),
                              departure_datetime = departure_datetime,
                              max_trip_duration = 60,
                              max_walk_time = 15,
                              max_car_time = 15,
                              time_window = 30,
                              verbose = FALSE,
                              progress = TRUE)

  mapview::mapview(dit, zcol="mode")

#' Class to internally handle Java reference to R5RCore
#'
#' @family r5r_network
#'
#' @keywords internal
#' @importClassesFrom rJava jobjRef
setClass(
  "r5r_network",
  slots = list(jcore = "jobjRef")
)

#' Constructor for r5r_network object
#'
#' @description
#' Wraps a Java R5RCore as an r5r_network.
#'
#' @param jcore A \code{jobjRef} Java object reference to R5RCore.
#' @return \code{r5r_network}
#'
#' @family r5r_network
#'
#' @keywords internal
wrap_r5r_network <- function(jcore) { # nocov start
  if (!identical(jcore$identify(), "I am an R5R core!")) {
    stop('Provided object is not a valid reference to a java R5R core.')
  }
  methods::new("r5r_network", jcore = jcore)
} # nocov end

# options(java.parameters = '-Xmx16384m')
# options(java.parameters = c("-XX:+UseConcMarkSweepGC", "-Xmx16384m"))

# library(r5r)
devtools::load_all(".")
library(tidyverse)
library(sf)
library(h3jsr)
library(data.table)

# build transport network
data_path <- "~/Repos/r5r_fares/poa"
r5r_core <- setup_r5(data_path = data_path, verbose = FALSE)



# load origin/destination points

poi <- read.csv(system.file("extdata/poa/poa_points_of_interest.csv", package = "r5r"))
points_small <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))
points_small$unit <- 1
points <- read_csv("~/Repos/r5r_fares/poa/poa_points.csv") %>%
  rename(id = id_hex)
departure_datetime <- as.POSIXct("13-05-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S")

# Accessibility -----------------------------------------------------------

calculate_access <- function(fares) {

  access_df <- map_df(fares, function(f) {
    f <- as.integer(f)

    r5r_core$setMaxFare(f, "porto-alegre")

    access <- accessibility(r5r_core,
                            origins = points_small,
                            destinations = points_small,
                            departure_datetime = departure_datetime,
                            opportunities_colname = "unit",
                            mode = c("WALK", "TRANSIT"),
                            cutoffs = c(30, 45),
                            max_trip_duration = 45,
                            max_walk_dist = 800,
                            time_window = 1,
                            percentiles = 50,
                            verbose = FALSE)

    access$max_fare <- f

    return(access)
  })

  return(access_df)
}

access_df <- calculate_access(c(450, 480, 720)) %>%
  # access_df <- calculate_access(c(240, 480, 720, 960, -1)) %>%
  left_join(points, by = c("from_id" = "id"))

access_df %>%
  filter(cutoff == 30) %>%
  drop_na() %>%
  ggplot(aes(x=lon, y=lat)) +
  geom_point(size=1, aes(color=accessibility)) +
  geom_point(data=poi[c(1, 10),], color = "blue") +
  facet_grid(cutoff~max_fare) +
  scale_color_distiller(palette = "Spectral") +
  coord_map()



# Travel Times ------------------------------------------------------------

r5r_core$setMaxFare(-1L, "porto-alegre")
r5r_core$setMaxFare(80L, "porto-alegre")

ttm <- travel_time_matrix(r5r_core,
                        origins = poi[1,],
                        destinations = poi[10,],
                        departure_datetime = departure_datetime,
                        mode = c("WALK", "TRANSIT"),
                        breakdown = T,
                        max_trip_duration = 45,
                        max_walk_dist = 800,
                        time_window = 1,
                        percentiles = 50,
                        verbose = FALSE)


# Pareto ------------------------------------------------------------------

pareto_df <- pareto_frontier(r5r_core,
                             origins = poi[2,],
                             destinations = poi[3,],
                             mode = c("WALK", "TRANSIT"),
                             departure_datetime = departure_datetime,
                             monetary_cost_cutoffs = seq(0, 1000, 50),
                             fare_calculator = "porto-alegre",
                             max_trip_duration = 60,
                             max_walk_dist = 8000,
                             time_window = 1, #30,
                             percentiles = 50, # c(5, 50, 95),
                             max_rides = 5,
                             verbose = FALSE,
                             progress = TRUE)

pareto_df$monetary_cost <- pareto_df$monetary_cost / 100
pareto_df$monetary_cost_upper <- pareto_df$monetary_cost_upper / 100
View(pareto_df)

pareto_df %>%
  mutate(percentile = factor(percentile)) %>%
  pivot_longer(cols=starts_with("monetary"), names_to = "mon", values_to="cost") %>%
  ggplot(aes(x=cost, y=travel_time, color=percentile, group=percentile)) +
  geom_step() +
  # geom_path() +
  scale_color_brewer(palette = "Set1") +
  scale_x_continuous(breaks = 1:10) +
  scale_y_continuous(breaks = seq(0, 50, 10), limits = c(0, 50)) +
  facet_grid(from_id~to_id)


r5r_core$setMaxFare(10L, "rio-de-janeiro")
r5r_core$verboseMode()



# Transit Network ---------------------------------------------------------

tn = transit_network_to_sf(r5r_core)
tn$routes %>% View()

mapview::mapview(tn$routes)


## Fare structure -------



f_struct <- setup_fare_calculator(r5r_core, 4.7, "MODE")

View(f_struct$fares_per_mode)

f_struct$base_fare
f_struct$max_discounted_transfers

f_struct$fare_per_mode
f_struct$fare_per_transfer
f_struct$routes_info

access <- accessibility(r5r_core,
                        origins = points_small,
                        destinations = points_small,
                        departure_datetime = departure_datetime,
                        opportunities_colname = "unit",
                        mode = c("WALK", "TRANSIT"),
                        cutoffs = c(30, 45),
                        fare_calculator_settings = f_struct,
                        max_fare = 10,
                        max_trip_duration = 45,
                        max_walk_dist = 800,
                        time_window = 1,
                        percentiles = 50,
                        verbose = FALSE) %>%
  left_join(points, by = c("from_id" = "id"))


access %>%
  filter(cutoff == 30) %>%
  drop_na() %>%
  ggplot(aes(x=lon, y=lat)) +
  geom_point(size=1, aes(color=accessibility)) +
  geom_point(data=poi[c(1, 10),], color = "blue") +
  facet_grid(~cutoff) +
  scale_color_distiller(palette = "Spectral") +
  coord_map()

#' @param origins,destinations Either a `POINT sf` object with WGS84 CRS, or a
#'   `data.frame` containing the columns `id`, `lon` and `lat`.
#' @param mode A character vector. The transport modes allowed for access,
#'   transfer and vehicle legs of the trips. Defaults to `WALK`. Please see
#'   details for other options.
#' @param mode_egress A character vector. The transport mode used after egress
#'   from the last public transport. It can be either `WALK`, `BICYCLE` or
#'   `CAR`. Defaults to `WALK`. Ignored when public transport is not used.
#' @param departure_datetime A POSIXct object. Please note that the departure
#'   time only influences public transport legs. When working with public
#'   transport networks, please check the `calendar.txt` within your GTFS feeds
#'   for valid dates. Please see details for further information on how
#'   datetimes are parsed.
#' @param max_walk_time An integer. The maximum walking time (in minutes) to
#'   access and egress the transit network, to make transfers within the network
#'   or to complete walk-only trips. Defaults to no restrictions (numeric value
#'   of `Inf`), as long as `max_trip_duration` is respected. When routing
#'   transit trips, the max time is considered separately for each leg (e.g. if
#'   you set `max_walk_time` to 15, you could get trips with an up to 15 minutes
#'   walk leg to reach transit and another up to 15 minutes walk leg to reach
#'   the destination after leaving transit. In walk-only trips, whenever
#'   `max_walk_time` differs from `max_trip_duration`, the lowest value is
#'   considered.
#' @param max_bike_time An integer. The maximum cycling time (in minutes) to
#'   access and egress the transit network, to make transfers within the network
#'   or to complete bicycle-only trips. Defaults to no restrictions (numeric
#'   value of `Inf`), as long as `max_trip_duration` is respected. When routing
#'   transit trips, the max time is considered separately for each leg (e.g. if
#'   you set `max_bike_time` to 15, you could get trips with an up to 15 minutes
#'   cycle leg to reach transit and another up to 15 minutes cycle leg to reach
#'   the destination after leaving transit. In bicycle-only trips, whenever
#'   `max_bike_time` differs from `max_trip_duration`, the lowest value is
#'   considered.
#' @param max_car_time An integer. The maximum driving time (in minutes) to
#'   access and egress the transit network. Defaults to no restrictions, as long
#'   as `max_trip_duration` is respected. The max time is considered separately
#'   for each leg (e.g. if you set `max_car_time` to 15 minutes, you could
#'   potentially drive up to 15 minutes to reach transit, and up to _another_ 15
#'   minutes to reach the destination after leaving transit). Defaults to `Inf`,
#'   no limit.
#' @param max_trip_duration An integer. The maximum trip duration in minutes.
#'   Defaults to 120 minutes (2 hours).
#' @param walk_speed A numeric. Average walk speed in km/h. Defaults to 3.6
#'   km/h.
#' @param bike_speed A numeric. Average cycling speed in km/h. Defaults to 12
#'   km/h.
#' @param max_rides An integer. The maximum number of public transport rides
#'   allowed in the same trip. Defaults to 3.
#' @param max_lts An integer between 1 and 4. The maximum level of traffic
#'   stress that cyclists will tolerate. A value of 1 means cyclists will only
#'   travel through the quietest streets, while a value of 4 indicates cyclists
#'   can travel through any road. Defaults to 2. Please see details for more
#'   information.
#' @param n_threads An integer. The number of threads to use when running the
#'   router in parallel. Defaults to use all available threads (Inf).
#' @param progress A logical. Whether to show a progress counter when running
#'   the router. Defaults to `FALSE`. Only works when `verbose` is set to
#'   `FALSE`, so the progress counter does not interfere with `R5`'s output
#'   messages. Setting `progress` to `TRUE` may impose a small penalty for
#'   computation efficiency, because the progress counter must be synchronized
#'   among all active threads.
#' @param output_dir Either `NULL` or a path to an existing directory. When not
#'   `NULL` (the default), the function will write one `.csv` file with the
#'   results for each origin in the specified directory. In such case, the
#'   function returns the path specified in this parameter. This parameter is
#'   particularly useful when running on memory-constrained settings because
#'   writing the results directly to disk prevents `r5r` from loading them to
#'   RAM memory.

# options(java.parameters = '-Xmx16384m')
# options(java.parameters = c("-XX:+UseConcMarkSweepGC", "-Xmx16384m"))

# library(r5r)
devtools::load_all(".")
library(tidyverse)
library(sf)
library(h3jsr)
library(data.table)
library(viridis)
library(patchwork)
library(googlesheets4)

# build transport network
data_path <- "~/Repos/r5r_fares/rio"
r5r_core <- setup_r5(data_path = data_path, verbose = FALSE)

# load origin/destination points

poi <- tribble(
  ~id, ~lat, ~lon,
  "centro", -22.9064720147941, -43.177139635807734,
  "catete", -22.930673881789726, -43.17764097392119,
  "copacabana", -22.96915219120712, -43.18463648168664,
  "pavuna", -22.814726, -43.362337,
  "jardim_oceanico", -23.009335, -43.312220,
  "santa_cruz", -22.914807, -43.688556,
  "madureira", -22.875438, -43.339674
)

points <- read_csv("~/Repos/r5r_fares/rio/points_rio_09_2019.csv") %>%
  rename(id = id_hex, lon=X, lat=Y) %>%
  mutate(unit = 1)
departure_datetime <- as.POSIXct("13-05-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S")

points_r8 <- h3jsr::get_parent(points$id, res = 8, simple = TRUE)
points_r8 <- unique(points_r8)
points_r8 <- h3jsr::h3_to_point(points_r8, simple = FALSE)
points_r8$id <- points_r8$h3_address
points_r8$unit <- 1
# area_sf <- h3jsr::h3_to_polygon(points$id, simple = FALSE) %>%
#   summarise()


# Setup fare calculator ---------------------------------------------------

fare_settings <- setup_fare_calculator(r5r_core,
                                       base_fare = 405,
                                       by = "MODE")


write_fare_calculator(fare_settings, file_path = here::here("tests_marcus", "rio_fares_v3.zip"))
fare_settings <- read_fare_calculator(file_path = here::here("tests_marcus", "rio_fares_v3.zip"))



fare_settings$base_fare
fare_settings$max_discounted_transfers
fare_settings$transfer_time_allowance
fare_settings$fare_cap
fare_settings$fare_per_mode %>% View()
fare_settings$fare_per_transfer %>% View()
fare_settings$routes_info %>% View()
fare_settings$debug_settings


unique(fare_settings$fares_per_route$route_id)

clipr::write_clip()
clipr::write_clip(fare_settings$fares_per_mode)
clipr::write_clip(fare_settings$fares_per_transfer)
clipr::write_clip(fare_settings$fares_per_route)

### Test --------------------------------------------------------

fare_settings <- read_fare_calculator(file_path = here::here("tests_marcus", "rio_fares_v3.zip"))

fare_settings$debug_settings$output_file <- here::here("tests_marcus", "rio_fare_calculator_output_v7.csv")

# r5r_core$setFareCalculatorDebugOutput(here::here("tests_marcus", "rio_fare_calculator_output_v4.csv"))

access <- accessibility(r5r_core,
                        origins = poi,
                        destinations = points_r8,
                        departure_datetime = departure_datetime,
                        opportunities_colname = "unit",
                        mode = c("WALK", "TRANSIT"),
                        cutoffs = c(30, 45),
                        fare_calculator_settings = fare_settings,
                        max_fare = 10,
                        max_trip_duration = 90,
                        max_walk_dist = 800,
                        time_window = 1,
                        percentiles = 50,
                        verbose = FALSE,
                        progress = TRUE)


rio_debug_v1 <- read_csv(here::here("tests_marcus", "rio_fare_calculator_output.csv"))
rio_debug_v4 <- read_csv(here::here("tests_marcus", "rio_fare_calculator_output_v4.csv"))
rio_debug_v5 <- read_csv(here::here("tests_marcus", "rio_fare_calculator_output_v5.csv"))
rio_debug_v6 <- read_csv(here::here("tests_marcus", "rio_fare_calculator_output_v6.csv"))
rio_debug_v7 <- read_csv(here::here("tests_marcus", "rio_fare_calculator_output_v7.csv"))

tn <- transit_network_to_sf(r5r_core)

mapview::mapview(points_r8, crs = 4326)
mapview::mapview(tn$routes)
rio_debug <- left_join(rio_debug_v1, rio_debug_v4, by = "pattern")

rio_debug %>%
  filter(fare.x != fare.y) %>%
  View()
# Accessibility -----------------------------------------------------------

calculate_access <- function(fares) {

  access_df <- map_df(fares, function(f) {
    f <- as.integer(f)

    r5r_core$setMaxFare(f, "rio-de-janeiro")

    access <- accessibility(r5r_core,
                            origins = points,
                            destinations = points,
                            departure_datetime = departure_datetime,
                            opportunities_colname = "unit",
                            mode = c("WALK", "TRANSIT"),
                            cutoffs = c(30, 45),
                            max_trip_duration = 45,
                            max_walk_dist = 800,
                            time_window = 1,
                            percentiles = 50,
                            verbose = FALSE)

    access$max_fare <- f

    return(access)
  })

  return(access_df)
}

access_df <- calculate_access(c(405, 500, 700)) %>%
  # access_df <- calculate_access(c(240, 480, 720, 960, -1)) %>%
  left_join(points, by = c("from_id" = "id"))

access_df %>%
  filter(cutoff == 45) %>%
  drop_na() %>%
  arrange(accessibility) %>%
  ggplot(aes(x=lon, y=lat)) +
  geom_point(size=1, aes(color=accessibility)) +
  facet_wrap(~max_fare, ncol = 2) +
  scale_color_distiller(palette = "Spectral") +
  coord_map()



# Travel Times ------------------------------------------------------------

r5r_core$setMaxFare(-1L, "rio-de-janeiro") # infinito
r5r_core$setMaxFare(405L, "rio-de-janeiro") # R$ 4,05
r5r_core$setMaxFare(500L, "rio-de-janeiro")
r5r_core$setMaxFare(700L, "rio-de-janeiro")

ttm <- travel_time_matrix(r5r_core,
                        origins = poi[1,],
                        destinations = poi[2,],
                        departure_datetime = departure_datetime,
                        mode = c("WALK", "TRANSIT"),
                        max_trip_duration = 60,
                        max_walk_dist = 800,
                        time_window = 1,
                        percentiles = 50,
                        verbose = FALSE)


# Pareto ------------------------------------------------------------------

fare_settings <- read_fare_calculator(file_path = here::here("tests_marcus", "rio_fares_v3.zip"))

pareto_cutoffs <- c(0, 3.80,  4.05,  4.70,  5.00,  6.05,  6.50,  7.10,  7.60,  8.10,  8.55,  9.40,  10.00)

system.time(
  pareto_df <- pareto_frontier(r5r_core,
                               origins = poi,
                               destinations = poi,
                               mode = c("WALK", "TRANSIT"),
                               departure_datetime = departure_datetime,
                               monetary_cost_cutoffs = pareto_cutoffs,
                               fare_calculator_settings = fare_settings,
                               max_trip_duration = 180,
                               max_walk_dist = 8000,
                               time_window = 1, #30,
                               percentiles = 50, # c(5, 50, 95),
                               max_rides = 5,
                               draws_per_minute = 5L,
                               verbose = FALSE,
                               progress = TRUE)
  )

r5r_core$getFareCalculatorCacheCalls()
r5r_core$getFareCalculatorFullCalls()

# [1] 570551246

pareto_df %>%
  mutate(percentile = factor(percentile),
         pair = paste(from_id, to_id)) %>%
  pivot_longer(cols=starts_with("monetary"), names_to = "mon", values_to="cost") %>%
  ggplot(aes(x=cost, y=travel_time, color=pair)) +
  geom_step() +
  geom_point() +
  # geom_path() +
  # scale_color_brewer(palette = "Set1") +
  # scale_x_continuous(breaks = 0:10, limits = c(0, 10)) +
  # scale_y_continuous(breaks = seq(0, 180, 30), limits = c(0, 180)) +
  theme(legend.position = "none") +
  facet_wrap(~pair)


r5r_core$setMaxFare(10L, "rio-de-janeiro")
r5r_core$verboseMode()


# Pareto Accessibility ----------------------------------------------------
# data <- pareto_df
# cost_threshold <- 1000
access_pareto_map <- function(data, cost_threshold) {
  data_filtered <- data %>%
    drop_na() %>%
    filter(monetary_cost <= cost_threshold)

  if (nrow(data_filtered) == 0) { return(NULL) }
  data_filtered$geometry <- h3jsr::h3_to_polygon(data_filtered$to_id)

  data_sf <- st_as_sf(data_filtered, crs = 4326)

  p <- data_sf %>%
    mutate(cost = cost_threshold/100) %>%
    ggplot() +
    geom_sf(data = area_sf, fill = "grey80", color = NA) +
    geom_sf(aes(fill=travel_time), color = NA) +
    coord_sf(datum = NA) +
    scale_fill_viridis(direction = -1) +
    facet_wrap(~cost) +
    theme(legend.position = "none")

  return(p)
}

pareto_cutoffs <- c(0, 380,  405,  470,  500,  605,  650,  710,  760,  810,  855,  940,  1000)

pareto_df <- pareto_frontier(r5r_core,
                             origins = poi[1,],
                             destinations = points,
                             mode = c("WALK", "TRANSIT"),
                             departure_datetime = departure_datetime,
                             monetary_cost_cutoffs = pareto_cutoffs, #seq(300, 900, 100),
                             fare_calculator = "rio-de-janeiro",
                             max_trip_duration = 90,
                             max_walk_dist = 4000,
                             time_window = 1, #30,
                             percentiles = 50, # c(5, 50, 95),
                             max_rides = 5,
                             verbose = FALSE,
                             progress = TRUE)


unique(pareto_df$monetary_cost)

plots <- map(sort(unique(pareto_df$monetary_cost)), access_pareto_map, data = pareto_df)
# plots <- map(seq(400, 900, 100), access_pareto_map, data = pareto_df)
plots[sapply(plots, is.null)] <- NULL

wrap_plots(plots, ncol = 4) + plot_annotation(title = "Centro")


# Transit Network ---------------------------------------------------------

tn = transit_network_to_sf(r5r_core)
tn$routes %>% View()

tn$routes %>% filter(route_id == 18895165)

devtools::load_all(".")

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path, elevation = "TOBLER", overwrite = T)

# load origin/destination points
points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))
# points <- read.csv(file.path(data_path, "poa_hexgrid.csv")) %>%
  # dplyr::sample_n(50)

# load fare structure object
fare_structure_path <- system.file(
  "extdata/poa/fares/fares_poa.zip",
  package = "r5r"
)
fare_structure <- read_fare_structure(fare_structure_path)

# inputs
departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
                                 format = "%d-%m-%Y %H:%M:%S")

# r5r_core$setDetailedItinerariesV2(TRUE)
# r5r_core$setDetailedItinerariesV2(FALSE)

# a <- capture.output(
system.time(
  det_new <- detailed_itineraries(r5r_core,
                              origins = points[1:5,],
                              destinations = points[1:5,],
                              mode = c("WALK", "TRANSIT"),
                              departure_datetime = departure_datetime,
                              max_walk_time = 30,
                              max_trip_duration = 90,
                              suboptimal_minutes = 5,
                              # fare_structure = fare_structure,
                              # max_fare = 9,
                              time_window = 1,
                              all_to_all = T,
                              progress = T,
                              shortest_path = F,
                              verbose = F,
                              drop_geometry = F)
  )
# )

det_new$dist <- sf::st_length(det_new)


mapview::mapview(det_new, zcol = "option") +
  mapview::mapview(points, xcol="lon", ycol="lat", crs=4326)

mapview::mapview(dplyr::filter(det_new, option == 2), zcol = "route")
mapview::mapview(dplyr::filter(det2,
                               from_id == "beira_rio_stadium",
                               to_id == "bus_central_station",
                               option == 2), zcol = "mode")

r5r::select_mode(c("WALK", "BICYCLE"), "WALK", style = "dit")
set_fare_structure(r5r_core, fare_structure)
r5r_core$setMaxFare(rJava::.jfloat(10.0))

r5r_core$dropFareCalculator()

library("tidyverse")

saveRDS(det, "det_v2.rds")
saveRDS(det2, "det_v1.rds")

library(sf)
library(tidyverse)

det_new %>%
  st_set_geometry(NULL) %>%
  group_by(option) %>%
  summarise(mode = paste(mode, collapse = "|"),
            routes = paste(route, collapse = "|"))

det_new %>%
  st_set_geometry(NULL) %>%
  group_by(option) %>%
  summarise(mode = paste(mode, collapse = "|"),
            routes = paste(route, collapse = "|"),
            total_fare = mean(total_fare))

a <- det_new %>%
  st_set_geometry(NULL) %>%
  group_by(option) %>%
  mutate(sum_dur = sum(segment_duration + wait),
         is_diff = sum_dur != total_duration)

suppressWarnings()

a <- r5r_core$message("bla")








# load libraries
library("r5r")
library("data.table")
library("tidyverse")

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path)

# inputs
departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
                                 format = "%d-%m-%Y %H:%M:%S")

# size <- 15
compute_paths <- function(sm, tw) {
  # sample_data <- fread(here::here("data", "sample_15.csv"))

  t <- system.time(
    dit <- detailed_itineraries(r5r_core,
                                origins = sample_data[1,],
                                destinations = sample_data[12,],
                                mode = c("WALK", "TRANSIT"),
                                departure_datetime = departure_datetime,
                                suboptimal_minutes = sm,
                                time_window = tw,
                                max_walk_dist = 1000,
                                max_trip_duration = 120,
                                progress = T,
                                shortest_path = F)
  )

  l <- dit$option |> unique() |> length()

  rm(dit)
  rJava::.jgc()

  return(data.table(suboptimal_minutes = sm,
                    time_window = tw,
                    n_options = l,
                    time = t[3]))
}

# compute paths
# times_old <- lapply(c(15, 25), compute_paths) |> rbindlist()

df <- NULL
for (sm in 0:15) {
  for (tw in 1:15) {
    df1 <- compute_paths(sm, tw)

    if (is.null(df)) {
      df <- df1
    } else {
      df <- rbind(df, df1)
    }
  }
}

#' Detailed itineraries between origin-destination pairs
#'
#' Returns detailed trip information between origin-destination pairs. The
#' output includes the waiting and moving time in each trip leg, as well as some
#' info such as the distance traveled, the routes used and the geometry of each
#' leg. Please note that this function was originally conceptualized as a trip
#' planning functionality, similar to other commercial and non-commercial APIs
#' and apps (e.g. Moovit, Google's Directions API, OpenTripPlanning's
#' PlannerResource API). Thus, it consumes much more time and memory than the
#' other (more analytical) routing functions included in the package.
#'
#' @template r5r_network
#' @template r5r_core
#' @template common_arguments
#' @template scenarios
#' @template verbose
#' @template fare_structure
#' @template max_fare
#' @param time_window An integer. The time window in minutes for which `r5r`
#'   will calculate multiple itineraries departing each minute. Defaults to 10
#'   minutes. If the same sequence of routes appear in different minutes of the
#'   time window, only the fastest of them will be kept in the output. This
#'   happens because the result is not aggregated by percentile, as opposed to
#'   other routing functions in the package. Because of that, the output may
#'   contain trips departing after the specified `departure_datetime`, but
#'   still within the time window. Please read the time window vignette for
#'   more details on how this argument affects the results of each routing
#'   function: `vignette("time_window", package = "r5r")`.
#' @param suboptimal_minutes A number. The difference in minutes that each
#'   non-optimal RAPTOR branch can have from the optimal branch without being
#'   disregarded by the routing algorithm. If, for example, users set
#'   `suboptimal_minutes = 10`, the routing algorithm will consider sub-optimal
#'   routes that arrive up to 10 minutes after the arrival of the optimal one.
#'   This argument emulates the real-life behaviour that makes people want to
#'   take a path that is technically not optimal in terms of travel time, for
#'   example, for some practical reasons (e.g. mode preference, safety, etc).
#'   In practice, the higher this value, the more itineraries will be returned
#'   in the final result.
#' @param shortest_path A logical. Whether the function should only return the
#'   fastest itinerary between each origin and destination pair (the default)
#'   or multiple alternatives.
#' @param all_to_all A logical. Whether to query routes between the 1st origin
#'   to the 1st destination, then the 2nd origin to the 2nd destination, and so
#'   on (`FALSE`, the default) or to query routes between all origins to all
#'   destinations (`TRUE`).
#' @param drop_geometry A logical. Whether the output should include the
#'   geometry of each trip leg or not. The default value of `FALSE` keeps the
#'   geometry column in the result.
#' @param osm_link_ids A logical. Whether the output should include additional
#'   columns: `osm_id_list` for the OSM ids of the road
#'   segments used along the trip geometry, and `board_stop_id` and
#'   `alight_stop_id` for the OSM ids of transit boarding and alighting stops.
#'   Defaults to `FALSE`.
#'
#'   Keep in mind that `osm_id_list` will contain an id even if the route only
#'   uses a small stretch of the road (e.g. 5m of a 600m street segment). For
#'   more precision, use `edge_id_list`, which returns the exact internal edge
#'   segments used in the trip. You can inspect these edge ids and their
#'   associated properties, including OSM ids, with
#'   [street_network_to_sf()].
#'
#' @template transport_modes_section
#' @template lts_section
#' @template datetime_parsing_section
#' @template mcraptor_algorithm_section
#'
#' @return When `drop_geometry` is `FALSE`, the function outputs a `LINESTRING
#'   sf` with detailed information on the itineraries between the specified
#'   origins and destinations. When `TRUE`, the output is a `data.table`. All
#'   distances are in meters and travel times are in minutes. If `output_dir`
#'   is not `NULL`, the function returns the path specified in that parameter,
#'   in which the `.csv` files containing the results are saved.
#'
#' @family routing
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' # build transport network
#' data_path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(data_path)
#'
#' # load origin/destination points
#' points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))
#'
#' # inputs
#' departure_datetime <- as.POSIXct(
#'   "13-05-2019 14:00:00",
#'   format = "%d-%m-%Y %H:%M:%S"
#' )
#'
#' det <- detailed_itineraries(
#'   r5r_network,
#'   origins = points[10,],
#'   destinations = points[12,],
#'   mode = c("WALK", "TRANSIT"),
#'   departure_datetime = departure_datetime,
#'   max_trip_duration = 60
#' )
#' head(det)
#'
#' stop_r5(r5r_network)
#' @export
detailed_itineraries <- function(r5r_network,
                                 r5r_core = deprecated(),
                                 origins,
                                 destinations,
                                 mode = "WALK",
                                 mode_egress = "WALK",
                                 departure_datetime = Sys.time(),
                                 time_window = 10L,
                                 suboptimal_minutes = 0L,
                                 max_walk_time = Inf,
                                 max_bike_time = Inf,
                                 max_car_time = Inf,
                                 max_trip_duration = 120L,
                                 walk_speed = 3.6,
                                 bike_speed = 12,
                                 max_rides = 3,
                                 max_lts = 2,
                                 shortest_path = TRUE,
                                 all_to_all = FALSE,
                                 fare_structure = NULL,
                                 max_fare = Inf,
                                 new_carspeeds = NULL,
                                 carspeed_scale = 1,
                                 new_lts = NULL,
                                 n_threads = Inf,
                                 verbose = FALSE,
                                 progress = FALSE,
                                 drop_geometry = FALSE,
                                 osm_link_ids = FALSE,
                                 output_dir = NULL) {

  # deprecating r5r_core --------------------------------------
  if (lifecycle::is_present(r5r_core)) {

    cli::cli_warn(c(
      "!" = "The `r5r_core` argument is deprecated as of r5r v2.3.0.",
      "i" = "Please use the `r5r_network` argument instead."
    ))

    r5r_network <- r5r_core
  }

  old_options <- options(datatable.optimize = Inf)
  on.exit(options(old_options), add = TRUE)

  old_dt_threads <- data.table::getDTthreads()
  dt_threads <- ifelse(is.infinite(n_threads), 0, n_threads)
  data.table::setDTthreads(dt_threads)
  on.exit(data.table::setDTthreads(old_dt_threads), add = TRUE)

  # check inputs and set r5r options --------------------------------------

  checkmate::assert_class(r5r_network, "r5r_network")

  origins <- assign_points_input(origins, "origins")
  destinations <- assign_points_input(destinations, "destinations")
  od_list <- expand_od_pairs(origins, destinations, all_to_all)
  origins <- od_list$origins
  destinations <- od_list$destinations

  mode_list <- assign_mode(mode, mode_egress)
  departure <- assign_departure(departure_datetime)

  # check availability of transit services on the selected date
  if (mode_list$transit_mode %like% 'TRANSIT|TRAM|SUBWAY|RAIL|BUS|CABLE_CAR|GONDOLA|FUNICULAR') {
    check_transit_availability_on_date(r5r_network, departure_date = departure$date)
  }

  r5r_network <- r5r_network@jcore

  # detailed itineraries via public transport cannot be computed on frequencies-based GTFS
  if (mode_list$transit_mode != "" & r5r_network$hasFrequencies()) {
    stop(
      "Assertion on 'r5r_network' failed: None of the GTFS feeds used to create ",
      "the transit network can contain a 'frequencies' table. Try using ",
      "gtfstools::frequencies_to_stop_times() to create a suitable feed."
    )
  }

  max_walk_time <- assign_max_street_time(
    max_walk_time,
    walk_speed,
    max_trip_duration,
    "walk"
  )
  max_bike_time <- assign_max_street_time(
    max_bike_time,
    bike_speed,
    max_trip_duration,
    "bike"
  )
  max_car_time <- assign_max_street_time(
    max_car_time,
    8, # 8 km/h, R5's default.
    max_trip_duration,
    "car"
  )
  max_trip_duration <- assign_max_trip_duration(
    max_trip_duration,
    mode_list,
    max_walk_time,
    max_bike_time
  )
  shortest_path <- assign_shortest_path(shortest_path)
  drop_geometry <- assign_drop_geometry(drop_geometry)
  osm_link_ids <- assign_osm_link_ids(osm_link_ids, drop_geometry)

  set_time_window(r5r_network, time_window)
  set_monte_carlo_draws(r5r_network, 1, time_window)
  set_speed(r5r_network, walk_speed, "walk")
  set_speed(r5r_network, bike_speed, "bike")
  set_max_rides(r5r_network, max_rides)
  set_max_lts(r5r_network, max_lts)
  set_n_threads(r5r_network, n_threads)
  set_verbose(r5r_network, verbose)
  set_progress(r5r_network, progress)
  set_fare_structure(r5r_network, fare_structure)
  set_max_fare(r5r_network, max_fare)
  set_output_dir(r5r_network, output_dir)
  set_suboptimal_minutes(
    r5r_network,
    suboptimal_minutes,
    fare_structure,
    shortest_path
  )

  # SCENARIOS -------------------------------------------
  set_new_congestion(r5r_network, new_carspeeds, carspeed_scale)
  set_new_lts(r5r_network, new_lts)

  # call r5r_network method and process result -------------------------------

  path_options <- r5r_network$detailedItineraries(
    origins$id,
    origins$lat,
    origins$lon,
    destinations$id,
    destinations$lat,
    destinations$lon,
    mode_list$direct_modes,
    mode_list$transit_mode,
    mode_list$access_mode,
    mode_list$egress_mode,
    departure$date,
    departure$time,
    max_walk_time,
    max_bike_time,
    max_car_time,
    max_trip_duration,
    drop_geometry,
    shortest_path,
    osm_link_ids
  )

  if (!is.null(output_dir)) return(output_dir)

  path_options <- java_to_dt(path_options)

  if (!drop_geometry) {
    if (nrow(path_options) > 0) {
      path_options[, geometry := sf::st_as_sfc(geometry)]
    } else {
      path_options[, geometry := sf::st_sfc(sf::st_linestring(), crs = 4326)[0]]
    }

    path_options <- sf::st_sf(path_options, crs = 4326)
  }

  return(path_options)
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

tester <- function(max_rides) set_max_rides(r5r_network, max_rides)

test_that("input is correct", {
  expect_error(tester("1"))
  expect_error(tester(0))
  expect_error(tester(c(1, 1)))
  expect_error(tester(Inf))
})

test_that("max_rides argument works in travel_time_matrix()", {
  expr <- "travel_time_matrix(
    r5r_network,
    origins=pois,
    destinations=pois,
    mode = c('WALK', 'TRANSIT'),
    departure_datetime = departure_datetime
  )"

  one_ride_expr <- sub("\\)$", ", max_rides = 1\\)", expr)
  many_rides_expr <- sub("\\)$", ", max_rides = 2\\)", expr)

  one_ride <- eval(parse(text = one_ride_expr))
  many_rides <- eval(parse(text = many_rides_expr))

  expect_true(nrow(many_rides) >= nrow(one_ride))
  one_ride[
    many_rides,
    on = c("from_id", "to_id"),
    travel_time_many_rides := i.travel_time_p50
  ]
  expect_true(
    all(one_ride$travel_time_p50 >= one_ride$travel_time_many_rides)
  )
  expect_true(
    any(one_ride$travel_time_p50 > one_ride$travel_time_many_rides)
  )
})

test_that("max_rides argument works in accessibility()", {
  expr <- "accessibility(
    r5r_network,
    origins=points[1:15],
    destinations=points[1:15],
    mode = c('WALK', 'TRANSIT'),
    departure_datetime = departure_datetime,
    opportunities_colnames = 'population',
    cutoffs = 60
  )"

  one_ride_expr <- sub("\\)$", ", max_rides = 1\\)", expr)
  many_rides_expr <- sub("\\)$", ", max_rides = 2\\)", expr)

  one_ride <- eval(parse(text = one_ride_expr))
  many_rides <- eval(parse(text = many_rides_expr))

  one_ride[many_rides, on = "id", accessibility_many_rides := i.accessibility]
  expect_true(all(one_ride$accessibility <= one_ride$accessibility_many_rides))
  expect_true(any(one_ride$accessibility < one_ride$accessibility_many_rides))
})

test_that("max_rides argument works in expanded_travel_time_matrix()", {
  expr <- "expanded_travel_time_matrix(
    r5r_network,
    origins=pois[1:5],
    destinations=pois[1:5],
    mode = c('WALK', 'TRANSIT'),
    departure_datetime = departure_datetime,
    draws_per_minute = 1
  )"

  one_ride_expr <- sub("\\)$", ", max_rides = 1\\)", expr)
  many_rides_expr <- sub("\\)$", ", max_rides = 2\\)", expr)

  one_ride <- eval(parse(text = one_ride_expr))
  many_rides <- eval(parse(text = many_rides_expr))

  count_transit <- function(strings) {
    split_vector <- strsplit(strings, "\\|")
    walkless_vector <- lapply(
      split_vector,
      function(s) {
        if (identical(s, "[WALK]")) {
          character(0)
        } else {
          s
        }
      }
    )
    count <- vapply(walkless_vector, length, integer(1))
  }

  one_ride[, transit_count := count_transit(routes)]
  expect_true(all(one_ride$transit_count <= 1))
  expect_true(any(one_ride$transit_count > 0))

  many_rides[, transit_count := count_transit(routes)]
  expect_true(all(many_rides$transit_count <= 2))
  expect_true(any(many_rides$transit_count > 0))
  expect_true(any(many_rides$transit_count > 1))
})

test_that("max_rides argument works in pareto_frontier()", {
  expr <- "pareto_frontier(
    r5r_network,
    origins=pois[1:5],
    destinations=pois[1:5],
    mode = c('WALK', 'TRANSIT'),
    departure_datetime = departure_datetime,
    fare_structure = fare_structure,
    fare_cutoffs  = c(0, 5, 10)
  )"

  one_ride_expr <- sub("\\)$", ", max_rides = 1\\)", expr)
  many_rides_expr <- sub("\\)$", ", max_rides = 2\\)", expr)

  one_ride <- eval(parse(text = one_ride_expr))
  many_rides <- eval(parse(text = many_rides_expr))

  expect_true(all(one_ride$monetary_cost <= 5))
  expect_true(any(one_ride$monetary_cost > 0))

  expect_true(all(many_rides$monetary_cost <= 10))
  expect_true(any(many_rides$monetary_cost > 0))
  expect_true(any(many_rides$monetary_cost > 5))
})

test_that("max_rides argument works in detailed_itineraries()", {
  expr <- "detailed_itineraries(
    r5r_network,
    origins=pois,
    destinations=pois[15:1],
    mode = c('WALK', 'TRANSIT'),
    departure_datetime = departure_datetime,
    drop_geometry = TRUE,
    shortest_path = FALSE
  )"

  one_ride_expr <- sub("\\)$", ", max_rides = 1\\)", expr)
  many_rides_expr <- sub("\\)$", ", max_rides = 2\\)", expr)

  one_ride <- eval(parse(text = one_ride_expr))
  many_rides <- eval(parse(text = many_rides_expr))

  count_transit <- function(routes) {
    non_walk_legs <- routes != ""
    sum(non_walk_legs)
  }

  one_ride <- one_ride[
    ,
    .(transit_legs = count_transit(route)),
    by = .(from_id, to_id, option)
  ]
  expect_true(all(one_ride$transit_legs <= 1))
  expect_true(any(one_ride$transit_legs > 0))

  many_rides <- many_rides[
    ,
    .(transit_legs = count_transit(route)),
    by = .(from_id, to_id, option)
  ]
  expect_true(all(many_rides$transit_legs <= 2))
  expect_true(any(many_rides$transit_legs > 0))
  expect_true(any(many_rides$transit_legs > 1))
})

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

tester <- function(time_window) set_time_window(r5r_network, time_window)

test_that("input is correct", {
  expect_error(tester("1"))
  expect_error(tester(0))
  expect_error(tester(c(1, 1)))
  expect_error(tester(Inf))
})

test_that("set_time_window argument works in expanded_travel_time_matrix()", {
  basic_expr <- call(
    "expanded_travel_time_matrix",
    r5r_network,
    origins=pois,
    destinations=pois,
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    draws_per_minute = 1,
    max_trip_duration = 60
  )

  small_window_expr <- big_window_expr <- basic_expr
  small_window_expr$time_window <- 1
  big_window_expr$time_window <- 60

  small_window <- eval(small_window_expr)
  expect_true(
    all(small_window$departure_time == format(departure_datetime, "%H:%M:%S"))
  )

  big_window <- eval(big_window_expr)
  departure_range <- departure_datetime + c(0:59) * 60
  departure_range_char <- format(departure_range, "%H:%M:%S")
  time_window_summary <- big_window[
    ,
    .(contains_all_times = identical(departure_time, departure_range_char)),
    by = .(from_id, to_id)
  ]

  # should include all departure times even if in some of them no trip can be
  # completed within the specified constraints

  incomplete_trips <- big_window[
    from_id == "gasometer_museum" & to_id == "iguatemi_shopping_center"
  ]
  expect_true(any(is.na(incomplete_trips$routes)))
  expect_identical(incomplete_trips$departure_time, departure_range_char)
})

# travel_time_matrix(), accessibility() and pareto_frontier(): we're gonna
# assume the argument is working if calculating the times with time_window = 1
# and draws_per_minute = 1 results in all percentiles with the same value, but
# larger time windows doesn't

percentiles <- c(25, 50, 75)

test_that("set_time_window argument works in travel_time_matrix()", {
  basic_expr <- call(
    "travel_time_matrix",
    r5r_network,
    origins=pois,
    destinations=pois,
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    draws_per_minute = 1,
    percentiles = percentiles
  )

  small_window_expr <- big_window_expr <- basic_expr
  small_window_expr$time_window <- 1
  big_window_expr$time_window <- 60

  small_window <- eval(small_window_expr)
  expect_identical(small_window$travel_time_p25, small_window$travel_time_p50)
  expect_identical(small_window$travel_time_p50, small_window$travel_time_p75)

  big_window <- eval(big_window_expr)
  expect_false(
    identical(big_window$travel_time_p25, big_window$travel_time_p50)
  )
  expect_false(
    identical(big_window$travel_time_p50, big_window$travel_time_p75)
  )
})

test_that("set_time_window argument works in accessibility()", {
  basic_expr <- call(
    "accessibility",
    r5r_network,
    origins=points[1:30],
    destinations=points[1:30],
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    draws_per_minute = 1,
    percentiles = percentiles,
    opportunities_colnames = "population",
    decay_function = "step",
    cutoffs = 30,
    max_trip_duration = 60
  )

  small_window_expr <- big_window_expr <- basic_expr
  small_window_expr$time_window <- 1
  big_window_expr$time_window <- 60

  small_window <- eval(small_window_expr)
  small_window <- data.table::dcast(
    small_window,
    id ~ percentile,
    value.var = "accessibility"
  )
  data.table::setnames(
    small_window,
    old = as.character(percentiles),
    new = paste0("access_", percentiles)
  )
  expect_identical(small_window$access_25, small_window$access_50)
  expect_identical(small_window$access_50, small_window$access_75)

  big_window <- eval(big_window_expr)
  big_window <- data.table::dcast(
    big_window,
    id ~ percentile,
    value.var = "accessibility"
  )
  data.table::setnames(
    big_window,
    old = as.character(percentiles),
    new = paste0("access_", percentiles)
  )
  expect_false(identical(big_window$access_25, big_window$access_50))
  expect_false(identical(big_window$access_50, big_window$access_75))
})

test_that("set_time_window argument works in pareto_frontier()", {
  basic_expr <- call(
    "pareto_frontier",
    r5r_network,
    origins=points[1:5],
    destinations=points[1:5],
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    percentiles = percentiles,
    fare_structure = fare_structure,
    fare_cutoffs = c(0, 5, 10),
    max_trip_duration = 60
  )

  small_window_expr <- big_window_expr <- basic_expr
  small_window_expr$time_window <- 1
  big_window_expr$time_window <- 60

  small_window <- eval(small_window_expr)
  small_window <- small_window[
    ,
    .(data = list(.SD)),
    by = .(from_id, to_id, percentile)
  ]
  small_window <- data.table::dcast(
    small_window,
    from_id + to_id ~ percentile,
    value.var = "data"
  )
  data.table::setnames(
    small_window,
    old = as.character(percentiles),
    new = paste0("frontier_", percentiles)
  )
  expect_identical(small_window$frontier_25, small_window$frontier_50)
  expect_identical(small_window$frontier_50, small_window$frontier_75)

  big_window <- eval(big_window_expr)
  big_window <- big_window[
    ,
    .(data = list(.SD)),
    by = .(from_id, to_id, percentile)
  ]
  big_window <- data.table::dcast(
    big_window,
    from_id + to_id ~ percentile,
    value.var = "data"
  )
  data.table::setnames(
    big_window,
    old = as.character(percentiles),
    new = paste0("frontier_", percentiles)
  )
  expect_false(identical(big_window$frontier_25, big_window$frontier_50))
  expect_false(identical(big_window$frontier_50, big_window$frontier_75))
})

# detailed_itineraries(): no draws_per_minute and percentiles arguments, so we
# can only check if using a larger time window brings more  information

test_that("set_time_window argument works in detailed_itineraries()", {
  basic_expr <- call(
    "detailed_itineraries",
    r5r_network,
    origins=points[1:5],
    destinations=points[5:1],
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    drop_geometry = TRUE,
    shortest_path = FALSE
  )

  small_window_expr <- big_window_expr <- basic_expr
  small_window_expr$time_window <- 1
  big_window_expr$time_window <- 30

  small_window <- eval(small_window_expr)
  small_window <- small_window[
    ,
    .(n_itin = max(option)),
    by = .(from_id, to_id)
  ]
  big_window <- eval(big_window_expr)
  big_window <- big_window[, .(n_itin = max(option)), by = .(from_id, to_id)]

  expect_true(all(big_window$n_itin > small_window$n_itin))
})

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

tester <- function(speed, mode) set_speed(r5r_network, speed, mode)

test_that("input is correct", {
  expect_error(tester(10, 1))
  expect_error(tester(10, c("bike", "bike")))
  expect_error(tester(10, "oie"))

  expect_error(tester("a", "walk"))
  expect_error(tester(c(1, 1), "walk"))
  expect_error(tester(0, "walk"))
  expect_error(tester(Inf, "walk"))
})

test_that("set_speed argument works in travel_time_matrix()", {
  walk_expr <- call("travel_time_matrix", r5r_network, origins=pois, destinations=pois, mode = "WALK")

  slow_walk_expr <- fast_walk_expr <- walk_expr
  slow_walk_expr$walk_speed <- 2
  fast_walk_expr$walk_speed <- 4

  slow_walk <- eval(slow_walk_expr)
  fast_walk <- eval(fast_walk_expr)

  expect_true(nrow(fast_walk) > nrow(slow_walk))
  slow_walk[
    fast_walk,
    on = c("from_id", "to_id"),
    travel_time_fast := i.travel_time_p50
  ]
  expect_true(all(slow_walk$travel_time_p50 >= slow_walk$travel_time_fast))
  expect_true(any(slow_walk$travel_time_p50 > slow_walk$travel_time_fast))

  bike_expr <- walk_expr
  bike_expr$mode <- "BICYCLE"
  slow_bike_expr <- fast_bike_expr <- bike_expr
  slow_bike_expr$bike_speed <- 5
  fast_bike_expr$bike_speed <- 10

  slow_bike <- eval(slow_bike_expr)
  fast_bike <- eval(fast_bike_expr)

  expect_true(nrow(fast_bike) > nrow(slow_bike))
  slow_bike[
    fast_bike,
    on = c("from_id", "to_id"),
    travel_time_fast := i.travel_time_p50
  ]
  expect_true(all(slow_bike$travel_time_p50 >= slow_bike$travel_time_fast))
  expect_true(any(slow_bike$travel_time_p50 > slow_bike$travel_time_fast))
})

test_that("set_speed argument works in expanded_travel_time_matrix()", {
  walk_expr <- call(
    "expanded_travel_time_matrix",
    r5r_network,
    origins=pois,
    destinations=pois,
    mode = "WALK"
  )

  slow_walk_expr <- fast_walk_expr <- walk_expr
  slow_walk_expr$walk_speed <- 2
  fast_walk_expr$walk_speed <- 4

  slow_walk <- eval(slow_walk_expr)
  fast_walk <- eval(fast_walk_expr)

  expect_true(nrow(fast_walk) > nrow(slow_walk))
  slow_walk[
    fast_walk,
    on = c("from_id", "to_id"),
    total_time_fast := i.total_time
  ]
  expect_true(all(slow_walk$total_time >= slow_walk$total_time_fast))
  expect_true(any(slow_walk$total_time > slow_walk$total_time_fast))

  bike_expr <- walk_expr
  bike_expr$mode <- "BICYCLE"
  slow_bike_expr <- fast_bike_expr <- bike_expr
  slow_bike_expr$bike_speed <- 5
  fast_bike_expr$bike_speed <- 10

  slow_bike <- eval(slow_bike_expr)
  fast_bike <- eval(fast_bike_expr)

  expect_true(nrow(fast_bike) > nrow(slow_bike))
  slow_bike[
    fast_bike,
    on = c("from_id", "to_id"),
    total_time_fast := i.total_time
  ]
  expect_true(all(slow_bike$total_time >= slow_bike$total_time_fast))
  expect_true(any(slow_bike$total_time > slow_bike$total_time_fast))
})

test_that("set_speed argument works in accessibility()", {
  walk_expr <- call(
    "accessibility",
    r5r_network,
    origins=points[1:15],
    destinations=points[1:15],
    mode = "WALK",
    opportunities_colnames = "population",
    decay_function = "step",
    cutoffs = 30
  )

  slow_walk_expr <- fast_walk_expr <- walk_expr
  slow_walk_expr$walk_speed <- 2
  fast_walk_expr$walk_speed <- 5

  slow_walk <- eval(slow_walk_expr)
  fast_walk <- eval(fast_walk_expr)

  slow_walk[fast_walk, on = "id", accessibility_fast := i.accessibility]
  expect_true(all(slow_walk$accessibility <= slow_walk$accessibility_fast))
  expect_true(any(slow_walk$accessibility < slow_walk$accessibility_fast))

  bike_expr <- walk_expr
  bike_expr$mode <- "BICYCLE"
  slow_bike_expr <- fast_bike_expr <- bike_expr
  slow_bike_expr$bike_speed <- 5
  fast_bike_expr$bike_speed <- 10

  slow_bike <- eval(slow_bike_expr)
  fast_bike <- eval(fast_bike_expr)

  slow_bike[fast_bike, on = "id", accessibility_fast := i.accessibility]
  expect_true(all(slow_bike$accessibility <= slow_bike$accessibility_fast))
  expect_true(any(slow_bike$accessibility < slow_bike$accessibility_fast))
})

test_that("set_speed argument works in pareto_frontier()", {
  walk_expr <- call(
    "pareto_frontier",
    r5r_network,
    origins=points[1:5],
    destinations=points[1:5],
    mode = "WALK",
    fare_structure = fare_structure,
    fare_cutoffs = 0
  )

  slow_walk_expr <- fast_walk_expr <- walk_expr
  slow_walk_expr$walk_speed <- 2
  fast_walk_expr$walk_speed <- 4

  slow_walk <- eval(slow_walk_expr)
  fast_walk <- eval(fast_walk_expr)

  expect_true(nrow(fast_walk) > nrow(slow_walk))
  slow_walk[
    fast_walk,
    on = c("from_id", "to_id"),
    travel_time_fast := i.travel_time
  ]
  expect_true(all(slow_walk$travel_time >= slow_walk$travel_time_fast))
  expect_true(any(slow_walk$travel_time > slow_walk$travel_time_fast))

  bike_expr <- walk_expr
  bike_expr$mode <- "BICYCLE"
  slow_bike_expr <- fast_bike_expr <- bike_expr
  slow_bike_expr$bike_speed <- 5
  fast_bike_expr$bike_speed <- 10

  slow_bike <- eval(slow_bike_expr)
  fast_bike <- eval(fast_bike_expr)

  expect_true(nrow(fast_bike) > nrow(slow_bike))
  slow_bike[
    fast_bike,
    on = c("from_id", "to_id"),
    travel_time_fast := i.travel_time
  ]
  expect_true(all(slow_bike$travel_time >= slow_bike$travel_time_fast))
  expect_true(any(slow_bike$travel_time > slow_bike$travel_time_fast))
})

# FIXME: currently, the speeds calculated from detailed_itineraries() distance
# and duration fields don't result in the same speed specified in the function
# call. some investigation of why is on issue #276
test_that("set_speed argument works in detailed_itineraries()", {
  walk_expr <- call(
    "detailed_itineraries",
    r5r_network,
    origins=pois,
    destinations=pois[15:1],
    mode = "WALK",
    drop_geometry = TRUE
  )

  slow_walk_expr <- fast_walk_expr <- walk_expr
  slow_walk_expr$walk_speed <- 2
  fast_walk_expr$walk_speed <- 4.68

  slow_walk <- eval(slow_walk_expr)
  fast_walk <- eval(fast_walk_expr)

  slow_walk[
    fast_walk,
    on = c("from_id", "to_id"),
    segment_duration_fast := i.segment_duration
  ]
  expect_true(
    all(slow_walk$segment_duration >= slow_walk$segment_duration_fast)
  )
  expect_true(any(slow_walk$segment_duration > slow_walk$segment_duration_fast))

  bike_expr <- walk_expr
  bike_expr$mode <- "BICYCLE"
  bike_expr$max_lts < 4
  slow_bike_expr <- fast_bike_expr <- bike_expr
  slow_bike_expr$bike_speed <- 5
  fast_bike_expr$bike_speed <- 10

  slow_bike <- eval(slow_bike_expr)
  fast_bike <- eval(fast_bike_expr)

  slow_bike[
    fast_bike,
    on = c("from_id", "to_id"),
    segment_duration_fast := i.segment_duration
  ]
  expect_true(
    all(slow_bike$segment_duration >= slow_bike$segment_duration_fast)
  )
  expect_true(any(slow_bike$segment_duration > slow_bike$segment_duration_fast))
})

context("Detailed itineraries function")

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

default_tester <- function(r5r_network,
                           origins = points[1:2, ],
                           destinations = points[2:1, ],
                           mode = c("WALK", "TRANSIT"),
                           departure_datetime = as.POSIXct("13-05-2019 14:00:00",
                                                           format = "%d-%m-%Y %H:%M:%S"),
                           max_walk_time = Inf,
                           max_bike_time = Inf,
                           max_trip_duration = 120L,
                           walk_speed = 3.6,
                           bike_speed = 12,
                           max_rides = 3,
                           shortest_path = TRUE,
                           n_threads = Inf,
                           verbose = FALSE,
                           drop_geometry = FALSE,
                           osm_link_ids = FALSE) {

 results <- detailed_itineraries(
   r5r_network,
   origins = origins,
   destinations = destinations,
   mode = mode,
   departure_datetime = departure_datetime,
   max_walk_time = max_walk_time,
   max_bike_time = max_bike_time,
   max_trip_duration = max_trip_duration,
   walk_speed = walk_speed,
   bike_speed = bike_speed,
   max_rides = max_rides,
   shortest_path = shortest_path,
   n_threads = n_threads,
   verbose = verbose,
   drop_geometry = drop_geometry,
   osm_link_ids = osm_link_ids
 )

 return(results)

}


# errors and warnings -----------------------------------------------------


test_that("detailed_itineraries adequately raises errors", {

  # error related to using object with wrong type as r5r_network
  expect_error(default_tester("r5r_network"))

  # error related to using wrong origins/destinations object type
  multipoint_origins      <- sf::st_cast(sf::st_as_sf(points[1:2,], coords = c("lon", "lat")), "MULTIPOINT")
  multipoint_destinations <- sf::st_cast(sf::st_as_sf(points[2:1,], coords = c("lon", "lat")), "MULTIPOINT")
  list_origins      <- list(id = c("1", "2"), lat = c(-30.02756, -30.02329), long = c(-51.22781, -51.21886))
  list_destinations <- list(id = c("2", "1"), lat = c(-30.02329, -30.02756), long = c(-51.21886, -51.22781))

  expect_error(default_tester(r5r_network, origins = multipoint_origins))
  expect_error(default_tester(r5r_network, destinations = multipoint_destinations))
  expect_error(default_tester(r5r_network, origins = list_origins))
  expect_error(default_tester(r5r_network, destinations = list_destinations))
  expect_error(default_tester(r5r_network, origins = "origins"))
  expect_error(default_tester(r5r_network, destinations = "destinations"))

  # error related to using wrong origins/destinations column types
  origins <- points[1:2, ]
  destinations <- points[2:1, ]

  origins_char_lat   <- data.frame(id = origins$id, lat = as.character(origins$lat), lon = origins$lon)
  origins_char_lon   <- data.frame(id = origins$id, lat = origins$lat, lon = as.character(origins$lon))
  destinations_char_lat   <- data.frame(id = destinations$id, lat = as.character(destinations$lat), lon = destinations$lon)
  destinations_char_lon   <- data.frame(id = destinations$id, lat = destinations$lat, lon = as.character(destinations$lon))

  expect_error(default_tester(r5r_network, origins = origins_char_lat))
  expect_error(default_tester(r5r_network, origins = origins_char_lon))
  expect_error(default_tester(r5r_network, destinations = destinations_char_lat))
  expect_error(default_tester(r5r_network, destinations = destinations_char_lon))

  # error related to 'origins' and 'destinations' with distinct number of rows
  expect_error(default_tester(r5r_network, origins = points[1:3,], destinations = points[2:1,]))
  expect_error(default_tester(r5r_network, origins = points[2:1,], destinations = points[1:3,]))

  # error related to nonexistent mode
  expect_error(default_tester(r5r_network, mode = "all"))

  # errors related to date formatting
  numeric_datetime <- as.numeric(as.POSIXct("13-05-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S"))

  expect_error(default_tester(r5r_network, departure_datetime = "13-05-2019 14:00:00"))
  expect_error(default_tester(r5r_network, numeric_datetime))

  # errors related to max_walk_time
  expect_error(default_tester(r5r_network, max_walk_time = "1"))
  expect_error(default_tester(r5r_network, max_walk_time = NULL))

  # errors related to max_bike_time
  expect_error(default_tester(r5r_network, max_bike_time = "1"))
  expect_error(default_tester(r5r_network, max_bike_time = NULL))

    # error/warning related to max_street_time
  expect_error(default_tester(r5r_network, max_trip_duration = "120"))

  # error related to non-numeric walk_speed
  expect_error(default_tester(r5r_network, walk_speed = "3.6"))

  # error related to non-numeric bike_speed
  expect_error(default_tester(r5r_network, bike_speed = "12"))

  # error related to non-logical shortest_path
  expect_error(default_tester(r5r_network, shortest_path = "TRUE"))
  expect_error(default_tester(r5r_network, shortest_path = 1))
  expect_error(default_tester(r5r_network, shortest_path = NULL))

  # error related to non-logical verbose
  expect_error(default_tester(r5r_network, verbose = "TRUE"))
  expect_error(default_tester(r5r_network, verbose = 1))
  expect_error(default_tester(r5r_network, verbose = NULL))

  # error related to non-logical drop_geometry
  expect_error(default_tester(r5r_network, drop_geometry = "TRUE"))
  expect_error(default_tester(r5r_network, drop_geometry = 1))

  # error related to non-logical osm_link_ids
  expect_error(default_tester(r5r_network, osm_link_ids = "TRUE"))
  expect_error(default_tester(r5r_network, osm_link_ids = 1))
  expect_error(default_tester(r5r_network, osm_link_ids = TRUE, drop_geometry=TRUE))

  })

test_that("detailed_itineraries adequately raises warnings and messages", {

  # message related to expanding origins/destinations dataframe
  expect_message(default_tester(r5r_network, origins = points[1, ]))
  expect_message(default_tester(r5r_network, destinations = points[1, ]))

  # error/warning related to using wrong origins/destinations column types
  origins <- points[1:2, ]
  destinations <- points[2:1, ]

  origins_numeric_id <- data.frame(id = 1:2, lat = origins$lat, lon = origins$lon)
  destinations_numeric_id <- data.frame(id = 1:2, lat = destinations$lat, lon = destinations$lon)

  expect_warning(default_tester(r5r_network, origins = origins_numeric_id))
  expect_warning(default_tester(r5r_network, destinations = destinations_numeric_id))


})


# adequate behaviour ------------------------------------------------------


test_that("detailed_itineraries output is correct", {

  #  * output class ---------------------------------------------------------


  # expect results to be of class 'sf' and 'data.table', independently of the
  # class of 'origins'/'destinations', when drop_geometry = FALSE

  origins_sf <- sf::st_as_sf(
    points[1:2,],
    coords = c("lon", "lat"),
    crs = 4326
  )
  destinations_sf <- sf::st_as_sf(
    points[2:1,],
    coords = c("lon", "lat"),
    crs = 4326
  )

  result_df_input <- default_tester(r5r_network)
  result_sf_input <- default_tester(r5r_network, origins_sf, destinations_sf)

  expect_true(is(result_df_input, "sf"))
  expect_true(is(result_df_input, "data.table"))
  expect_true(is(result_sf_input, "sf"))
  expect_true(is(result_sf_input, "data.table"))

  # expect results to be of class 'data.table', but not 'sf', independently of
  # the class of 'origins'/'destinations', when drop_geometry = TRUE

  result_df_input <- default_tester(r5r_network, drop_geometry = TRUE)
  result_sf_input <- default_tester(r5r_network, origins_sf, destinations_sf, drop_geometry = TRUE)

  expect_false(is(result_df_input, "sf"))
  expect_true(is(result_df_input, "data.table"))
  expect_false(is(result_sf_input, "sf"))
  expect_true(is(result_sf_input, "data.table"))

  # bring osm ids
  result_df_input_ids <- default_tester(r5r_network,
                                        osm_link_ids = TRUE,
                                        drop_geometry = FALSE)

  expect_true(
    all(c('edge_id_list', 'board_stop_id', 'alight_stop_id') %in% names(result_df_input_ids))
    )



  # expect each column to be of right class
  expect_true(typeof(result_df_input$from_id) == "character")
  expect_true(typeof(result_df_input$from_lat) == "double")
  expect_true(typeof(result_df_input$from_lon) == "double")
  expect_true(typeof(result_df_input$to_id) == "character")
  expect_true(typeof(result_df_input$to_lat) == "double")
  expect_true(typeof(result_df_input$to_lon) == "double")
  expect_true(typeof(result_df_input$option) == "integer")
  expect_true(typeof(result_df_input$segment) == "integer")
  expect_true(typeof(result_df_input$mode) == "character")
  expect_true(typeof(result_df_input$total_duration) == "double")
  expect_true(typeof(result_df_input$segment_duration) == "double")
  expect_true(typeof(result_df_input$wait) == "double")
  expect_true(typeof(result_df_input$distance) == "integer")
  expect_true(typeof(result_df_input$route) == "character")
  expect_true(typeof(result_df_input_ids$edge_id_list) == "character")
  expect_true(typeof(result_df_input_ids$board_stop_id) == "character")
  expect_true(typeof(result_df_input_ids$alight_stop_id) == "character")


  #  * r5r options ----------------------------------------------------------


  # expect walking segments to be shorter when setting higher walk speeds
  # ps: note that if a very high speed is set then the routes change completely
  # and we lose the ability to use lower walking speeds routes as reference

  origins      <- points[1,]
  destinations <- points[3,]

  df <- default_tester(r5r_network, origins = origins, destinations = destinations, mode = "WALK", walk_speed = 4)
  duration_lower_speed <- data.table::setDT(df)$segment_duration

  df <- default_tester(r5r_network, origins = origins, destinations = destinations, mode = "WALK", walk_speed = 5)
  duration_higher_speed <- data.table::setDT(df)$segment_duration

  expect_true(duration_higher_speed < duration_lower_speed)

  # expect bike segments to be shorter when setting higher cycling speeds
  # ps: same as with walk_speeds

  df <- default_tester(r5r_network, origins = origins, destinations = destinations,
                       mode = "BICYCLE", bike_speed = 12)
  duration_lower_speed <- data.table::setDT(df)$segment_duration

  df <- default_tester(r5r_network, origins = origins, destinations = destinations,
                       mode = "BICYCLE", bike_speed = 13)
  duration_higher_speed <- data.table::setDT(df)$segment_duration

  expect_true(duration_higher_speed < duration_lower_speed)


  #  * arguments ------------------------------------------------------------


  # expect each OD pair to have only option when shortest_path == TRUE

  df <- default_tester(r5r_network, shortest_path = TRUE)
  max_n_options <- data.table::setDT(df)[, length(unique(option)), by = .(from_id, to_id)][, max(V1)]

  expect_true(max_n_options == 1)

  # # expect each OD pair to have (possibly) more than one option when shortest_path == FALSE
  #
  # df <- default_tester(r5r_network, shortest_path = FALSE)
  # max_n_options <- data.table::setDT(df)[, length(unique(option)), by = .(from_id, to_id)][, max(V1)]
  #
  # expect_true(max_n_options > 1)

  # expect all route options to have lower total duration than max_trip_duration

  max_trip_duration <- 60L
  origins <- points[1:15,]
  destinations <- points[15:1,]

  df <- default_tester(r5r_network, origins, destinations,
                       max_trip_duration = max_trip_duration, shortest_path = FALSE)

  max_duration <- data.table::setDT(df)[, sum(segment_duration), by = .(from_id, to_id, option)][, max(V1)]

  expect_true(max_duration < max_trip_duration)

  # expect an empty data.table as output when no routes are found between the pairs

  df <- default_tester(r5r_network, origins = points[10,], destinations = points[12,],
                       mode = "WALK", max_trip_duration = 30L)

  expect_true(nrow(df) == 0)

  df <- default_tester(r5r_network, origins = points[10:11,], destinations = points[12,],
                       mode = "WALK", max_trip_duration = 30L)

  expect_true(nrow(df) == 0)

})


test_that("using transit outside the gtfs dates throws an error", {
  expect_error(
    tester(r5r_network,
           mode='transit',
           departure_datetime = as.POSIXct("13-05-2025 14:00:00",
                                           format = "%d-%m-%Y %H:%M:%S")
    )
  )
})

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

tester <- function(max_lts) set_max_lts(r5r_network, max_lts)

test_that("input is correct", {
  expect_error(tester("1"))
  expect_error(tester(c(1, 1)))
  expect_error(tester(0))
  expect_error(tester(5))
})

test_that("max_lts argument works in travel_time_matrix()", {
  expr <- "travel_time_matrix(
    r5r_network,
    origins=pois[1:5],
    destinations=pois[5:1],
    mode = \"BICYCLE\"
  )"

  fast_expr <- sub("\\)$", ", max_lts = 4\\)", expr)
  slow_expr <- sub("\\)$", ", max_lts = 1\\)", expr)

  fast_res <- eval(parse(text = fast_expr))
  slow_res <- eval(parse(text = slow_expr))

  expect_true(all(slow_res$travel_time_p50 >= fast_res$travel_time_p50))
})

test_that("max_lts argument works in expanded_travel_time_matrix()", {
  expr <- "expanded_travel_time_matrix(
    r5r_network,
    origins=pois[1:5],
    destinations=pois[1:5],
    mode = \"BICYCLE\"
  )"

  fast_expr <- sub("\\)$", ", max_lts = 4\\)", expr)
  slow_expr <- sub("\\)$", ", max_lts = 1\\)", expr)

  fast_res <- eval(parse(text = fast_expr))
  slow_res <- eval(parse(text = slow_expr))

  expect_true(all(slow_res$total_time >= fast_res$total_time))
})

test_that("max_lts argument works in pareto_frontier()", {
  expr <- "pareto_frontier(
    r5r_network,
    origins=pois[1:5],
    destinations=pois[1:5],
    mode = \"BICYCLE\",
    departure_datetime = departure_datetime,
    fare_structure = fare_structure,
    fare_cutoffs = 0
  )"

  fast_expr <- sub("\\)$", ", max_lts = 4\\)", expr)
  slow_expr <- sub("\\)$", ", max_lts = 1\\)", expr)

  fast_res <- eval(parse(text = fast_expr))
  slow_res <- eval(parse(text = slow_expr))

  expect_true(all(slow_res$travel_time >= fast_res$travel_time))
})

test_that("max_lts argument works in detailed_itineraries()", {
  expr <- "detailed_itineraries(
    r5r_network,
    origins=pois[1:5],
    destinations=pois[5:1],
    mode = 'BICYCLE'
  )"

  fast_expr <- sub("\\)$", ", max_lts = 4\\)", expr)
  slow_expr <- sub("\\)$", ", max_lts = 1\\)", expr)

  fast_res <- eval(parse(text = fast_expr))
  slow_res <- eval(parse(text = slow_expr))

  expect_true(all(slow_res$segment_duration >= fast_res$segment_duration))
})

test_that("max_lts argument works in accessibility()", {
  expr <- "accessibility(
    r5r_network,
    origins=points[1:15],
    destinations=points[1:15],
    mode = 'BICYCLE',
    opportunities_colnames = 'population',
    decay_function = 'step',
    cutoffs = 30
  )"

  fast_expr <- sub("\\)$", ", max_lts = 4\\)", expr)
  slow_expr <- sub("\\)$", ", max_lts = 1\\)", expr)

  fast_res <- eval(parse(text = fast_expr))
  slow_res <- eval(parse(text = slow_expr))

  expect_true(all(slow_res$accessibility <= fast_res$accessibility))
})

options(java.parameters = "-Xm10G")

devtools::load_all(".")
library(bench)

path <- system.file("extdata/poa", package = "r5r")
r5r_network <- setup_r5(data_path = path, verbose = FALSE)

# 2) load origin/destination points and set arguments
points <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))
points <- rbind(points, points, points)
mode <- c("WALK", "TRANSIT")
max_walk_time <- 30   # minutes
max_trip_duration <- 60 # minutes
departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
                                 format = "%d-%m-%Y %H:%M:%S")


bench::mark(iterations = 3, check = FALSE,
            ttm <- travel_time_matrix(r5r_network = r5r_network,
                                      origins = points,
                                      destinations = points,
                                      mode = mode,
                                      departure_datetime = departure_datetime,
                                      max_walk_time = max_walk_time,
                                      max_trip_duration = max_trip_duration,
                                      progress = T)
)


# expression    min median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time result
# java_to_dt    29s  30.6s    0.0332     229MB   0.0774     3     7      1.51m <NULL>
#  arrow df   24.3s  25.2s    0.0389     570MB   0.0519     3     4      1.28m <NULL>
#arrow arrow  25.3s  26.7s    0.0363     569MB   0.0485     3     4      1.38m <NULL>



# # Define the dataset
# DS <- arrow::open_csv_dataset(sources = csv_files)
#
# # Create a scanner
# O <- Scanner$create(DS)
#
# # Load it as Arrow Table in memory
# AT <- SO$ToTable()
#
#
# # prep land use data
# jobs_df <- select(points, c('id', 'jobs', 'healthcare'))
#
# # merge jobs
# AT <- left_join(AT, jobs_df, by = c('to_id' = 'id'))
#
# head(AT) |> collect()
#
# # calculate cumulative access in less than 20 min.
# access_df <- AT |>
#               filter(travel_time_p50 <= 20) |>
#               group_by(from_id) |>
#               summarise(access = sum(jobs)) |>
#               collect()
#
# head(access_df)
#
#
# # calculate access to closest healthcare facility
# access_df2 <- AT |>
#   filter(healthcare > 0) |>
#   group_by(from_id) |>
#   summarise(access = min(jobs)) |>
#   collect()
#
# head(access_df2)
#

library(r5r)
library(dplyr)
library(data.table)
library(ggplot2)
library(dodgr)

RcppParallel::setThreadOptions (numThreads = 4L) # or desired number

# data path where the .pbf file is located
data_path <- system.file("extdata/poa", package = "r5r")
pbf_path <- file.path(data_path,'/poa_osm.pbf')

# read layer of lines from pbf
edges <- sf::st_read(
  pbf_path,
  layer = 'lines',
  quiet = TRUE
)

nodes <- sf::st_read(
  pbf_path,
  layer = 'points',
  quiet = TRUE
)

edges$oneway <- ifelse(
  edges$other_tags %like% "\"oneway\"=>\"yes\"",
  "yes",
  NA)

# keep only osm ids
edges <- edges[,c("osm_id","highway", "oneway")]
nodes <- nodes[,c("osm_id","highway")]


# build car network
car_network <- dodgr::weight_streetnet(
  x = edges,
  id_col = 'osm_id',
  wt_profile = "motorcar"
  )

# convert node ids to numeric
data.table::setDT(car_network)
car_network[, from_id := as.numeric(from_id)]
car_network[, to_id := as.numeric(to_id)]

# unique id, mas a edge_id ja eh uma unique id de cada edge
car_network[,
            unique_id := paste(way_id, geom_num, edge_id, from_id, to_id, sep = "_"),
            by = .I
            ]

head(car_network)


# bring info about one-way or two-way
data.table::setDT(edges)
car_network[edges, on = .(way_id=osm_id), oneway := i.oneway]
car_network$oneway

# 26786730 one way
# 27126547 two ways
car_network2 <- car_network[way_id %in% c(26786730, 27126547), ]
car_network2 <- car_network

# edges_sf <- sf::st_sf(edges)
# mapview::mapview(
#   filter(edges_sf, osm_id==27126547)
# )

one_way_net <- car_network2 |>
  filter(oneway=="yes") |>
  rename(osm_id = way_id) |>
  group_by(osm_id, oneway)  |>
  summarise(from_id = min(from_id),
            from_lat = from_lat[which.min(from_id)],
            from_lon = from_lon[which.min(from_id)],
            to_id =  max(to_id),
            to_lat = from_lat[which.max(to_id)],
            to_lon = from_lon[which.max(to_id)]
            ) |>
  ungroup()


two_way_net_forwards <- car_network2 |>
  filter(is.na(oneway)) |>
  rename(osm_id = way_id) |>
  group_by(osm_id)  |>
  summarise(oneway = "forwards",
            from_id = min(from_id),
            from_lat = from_lat[which.min(from_id)],
            from_lon = from_lon[which.min(from_id)],
            to_id =  max(to_id),
            to_lat = from_lat[which.max(to_id)],
            to_lon = from_lon[which.max(to_id)]
  ) |>
  ungroup()



two_way_net_backwards <- car_network2 |>
  filter(is.na(oneway)) |>
  rename(osm_id = way_id) |>
  group_by(osm_id)  |>
  summarise(oneway = "backwards",
            from_id = max(from_id),
            from_lat = from_lat[which.max(from_id)],
            from_lon = from_lon[which.max(from_id)],
            to_id =  min(to_id),
            to_lat = from_lat[which.min(to_id)],
            to_lon = from_lon[which.min(to_id)]
  ) |>
  ungroup()




my_net <- rbind(one_way_net, two_way_net_forwards, two_way_net_backwards)
my_net

nrow(car_network2)
nrow(my_net)

# another alternative would be to use sfnetworks to remove simples edges wi

#' Extract OpenStreetMap network in sf format
#'
#' Extracts the OpenStreetMap network in `sf` format from a routable transport
#' network built with [build_network()]).
#'
#' @template r5r_network
#' @template r5r_core
#'
#' @return A list with two components of a street network in sf format: vertices
#'         (POINT) and edges (LINESTRING).
#'
#' @family network functions
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' # build transport network
#' path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(path)
#'
#' # extract street network from r5r_network
#' street_net <- street_network_to_sf(r5r_network)
#'
#' stop_r5(r5r_network)
#' @export
street_network_to_sf <- function(r5r_network,
                                 r5r_core = deprecated()) {

  # deprecating r5r_core --------------------------------------
  if (lifecycle::is_present(r5r_core)) {

    cli::cli_warn(c(
      "!" = "The `r5r_core` argument is deprecated as of r5r v2.3.0.",
      "i" = "Please use the `r5r_network` argument instead."
    ))

    r5r_network <- r5r_core
  }

  # check input
  checkmate::assert_class(r5r_network, "r5r_network")

  # Get street network from R5R network
  network <- r5r_network@jcore$getStreetNetwork()

  # Convert vertices to SF (point)
  vertices_df <- java_to_dt(network$get(0L))
  vertices_sf <- sfheaders::sf_point(vertices_df, x='lon', y='lat', keep = TRUE)
  sf::st_crs(vertices_sf) <- 4326 # WGS 84

  # Convert edges to SF (linestring)
  edges_df <- java_to_dt(network$get(1L))
  edges_df[, geometry := sf::st_as_sfc(geometry)]
  edges_sf <- sf::st_sf(edges_df, crs = 4326) # WGS 84

  # gather in a list
  street_network <- list(vertices = vertices_sf, edges = edges_sf)

  return(street_network)
}

context("street_network_to_sf")

testthat::skip_on_cran()


# expected behavior
test_that("street_network_to_sf - expected behavior", {

  testthat::expect_type( street_network_to_sf(r5r_network), 'list')

})



# Expected errors
test_that("street_network_to_sf - expected errors", {

  # invalid input
  testthat::expect_error( street_network_to_sf('a') )


  # warning due to deprecated argument
  testthat::expect_warning( street_network_to_sf(r5r_core=r5r_network))

  })

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

# TODO
# test if setting max_speed to 0 closes the road

# data.frame with new speed info
edge_speeds <- read.csv(file.path(data_path, "poa_osm_congestion.csv"))

# sf with congestion polygons
congestion_poly <- readRDS(file.path(data_path, "poa_poly_congestion.rds"))
congestion_poly$scale <- c(0.2, 0.5)

# get origin and destination points in a single road
network <- r5r::street_network_to_sf(r5r_network)

point_orig <- network$vertices |>
  dplyr::filter(index == 18050 ) |> # 7772
  sfheaders::sf_to_df() |>
  dplyr::select(id=sfg_id, lon=x, lat=y)

point_dest <- network$vertices |>
  dplyr::filter(index == 16844 )  |> #8128
  sfheaders::sf_to_df() |>
  dplyr::select(id=sfg_id, lon=x, lat=y)

point_orig$id <- as.character(point_orig$id)
point_dest$id <- as.character(point_dest$id)

meta_fun <- function(
    fun = r5r::travel_time_matrix,
    new_carspeeds= NULL,
    carspeed_scale = 1,
    point_origg = point_orig,
    point_destt = point_dest){

      fun(
        r5r_network = r5r_network,
        origins = point_origg,
        destinations = point_destt,
        mode = 'car',
        departure_datetime = Sys.time(),
        max_trip_duration = 60,
        new_carspeeds = new_carspeeds,
        carspeed_scale = carspeed_scale
      )
  }



# car speeds with osm ids -------------------------------------------------------------------
test_that("success in increasing travel times with osm ids", {

  # calculate travel times / access *before* changing road speeds
  ttm_pre <- meta_fun(r5r::travel_time_matrix)
  expanded_ttm_pre <- meta_fun(r5r::expanded_travel_time_matrix)
  det_pre <- meta_fun(r5r::detailed_itineraries)
  arrival_ttm_pre <- r5r::arrival_travel_time_matrix(
    r5r_network = r5r_network,
    origins = point_orig,
    destinations = point_dest,
    mode = 'car',
    arrival_datetime = Sys.time(),
    max_trip_duration = 60
    )
  # to do: r5r::accessibility
  # plot(det_pre['total_duration'])
  # mapview::mapview(network$edges) + network$vertices + det_pre
  # mapview::mapview(network$edges) + det_pre

  # changing CARSPEED_SCALE without changing new_carspeeds
  ttm_pos <- meta_fun(r5r::travel_time_matrix, carspeed_scale = 0.1)
  expanded_ttm_pos <- meta_fun(r5r::expanded_travel_time_matrix, carspeed_scale = 0.1)
  det_pos <- meta_fun(r5r::detailed_itineraries, carspeed_scale = 0.5)
  arrival_ttm_pos <- r5r::arrival_travel_time_matrix(
    r5r_network = r5r_network,
    origins = point_orig,
    destinations = point_dest,
    mode = 'car',
    arrival_datetime = Sys.time(),
    max_trip_duration = 60,
    carspeed_scale = 0.1
  )


  #  mapview::mapview(det_pre) + det_pos

  # checking for longer travel times
  testthat::expect_true(ttm_pos$travel_time_p50 > ttm_pre$travel_time_p50)
  testthat::expect_true(all(expanded_ttm_pos$total_time > expanded_ttm_pre$total_time))
  testthat::expect_true(arrival_ttm_pos$total_time > arrival_ttm_pre$total_time)
  testthat::expect_true(det_pos$total_duration > det_pre$total_duration)
  # testthat::expect_true(det_pos$total_distance == det_pre$total_distance)

  # setting NEW_CARSPEEDS without changing carspeed_scale
  fast_carspeeds <- data.frame(osm_id = c(450002312, 390862071), max_speed = 1.5, speed_type = "scale")
  ttm_3 <- meta_fun(r5r::travel_time_matrix, new_carspeeds = fast_carspeeds)
  testthat::expect_true(ttm_3$travel_time_p50 < ttm_pre$travel_time_p50)

  # test closing road with speed = 0"
  closed_road <- data.frame(osm_id = 450002312, max_speed = 0, speed_type = "scale")
  det_closed <- meta_fun(r5r::detailed_itineraries, new_carspeeds = closed_road)
  testthat::expect_true(det_closed$total_duration > det_pre$total_duration)
  # mapview::mapview(det_closed) + det_pre

})



# car speeds with polygons -------------------------------------------------------------------
test_that("success in increasing travel times with polygons", {

  # calculate travel times / access *before* changing road speeds
  ttm_pre <- meta_fun(r5r::travel_time_matrix,
                      point_origg = pois[1], point_destt = pois[12])
  expanded_ttm_pre <- meta_fun(r5r::expanded_travel_time_matrix,
                               point_origg = pois[1], point_destt = pois[12])
  det_pre <- meta_fun(r5r::detailed_itineraries,
                      point_origg = pois[1], point_destt = pois[12])
  arrival_ttm_pre <- r5r::arrival_travel_time_matrix(
    r5r_network = r5r_network,
    origins = pois[1],
    destinations = pois[12],
    mode = 'car',
    arrival_datetime = Sys.time(),
    max_trip_duration = 60
  )
  # to do: r5r::accessibility
  # plot(det_pre['total_duration'])
  # mapview::mapview(network$edges) + network$vertices + det

  # changing CARSPEED_SCALE without changing new_carspeeds
  ttm_pos <- meta_fun(r5r::travel_time_matrix, new_carspeeds = congestion_poly,
                      point_origg = pois[1], point_destt = pois[12])
  expanded_ttm_pos <- meta_fun(r5r::expanded_travel_time_matrix, new_carspeeds = congestion_poly,
                               point_origg = pois[1], point_destt = pois[12])
  det_pos <- meta_fun(r5r::detailed_itineraries, new_carspeeds = congestion_poly,
                      point_origg = pois[1], point_destt = pois[12])
  arrival_ttm_pos <- r5r::arrival_travel_time_matrix(
    r5r_network = r5r_network,
    origins = pois[1],
    destinations = pois[12],
    mode = 'car',
    arrival_datetime = Sys.time(),
    max_trip_duration = 60,
    new_carspeeds = congestion_poly
  )


  #  mapview::mapview(det_pre) + det_pos

  # checking for longer travel times
  testthat::expect_true(ttm_pos$travel_time_p50 > ttm_pre$travel_time_p50)
  testthat::expect_true(all(expanded_ttm_pos$total_time > expanded_ttm_pre$total_time))
  testthat::expect_true(arrival_ttm_pos$total_time > arrival_ttm_pre$total_time)
  testthat::expect_true(det_pos$total_duration > det_pre$total_duration)
  # testthat::expect_true(det_pos$total_distance == det_pre$total_distance)

  # setting carspeed_scale without changing NEW_CARSPEEDS
  ttm_3 <- meta_fun(r5r::travel_time_matrix, carspeed_scale =  1.3)
  testthat::expect_true(ttm_3$travel_time_p50 < ttm_pre$travel_time_p50)
})

# errors in congestion polygon -------------------------------------------------------------------
test_that("errors in congestion polygon", {

  # wrong col names
  wrong_congestion_poly1 <- congestion_poly
  names(wrong_congestion_poly1) <- c("poly_id", "speed", "priority", "geometry")
  testthat::expect_error(meta_fun(new_carspeeds = wrong_congestion_poly1))

  # missing col
  wrong_congestion_poly2 <- congestion_poly
  wrong_congestion_poly2$poly_id <- NULL
  testthat::expect_error(meta_fun(new_carspeeds = wrong_congestion_poly2))

  # # Wrong geometry type
  # wrong_congestion_poly3 <- congestion_poly
  # wrong_congestion_poly3 <- sf::st_cast(wrong_congestion_poly3, to = 'MULTIPOINT')
  # testthat::expect_error(meta_fun(new_carspeeds = wrong_congestion_poly3))

  # Wrong projection
  wrong_congestion_poly4 <- sf::st_transform(congestion_poly, 3857)
  testthat::expect_error(meta_fun(new_carspeeds = wrong_congestion_poly4))

  })


test_that("errors due to incorrect input types", {

  testthat::expect_error(meta_fun(new_carspeeds = 'banana'))
  testthat::expect_error(meta_fun(carspeed_scale = 'banana'))

})


test_that("errors error in the new_carspeeds column names", {

  mock_data <- data.frame(osm_id = '27184648', max_speed = 10, speed_type="banana")
  testthat::expect_error(meta_fun(new_carspeeds =  mock_data))

  mock_data <- data.frame(my_osm_id = '9999', max_speed = 9999, speed_type="km/h")
  testthat::expect_error(meta_fun(new_carspeeds =  mock_data))

  testthat::expect_error(meta_fun(carspeed_scale = Inf))
  testthat::expect_error(meta_fun(carspeed_scale = -1))

})


test_that("message for missing OSM ids", {

  mock_data <- data.frame(osm_id = 45698769, max_speed = 100, speed_type="km/h")
  log_file <- file.path(r5r_network@jcore$getLogPath())
  expect_true(file.exists(log_file), info = paste("Log file not found at", log_file))

  meta_fun(new_carspeeds = mock_data)
  log <- readLines(log_file, warn = FALSE)
  expect_true(
    any(grepl("45698769", log, fixed = TRUE)),
    info = "Did not find warning for a bad OSM Id in log"
  )
})

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

# data.frame with new LTS info
network <- r5r::street_network_to_sf(r5r_network)
edge_lts <- data.frame(
  osm_id = network$edges$osm_id,
  lts = 1L
  )

# sf with new LTS
lts_lines <- readRDS(file.path(data_path, "poa_ls_lts.rds"))


meta_fun <- function(
    fun = r5r::travel_time_matrix,
    new_lts = NULL
    ){

  fun(
    r5r_network = r5r_network,
    origins = pois[1,],
    destinations = pois[13,],
    mode = 'BICYCLE',
    departure_datetime = Sys.time(),
    max_trip_duration = 60,
    new_lts = new_lts
  )
}



# LTS with osm ids -------------------------------------------------------------------
test_that("success in increasing travel times with osm ids", {

  # calculate travel times / access *before* changing road speeds
  ttm_pre <- meta_fun(r5r::travel_time_matrix)
  expanded_ttm_pre <- meta_fun(r5r::expanded_travel_time_matrix)
  det_pre <- meta_fun(r5r::detailed_itineraries)
  arrival_ttm_pre <- r5r::arrival_travel_time_matrix(
    r5r_network = r5r_network,
    origins = pois[1,],
    destinations = pois[13,],
    mode = 'car',
    arrival_datetime = Sys.time(),
    max_trip_duration = 60
  )
  # to do: r5r::accessibility

  # plot(det_pre['total_duration'])
  # mapview(network$edges) + network$vertices + det


  # calculate travel times / access *before* changing road speeds
  ttm_pos <- meta_fun(r5r::travel_time_matrix, new_lts = edge_lts)
  expanded_ttm_pos <- meta_fun(r5r::expanded_travel_time_matrix, new_lts = edge_lts)
  det_pos <- meta_fun(r5r::detailed_itineraries, new_lts = edge_lts)
  arrival_ttm_pos <- r5r::arrival_travel_time_matrix(
    r5r_network = r5r_network,
    origins = pois[1,],
    destinations = pois[13,],
    mode = 'BICYCLE',
    arrival_datetime = Sys.time(),
    max_trip_duration = 60,
    new_lts = edge_lts
  )


  #  mapview::mapview(det_pre) + det_pos

  # checking for longer travel times
  testthat::expect_true(ttm_pos$travel_time_p50 < ttm_pre$travel_time_p50)
  testthat::expect_true(all(expanded_ttm_pos$total_time < expanded_ttm_pre$total_time))
  # testthat::expect_true(arrival_ttm_pos$total_time < arrival_ttm_pre$total_time)
  testthat::expect_true(det_pos$total_duration < det_pre$total_duration)
  testthat::expect_true(det_pos$total_distance < det_pre$total_distance)

})


# LTS with sf linestrings -------------------------------------------------------------------
test_that("success in increasing travel times with sf linestrings", {

  # calculate travel times / access *before* changing road speeds
  ttm_pre <- meta_fun(r5r::travel_time_matrix)
  expanded_ttm_pre <- meta_fun(r5r::expanded_travel_time_matrix)
  det_pre <- meta_fun(r5r::detailed_itineraries)
  arrival_ttm_pre <- r5r::arrival_travel_time_matrix(
    r5r_network = r5r_network,
    origins = pois[1,],
    destinations = pois[13,],
    mode = 'car',
    arrival_datetime = Sys.time(),
    max_trip_duration = 60
  )
  # to do: r5r::accessibility

  # plot(det_pre['total_duration'])
  # mapview(network$edges) + network$vertices + det


  # calculate travel times / access *before* changing road speeds
  ttm_pos <- meta_fun(r5r::travel_time_matrix, new_lts = lts_lines)
  expanded_ttm_pos <- meta_fun(r5r::expanded_travel_time_matrix, new_lts = lts_lines)
  det_pos <- meta_fun(r5r::detailed_itineraries, new_lts = lts_lines)
  arrival_ttm_pos <- r5r::arrival_travel_time_matrix(
    r5r_network = r5r_network,
    origins = pois[1,],
    destinations = pois[13,],
    mode = 'BICYCLE',
    arrival_datetime = Sys.time(),
    max_trip_duration = 60,
    new_lts = lts_lines
  )


  #  mapview::mapview(det_pre) + det_pos

  # checking for longer travel times
  testthat::expect_true(ttm_pos$travel_time_p50 < ttm_pre$travel_time_p50)
  testthat::expect_true(all(expanded_ttm_pos$total_time < expanded_ttm_pre$total_time))
  # testthat::expect_true(arrival_ttm_pos$total_time < arrival_ttm_pre$total_time)
  testthat::expect_true(det_pos$total_duration < det_pre$total_duration)
  testthat::expect_true(det_pos$total_distance < det_pre$total_distance)

})

# LTS with spatial sf -------------------------------------------------------------------

test_that("errors in lts sf", {

  # wrong col names
  wrong_lts_lines1 <- lts_lines
  names(wrong_lts_lines1) <- c("my_line_id", "lts", "priority", "geometry")
  testthat::expect_error(meta_fun(new_lts = wrong_lts_lines1))

  # missing col
  wrong_lts_lines2 <- lts_lines
  wrong_lts_lines2$line_id <- NULL
  testthat::expect_error(meta_fun(new_lts = wrong_lts_lines2))

  # # Wrong geometry type
  # wrong_lts_lines3 <- lts_lines
  # wrong_lts_lines3 <- sf::st_cast(wrong_lts_lines3, to = 'MULTIPOINT')
  # testthat::expect_error(meta_fun(edge_lts = wrong_lts_lines3))

  # Wrong projection
  wrong_lts_lines4 <- sf::st_transform(lts_lines, 3857)
  testthat::expect_error(meta_fun(new_lts = wrong_lts_lines4))

})


test_that("errors due to incorrect input types", {

  testthat::expect_error(meta_fun(new_lts = 'banana'))

})





# test_that("message for missing OSM ids", {
#
#   mock_data <- data.frame(
#     osm_id = 123,
#     lts = 1L
#   )
#
#   testthat::expect_message(
#     meta_fun(new_lts = mock_data),
#     regexp = "Cannot find the following OSM IDs in network"
#   )
# })

#' @section Decay functions:
#'
#' `R5` allows one to use different decay functions when calculating
#' accessibility. Please see the original `R5` documentation from Conveyal for
#' more information on each one one
#' (<https://docs.conveyal.com/learn-more/decay-functions>). A summary of each
#' available option, as well as the value passed to `decay_function` to use it
#' (inside parentheses) are listed below:
#'
#' - **Step**, also known as cumulative opportunities (`"step"`): \cr
#' a binary decay function used to find the sum of available opportunities
#' within a specific travel time cutoff.
#'
#' - **Logistic CDF** (`"logistic"`): \cr
#' This is the logistic function, i.e. the cumulative distribution function of
#' the logistic distribution, expressed such that its parameters are the median
#' (inflection point) and standard deviation. This function applies a sigmoid
#' rolloff that has a convenient relationship to discrete choice theory. Its
#' parameters can be set to reflect a whole population's tolerance for making
#' trips with different travel times. The function's value represents the
#' probability that a randomly chosen member of the population would accept
#' making a trip, given its duration. Opportunities are then weighted by how
#' likely it is that a person would consider them "reachable".
#'   - Calibration: The median parameter is controlled by the `cutoff`
#'   parameter, leaving only the standard deviation to configure through the
#'   `decay_value` parameter.
#'
#' - **Fixed Exponential** (`"fixed_exponential"`): \cr
#' This function is of the form `exp(-Lt)` where L is a single fixed decay
#' constant in the range (0, 1). It is constrained to be positive to ensure
#' weights decrease (rather than grow) with increasing travel time.
#'   - Calibration: This function is controlled exclusively by the `L` constant,
#'   given by the `decay_value` parameter. Values provided in `cutoffs` are
#'   ignored.
#'
#' - **Half-life Exponential Decay** (`"exponential"`): \cr
#' This is similar to the fixed-exponential option above, but in this case the
#' decay parameter is inferred from the `cutoffs` parameter values, which is
#' treated as the half-life of the decay.
#'
#' - **Linear** (`"linear"`): \cr
#' This is a simple, vaguely sigmoid option, which may be useful when you have
#' a sense of a maximum travel time that would be tolerated by any traveler,
#' and a minimum time below which all travel is perceived to be equally easy.
#'   - Calibration: The transition region is transposable and symmetric around
#'   the `cutoffs` parameter values, taking `decay_value` minutes to taper down
#'   from one to zero.

#' Stop running r5r network
#'
#' @description Stops running r5r network
#'
#' @param ... \code{r5r_network} objects currently running. By default, if no
#'            r5r network is supplied all running networks are stopped.
#' @return No return value, called for side effects.
#'
#' @family support functions
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' path <- system.file("extdata/poa", package = "r5r")
#'
#' r5r_network <- build_network(path)
#'
#' stop_r5(r5r_network)
#' @export
stop_r5 <- function(...) {

  supplied_cores <- list(...)

  # find all running r5r networks in the parent frame

  current_objects <- mget(ls(envir = parent.frame()), envir = parent.frame())

  classes_list <- lapply(current_objects, class)

  running_cores <- current_objects[which(classes_list == "r5r_network")]

  # if no networks have been supplied, remove all running networks
  # else, remove matches between running and supplied networks

  if (length(supplied_cores) == 0) {

    rm(list = names(running_cores), envir = parent.frame())

    message("All r5r networks have been successfully stopped.")

  } else {

    matches_supplied <- running_cores[running_cores %in% supplied_cores]

    # if a match has been found, stop it
    # else, a non r5r core object has been supplied, which raises a warning

    if (length(matches_supplied) >= 1) {

      rm(list = names(matches_supplied), envir = parent.frame())

      message(paste0(paste(names(matches_supplied), collapse = ", "),
                     " has been successfully stopped."))

    } else {

      unknown_object <- current_objects[current_objects %in% supplied_cores]

      warning(paste0(paste(names(unknown_object), collapse = ", "),
                     " is not a r5r core object."))

    }

  }

  rJava::.jgc()

}

context("Stop r5r network")

testthat::skip_on_cran()

test_that("stop_r5 raises warnings and errors when supplied wrong objects", {
  path <- system.file("extdata/poa", package = "r5r")
  old_env <- environment()
  r5r_network <- build_network(data_path = path, verbose=FALSE, temp_dir = TRUE)

  # tries to stop a non-r5r network object
  expect_warning(stop_r5(path))

  # tries to stop a nonexistent object
  expect_error(stop_r5(nonexistent_object))

  # stops the running r5r network object
  expect_message(stop_r5())
  expect_identical(environment(), old_env)
})

test_that("stop_r5 successfully stops multiple running r5r networks", {
  path <- system.file("extdata/poa", package = "r5r")
  old_env <- environment()

  # stops all running r5r networks

  r5r_network_1 <- build_network(data_path = path, verbose=FALSE, temp_dir = TRUE)
  mid_env <- environment()
  r5r_network_2 <- build_network(data_path = path, verbose=FALSE, temp_dir = TRUE)

  expect_message(stop_r5())
  expect_identical(environment(), old_env)

  # stops each network separately

  r5r_network_1 <- build_network(data_path = path, verbose=FALSE, temp_dir = TRUE)
  r5r_network_2 <- build_network(data_path = path, verbose=FALSE, temp_dir = TRUE)

  expect_message(stop_r5(r5r_network_2))
  expect_identical(environment(), mid_env)
  expect_message(stop_r5(r5r_network_1))
  expect_identical(environment(), old_env)
})

context("r5r_cache")

# skip tests because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()

try(silent = TRUE, r5r::stop_r5())

# Reading the data -----------------------

test_that("r5r_cache", {

  # simply list files
  testthat::expect_message( r5r::r5r_cache() )

  ## delete existing

  # download
  r5r::download_r5(force_update = FALSE)

  # cache dir
  cache_d <- paste0('r5r/r5_jar_v', r5r_env$r5_jar_version)
  cache_dir <- tools::R_user_dir(cache_d, which = 'cache')

  # list cached files
  fname_full <- list.files(cache_dir, full.names = TRUE)
  fname <- basename(fname_full)

  testthat::expect_true( file.exists(fname_full) )
  testthat::expect_message( r5r::r5r_cache(delete_file = fname) )
  # testthat::expect_false( file.exists(fname_full) )

  ## delete ALL
  # download
  r5r::download_r5(force_update = FALSE)

  testthat::expect_true( file.exists(fname_full) )
  testthat::expect_message( r5r::r5r_cache(delete_file = 'all') )
  # testthat::expect_true( length(list.files(cache_dir)) == 0 )

  # if file does not exist, simply print message
  testthat::expect_message( r5r::r5r_cache(delete_file ='aaa') )

 })


# ERRORS and messages  -----------------------
test_that("r5r_cache", {

  testthat::expect_error(r5r_cache(list_files= 999))
  testthat::expect_error(r5r_cache(delete_file = 999))
  })


# clean cache
r5r_cache(delete_file = 'all')

# options(java.parameters = '-Xmx2G')

# library(r5r)
devtools::load_all(".")
library(data.table)
library(tidyverse)
library(sf)
library(mapview)

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path = data_path, verbose = FALSE,
                     overwrite = FALSE,
                     temp_dir = FALSE,
                     elevation = "none")



# load points
poi <- read.csv(system.file("extdata/poa/poa_points_of_interest.csv", package = "r5r"))

# load fare structure
fare_structure_path <- system.file("extdata/poa/fares/fares_poa.zip", package = "r5r")
fare_structure <- read_fare_structure(fare_structure_path)

calculate_and_plot_frontiers <- function() {
  frontiers_df <- pareto_frontier(r5r_core,
                                  origins = poi[10, ],
                                  destinations = poi[c(9, 12), ],
                                  departure_datetime = as.POSIXct("13-05-2019 14:00:00",
                                                                  format = "%d-%m-%Y %H:%M:%S"),
                                  mode = c("WALK", "TRANSIT"),
                                  max_trip_duration = 180,
                                  fare_structure = fare_structure,
                                  monetary_cost_cutoffs = c(1, 4.5, 4.8, 7.20, 8.37, 11.4, 12.57),
                                  max_rides = 5,
                                  progress = TRUE
  )

  # recode modes
  frontiers_df[, monetary_cost := round(monetary_cost, 2)]
  frontiers_df[, modes := fcase(monetary_cost == 1,   'Walk (free)',
                                monetary_cost == 4.5, 'Train (4.50)',
                                monetary_cost == 4.8, 'Bus (4.80)',
                                monetary_cost == 7.2, 'Bus + Bus (7.20)',
                                monetary_cost == 8.37, 'Bus + Train (8.37)')]

  p <- ggplot(data=frontiers_df, aes(x=monetary_cost, y=travel_time, color=to_id,
                                label = modes)) +
    geom_step(linetype = "dashed") +
    geom_point() +
    geom_text(color='gray30', hjust = -.2, nudge_x = 0.05, angle = 45) +
    labs(title='Pareto frontier of alternative routes from Farrapos station to:',
         subtitle = 'Praia de Belas shopping mall and Moinhos hospital',
         color='Destination') +
    scale_x_continuous(name="Travel cost ($)", breaks=c(0, 2, 4, 4.8, 6, 7.2, 8.37, 10)) +
    scale_y_continuous(name="Travel time (minutes)", breaks=seq(0,160,20)) +
    coord_cartesian(xlim = c(0,10), ylim = c(0, 160)) +
    theme_classic() + theme(legend.position=c(.8, 0.9))

  return(p)
}

# fare_structure$transfer_time_allowance <- 60
# fare_structure$fare_cap <- 8
# fare_structure$debug_settings$output_file <- here::here("debug.csv")
# calculate_and_plot_frontiers()

# r5r_core$startServer("")


# debug_df <- read.csv(here::here("debug.csv"))
# r5r_core$getFareStructure() %>% clipr::write_clip()

orig <- poi %>% sample_n(size = nrow(poi))
dest <- poi %>% sample_n(size = nrow(poi))

system.time(
  frontiers_df <- pareto_itineraries(r5r_core,
                                     # origins = poi[10, ],
                                     # destinations = poi[12, ],
                                     # origins = poi[c(10, 10), ],
                                     # destinations = poi[c(9, 12), ],
                                     origins = orig,
                                     destinations = dest,
                                     departure_datetime = as.POSIXct("13-05-2019 14:00:00",
                                                                     format = "%d-%m-%Y %H:%M:%S"),
                                     time_window = 30,
                                     mode = c("WALK", "TRANSIT"),
                                     max_trip_duration = 90,
                                     max_walk_dist = 1000,
                                     fare_structure = fare_structure,
                                     max_fare = 10.0,
                                     max_rides = 5,
                                     progress = TRUE)
)

frontiers_df$geometry <- st_as_sfc(frontiers_df$geometry)
frontiers_df <- st_as_sf(frontiers_df, crs = 4326)


# df <- frontiers_df
# trip = 2
view_results <- function(df, trip = 1) {
  df_filtered <- filter(df, trip_id == trip)

  df_filtered %>%
    select(from_id, to_id, departure_time, duration, total_fare,
           leg_id, leg_type, route_id, origin_stop_name, destination_stop_name, cumulative_fare,
           allowance_value, allowance_number, allowance_time) %>%
    st_set_geometry(NULL) %>%
    View()

  mv1 <- mapview(df_filtered, zcol = "route_id")

  df_stops <- st_set_geometry(df_filtered, NULL)

  df_stops1 <- df_stops %>% select(lon = origin_lon, lat = origin_lat, stop = origin_stop_name)
  df_stops2 <- df_stops %>% select(lon = destination_lon, lat = destination_lat, stop = destination_stop_name)
  df_stops3 <- rbind(df_stops1, df_stops2) %>% distinct()

  mv2 <- mapview(df_stops3, xcol = "lon", ycol = "lat", zcol = "stop", crs = 4326)

  mv1 + mv2
}
view_results(frontiers_df, 1)
view_results(frontiers_df, 2)



# Pareto Frontiers Plot ---------------------------------------------------


pareto_frontiers_df <- frontiers_df %>%
  st_set_geometry(NULL) %>%
  select(from_id, to_id, trip_id, duration, total_fare) %>%
  distinct()

pareto_frontiers_df <- pareto_frontiers_df %>%
  group_by(from_id, to_id) %>%
  arrange(total_fare, duration) %>%
  mutate(is_better = duration < lag(duration, default = Inf)) %>%
  filter(is_better == T)


pareto_frontiers_df %>%
  ggplot(aes(x=total_fare, y=duration)) +
  geom_vline(xintercept = c(0, 4.5, 4.8, 7.2, 8.37, 9.3 ), color = "grey80") +
  geom_step() +
  geom_point() +
  scale_x_continuous(limits = c(0, 10)) +
  scale_y_continuous(limits = c(0, 90)) +
  facet_wrap(~from_id+to_id)


# comparing TA and no TA --------------------------------------------------

frontiers_no_ta <- frontiers_no_ta %>% st_set_geometry(NULL)
frontiers_with_ta <- frontiers_with_ta %>% st_set_geometry(NULL)

full_frontiers <-
  full_join(frontiers_with_ta %>% mutate(departure_time = str_sub(departure_time, 1, 5)),
            frontiers_no_ta %>% mutate(departure_time = str_sub(departure_time, 1, 5)),
          by = c("from_id", "to_id", "departure_time", "origin_stop_id", "destination_stop_id",
                 "agency_id", "route_id", "route_short_name"))

full_frontiers %>%
  filter(is.na(trip_id.x)) %>%
  slice(200) %>%
  glimpse()



# full frontiers ----------------------------------------------------------

points <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))

r5r_core$setTravelAllowance(TRUE)
full_frontiers_with_ta_df <- pareto_frontier(r5r_core,
                                origins = points,
                                destinations = points,
                                departure_datetime = as.POSIXct("13-05-2019 14:00:00",
                                                                format = "%d-%m-%Y %H:%M:%S"),
                                mode = c("WALK", "TRANSIT"),
                                max_trip_duration = 180,
                                fare_structure = fare_structure,
                                monetary_cost_cutoffs = c(1, 4.5, 4.8, 7.20, 8.37, 11.4, 12.57),
                                max_rides = 5,
                                progress = TRUE
)

r5r_core$setTravelAllowance(FALSE)
full_frontiers_no_ta_df <- pareto_frontier(r5r_core,
                                             origins = points,
                                             destinations = points,
                                             departure_datetime = as.POSIXct("13-05-2019 14:00:00",
                                                                             format = "%d-%m-%Y %H:%M:%S"),
                                             mode = c("WALK", "TRANSIT"),
                                             max_trip_duration = 180,
                                             fare_structure = fare_structure,
                                             monetary_cost_cutoffs = c(1, 4.5, 4.8, 7.20, 8.37, 11.4, 12.57),
                                             max_rides = 5,
                                             progress = TRUE)

## combine frontiers
points_wta <- full_frontiers_with_ta_df %>%
  select(from_id, to_id, monetary_cost) %>%
  distinct()

points_nta <- full_frontiers_no_ta_df %>%
  select(from_id, to_id, monetary_cost) %>%
  distinct()

full_points <- rbind(points_wta, points_nta) %>%
  distinct()

full_points %>%
  left_join(full_frontiers_with_ta_df, suffix = c("", ".wta"))


full_frontiers_df <-
  full_join(full_frontiers_with_ta_df, full_frontiers_no_ta_df,
          by = c("from_id", "to_id", "monetary_cost", "percentile"),
          suffix = c(".wta", ".nta")) %>%
  select(from_id, to_id, monetary_cost, travel_time.wta, travel_time.nta)

diff_frontiers_df <- filter(full_frontiers_df,
                            travel_time.wta != travel_time.nta |
                              is.na(travel_time.wta) | is.na(travel_time.nta))

write_csv(diff_frontiers_df, here::here("diff_frontiers.csv"))

deg2rad <- function(deg) {
    return(deg * pi / 180)
}

rad2deg <- function(rad) {
    return(rad * 180 / pi)
}

# these are web mercator _pixels_ not tile numbers
lon_to_webmercator_pixel <- function (lon, zoom) {
    return((lon + 180) / 360 * 2 ^ zoom * 256)
}

lat_to_webmercator_pixel <- function (lat, zoom) {
    latr = deg2rad(lat)
    return(
        (1 - 
            log(tan(latr) + 1/cos(latr)) / pi
        ) * 2 ^ (zoom - 1) * 256
    )
}

webmercator_pixel_to_lon <- function(px, zoom) {
    return(px / 256 / 2^zoom * 360 - 180)
}

webmercator_pixel_to_lat <- function(px, zoom) {
    return(rad2deg(atan(sinh(pi - px / 256 / 2^zoom * 2 * pi))))
}

# load packages
#library("r5r")
devtools::load_all('.')
library("dplyr")
library("ggplot2")

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path)

# load origin/destination points
points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))

# load fare structure object
fare_structure_path <- system.file(
  "extdata/poa/fares/fares_poa.zip",
  package = "r5r"
)
fare_structure <- read_fare_structure(fare_structure_path)

# inputs
departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
                                 format = "%d-%m-%Y %H:%M:%S")
# plan itineraries
dit <- detailed_itineraries(r5r_core,
                            origins = points[10,],
                            destinations = points[12,],
                            mode = c("WALK", "TRANSIT"),
                            departure_datetime = departure_datetime,
                            max_trip_duration = 140,
                            fare_structure = fare_structure,
                            max_fare = 9,
                            shortest_path = FALSE,
                            drop_geometry = TRUE)


# build Pareto frontiers
pareto_df <- dit  |>
  group_by(option) |>
  summarise(travel_time = max(total_duration),
            fare = max(total_fare),
            mode = paste(if_else(mode == "WALK", "", mode), collapse = " ")) |>
  mutate(mode = if_else(mode == "", "WALK", mode))


# plot Pareto frontiers
ggplot(pareto_df , aes(x=fare, y=travel_time)) +
  geom_step() +
  geom_point() +
  geom_text(aes(label=mode), hjust = -0.1, vjust=-0.5) +
  geom_text(aes(label=paste("BRL", scales::comma(fare, accuracy = 0.01))),
            hjust = 0.3, vjust=1.5) +
  scale_y_continuous(breaks = seq(0, 140, 20), limits = c(0, 140)) +
  scale_x_continuous(breaks = 0:10, limits = c(0, 10),
                     labels = scales::comma_format(accuracy = 0.01)) +
  theme_light() +
  labs(x = "fare (BRL)", y = "travel time (minutes)",
       subtitle = "Travel time and monetary cost of trips between\nFarrapos Station and Praia de Belas Shopping Mall")

#' @section Transport modes:
#'
#' `R5` allows for multiple combinations of transport modes. The options
#' include:
#'
#' - **Transit modes:** `TRAM`, `SUBWAY`, `RAIL`, `BUS`, `FERRY`, `CABLE_CAR`,
#' `GONDOLA`, `FUNICULAR`. The option `TRANSIT` automatically considers all
#' public transport modes available.
#'
#' - **Non transit modes:** `WALK`, `BICYCLE`, `CAR`, `BICYCLE_RENT`,
#' `CAR_PARK`.

#' Calculate travel time and monetary cost Pareto frontier
#'
#' Fast computation of travel time and monetary cost Pareto frontier between
#' origin and destination pairs.
#'
#' @template r5r_network
#' @template r5r_core
#' @template common_arguments
#' @template time_window_related_args
#' @template fare_structure
#' @template verbose
#' @param percentiles An integer vector (max length of 5). Specifies the
#'   percentile to use when returning travel time estimates within the given
#'   time window. Please note that this parameter is applied to the travel time
#'   estimates only (e.g. if the 25th percentile is specified, and the output
#'   between A and B is 15 minutes and 10 dollars, 25% of all trips cheaper
#'   than 10 dollars taken between these points are shorter than 15 minutes).
#'   Defaults to 50, returning the median travel time. If a vector with length
#'   bigger than 1 is passed, the output contains an additional column that
#'   specifies the percentile of each travel time and monetary cost
#'   combination. Due to upstream restrictions, only 5 percentiles can be
#'   specified at a time. For more details, please see R5 documentation at
#'   <https://docs.conveyal.com/analysis/methodology#accounting-for-variability>.
#' @param fare_cutoffs A numeric vector. The monetary cutoffs that
#'   should be considered when calculating the Pareto frontier. Most of the
#'   time you'll want this parameter to be the combination of all possible
#'   fares listed in you `fare_structure`. Choosing a coarse distribution of
#'   cutoffs may result in many different trips falling within the same cutoff.
#'   For example, if you have two different routes in your GTFS, one costing $3
#'   and the other costing $4, and you set this parameter to `5`, the output
#'   will tell you the fastest trips that costed up to $5, but you won't be
#'   able to identify which route was used to complete such trips. In this
#'   case, it would be more beneficial to set the parameter as `c(3, 4)` (you
#'   could also specify combinations of such values, such as 6, 7, 8 and so on,
#'   because a transit user could hypothetically benefit from making transfers
#'   between the available routes).
#'
#' @return A `data.table` with the travel time and monetary cost Pareto frontier
#'   between the specified origins and destinations. An additional column
#'   identifying the travel time percentile is present if more than one value
#'   was passed to `percentiles`. Origin and destination pairs whose trips
#'   couldn't be completed within the maximum travel time using less money than
#'   the specified monetary cutoffs are not returned in the `data.table`. If
#'   `output_dir` is not `NULL`, the function returns the path specified in
#'   that parameter, in which the `.csv` files containing the results are
#'   saved.
#'
#' @template transport_modes_section
#' @template lts_section
#' @template datetime_parsing_section
#' @template mcraptor_algorithm_section
#'
#' @family routing
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' # build transport network
#' data_path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(data_path = data_path)
#'
#' # load origin/destination points
#' points <- read.csv(file.path(data_path, "poa_hexgrid.csv"))[1:5,]
#'
#' # load fare structure object
#' fare_structure_path <- system.file(
#'   "extdata/poa/fares/fares_poa.zip",
#'   package = "r5r"
#' )
#' fare_structure <- read_fare_structure(fare_structure_path)
#'
#' departure_datetime <- as.POSIXct(
#'   "13-05-2019 14:00:00",
#'   format = "%d-%m-%Y %H:%M:%S"
#' )
#'
#' pf <- pareto_frontier(
#'   r5r_network,
#'   origins = points,
#'   destinations = points,
#'   mode = c("WALK", "TRANSIT"),
#'   departure_datetime = departure_datetime,
#'   fare_structure = fare_structure,
#'   fare_cutoffs = c(4.5, 4.8, 9, 9.3, 9.6)
#' )
#' head(pf)
#'
#' stop_r5(r5r_network)
#' @export
pareto_frontier <- function(r5r_network,
                            r5r_core = deprecated(),
                            origins,
                            destinations,
                            mode = c("WALK", "TRANSIT"),
                            mode_egress = "WALK",
                            departure_datetime = Sys.time(),
                            time_window = 10L,
                            percentiles = 50L,
                            max_walk_time = Inf,
                            max_bike_time = Inf,
                            max_car_time = Inf,
                            max_trip_duration = 120L,
                            fare_structure = NULL,
                            fare_cutoffs = -1L,
                            walk_speed = 3.6,
                            bike_speed = 12,
                            max_rides = 3,
                            max_lts = 2,
                            n_threads = Inf,
                            verbose = FALSE,
                            progress = FALSE,
                            output_dir = NULL) {

  # deprecating r5r_core --------------------------------------
  if (lifecycle::is_present(r5r_core)) {

    cli::cli_warn(c(
      "!" = "The `r5r_core` argument is deprecated as of r5r v2.3.0.",
      "i" = "Please use the `r5r_network` argument instead."
    ))

    r5r_network <- r5r_core
  }

  old_options <- options(datatable.optimize = Inf)
  on.exit(options(old_options), add = TRUE)

  old_dt_threads <- data.table::getDTthreads()
  dt_threads <- ifelse(is.infinite(n_threads), 0, n_threads)
  data.table::setDTthreads(dt_threads)
  on.exit(data.table::setDTthreads(old_dt_threads), add = TRUE)

  # check inputs and set r5r options --------------------------------------

  origins <- assign_points_input(origins, "origins")
  destinations <- assign_points_input(destinations, "destinations")
  mode_list <- assign_mode(mode, mode_egress)
  departure <- assign_departure(departure_datetime)

  # check availability of transit services on the selected date
  if (mode_list$transit_mode %like% 'TRANSIT|TRAM|SUBWAY|RAIL|BUS|CABLE_CAR|GONDOLA|FUNICULAR') {
    check_transit_availability_on_date(r5r_network, departure_date = departure$date)
  }

  checkmate::assert_class(r5r_network, "r5r_network")
  r5r_network <- r5r_network@jcore

  max_walk_time <- assign_max_street_time(
    max_walk_time,
    walk_speed,
    max_trip_duration,
    "walk"
  )
  max_bike_time <- assign_max_street_time(
    max_bike_time,
    bike_speed,
    max_trip_duration,
    "bike"
  )
  max_car_time <- assign_max_street_time(
    max_car_time,
    8, # 8 km/h, R5's default.
    max_trip_duration,
    "car"
  )
  max_trip_duration <- assign_max_trip_duration(
    max_trip_duration,
    mode_list,
    max_walk_time,
    max_bike_time
  )

  set_time_window(r5r_network, time_window)
  set_percentiles(r5r_network, percentiles)
  set_monte_carlo_draws(r5r_network, 1, time_window)
  set_speed(r5r_network, walk_speed, "walk")
  set_speed(r5r_network, bike_speed, "bike")
  set_max_rides(r5r_network, max_rides)
  set_max_lts(r5r_network, max_lts)
  set_n_threads(r5r_network, n_threads)
  set_verbose(r5r_network, verbose)
  set_progress(r5r_network, progress)
  set_fare_structure(r5r_network, fare_structure)
  set_output_dir(r5r_network, output_dir)
  set_fare_cutoffs(r5r_network, fare_cutoffs)

  # call r5r_network method and process result -------------------------------

  frontier <- r5r_network$paretoFrontier(
    origins$id,
    origins$lat,
    origins$lon,
    destinations$id,
    destinations$lat,
    destinations$lon,
    mode_list$direct_modes,
    mode_list$transit_mode,
    mode_list$access_mode,
    mode_list$egress_mode,
    departure$date,
    departure$time,
    max_walk_time,
    max_bike_time,
    max_car_time,
    max_trip_duration
  )

  if (!verbose & progress) cat("Preparing final output...", file = stderr())

  frontier <- java_to_dt(frontier)

  if (nrow(frontier) > 0) {
    # replace travel-times of nonviable trips with NAs
    frontier[travel_time > max_trip_duration, travel_time := NA_integer_]
  }

  if (!verbose & progress) cat(" DONE!\n", file = stderr())

  if (!is.null(output_dir)) return(output_dir)
  return(frontier[])
}

# # if running manually, please run the following line first:
# # source("tests/testthat/setup.R")
#
testthat::skip_on_cran()
#
# tester <- function(draws_per_minute) {
#   set_monte_carlo_draws(r5r_network, draws_per_minute, time_window = 30)
# }
#
# test_that("input is correct", {
#   expect_error(tester("1"))
#   expect_error(tester(0))
#   expect_error(tester(c(1, 1)))
#   expect_error(tester(Inf))
# })
#
# # only one monte carlo draw is performed per minute if the routing function is
# # using the fastraptor algorithm and the gtfs feeds used to create the network
# # don't include a frequencies table (independently of the value set in
# # draws_per_minute). the number of monte carlo draws is respected when any of
# # the feeds include a frequencies table and when using the mcraptor algorithm.
#
# test_that("works in expanded_travel_time_matrix() - poa", {
#   basic_expr <- call(
#     "expanded_travel_time_matrix",
#     r5r_network = r5r_network,
#     origins = pois,
#     destinations = pois,
#     mode = c("TRANSIT", "WALK"),
#     departure_datetime = departure_datetime,
#     max_trip_duration = 60,
#     time_window = 30
#   )
#
#   small_draws_expr <- big_draws_expr <- basic_expr
#   small_draws_expr$draws_per_minute <- 1
#   big_draws_expr$draws_per_minute <- 5
#
#   small_draws <- eval(small_draws_expr)
#   big_draws <- eval(big_draws_expr)
#  # expect_identical(small_draws, big_draws)
#   expect_identical(unique(small_draws$draw_number), 1L)
# })
#
# test_that("works in expanded_travel_time_matrix() - spo", {
#   basic_expr <- call(
#     "expanded_travel_time_matrix",
#     r5r_network = spo_network,
#     origins = spo_points[51:65],
#     destinations = spo_points[51:65],
#     mode = c("TRANSIT", "WALK"),
#     departure_datetime = departure_datetime,
#     max_trip_duration = 60,
#     time_window = 30
#   )
#
#   small_draws_expr <- big_draws_expr <- basic_expr
#   small_draws_expr$draws_per_minute <- 1
#   big_draws_expr$draws_per_minute <- 5
#
#   small_draws <- eval(small_draws_expr)
#   expect_identical(unique(small_draws$draw_number), 1L)
#
#   big_draws <- eval(big_draws_expr)
#   draws_summary <- big_draws[
#     ,
#     .(contains_all_draws = identical(draw_number, 1:5)),
#     by = .(from_id, to_id, departure_time)
#   ]
#   expect_true(all(draws_summary$contains_all_draws))
#
#   big_draws_2 <- eval(big_draws_expr)
#   expect_false(identical(big_draws, big_draws_2))
# })
#
# # there's no way to actually know how many draws were performed with the other
# # functions, so we assume *some* draws are done and the randomization happens
# # if the results of the same expression are different in two different
# # evaluations
#
# test_that("works in travel_time_matrix() - poa", {
#   basic_expr <- call(
#     "travel_time_matrix",
#     r5r_network = r5r_network,
#     origins = pois,
#     destinations = pois,
#     mode = c("TRANSIT", "WALK"),
#     departure_datetime = departure_datetime,
#     max_trip_duration = 60,
#     time_window = 30,
#     percentiles = c(1, 50, 99)
#   )
#
#   small_draws_expr <- big_draws_expr <- basic_expr
#   small_draws_expr$draws_per_minute <- 1
#   big_draws_expr$draws_per_minute <- 5
#
#   small_draws <- eval(small_draws_expr)
#   big_draws <- eval(big_draws_expr)
#   expect_identical(small_draws, big_draws)
#
#   mcrap_basic_expr <- basic_expr
#   mcrap_basic_expr$fare_structure <- fare_structure
# test_that("works in expanded_travel_time_matrix() - spo", {
#   basic_expr <- call(
#     "expanded_travel_time_matrix",
#     r5r_network = spo_network,
#     origins = spo_points[51:65],
#     destinations = spo_points[51:65],
#     mode = c("TRANSIT", "WALK"),
#     departure_datetime = departure_datetime,
#     max_trip_duration = 60,
#     time_window = 30
#   )
#
#   small_draws_expr <- big_draws_expr <- basic_expr
#   small_draws_expr$draws_per_minute <- 1
#   big_draws_expr$draws_per_minute <- 5
#
#   small_draws <- eval(small_draws_expr)
#   expect_identical(unique(small_draws$draw_number), 1L)
#
#   big_draws <- eval(big_draws_expr)
#   draws_summary <- big_draws[
#     ,
#     .(contains_all_draws = identical(draw_number, 1:5)),
#     by = .(from_id, to_id, departure_time)
#   ]
#   expect_true(all(draws_summary$contains_all_draws))
#
#   big_draws_2 <- eval(big_draws_expr)
#   expect_false(identical(big_draws, big_draws_2))
# })

# there's no way to actually know how many draws were performed with the other
# functions, so we assume *some* draws are done and the randomization happens
# if the results of the same expression are different in two different
# evaluations

test_that("works in travel_time_matrix() - poa", {
  basic_expr <- call(
    "travel_time_matrix",
    r5r_network = r5r_network,
    origins = pois,
    destinations = pois,
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    max_trip_duration = 60,
    time_window = 30,
    percentiles = c(1, 50, 99)
  )

  small_draws_expr <- big_draws_expr <- basic_expr
  small_draws_expr$draws_per_minute <- 1
  big_draws_expr$draws_per_minute <- 5

  small_draws <- eval(small_draws_expr)
  big_draws <- eval(big_draws_expr)
  expect_identical(small_draws, big_draws)

  mcrap_basic_expr <- basic_expr
  mcrap_basic_expr$fare_structure <- fare_structure
  mcrap_basic_expr$max_fare <- 5
  mcrap_small_draws_expr <- mcrap_big_draws_expr <- mcrap_basic_expr
  mcrap_small_draws_expr$draws_per_minute <- 1
  mcrap_big_draws_expr$draws_per_minute <- 5

  mcrap_small_draws <- eval(mcrap_small_draws_expr)
  mcrap_big_draws <- eval(mcrap_big_draws_expr)
  expect_false(identical(mcrap_small_draws, mcrap_big_draws))
})

test_that("works in travel_time_matrix() - spo", {
  basic_expr <- call(
    "travel_time_matrix",
    r5r_network = spo_network,
    origins = spo_points[51:80],
    destinations = spo_points[51:80],
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    max_trip_duration = 60,
    time_window = 30,
    percentiles = c(1, 50, 99)
  )

  small_draws_expr <- big_draws_expr <- basic_expr
  small_draws_expr$draws_per_minute <- 1
  big_draws_expr$draws_per_minute <- 5

  small_draws <- eval(small_draws_expr)
  big_draws <- eval(big_draws_expr)
  expect_false(identical(small_draws, big_draws))

  big_draws2 <- eval(big_draws_expr)
  expect_false(identical(big_draws, big_draws2))
})

test_that("works in accessibility() - poa", {
  basic_expr <- call(
    "accessibility",
    r5r_network = r5r_network,
    origins = points[1:30],
    destinations = points[1:30],
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    opportunities_colnames = "population",
    decay_function = "step",
    cutoffs = 30,
    max_trip_duration = 60,
    time_window = 30,
    percentiles = c(1, 50, 99)
  )

  small_draws_expr <- big_draws_expr <- basic_expr
  small_draws_expr$draws_per_minute <- 1
  big_draws_expr$draws_per_minute <- 5

  small_draws <- eval(small_draws_expr)
  big_draws <- eval(big_draws_expr)
  expect_identical(small_draws, big_draws)

  mcrap_basic_expr <- basic_expr
  mcrap_basic_expr$fare_structure <- fare_structure
  mcrap_basic_expr$max_fare <- 5
  mcrap_small_draws_expr <- mcrap_big_draws_expr <- mcrap_basic_expr
  mcrap_small_draws_expr$draws_per_minute <- 1
  mcrap_big_draws_expr$draws_per_minute <- 5

  mcrap_small_draws <- eval(mcrap_small_draws_expr)
  mcrap_big_draws <- eval(mcrap_big_draws_expr)
  expect_false(identical(mcrap_small_draws, mcrap_big_draws))
})

test_that("works in accessibility() - spo", {
  basic_expr <- call(
    "accessibility",
    r5r_network = spo_network,
    origins = spo_points[51:80],
    destinations = spo_points[51:80],
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    opportunities_colnames = "opportunities",
    decay_function = "step",
    cutoffs = 30,
    max_trip_duration = 60,
    time_window = 30,
    percentiles = c(1, 50, 99)
  )

  small_draws_expr <- big_draws_expr <- basic_expr
  small_draws_expr$draws_per_minute <- 1
  big_draws_expr$draws_per_minute <- 5

  small_draws <- eval(small_draws_expr)
  big_draws <- eval(big_draws_expr)
  expect_false(identical(small_draws, big_draws))

  big_draws2 <- eval(big_draws_expr)
  expect_false(identical(big_draws, big_draws2))
})

test_that("poa draws_per_minute should fail in pareto_frontier() issue #281", {
  basic_expr <- call(
    "pareto_frontier",
    r5r_network = r5r_network,
    origins = points[1:10],
    destinations = points[1:10],
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    max_trip_duration = 60,
    time_window = 30,
    percentiles = c(1, 50, 99),
    fare_structure = fare_structure,
    fare_cutoffs = c(0, 5, 10)
  )

  small_draws_expr <- big_draws_expr <- basic_expr
  small_draws_expr$draws_per_minute <- 1
  big_draws_expr$draws_per_minute <- 5

  expect_error( eval(small_draws_expr) )
  expect_error( eval(big_draws_expr) )

})

test_that("spo draws_per_minute should fail in pareto_frontier() issue #281", {
  basic_expr <- call(
    "pareto_frontier",
    r5r_network = spo_network,
    origins = spo_points[51:80],
    destinations = spo_points[51:80],
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    max_trip_duration = 60,
    time_window = 30,
    percentiles = c(1, 50, 99),
    fare_structure = spo_fare_struc,
    fare_cutoffs = c(0, 5, 10, 15, Inf)
  )

  small_draws_expr <- big_draws_expr <- basic_expr
  small_draws_expr$draws_per_minute <- 1
  big_draws_expr$draws_per_minute <- 5

  expect_error( eval(small_draws_expr) )
  expect_error( eval(big_draws_expr) )
})

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

tester <- function(percentiles) {
  set_percentiles(r5r_network, percentiles)
  }

test_that("input is correct", {
  expect_error(tester("1"))
  expect_error(tester(0))
  expect_error(tester(100))
  expect_error(tester(c(1, 2, 3, 4, 5, 6)))
  expect_error(tester(c(1, 1)))
  expect_error(tester(NA))
  expect_error(tester(Inf))
})

test_that("set_percentiles argument works in travel_time_matrix()", {
  basic_expr <- call(
    "travel_time_matrix",
    r5r_network,
    origins=pois,
    destinations=pois,
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    draws_per_minute = 1,
    time_window = 30,
    max_trip_duration = 60
  )

  one_percent_expr <- many_percent_expr <- basic_expr
  one_percent_expr$percentiles <- 50
  many_percent_expr$percentiles <- c(1, 50, 99)

  one_percent <- eval(one_percent_expr)
  one_percent[, c("from_id", "to_id") := NULL]
  expect_true(ncol(one_percent) == 1)
  expect_true(names(one_percent) == "travel_time_p50")
  expect_type(one_percent$travel_time_p50, "integer")

  many_percent <- eval(many_percent_expr)
  many_percent[, c("from_id", "to_id") := NULL]
  expect_true(ncol(many_percent) == 3)
  expect_identical(
    names(many_percent),
    paste0("travel_time_p", c("01", "50", "99"))
  )
  expect_type(many_percent$travel_time_p01, "integer")
  expect_type(many_percent$travel_time_p50, "integer")
  expect_type(many_percent$travel_time_p99, "integer")

  # travel time may be NA if less than x% of the trips between a pair can be
  # completed within max_trip_duration. the lowest percentile can't have any NAs
  # though

  expect_false(any(is.na(many_percent$travel_time_p01)))
  expect_true(any(is.na(many_percent$travel_time_p50)))
  expect_true(any(is.na(many_percent$travel_time_p99)))

  many_percent[is.na(travel_time_p50), travel_time_p50 := 1000]
  many_percent[is.na(travel_time_p99), travel_time_p99 := 1000]
  expect_true(all(many_percent$travel_time_p01 <= many_percent$travel_time_p50))
  expect_true(all(many_percent$travel_time_p50 <= many_percent$travel_time_p99))
})

test_that("set_percentiles argument works in accessibility()", {
  basic_expr <- call(
    "accessibility",
    r5r_network,
    origins=points[1:30],
    destinations=points[1:30],
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    opportunities_colnames = "population",
    decay_function = "step",
    cutoffs = 30,
    draws_per_minute = 1,
    time_window = 30,
    max_trip_duration = 60
  )

  one_percent_expr <- many_percent_expr <- basic_expr
  one_percent_expr$percentiles <- 50
  many_percent_expr$percentiles <- c(1, 50, 99)

  one_percent <- eval(one_percent_expr)
  expect_type(one_percent$percentile, "integer")
  expect_true(unique(one_percent$percentile) == 50)

  many_percent <- eval(many_percent_expr)
  expect_type(many_percent$percentile, "integer")
  expect_equal(unique(many_percent$percentile), c(1, 50, 99))

  many_percent <- data.table::dcast(
    many_percent,
    id ~ percentile,
    value.var = "accessibility"
  )
  data.table::setnames(
    many_percent,
    old = as.character(c(1, 50, 99)),
    new = paste0("access_", c(1, 50, 99))
  )
  expect_true(all(many_percent$access_1 >= many_percent$access_50))
  expect_true(all(many_percent$access_50 >= many_percent$access_99))
})

test_that("set_percentiles argument works in pareto_frontier()", {
  basic_expr <- call(
    "pareto_frontier",
    r5r_network,
    origins=points[1:5],
    destinations=points[1:5],
    mode = c("TRANSIT", "WALK"),
    departure_datetime = departure_datetime,
    fare_structure = fare_structure,
    fare_cutoffs = c(0, 5, 10),
    time_window = 30,
    max_trip_duration = 60
  )

  one_percent_expr <- many_percent_expr <- basic_expr
  one_percent_expr$percentiles <- 50
  many_percent_expr$percentiles <- c(1, 50, 99)

  one_percent <- eval(one_percent_expr)
  expect_type(one_percent$percentile, "integer")
  expect_true(unique(one_percent$percentile) == 50)

  many_percent <- eval(many_percent_expr)
  expect_type(many_percent$percentile, "integer")
  expect_equal(unique(many_percent$percentile), c(1, 50, 99))

  # all 0 cost trips must be identical

  zero_cost <- many_percent[monetary_cost == 0]
  zero_cost <- data.table::dcast(
    zero_cost,
    from_id + to_id ~ percentile,
    value.var = "travel_time"
  )
  data.table::setnames(
    zero_cost,
    old = as.character(c(1, 50, 99)),
    new = paste0("travel_time_", c(1, 50, 99))
  )
  expect_identical(zero_cost$travel_time_1, zero_cost$travel_time_50)
  expect_identical(zero_cost$travel_time_50, zero_cost$travel_time_99)

  # each percentile-value defines their own frontiers

  is_pareto_frontier <- function(df) {
    travel_times_decrease <- all(diff(df$travel_time) < 0)
    monetary_costs_increase <- all(diff(df$monetary_cost) > 0)
    return(travel_times_decrease && monetary_costs_increase)
  }

  many_percent <- many_percent[
    ,
    .(data = list(.SD)),
    by = .(from_id, to_id, percentile)
  ]
  many_percent[
    ,
    is_pareto_frontier := vapply(data, is_pareto_frontier, logical(1))
  ]
  expect_true(all(many_percent$is_pareto_frontier))
})

devtools::load_all(".")

spo_path <- system.file("extdata/spo", package = "r5r")
spo_core <- setup_r5(spo_path, verbose = FALSE)
spo_points <- data.table::fread(file.path(spo_path, "spo_hexgrid.csv"))
spo_points[, opportunities := 1]
spo_fare_struc <- setup_fare_structure(spo_core, 5, by = "GENERIC")
spo_fare_struc$fares_per_transfer <- data.table::data.table(NULL)

departure_datetime <- as.POSIXct("13-05-2019 14:07:00",
                                 format = "%d-%m-%Y %H:%M:%S")

basic_expr <- call(
  "pareto_frontier",
  r5r_core = spo_core,
  origins = spo_points[id == "89a8100c08fffff"],
  destinations = spo_points[id == "89a8100c557ffff"],
  mode = c("TRANSIT", "WALK"),
  departure_datetime = departure_datetime,
  max_trip_duration = 60,
  time_window = 30,
  percentiles = c(1, 50, 99),
  fare_structure = spo_fare_struc,
  monetary_cost_cutoffs = c(0, 5, 10, 15, Inf)*100,
  draws_per_minute = 10,
  verbose=F
)
frontier <- eval(basic_expr)
print(frontier$monetary_cost |> unique())


ttm_expr <- call(
  "expanded_travel_time_matrix",
  r5r_core = spo_core,
  origins = spo_points[id == "89a8100c08fffff"],
  destinations = spo_points[id == "89a8100c557ffff"],
  mode = c("TRANSIT", "WALK"),
  departure_datetime = departure_datetime,
  max_trip_duration = 60,
  time_window = 30,
  draws_per_minute = 1
)
ttm <- eval(ttm_expr)
print(ttm$routes |> unique())

spo_core$setDetailedItinerariesV2(TRUE)


spo_core$hasFrequencies()
spo_core$hasS
dit_expr <- call(
  "detailed_itineraries",
  r5r_core = spo_core,
  origins = spo_points[id == "89a8100c08fffff"],
  destinations = spo_points[id == "89a8100c557ffff"],
  mode = c("TRANSIT", "WALK"),
  fare_structure = spo_fare_struc,
  max_fare=10,
  departure_datetime = departure_datetime,
  max_trip_duration = 60,
  time_window = 10,
  verbose = T
)
dit <- eval(dit_expr)

spo_core$verboseMode()
spo_core$printHasSchedules()

#' @section Level of Traffic Stress (LTS):
#'
#' When cycling is enabled in `R5` (by passing the value `BIKE` to either
#' `mode` or `mode_egress`), setting `max_lts` will allow cycling only on
#' streets with a given level of danger/stress. Setting `max_lts` to 1, for
#' example, will allow cycling only on separated bicycle infrastructure or
#' low-traffic streets and routing will revert to walking when traversing any
#' links with LTS exceeding 1. Setting `max_lts` to 3 will allow cycling on
#' links with LTS 1, 2 or 3. Routing also reverts to walking if the street
#' segment is tagged as non-bikable in OSM (e.g. a staircase), independently of
#' the specified max LTS.
#'
#' The default methodology for assigning LTS values to network edges is based
#' on commonly tagged attributes of OSM ways. See more info about LTS in the
#' original documentation of R5 from Conveyal at
#' <https://docs.conveyal.com/learn-more/traffic-stress>. In summary:
#'
#' - **LTS 1**: Tolerable for children. This includes low-speed, low-volume
#' streets, as well as those with separated bicycle facilities (such as
#' parking-protected lanes or cycle tracks).
#' - **LTS 2**: Tolerable for the mainstream adult population. This includes
#' streets where cyclists have dedicated lanes and only have to interact with
#' traffic at formal crossing.
#' - **LTS 3**: Tolerable for "enthused and confident" cyclists. This includes
#' streets which may involve close proximity to moderate- or high-speed
#' vehicular traffic.
#' - **LTS 4**: Tolerable only for "strong and fearless" cyclists. This
#' includes streets where cyclists are required to mix with moderate- to
#' high-speed vehicular traffic.
#'
#' For advanced users, you can provide custom LTS values by adding a tag `<key
#' = "lts">` to the `osm.pbf` file.

# create polygons with new speeds for sample data - Porto Alegre

library(sf)
library(dplyr)
library(mapview)

# 1. City-centre reference point (lon, lat in EPSG:4326) -----------------------

# Porto Alegre city-centre coords
poa_center <- st_sf(
  geometry = st_sfc(
    st_point(c(-51.2350, -30.0345)),
    crs = 4326
  )
)

mapview(poa_center)

# Re-project to a metric CRS (SIRGAS-2000 / UTM 22 S, EPSG:31982)
poa_center_utm <- st_transform(poa_center, 31982)



# 2. Build 1st buffer of 1 Km) -------------------------------------------------

# (a) 1 km radius, centred on city centre
poly_1km <- st_buffer(poa_center_utm, dist = 1000)
mapview(poly_1km)

# (b) 1000 km radius, whose centre is 1.5 km due-east (4 km + 2 km) of (a)
# shift in metres (E, N)
offset_east <- c(1500, 0)

poly_500m_east <- poa_center_utm
st_geometry(poly_500m_east) <- st_geometry(poly_500m_east) + offset_east
poly_500m_east <- st_buffer(poly_500m_east, dist = 500)

st_crs(poly_500m_east) <- st_crs(poly_1km)

# 3. Combine into a two-row sf object) -------------------------------------------------
multipolys <- rbind(
  st_sf(id = 1,      geometry = st_geometry(poly_1km)),
  st_sf(id = 2, geometry = st_geometry(poly_500m_east))
)

# (Optional) Transform back to WGS-84 for leaflet/GeoJSON export
multipolys_wgs84 <- st_transform(multipolys, 4326)

mapview(multipolys_wgs84)

# set max speed
multipolys_wgs84$max_speed <- c(0.7, 0.8)

# reorder columns
multipolys_wgs84 <- multipolys_wgs84 |>
  select(poly_id = id, max_speed)

# save data
saveRDS(multipolys_wgs84, 'poa_poly_congestion.rds')

#' @section Datetime parsing:
#'
#' `r5r` ignores the timezone attribute of datetime objects when parsing dates
#' and times, using the study area's timezone instead. For example, let's say
#' you are running some calculations using Rio de Janeiro, Brazil, as your study
#' area. The datetime `as.POSIXct("13-05-2019 14:00:00",
#' format = "%d-%m-%Y %H:%M:%S")` will be parsed as May 13th, 2019, 14:00h in
#' Rio's local time, as expected. But `as.POSIXct("13-05-2019 14:00:00",
#' format = "%d-%m-%Y %H:%M:%S", tz = "Europe/Paris")` will also be parsed as
#' the exact same date and time in Rio's local time, perhaps surprisingly,
#' ignoring the timezone attribute.

#' @param verbose A logical. Whether to show `R5` informative messages when
#'   running the function. Defaults to `FALSE` (please note that in such case
#'   `R5` error messages are still shown). Setting `verbose` to `TRUE` shows
#'   detailed output, which can be useful for debugging issues not caught by
#'   `r5r`.

# options(java.parameters = '-Xmx16384m')
# options(java.parameters = c("-XX:+UseConcMarkSweepGC", "-Xmx16384m"))

# library(r5r)
devtools::load_all(".")
library(tidyverse)
library(sf)
library(h3jsr)
library(data.table)
library(viridis)
library(patchwork)
library(googlesheets4)

# build transport network
data_path <- "~/Repos/r5r_fares/rio"
r5r_core <- setup_r5(data_path = data_path, verbose = FALSE)

# load origin/destination points

poi <- tribble(
  ~id, ~lat, ~lon,
  "centro", -22.9064720147941, -43.177139635807734,
  "catete", -22.930673881789726, -43.17764097392119,
  "copacabana", -22.96915219120712, -43.18463648168664,
  "pavuna", -22.814726, -43.362337,
  "jardim_oceanico", -23.009335, -43.312220,
  "santa_cruz", -22.914807, -43.688556,
  "madureira", -22.875438, -43.339674
)

points <- read_csv("~/Repos/r5r_fares/rio/points_rio_09_2019.csv") %>%
  rename(id = id_hex, lon=X, lat=Y) %>%
  mutate(unit = 1)
departure_datetime <- as.POSIXct("13-05-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S")

points_r8 <- h3jsr::get_parent(points$id, res = 8, simple = TRUE)
points_r8 <- unique(points_r8)
points_r8 <- h3jsr::h3_to_point(points_r8, simple = FALSE)
points_r8$id <- points_r8$h3_address
points_r8$unit <- 1
# area_sf <- h3jsr::h3_to_polygon(points$id, simple = FALSE) %>%
#   summarise()

# Pareto ------------------------------------------------------------------

fare_settings <- read_fare_calculator(file_path = here::here("tests_marcus", "rio_fares_v3.zip"))
pareto_cutoffs <- c(0, 3.80,  4.05,  4.70,  5.00,  6.05,  6.50,  7.10,  7.60,  8.10,  8.55,  9.40,  10.00)

fare_settings <- setup_fare_calculator(r5r_core, base_fare = 0, by = "GENERIC")
fare_settings$fares_per_mode$allow_same_route_transfer <- TRUE
fare_settings$fares_per_mode$unlimited_transfers <- TRUE

fare_settings$debug_settings$output_file <- here::here("tests_marcus", "rio_fare_calculator_output.csv")

points_sample <- sample_n(points_r8, 50)

system.time(
  pareto_df <- pareto_frontier(r5r_core,
                               origins = poi,
                               destinations = poi,
                               mode = c("WALK", "TRANSIT"),
                               departure_datetime = departure_datetime,
                               monetary_cost_cutoffs = pareto_cutoffs,
                               fare_calculator_settings = fare_settings,
                               max_trip_duration = 180,
                               max_walk_dist = 8000,
                               time_window = 1, #30,
                               percentiles = 50, # c(5, 50, 95),
                               max_rides = 3,
                               draws_per_minute = 1L,
                               verbose = FALSE,
                               progress = TRUE)
)

rio_debug_1 <- read_csv(here::here("tests_marcus", "rio_fare_calculator_debug_anterior.csv"))
rio_debug_2 <- read_csv(here::here("tests_marcus", "rio_fare_calculator_output.csv"))

rio_debug_df <-full_join(rio_debug_1, rio_debug_2, by="pattern") %>%
  mutate(fare.x = fare.x / 100)
View(rio_debug_df)

pareto_cutoffs <- c(rio_debug_df$fare.x, unique(rio_debug_df$fare.y))
pareto_cutoffs <- unique(pareto_cutoffs)
pareto_cutoffs <- pareto_cutoffs[!is.na(pareto_cutoffs)]
pareto_cutoffs <- sort(pareto_cutoffs)


pareto_df %>%
  # filter(from_id == "catete") %>%
  mutate(percentile = factor(percentile),
         pair = paste(from_id, to_id)) %>%
  pivot_longer(cols=starts_with("monetary"), names_to = "mon", values_to="cost") %>%
  ggplot(aes(x=cost, y=travel_time, color=pair)) +
  geom_step() +
  geom_point() +
  # geom_text(aes(label = paste0("(", monetary_cost, ",", travel_time, ")"))) +
  # geom_path() +
  # scale_color_brewer(palette = "Set1") +
  scale_x_continuous(breaks = 0:10, limits = c(0, 10)) +
  scale_y_continuous(breaks = seq(0, 180, 30), limits = c(0, 180)) +
  theme(legend.position = "none") +
  facet_wrap(~pair)

write_csv(rio_debug_df, "rio_debug_novo.csv")


##########
density_test <- function(t) {
  ttm_1 <- travel_time_matrix(r5r_core,
                              origins = points_r8,
                              destinations = points_r8,
                              mode = c("WALK", "TRANSIT"),
                              departure_datetime = departure_datetime,
                              fare_calculator_settings = NULL,
                              max_trip_duration = 180,
                              max_walk_dist = t,
                              time_window = 1, #30,
                              percentiles = 50, # c(5, 50, 95),
                              max_rides = 3,
                              draws_per_minute = 1L,
                              verbose = FALSE,
                              progress = TRUE)

  ttm_2 <- travel_time_matrix(r5r_core,
                              origins = points_r8,
                              destinations = points_r8,
                              mode = c("WALK", "TRANSIT"),
                              departure_datetime = departure_datetime,
                              max_fare = 50,
                              fare_calculator_settings = fare_settings,
                              max_trip_duration = 180,
                              max_walk_dist = t,
                              time_window = 1, #30,
                              percentiles = 50, # c(5, 50, 95),
                              max_rides = 3,
                              draws_per_minute = 1L,
                              verbose = FALSE,
                              progress = TRUE)


  ttm_c <-
    rbind(mutate(ttm_1, method = "null"),
          mutate(ttm_2, method = "normal")
    )


  unit_access <- ttm_c %>%
    filter(travel_time <= 120) %>%
    count(from_id, method)
  # pivot_wider(names_from = method, values_from = n)

  p <- unit_access %>%
    ggplot(aes(x=n, fill = method)) +
    geom_density(alpha = 0.5) +
    labs(title = paste0("max_walk_dist = ", t))

  return(p)

}

plots <- map(c(500, 1000, 1500, 2000, 3000, 5000),
    density_test)

library(patchwork)

wrap_plots(plots, ncol = 2) +
  patchwork::plot_annotation(title = "After Fix",
                             subtitle = "Density plots of unitary accessibility")

density_test(500)
density_test(1000)
density_test(1500)
density_test(2000)
density_test(3000)
density_test(5000)


unit_access

unit_access_b %>%
  ggplot(aes(x=n, fill = method)) +
  geom_density(alpha = 0.5) +
  labs(title = "max_walk_dist = 2500")
facet_wrap(~method)


ttm_b <- full_join(ttm_1, ttm_2, by=c("from_id", "to_id")) %>%
  full_join(ttm_3, by=c("from_id", "to_id")) %>%
  mutate(diff_1 = travel_time.x - travel_time.y,
         diff_2 = travel_time.x - travel_time)

ttm_b %>%
  filter(is.na(travel_time.y)) %>% View()

pt <- ttm_b %>%
  filter(is.na(travel_time.y)) %>%
  count(from_id) %>%
  filter(n > 50)

View(ttm_b)

ttm_b %>%
  pivot_longer(cols = starts_with("travel_time"))

points_r8 %>%
  # filter(h3_address == "88a8a062b3fffff") %>%
  filter(h3_address %in% pt$from_id) %>%
  mapview::mapview(crs = 4326)

#' Setup a fare structure to calculate the monetary costs of trips
#'
#' Creates a basic fare structure that describes how transit fares should be
#' calculated in [travel_time_matrix()], [expanded_travel_time_matrix()],
#' [accessibility()] and [pareto_frontier()]. This fare structure can be
#' manually edited and adjusted to the existing rules in your study area, as
#' long as they stick to some basic premises. Please see the \href{../doc/fare_structure.html}{fare-structure vignette} for more information.
#'
#' @template r5r_network
#' @template r5r_core
#' @param base_fare A numeric. A base value used to populate the fare
#'   structure.
#' @param by A string. Describes how `fare_type`s (a classification we created
#'   to assign fares to different routes) are distributed among routes.
#'   Possible values are `MODE`, `AGENCY` and `GENERIC`. `MODE` is used when
#'   the mode is what determines the price of a route (e.g. if all the buses of
#'   a given city cost $5). `AGENCY` is used when the agency that operates each
#'   route is what determines its price (i.e. when two different routes/modes
#'   operated by a single agency cost the same; note that you can also use
#'   `AGENCY_NAME`, if the agency_ids listed in your GTFS cannot be easily
#'   interpreted). `GENERIC` is used when all the routes cost the same. Please
#'   note that this classification can later be edited to better suit your
#'   needs (when, for example, two types of buses cost the same, but one offers
#'   discounts after riding the subway and the other one doesn't), but this
#'   parameter may save you some work.
#' @param debug_path Either a path to a `.csv` file or `NULL`. When `NULL` (the
#'   default), fare debugging capabilities are disabled - i.e. there's no way
#'   to check if the fare calculation is correct. When a path is provided,
#'   `r5r` saves different itineraries and their respective fares to the
#'   specified file. How each itinerary is described is controlled by
#'   `debug_info`.
#' @param debug_info Either a string (when `debug_path` is a path) or `NULL`
#'   (the default). Doesn't have any effect if `debug_path` is `NULL`. When a
#'   string, accepts the values `MODE`, `ROUTE` and `MODE_ROUTE`. These values
#'   dictates how itinerary information is written to the output. Let's suppose
#'   we have an itinerary composed by two transit legs: first a subway leg
#'   whose route_id is 001, and then a bus legs whose route_id is 007. If
#'   `debug_info` is `MODE`, then this itinerary will be described as
#'   `SUBWAY|BUS`. If `ROUTE`, as `001|007`. If `MODE_ROUTE`, as `SUBWAY
#'   001|BUS 007`. Please note that the final debug information will contain
#'   not only the itineraries that were in fact used in the itineraries
#'   returned in [travel_time_matrix()], [accessibility()] and
#'   [pareto_frontier()], but all the itineraries that `R5` checked when
#'   calculating the routes. This imposes a performance penalty when tracking
#'   debug information (but has the positive effect of returning a larger
#'   sample of itineraries, which might help finding some implementation issues
#'   on the fare structure).
#'
#' @return A fare structure object.
#'
#' @family fare structure
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' data_path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(data_path)
#'
#' fare_structure <- setup_fare_structure(r5r_network, base_fare = 5)
#'
#' # to debug fare calculation
#' fare_structure <- setup_fare_structure(
#'   r5r_network,
#'   base_fare = 5,
#'   debug_path = "fare_debug.csv",
#'   debug_info = "MODE"
#' )
#'
#' fare_structure$debug_settings
#'
#' # debugging can be manually turned off by setting output_file to ""
#' fare_structure$debug_settings <- ""
#'
#' @export
setup_fare_structure <- function(r5r_network,
                                 r5r_core = deprecated(),
                                 base_fare,
                                 by = "MODE",
                                 debug_path = NULL,
                                 debug_info = NULL) {

  # deprecating r5r_core --------------------------------------
  if (lifecycle::is_present(r5r_core)) {

    cli::cli_warn(c(
      "!" = "The `r5r_core` argument is deprecated as of r5r v2.3.0.",
      "i" = "Please use the `r5r_network` argument instead."
    ))

    r5r_network <- r5r_core
  }

  # check inputs and set r5r options --------------------------------------
  checkmate::assert_class(r5r_network, "r5r_network")
  r5r_network <- r5r_network@jcore
  checkmate::assert_numeric(base_fare, lower = 0, len = 1, any.missing = FALSE)

  by_options <- c("MODE", "AGENCY_ID", "AGENCY_NAME", "GENERIC")
  by <- toupper(by)
  checkmate::assert(
    checkmate::check_string(by),
    checkmate::check_names(by, subset.of = by_options),
    combine = "and"
  )

  checkmate::assert_string(debug_path, pattern = "\\.csv$", null.ok = TRUE)

  if (is.null(debug_path) && !is.null(debug_info)) {
    stop("Please specify a file to write debug info to with 'debug_path'.")
  } else if (!is.null(debug_path) && is.null(debug_info)) {
    debug_info <- "ROUTE"
  }

  debug_info_options <- c("MODE", "ROUTE", "MODE_ROUTE")
  checkmate::assert_string(debug_info, null.ok = TRUE)
  if (!is.null(debug_info)) {
    debug_info <- toupper(debug_info)
    checkmate::assert_names(debug_info, subset.of = debug_info_options)
  }

  # r5r_network method to build fare structure returns a json

  f_struct <- r5r_network$buildFareStructure(rJava::.jfloat(base_fare), by)
  json_string <- f_struct$toJson()

  fare_structure <- jsonlite::parse_json(json_string, simplifyVector = TRUE)

  # Inf values are not supported by Java, so we use -1 to represent them

  if (fare_structure$fare_cap <= 0) fare_structure$fare_cap <- Inf

  if (!is.null(debug_path)) {
    debug <- list(
      output_file = debug_path,
      trip_info = debug_info
    )
  } else {
    debug <- list(
      output_file = "",
      trip_info = "MODE"
    )
  }
  fare_structure$debug_settings <- debug

  data.table::setDT(fare_structure$fares_per_type)
  data.table::setDT(fare_structure$fares_per_transfer)
  data.table::setDT(fare_structure$fares_per_route)

  return(fare_structure)
}


#' Write a fare structure object to disk
#'
#' Writes a fare structure object do disk. Fare structure is saved as a
#' collection of `.csv` files inside a `.zip` file.
#'
#' @template fare_structure
#' @param file_path A path to a `.zip` file. Where the fare structure should be
#'   written to.
#'
#' @return The path passed to `file_path`, invisibly.
#'
#' @family fare structure
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' data_path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(data_path)
#'
#' fare_structure <- setup_fare_structure(r5r_network, base_fare = 5)
#'
#' tmpfile <- tempfile("sample_fare_structure", fileext = ".zip")
#' write_fare_structure(fare_structure, tmpfile)
#'
#' @export
write_fare_structure <- function(fare_structure, file_path) {
  checkmate::assert_string(file_path, pattern = "\\.zip$", null.ok = TRUE)

  fare_global_settings <- data.table::data.table(
    setting = c(
      "max_discounted_transfers",
      "transfer_time_allowance",
      "fare_cap"
    ),
    value = c(
      fare_structure$max_discounted_transfers,
      fare_structure$transfer_time_allowance,
      fare_structure$fare_cap
    )
  )

  fare_debug_settings <- data.table::data.table(
    setting = c("output_file", "trip_info"),
    value = c(
      fare_structure$debug_settings$output_file,
      fare_structure$debug_settings$trip_info
    )
  )

  tmpdir <- tempfile(pattern = "r5r_fare_structure")
  dir.create(tmpdir)
  tmpfile <- function(path) file.path(tmpdir, path)

  data.table::fwrite(fare_global_settings, tmpfile("global_settings.csv"))
  data.table::fwrite(
    fare_structure$fares_per_type,
    tmpfile("fares_per_type.csv")
  )
  data.table::fwrite(
    fare_structure$fares_per_transfer,
    tmpfile("fares_per_transfer.csv")
  )
  data.table::fwrite(
    fare_structure$fares_per_route,
    tmpfile("fares_per_route.csv")
  )
  data.table::fwrite(fare_debug_settings, tmpfile("debug_settings.csv"))

  zip::zip(
    zipfile = file_path,
    files = c(
      normalizePath(tmpfile("global_settings.csv")),
      normalizePath(tmpfile("fares_per_type.csv")),
      normalizePath(tmpfile("fares_per_transfer.csv")),
      normalizePath(tmpfile("fares_per_route.csv")),
      normalizePath(tmpfile("debug_settings.csv"))
    ),
    mode = "cherry-pick"
  )

  return(invisible(file_path))
}


#' Read a fare structure object from a file
#'
#' @param file_path A path pointing to a fare structure with a `.zip`
#'   extension.
#' @param encoding A string. Passed to [data.table::fread()], defaults to
#'   `"UTF-8"`. Other possible options are `"unknown"` and `"Latin-1"`. Please
#'   note that this is not used to re-encode the input, but to enable handling
#'   encoded strings in their native encoding.
#'
#' @return A fare structure object.
#'
#' @family fare structure
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' path <- system.file("extdata/poa/fares/fares_poa.zip", package = "r5r")
#' fare_structure <- read_fare_structure(path)
#'
#' @export
read_fare_structure <- function(file_path, encoding = "UTF-8") {
  checkmate::assert_file_exists(file_path, extension = "zip")
  val_enc <- c("unknown", "UTF-8", "Latin-1")
  checkmate::assert(
    checkmate::check_string(encoding),
    checkmate::check_names(encoding, subset.of = val_enc),
    combine = "and"
  )

  tmpdir <- tempfile("read_fare_structure")
  dir.create(tmpdir)
  zip::unzip(zipfile = file_path, exdir = tmpdir)

  tmpfile <- function(path) file.path(tmpdir, path)

  global_settings <- data.table::fread(
    tmpfile("global_settings.csv"),
    encoding = encoding
  )

  fare_structure <- list()

  fare_structure$max_discounted_transfers <- as.integer(
    global_settings[setting == "max_discounted_transfers"]$value
  )
  fare_structure$transfer_time_allowance <- as.integer(
    global_settings[setting == "transfer_time_allowance"]$value
  )
  fare_structure$fare_cap <- as.numeric(
    global_settings[setting == "fare_cap"]$value
  )

  fare_structure$fares_per_type <- data.table::fread(
    tmpfile("fares_per_type.csv"),
    select = c(
      type = "character",
      unlimited_transfers = "logical",
      allow_same_route_transfer = "logical",
      use_route_fare = "logical",
      fare = "numeric"
    ),
    encoding = encoding
  )

  fare_structure$fares_per_transfer <- data.table::fread(
    file = tmpfile("fares_per_transfer.csv"),
    select = c(
      first_leg = "character",
      second_leg = "character",
      fare = "numeric"
    ),
    encoding = encoding
  )

  fare_structure$fares_per_route <- data.table::fread(
    file = tmpfile("fares_per_route.csv"),
    select = c(
      agency_id = "character",
      agency_name = "character",
      route_id = "character",
      route_short_name = "character",
      route_long_name = "character",
      mode = "character",
      route_fare = "numeric",
      fare_type = "character"
    ),
    encoding = encoding
  )

  debug_options <- data.table::fread(
    tmpfile("debug_settings.csv"),
    encoding = encoding
  )
  fare_structure$debug_settings <- list(
    output_file = debug_options[setting == "output_file"]$value,
    trip_info = debug_options[setting == "trip_info"]$value
  )

  return(fare_structure)
}


#' Assert fare structure
#'
#' Asserts whether the specified fare structure object complies with the
#' structure set in [setup_fare_structure()].
#'
#' @template fare_structure
#'
#' @return Throws and error upon failure and invisibly returns `TRUE` on
#'   success.
#'
#' @keywords internal
assert_fare_structure <- function(fare_structure) {
  # TODO: all GTFS routes must be in fares_per_route
  # TODO: fares_per_route$route_id must be unique
  checkmate::assert_list(fare_structure, any.missing = FALSE)

  if ("type" %in% names(fare_structure)) {
    # this is an R5 built-in fare structure
    return(invisible(TRUE))
  } else {
    # This is an R5R fare structure object
    element_names <- c(
      "max_discounted_transfers",
      "transfer_time_allowance",
      "fare_cap",
      "fares_per_type",
      "fares_per_transfer",
      "fares_per_route",
      "debug_settings"
    )
    checkmate::assert_names(
      names(fare_structure),
      type = "unique",
      must.include = element_names,
      subset.of = element_names
    )

    checkmate::assert_number(fare_structure$max_discounted_transfers, lower = 0)
    checkmate::assert_number(fare_structure$transfer_time_allowance, lower = 0)
    checkmate::assert_number(fare_structure$fare_cap, lower = 0)

    checkmate::expect_data_frame(fare_structure$fares_per_type)
    checkmate::expect_character(
      fare_structure$fares_per_type$type,
      any.missing = FALSE,
      unique = TRUE
    )
    checkmate::expect_logical(
      fare_structure$fares_per_type$unlimited_transfers,
      any.missing = FALSE
    )
    checkmate::expect_logical(
      fare_structure$fares_per_type$allow_same_route_transfer,
      any.missing = FALSE
    )
    checkmate::expect_logical(
      fare_structure$fares_per_type$use_route_fare,
      any.missing = FALSE
    )
    checkmate::expect_numeric(
      fare_structure$fares_per_type$fare,
      any.missing = FALSE,
      lower = 0,
      finite = TRUE
    )

    checkmate::expect_data_frame(fare_structure$fares_per_transfer)
    if (length(names(fare_structure$fares_per_transfer)) > 0) {
      checkmate::expect_character(
        fare_structure$fares_per_transfer$first_leg,
        any.missing = FALSE
      )
      checkmate::assert_names(
        fare_structure$fares_per_transfer$first_leg,
        subset.of = unique(fare_structure$fares_per_type$type)
      )
      checkmate::expect_character(
        fare_structure$fares_per_transfer$second_leg,
        any.missing = FALSE
      )
      checkmate::assert_names(
        fare_structure$fares_per_transfer$second_leg,
        subset.of = unique(fare_structure$fares_per_type$type)
      )
      checkmate::expect_numeric(
        fare_structure$fares_per_transfer$fare,
        any.missing = FALSE,
        lower = 0,
        finite = TRUE
      )
    }

    checkmate::expect_data_frame(fare_structure$fares_per_route)
    checkmate::expect_character(
      fare_structure$fares_per_route$agency_id,
      any.missing = FALSE
    )
    checkmate::expect_character(
      fare_structure$fares_per_route$agency_name,
      any.missing = FALSE
    )
    checkmate::expect_character(
      fare_structure$fares_per_route$route_id,
      any.missing = FALSE
    )
    checkmate::expect_character(
      fare_structure$fares_per_route$route_short_name,
      any.missing = FALSE
    )
    checkmate::expect_character(
      fare_structure$fares_per_route$route_long_name,
      any.missing = FALSE
    )
    checkmate::expect_character(
      fare_structure$fares_per_route$mode,
      any.missing = FALSE
    )
    checkmate::expect_numeric(
      fare_structure$fares_per_route$route_fare,
      any.missing = FALSE,
      lower = 0,
      finite = TRUE
    )
    checkmate::expect_character(
      fare_structure$fares_per_route$fare_type,
      any.missing = FALSE
    )
    checkmate::assert_names(
      fare_structure$fares_per_route$fare_type,
      subset.of = unique(fare_structure$fares_per_type$type)
    )

    debug_elements <- c("output_file", "trip_info")
    checkmate::assert_list(fare_structure$debug_settings)
    checkmate::assert_names(
      names(fare_structure$debug_settings),
      type = "unique",
      must.include = debug_elements,
      subset.of = debug_elements
    )
    checkmate::assert_string(fare_structure$debug_settings$output_file)
    checkmate::assert_string(fare_structure$debug_settings$trip_info)

    return(invisible(TRUE))
  }
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

struc <- setup_fare_structure(r5r_network, base_fare = 5)
tmpfile <- tempfile(pattern = "fare_structure", fileext = ".zip")

tester <- function(fare_structure = struc, file_path = tmpfile) {
  write_fare_structure(fare_structure, file_path)
}

test_that("raises error due to incorrect input types", {
  expect_error(tester(file_path = tempfile(fileext = ".csv")))
})

test_that("return path invisibly", {
  expect_equal(tester(), tmpfile)
})

test_that("debug_info defaults to 'ROUTE' if non-specified", {
  new_tmpfile <- tempfile("fare_structure_test", fileext = ".zip")
  tester(file_path = new_tmpfile)
  expect_true(file.exists(new_tmpfile))
})

test_that("written structure is identical to original", {
  tester()
  written_struc <- read_fare_structure(tmpfile)
  expect_identical(struc, written_struc)
})

# clean cache
r5r::r5r_cache(delete_file = 'all')

path <- system.file("extdata/poa/fares/fares_poa.zip", package = "r5r")
struc <- read_fare_structure(path)

copied_element <- function(element) {
  results <- struc
  results[[element]] <- data.table::copy(struc[[element]])
  results
}

test_that("basic struture is right", {
  expect_error(assert_fare_structure("a"))

  struc_copy <- struc
  struc_copy$max_discounted_transfers <- NULL
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$fare_cap2 <- struc$fare_cap
  expect_error(assert_fare_structure(struc_copy))

  names(struc_copy) <- gsub("fare_cap2", "fare_cap", names(struc_copy))
  expect_error(assert_fare_structure(struc_copy))
})

test_that("max_discounted_transfers is right", {
  struc_copy <- struc
  struc_copy$max_discounted_transfers <- "1"
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$max_discounted_transfers <- -1
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$max_discounted_transfers <- NA
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$max_discounted_transfers <- c(2, 2)
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$max_discounted_transfers <- Inf
  expect_true(assert_fare_structure(struc_copy))
})

test_that("transfer_time_allowance is right", {
  struc_copy <- struc
  struc_copy$transfer_time_allowance <- "1"
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$transfer_time_allowance <- -1
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$transfer_time_allowance <- NA
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$transfer_time_allowance <- c(2, 2)
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$transfer_time_allowance <- Inf
  expect_true(assert_fare_structure(struc_copy))
})

test_that("fare_cap is right", {
  struc_copy <- struc
  struc_copy$fare_cap <- "1"
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$fare_cap <- -1
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$fare_cap <- NA
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$fare_cap <- c(2, 2)
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$fare_cap <- Inf
  expect_true(assert_fare_structure(struc_copy))
})

test_that("fares_per_type is right", {
  struc_copy <- struc
  struc_copy$fares_per_type <- "1"
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_type")
  struc_copy$fares_per_type[, type := as.factor(type)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_type")
  struc_copy$fares_per_type[1, type := NA]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy$fares_per_type[1, type := "RAIL"]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_type")
  struc_copy$fares_per_type[
    ,
    unlimited_transfers := as.factor(unlimited_transfers)
  ]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_type")
  struc_copy$fares_per_type[1, unlimited_transfers := NA]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_type")
  struc_copy$fares_per_type[
    ,
    allow_same_route_transfer := as.factor(allow_same_route_transfer)
  ]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_type")
  struc_copy$fares_per_type[1, allow_same_route_transfer := NA]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_type")
  struc_copy$fares_per_type[, use_route_fare := as.factor(use_route_fare)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_type")
  struc_copy$fares_per_type[1, use_route_fare := NA]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_type")
  struc_copy$fares_per_type[, fare := as.factor(fare)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_type")
  struc_copy$fares_per_type[1, fare := NA]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy$fares_per_type[1, fare := -1]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy$fares_per_type[1, fare := Inf]
  expect_error(assert_fare_structure(struc_copy))
})

test_that("fares_per_transfer is right", {
  struc_copy <- struc
  struc_copy$fares_per_transfer <- "1"
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- struc
  struc_copy$fares_per_transfer <- data.table::data.table(NULL)
  expect_true(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_transfer")
  struc_copy$fares_per_transfer[, first_leg := as.factor(first_leg)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_transfer")
  struc_copy$fares_per_transfer[1, first_leg := NA]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy$fares_per_transfer[1, first_leg := "oie"]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_transfer")
  struc_copy$fares_per_transfer[, second_leg := as.factor(second_leg)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_transfer")
  struc_copy$fares_per_transfer[1, second_leg := NA]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy$fares_per_transfer[1, second_leg := "oie"]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_transfer")
  struc_copy$fares_per_transfer[, fare := as.factor(fare)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_transfer")
  struc_copy$fares_per_transfer[1, fare := NA]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy$fares_per_transfer[1, fare := -1]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy$fares_per_transfer[1, fare := Inf]
  expect_error(assert_fare_structure(struc_copy))
})

test_that("fares_per_route is right", {
  struc_copy <- struc
  struc_copy$fares_per_route <- "1"
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[, agency_id := as.factor(agency_id)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[1, agency_id := NA]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[, agency_name := as.factor(agency_name)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[1, agency_name := NA]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[, route_id := as.factor(route_id)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[1, route_id := NA]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[, route_short_name := as.factor(route_short_name)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[1, route_short_name := NA]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[, route_long_name := as.factor(route_long_name)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[1, route_long_name := NA]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[, mode := as.factor(mode)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[1, mode := NA]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[, route_fare := as.factor(route_fare)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[1, route_fare := NA]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy$fares_per_route[1, route_fare := -1]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy$fares_per_route[1, route_fare := Inf]
  expect_error(assert_fare_structure(struc_copy))

  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[, fare_type := as.factor(fare_type)]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy <- copied_element("fares_per_route")
  struc_copy$fares_per_route[1, fare_type := NA]
  expect_error(assert_fare_structure(struc_copy))
  struc_copy$fares_per_route[1, fare_type := "oie"]
  expect_error(assert_fare_structure(struc_copy))
})

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

tmpfile <- tempfile(fileext = ".csv")

gtfs_eptc <- gtfstools::read_gtfs(
  system.file("extdata/poa/poa_eptc.zip", package = "r5r"),
  encoding = "UTF-8"
)
gtfs_trensurb <- gtfstools::read_gtfs(
  system.file("extdata/poa/poa_trensurb.zip", package = "r5r"),
  encoding = "UTF-8"
)
gtfs <- gtfstools::merge_gtfs(gtfs_eptc, gtfs_trensurb)

tester <- function(r5r_network = get("r5r_network", envir = parent.frame()),
                   base_fare = 5, by = "MODE",
                   debug_path = NULL,
                   debug_info = NULL) {
  setup_fare_structure(
    r5r_network = r5r_network,
    base_fare = base_fare,
    by = by,
    debug_path = debug_path,
    debug_info = debug_info
  )
}

test_that("raises error due to incorrect input types", {
  expect_error(tester(unclass(r5r_network)))

  expect_error(tester(base_fare = "5"))
  expect_error(tester(base_fare = -1))
  expect_error(tester(base_fare = c(2, 3)))
  expect_error(tester(base_fare = NA))

  expect_error(tester(by = 1))
  expect_error(tester(by = c("MODE", "ROUTE")))
  expect_error(tester(by = "bad_by"))

  expect_error(tester(debug_path = 1))
  expect_error(tester(debug_path = tempfile(fileext = ".pdf")))

  expect_error(tester(debug_path = tmpfile, debug_info = c("MODE", "ROUTE")))
  expect_error(tester(debug_path = tmpfile, debug_info = "oie"))
})

test_that("debug_info cannot be non-NULL if debug_path is NULL", {
  expect_error(tester(debug_info = "MODE"))
})

test_that("debug_info defaults to 'ROUTE' if non-specified", {
  struc <- tester(debug_path = tmpfile)
  expect_equal(struc$debug_settings$trip_info, "ROUTE")
})

test_that("outputs a list with correct elements", {
  struc <- tester()

  expect_type(struc, "list")
  expect_identical(
    names(struc),
    c(
      "max_discounted_transfers",
      "transfer_time_allowance",
      "fare_cap",
      "fares_per_type",
      "fares_per_transfer",
      "fares_per_route",
      "debug_settings"
    )
  )

  expect_type(struc$max_discounted_transfers, "integer")
  expect_type(struc$transfer_time_allowance, "integer")
  expect_type(struc$fare_cap, "double")

  expect_s3_class(struc$fares_per_type, "data.table")
  expect_type(struc$fares_per_type$type, "character")
  expect_type(struc$fares_per_type$unlimited_transfers, "logical")
  expect_type(struc$fares_per_type$allow_same_route_transfer, "logical")
  expect_type(struc$fares_per_type$use_route_fare, "logical")
  expect_type(struc$fares_per_type$fare, "double")

  expect_s3_class(struc$fares_per_transfer, "data.table")
  expect_type(struc$fares_per_transfer$first_leg, "character")
  expect_type(struc$fares_per_transfer$second_leg, "character")
  expect_type(struc$fares_per_transfer$fare, "double")

  expect_s3_class(struc$fares_per_route, "data.table")
  expect_type(struc$fares_per_route$agency_id, "character")
  expect_type(struc$fares_per_route$agency_name, "character")
  expect_type(struc$fares_per_route$route_id, "character")
  expect_type(struc$fares_per_route$route_short_name, "character")
  expect_type(struc$fares_per_route$route_long_name, "character")
  expect_type(struc$fares_per_route$mode, "character")
  expect_type(struc$fares_per_route$route_fare, "double")
  expect_type(struc$fares_per_route$fare_type, "character")

  expect_type(struc$debug_settings, "list")
  expect_type(struc$debug_settings$output_file, "character")
  expect_type(struc$debug_settings$trip_info, "character")
})

test_that("output includes all routes from the gtfs", {
  struc <- tester()
  expect_true(all(gtfs$routes$route_id %in% struc$fares_per_route$route_id))
})

test_that("uses the parameter 'by' to fill the structure", {
  struc <- tester(by = "AGENCY_ID")
  expect_true(all(gtfs$agency$agency_id %in% struc$fares_per_type$type))
  expect_true(
    all(gtfs$agency$agency_id %in% struc$fares_per_transfer$first_leg)
  )
  expect_true(
    all(gtfs$agency$agency_id %in% struc$fares_per_transfer$second_leg)
  )

  struc <- tester(by = "AGENCY_NAME")
  expect_true(all(gtfs$agency$agency_name %in% struc$fares_per_type$type))
  expect_true(
    all(gtfs$agency$agency_name %in% struc$fares_per_transfer$first_leg)
  )
  expect_true(
    all(gtfs$agency$agency_name %in% struc$fares_per_transfer$second_leg)
  )

  gtfs_modes <- gtfs$routes$route_type
  gtfs_modes <- ifelse(gtfs_modes == 3, "BUS", "RAIL")
  struc <- tester(by = "MODE")
  expect_true(all(gtfs_modes %in% struc$fares_per_type$type))
  expect_true(all(gtfs_modes %in% struc$fares_per_transfer$first_leg))
  expect_true(all(gtfs_modes %in% struc$fares_per_transfer$second_leg))

  struc <- tester(by = "GENERIC")
  expect_true(struc$fares_per_type$type == "GENERIC")
  expect_true(struc$fares_per_transfer$first_leg == "GENERIC")
  expect_true(struc$fares_per_transfer$second_leg == "GENERIC")
})

test_that("debug info is correctly set", {
  # debug disabled by default
  struc <- tester()
  expect_equal(struc$debug_settings$output_file, "")
  expect_equal(struc$debug_settings$trip_info, "MODE")

  # assigns trip_info as ROUTE when not specified
  struc <- tester(debug_path = tmpfile)
  expect_equal(struc$debug_settings$output_file, tmpfile)
  expect_equal(struc$debug_settings$trip_info, "ROUTE")

  # else assigns what is specified
  struc <- tester(debug_path = tmpfile, debug_info = "MODE_ROUTE")
  expect_equal(struc$debug_settings$output_file, tmpfile)
  expect_equal(struc$debug_settings$trip_info, "MODE_ROUTE")
})

test_that("fare_cap is infinite by default", {
  struc <- tester()
  expect_true(is.infinite(struc$fare_cap))
})

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

path <- system.file("extdata/poa/fares/fares_poa.zip", package = "r5r")

tester <- function(file_path = path, encoding = "UTF-8") {
  read_fare_structure(file_path, encoding)
}

test_that("raises error due to incorrect input types", {
  expect_error(tester(tempfile(fileext = ".csv")))
  expect_error(tester("DESCRIPTION"))
  expect_error(tester(encoding = as.factor("UTF-8")))
  expect_error(tester(encoding = c("UTF-8", "UTF-8")))
  expect_error(tester(encoding = "oie"))
})

test_that("outputs a list with correct elements", {
  struc <- tester()

  expect_type(struc, "list")
  expect_identical(
    names(struc),
    c(
      "max_discounted_transfers",
      "transfer_time_allowance",
      "fare_cap",
      "fares_per_type",
      "fares_per_transfer",
      "fares_per_route",
      "debug_settings"
    )
  )

  expect_type(struc$max_discounted_transfers, "integer")
  expect_type(struc$transfer_time_allowance, "integer")
  expect_type(struc$fare_cap, "double")

  expect_s3_class(struc$fares_per_type, "data.table")
  expect_type(struc$fares_per_type$type, "character")
  expect_type(struc$fares_per_type$unlimited_transfers, "logical")
  expect_type(struc$fares_per_type$allow_same_route_transfer, "logical")
  expect_type(struc$fares_per_type$use_route_fare, "logical")
  expect_type(struc$fares_per_type$fare, "double")

  expect_s3_class(struc$fares_per_transfer, "data.table")
  expect_type(struc$fares_per_transfer$first_leg, "character")
  expect_type(struc$fares_per_transfer$second_leg, "character")
  expect_type(struc$fares_per_transfer$fare, "double")

  expect_s3_class(struc$fares_per_route, "data.table")
  expect_type(struc$fares_per_route$agency_id, "character")
  expect_type(struc$fares_per_route$agency_name, "character")
  expect_type(struc$fares_per_route$route_id, "character")
  expect_type(struc$fares_per_route$route_short_name, "character")
  expect_type(struc$fares_per_route$route_long_name, "character")
  expect_type(struc$fares_per_route$mode, "character")
  expect_type(struc$fares_per_route$route_fare, "double")
  expect_type(struc$fares_per_route$fare_type, "character")

  expect_type(struc$debug_settings, "list")
  expect_type(struc$debug_settings$output_file, "character")
  expect_type(struc$debug_settings$trip_info, "character")
})

test_that("encoding argument works", {
  utf8_struc <- tester()
  latin1_struc <- tester(encoding = "Latin-1")
  expect_false(identical(utf8_struc, latin1_struc))
  expect_false(
    identical(
      utf8_struc$fares_per_route$agency_name,
      latin1_struc$fares_per_route$agency_name
    )
  )
  expect_false(
    identical(
      utf8_struc$fares_per_route$route_long_name,
      latin1_struc$fares_per_route$route_long_name
    )
  )
})

# options(java.parameters = '-Xmx2G')

# library(r5r)
devtools::load_all(".")
library(data.table)
library(tidyverse)

# build transport network
data_path <- system.file("extdata/spo", package = "r5r")
r5r_core <- setup_r5(data_path = data_path, verbose = TRUE, overwrite = TRUE,
                     temp_dir = FALSE)

# load origin/destination points
points <- read.csv(file.path(data_path, "spo_hexgrid.csv"))

# r5r_core$setTravelTimesBreakdown(FALSE)
ttm_flat_walk <- travel_time_matrix(r5r_core,
                            origins = points,
                            destinations = points,
                            mode = c("WALK", "TRANSIT"),
                            max_trip_duration = 60,
                            max_walk_dist = Inf,
                            time_window = 1,
                            percentiles = c(50),
                            verbose = FALSE,
                            progress = TRUE)


ttm_flat_bike <- travel_time_matrix(r5r_core,
                                    origins = points,
                                    destinations = points,
                                    breakdown = FALSE,
                                    mode = c("BICYCLE"),
                                    max_trip_duration = 60,
                                    max_walk_dist = Inf,
                                    time_window = 1,
                                    percentiles = c(50),
                                    verbose = FALSE,
                                    progress = TRUE)

ttm_flat_walk$scenario <- "flat"
ttm_flat_walk$mode <- "walk"
ttm_flat_bike$scenario <- "flat"
ttm_flat_bike$mode <- "bike"

write_csv(ttm_flat_walk, here::here("elevation", "walk_flat.csv"))
write_csv(ttm_flat_bike, here::here("elevation", "bike_flat.csv"))

#' @section Routing algorithm:
#'
#' The [travel_time_matrix()], [expanded_travel_time_matrix()],
#' [arrival_travel_time_matrix()]  and [accessibility()] functions use an
#' `R5`-specific extension to the RAPTOR routing algorithm (see Conway et al.,
#' 2017). This RAPTOR extension uses a systematic sample of one departure per
#' minute over the time window set by the user in the 'time_window' parameter.
#' A detailed description of base RAPTOR can be found in Delling et al (2015).
#' However, whenever the user includes transit fares inputs to these functions,
#' they automatically switch to use an `R5`-specific extension to the McRAPTOR
#' routing algorithm.
#'
#' - Conway, M. W., Byrd, A., & van der Linden, M. (2017). Evidence-based
#' transit and land use sketch planning using interactive accessibility methods
#' on combined schedule and headway-based networks. Transportation Research
#' Record, 2653(1), 45-53. \doi{10.3141/2653-06}
#'
#' - Delling, D., Pajor, T., & Werneck, R. F. (2015). Round-based public
#' transit routing. Transportation Science, 49(3), 591-604.
#' \doi{10.1287/trsc.2014.0534}

#' @param fare_structure A fare structure object, following the convention
#'   set in [setup_fare_structure()]. This object describes how transit fares
#'   should be calculated. Please see the fare structure vignette to
#'   understand how this object is structured:
#'   `vignette("fare_structure", package = "r5r")`.

#' Get most recent JAR file url from metadata
#'
#' Returns the most recent JAR file url from metadata, depending on the version.
#'
#' @param version A string. The version of R5 to be downloaded. When `NULL`, it
#'        defaults to the latest version.
#'
#' @return A url a string.
#'
#' @family support functions
#'
#' @keywords internal
fileurl_from_metadata <- function(version = NULL) { # nocov start

  # R5 version
  if(is.null(version)) {version = r5r_env$r5_jar_version}

  checkmate::assert_string(version)

  metadata <- system.file("extdata/metadata_r5r.csv", package = "r5r")
  metadata <- data.table::fread(metadata)

  # check for invalid 'version' input

  if (!(version %in% metadata$version)) {
    stop(
      "Error: Invalid value to argument 'version'. ",
      "Please use one of the following: ",
      paste(unique(metadata$version), collapse = "; ")
    )
  }

  # check which jar file to download based on the 'version' parameter

  env <- environment()
  metadata <- metadata[version == get("version", envir = env)]
  metadata <- metadata[release_date == max(release_date)]
  url <- metadata$download_path
  return(url)

} # nocov end


check_transit_availability_on_date <- function(r5r_network,
                                               departure_date){ # nocov start

  # check services Available on the departure date
  services <- check_transit_availability(r5r_network, dates = departure_date)
  services_available <- services$pct_active

  if (services_available == 0 | is.na(services_available)) {
    cli::cli_abort("There are no transit services available on the selected
                   departure date: {.val {departure_date}}. Please ensure there
                   is a GTFS in your data path & that the departure date falls
                   within the GTFS calendar."
                   )
    }


  if (services_available < 0.2) {
    cli::cli_alert_warning("Less than 20% of the transit services in the GTFS are running
                   on the selected departure date.")
  }
} # nocov end


#' Initialize Java and Check Version
#'
#' Sets up Java logging for r5r and ensures Java SE Development Kit 21 is installed.
#'
#' @param data_path A character string. The directory where the log file should be saved.
#' @param temp_dir A logical. Whether the jar file should be saved in a temporary
#'   directory. Defaults to `FALSE`.
#' @param verbose A logical. Whether to show informative messages. Defaults to `FALSE`.
#'
#' @return No return value. The function will stop execution with an error if Java 21 is not found.
#'
#' @details
#' This function initializes the Java Virtual Machine (JVM) with a log path for r5r, and checks that
#' the installed Java version is 21. If not, it stops with an informative error message and download links.
#'
#' @family support functions
#'
#' @keywords internal
start_r5r_java <- function(data_path,
                           temp_dir = FALSE,
                           verbose = FALSE) { # nocov start

  log_filename <- "r5r-log.log"
  log_path <- paste0("-DLOG_PATH=", file.path(data_path, log_filename))

  r5_version <- paste0("-DR5_VER=", r5r_env$r5_jar_version)
  r5r_version <- paste0("-DR5R_VER=", utils::packageVersion("r5r"))

  rJava::.jinit(parameters = c(log_path, r5_version, r5r_version))

  ver <- get_java_version()

  if (ver != 21) {
    # Helper that turns plain text into a clickable hyperlink (works in
    # RStudio ≥ 2023.12 and other terminals that support OSC-8 links)
    link <- cli::style_hyperlink

    cli::cli_abort(c(
      "This package requires {.val Java-SE Development Kit 21}.",
      "i" = "Please install JDK 21 from one of the sources below:",
      "*" = link("rJavaEnv",       "https://www.ekotov.pro/rJavaEnv"),
      "*" = link("Eclipse Temurin","https://adoptium.net/temurin"),
      "*" = link("Amazon Corretto","https://aws.amazon.com/corretto"),
      "*" = link("OpenJDK",        "https://jdk.java.net/java-se-ri/21"),
      "*" = link("Oracle",         "https://docs.oracle.com/en/java/javase/21/install/index.html")
    ))
  }

  # if (ver != 21) {
  #   cli::cli_abort(c(
  #     "This package requires {.val Java SE Development Kit 21}.",
  #     "i" = "Please install JDK 21 from one of the sources below:",
  #     "*" = "{.pkg rJavaEnv}: <https://www.ekotov.pro/rJavaEnv>",
  #     "*" = "Eclipse Temurin: <https://adoptium.net/temurin>",
  #     "*" = "Amazon Corretto: <https://aws.amazon.com/corretto>",
  #     "*" = "OpenJDK: <https://jdk.java.net/java-se-ri/21>",
  #     "*" = "Oracle: <https://docs.oracle.com/en/java/javase/21/install/index.html>"
  #   ))
  # }

  # r5r jar
  r5r_jar <- system.file("jar/r5r.jar", package = "r5r")
  rJava::.jaddClassPath(path = r5r_jar)

  # r5r jar
  # check if the most recent JAR release is stored already.
  fileurl <- fileurl_from_metadata( r5r_env$r5_jar_version )
  filename <- basename(fileurl)

  if (is.null(getOption("r5r.r5jar"))) {
    jar_file <- data.table::fifelse(
      temp_dir,
      file.path(tempdir(), filename),
      file.path( r5r_env$cache_dir, filename)
    )

    # If there isn't a JAR already larger than 60MB, download it
    if (checkmate::test_file_exists(jar_file) && file.info(jar_file)$size > r5r_env$r5_jar_size) {
      if (!verbose) message("Using cached R5 version from ", jar_file)
    } else {
      check  <- download_r5(temp_dir = temp_dir, quiet = !verbose)
      if (is.null(check)) { return(invisible(NULL)) }
    }
  } else {
    jar_file <- getOption("r5r.r5jar")
    if (checkmate::test_file_exists(jar_file)) {
      message("Overriding default R5 with option r5r.r5jar, using ", jar_file)
    } else {
      stop(paste0("Cannot find r5 jar \"", jar_file, "\", specified in options(r5r.r5jar)"))
    }
  }

  # R5 jar
  rJava::.jaddClassPath(path = jar_file)
} # nocov end


#' Return a temporary directory path that is unique with every call
#'
#' This is different from the built in tempdir() in that it does not return the same directory within a given runtime. Always returns a unique directory
#'
#' @return Path. Returns the path of the created temporary directory
#'
#' @family support functions
#'
#' @keywords internal
tempdir_unique <- function(){ # nocov start
  output_dir <- tempfile("r5rtemp_")
  if (!dir.create(output_dir)) {
    stop("Failed to create temporary directory.")
  }
  return(output_dir)
} # nocov end


#' Check if there is an elevation data file in `.tif` format in a given data path
#'
#' @param data_path A string pointing to a directory.
#' @return `TRUE` if there is a `.tif` file, `FALSE` otherwise
#'
#' @family support functions
#'
#' @keywords internal
exists_tiff <- function(data_path){ # nocov start

  all_files <- list.files(data_path)
  check <- ifelse(any(data.table::like(all_files, '.tif')), TRUE, FALSE)

  return(check)
} # nocov end


#' Validate OSM IDs returned from Java backend and print warnings
#'
#' Parses a Java-style array string (e.g., \code{"[id1, id2]"}), extracts OSM IDs,
#' and prints a pretty warning if any invalid IDs are found.
#'
#' If no invalid IDs are found (i.e., input is \code{"[]"}), prints nothing.
#'
#' @param bad_ids_string Character. A string formatted as a Java array (e.g., \code{"[id1, id2]"}).
#'
#' @return Warning if necessary.
#'
#' @family support functions
#' @keywords internal
validate_bad_osm_ids <- function(bad_ids_string) {
  # Remove brackets and trim whitespace
  ids <- gsub("^\\[|\\]$", "", bad_ids_string)
  ids <- trimws(ids)

  # Handle empty case: after removing brackets, should be empty string
  if (nchar(ids) > 0) {
    # Optionally, split into vector for pretty printing
    ids_vec <- unlist(strsplit(ids, ",\\s*"))
    cli::cli_alert_warning("Found invalid osm IDs in congestion data: {paste(ids_vec, collapse = ', ')}")
  }
}


#' Save speeds polygon to .geojson temporary file
#'
#' Support function that checks the input of speeds polygon passed to
#' `build_custom_network()` and saves it to a `.geojson` temporary file.
#'
#' @param new_speeds_poly An sf polygon
#'
#' @family Support functions
#'
#' @return The path to a `.geojson` saved as a temporary file.
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#'
#' # read polygons with new speeds
#' congestion_poly <- readRDS(
#'   system.file("extdata/poa/poa_poly_congestion.rds", package = "r5r")
#'   )
#'
#' geojson_path <- r5r:::congestion_poly2geojson(
#'   new_speeds_poly = congestion_poly
#'   )
#'
#' @keywords internal
congestion_poly2geojson <- function(new_speeds_poly){

  # check input class
  checkmate::assert_class(new_speeds_poly, "sf")

  # check input colnames
  checkmate::assert_names(
    x = names(new_speeds_poly),
    must.include = c("poly_id", "scale", "priority", "geometry")
  )

  # Check column types and check input geometry
  checkmate::assert_character(new_speeds_poly$poly_id, any.missing = FALSE)
  checkmate::assert_numeric(new_speeds_poly$scale, any.missing = FALSE)
  checkmate::assert_integer(new_speeds_poly$priority, any.missing = FALSE)
  checkmate::assert_subset(
    x = unique(as.character(sf::st_geometry_type(new_speeds_poly))),
    choices = c("POLYGON", "MULTIPOLYGON"),
    empty.ok = FALSE
  )

  # check input spatial projection
  if (sf::st_crs(new_speeds_poly) != sf::st_crs(4326)) {
    stop(
      "The CRS of parameter `new_speeds` must be WGS 84 (EPSG 4326). ",
      "Please use either sf::set_crs() to set it or ",
      "sf::st_transform() to reproject it."
    )
  }

  # save polygons to temp file
  file_path <- tempfile(
    pattern = 'r5r_congestion_poly',
    fileext = ".geojson"
  )

  sf::st_write(new_speeds_poly, file_path, quiet = TRUE)

  if (file.exists(file_path)) { return(file_path)}
}


#' Save LTS lines to shapefile temporary file
#'
#' Support function that checks the input of LTS lines passed to
#' and saves it to a `.shp` temporary file.
#'
#' @param new_lts_lines An sf LINESTRING or MULTILINESTRING
#'
#' @family Support functions
#'
#' @return The path to a `.shp` saved as a temporary file.
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#'
#' # read lines with new speeds
#' new_lts_lines <- readRDS(
#'   system.file("extdata/poa/poa_ls_lts.rds", package = "r5r")
#'   )
#'
#' shp_path <- r5r:::lts_lines2shp(
#'   new_lts_lines = new_lts_lines
#'   )
#'
#' @keywords internal
lts_lines2shp <- function(new_lts_lines){

  # check input class
  checkmate::assert_class(new_lts_lines, "sf")

  # check input colnames
  checkmate::assert_names(
    x = names(new_lts_lines),
    must.include = c("line_id", "lts", "priority", "geometry")
  )

  # Check column types and check input geometry
  checkmate::assert_character(new_lts_lines$line_id, any.missing = FALSE)
  checkmate::assert_integer(new_lts_lines$lts, any.missing = FALSE, lower = 1, upper = 4)
  checkmate::assert_integer(new_lts_lines$priority, any.missing = FALSE)
  checkmate::assert_subset(
    x = unique(as.character(sf::st_geometry_type(new_lts_lines))),
    choices = c("LINESTRING"),
    empty.ok = FALSE
  )

  # check input spatial projection
  if (sf::st_crs(new_lts_lines) != sf::st_crs(4326)) {
    stop(
      "The CRS of parameter `new_lts_lines` must be WGS 84 (EPSG 4326). ",
      "Please use either sf::set_crs() to set it or ",
      "sf::st_transform() to reproject it."
    )
  }

  # save lines to temp file
  file_path <- tempfile(
    pattern = 'r5r_lts_lines',
    fileext = ".shp"
  )

  sf::st_write(new_lts_lines, file_path, quiet = TRUE)

  if (file.exists(file_path)) { return(file_path)}
}

context("build_network")

# skips tests on CRAN since they require a specific version of java
testthat::skip_on_cran()

path <- system.file("extdata/poa", package = "r5r")

# expected behavior
test_that("build_network - expected behavior", {

  testthat::expect_message( build_network(data_path = path, verbose = F, temp_dir = TRUE) )
  testthat::expect_message( build_network(data_path = path, verbose = T, temp_dir = TRUE) )

  testthat::expect_message( build_network(data_path = path, elevation="TOBLER", temp_dir = TRUE) )

  testthat::expect_is( build_network(data_path = path, temp_dir = TRUE), "r5r_network" )

  # remove files GTFS
  #  file.rename(file.path(path, "poa.zip"), file.path(path, "poa.x"))
  # testthat::expect_message( build_network(data_path = path, verbose = F) )
  #  file.rename(file.path(path, "poa.x"), file.path(path, "poa.zip"))

  # testthat::expect_message(build_network(data_path = path, version='6.4.0', verbose = F, temp_dir = TRUE))

})

# Expected errors
test_that("build_network - expected errors", {

  testthat::expect_error( build_network(data_path = NULL) )
  testthat::expect_error( build_network(data_path = 'a') )
  testthat::expect_error(build_network(data_path = path, verbose = 'a'))
  testthat::expect_error(build_network(data_path = path, temp_dir = 'a'))
  testthat::expect_error(build_network(data_path = path, elevation = 'a'))

  # No OSM data
  r5r_temp <- r5r:::tempdir_unique()
  testthat::expect_error(
    build_network(data_path = r5r_temp)
    )

  })

test_that("'overwrite' parameter works correctly", {

  testthat::expect_error(build_network(path, overwrite = 1))

  # since a network was already created, if overwrite = FALSE it should use it
  testthat::expect_message(
    r5r_network <- build_network(path, verbose = FALSE, temp_dir = TRUE),
    regexp = "Using cached network from"
  )

  # but if overwrite = TRUE, then it should create a new network anyway
  testthat::expect_message(
    r5r_network <- build_network(path, verbose = FALSE, overwrite = TRUE, temp_dir = TRUE),
    regexp = "Finished building network at"
  )

})

test_that("throws error if write access to given dir is denied", {
  # this test only works correctly with unix OSes. not sure how to change
  # permissions from inside R in windows
  skip_if_not(.Platform$OS.type == "unix")

  invisible(file.copy(path, tempdir(), recursive = TRUE))

  tmpdir <- file.path(tempdir(), "poa")

  data_files <- list.files(tmpdir, full.names = TRUE)
  files_to_remove <- data_files[grepl("network|\\.pbf\\.mapdb", data_files)]
  if (length(files_to_remove) > 0) invisible(file.remove(files_to_remove))

  Sys.chmod(tmpdir, "555")

  expect_error(build_network(tmpdir), class = "dir_permission_denied")

  Sys.chmod(tmpdir, "755")
})


# mock test
test_that("throws error if Java is not 21", {

  local_mocked_bindings(
    get_java_version = function(...) 999
  )

  expect_error( r5r:::start_r5r_java(data_path = data_path) )

})

# TO DO: create a mock test
# test_that("throws error due to large geographic extent", {
#
#   my_wrapper <- function(...) {
#     rJava::.jcall(...)
#   }
#
#   local_mocked_bindings(
#     my_wrapper = function(...) "Geographic extent of street layer"
#   )
#
#   # expect_error( r5r:::start_r5r_java(data_path = data_path) )
#
#   build_network(data_path)
#
#   )
# }

test_that("build_network - feeds with errors", {

  # create a feed with an error
  temp_net_dir <- tempfile()
  dir.create(temp_net_dir, mode = "0700")
  gtfs_dir <- file.path(temp_net_dir, "gtfs")
  dir.create(gtfs_dir, mode = "0700")

  # network with medium priority errors: empty frequencies.txt
  # should not prevent network build
  net_medium_dir <- file.path(temp_net_dir, "network_medium")
  dir.create(net_medium_dir, mode = "0700")
  src_net_path <- system.file("extdata/poa", package = "r5r")
  unzip(file.path(src_net_path, "poa_eptc.zip"), exdir = gtfs_dir)
  file.create(file.path(gtfs_dir, "frequencies.txt"))

  # turn back to GTFS
  zip(
    file.path(net_medium_dir, "gtfs.zip"),
    file.path(gtfs_dir, list.files(gtfs_dir)),
    flags = "-j" # don't record directory names
  )

  file.copy(
    file.path(src_net_path, "poa_osm.pbf"),
    file.path(net_medium_dir, "osm.pbf")
  )

  # build network
  testthat::expect_message(
    {net <- build_network(net_medium_dir, temp_dir = TRUE)},
    # current R5 (7.4) records the empty table error twice ?
    regex = "2 errors found in GTFS"
  )

  errors <- get_gtfs_errors(net)

  testthat::expect_equal(
    errors,
    # note that the expected table is dependent on R5 internals. If this test
    # starts failing, check to see if the actual errors have the same meaning as
    # the expected error and update accordingly.
    data.table::data.table(
      # current R5 (7.4) records the empty table error twice ?
      file = "frequencies",
      line = c(0,0),
      type = "EmptyTableError",
      field = NA,
      id = NA,
      priority = "MEDIUM"
    )
  )

  # network with high priority errors: missing stops
  net_high_dir <- file.path(temp_net_dir, "network_high")
  dir.create(net_high_dir, mode = "0700")

  # clean up from medium
  file.remove(file.path(gtfs_dir, "frequencies.txt"))
  trips <- read.csv(file.path(gtfs_dir, "trips.txt"))
  write.csv(
    trips[-1,],
    file.path(gtfs_dir, "trips.txt")
  )

  # turn back to GTFS
  zip(
    file.path(net_high_dir, "gtfs.zip"),
    file.path(gtfs_dir, list.files(gtfs_dir)),
    flags = "-j" # don't record directory names
  )

  file.copy(
    file.path(src_net_path, "poa_osm.pbf"),
    file.path(net_high_dir, "osm.pbf")
  )

  # build network
  testthat::expect_error(
    build_network(net_high_dir, temp_dir = TRUE),
    # current R5 (7.4) records the empty table error twice ?
    regex = "High priority GTFS errors found; network build failed."
  )

  errors <- get_gtfs_errors(net_high_dir)
  testthat::expect_equal(
    errors,
    # note that the expected table is dependent on R5 internals. If this test
    # starts failing, check to see if the actual errors have the same meaning as
    # the expected error and update accordingly.
    data.table::data.table(
      # current R5 short-circuits after high priority error so we only get one
      file = "stop_times",
      line = 2,
      type = "ReferentialIntegrityError",
      field = "trip_id",
      id = NA,
      priority = "HIGH"
    )
  )
})

#' create lines with new LTS for sample data - Porto Alegre
#'
#' Here we simulate that local authorities would have build dedicated lanes
#' along all secondary roads in the city.

library(sf)
library(dplyr)
library(mapview)

# data path
data_path <- system.file("extdata/poa", package = "r5r")

# path to OSM pbf
pbf_path <- paste0(data_path, "/poa_osm.pbf")

# read layer of lines from pbf
allroads <- sf::st_read(
  pbf_path,
  layer = 'lines',
  quiet = TRUE
)

# Filter only road types of interest
roads <- allroads |>
  select(osm_id, highway) |>
  filter(highway %in% "secondary")

head(roads)

# Suppose your sf object is called `lines`
lines_multi <- st_sf(
  line_id = '1',
  lts = 2L,
  priority = 1L,
  geometry = st_combine(roads)
)

mapview::mapview(lines_multi)

saveRDS(lines_multi, 'poa_mls_lts.rds')

# https://gis.stackexchange.com/questions/277219/sf-equivalent-of-r-maptools-packages-spatiallinesmidpoints

get_mid_point <- function(a){
  new_shape <- sf::st_segmentize(x = a, dfMaxLength = 50)
  new_shape <- sfheaders::sf_cast(new_shape, "POINT")

  m_position <- (nrow(new_shape) /2) |> round(digits = 0)
  mid_point <- new_shape[m_position,]
  return(mid_point)
}

m <- get_mid_point(a)



mapview(m)
library(ggplot2)


c <- st_centroid(a)

ss <- st_line_midpoints(a)

ggplot() +
geom_sf(data = new_shape, color='blue')+
  geom_sf(data = a, color='gray')+
  geom_sf(data = m, color='red') +
  geom_sf(data = c, color='green') +
  geom_sf(data = ss, color='orange')


library(bench)

bench::mark(check=F,
    get_mid_point(a),
  st_line_midpoints(sf_lines = a)
)

expression              min  median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc
<bch:expr>          <bch:t> <bch:t>     <dbl> <bch:byt>    <dbl> <int> <dbl>
  1 get_mid_point(a)     25.2ms  27.1ms      36.7     322KB     2.30    16     1
2 st_line_midpoints(… 799.9µs 826.1µs    1139.     12.7KB     0      570     0
                     # ℹ 5 more variables: total_time <bch:tm>, result <list>, memory <list>,

devtools::load_all(".")

library(data.table)
# build transport network
data_path <- system.file("extdata/spo", package = "r5r")
r5r_core <- setup_r5(data_path, overwrite = F)

# load origin/destination points
points <- fread(file.path(data_path, "spo_hexgrid.csv")) |>
  dplyr::sample_n(15)
# points <- read.csv(file.path(data_path, "poa_hexgrid.csv")) %>%
  # dplyr::sample_n(50)


# inputs
departure_datetime <- as.POSIXct("13-05-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S")

r5r_core$setDetailedItinerariesV2(TRUE)
r5r_core$setDetailedItinerariesV2(FALSE)
#
# r5r::assign_mode(mode = c("WALK", "BUS", "SUBWAY", "RAIL"), mode_egress = "WALK", style = "dit")

system.time(
  det2 <- detailed_itineraries(r5r_core,
                               origins = points[id=="89a8100c527ffff"],
                               destinations = points[id=="89a8100c5a3ffff"],
                               mode = c("WALK", "BUS", "SUBWAY", "RAIL"),
                              departure_datetime = departure_datetime,
                              max_walk_dist = 1500,
                              max_trip_duration = 120,
                              suboptimal_minutes = 5,
                              # fare_structure = fare_structure,
                              # max_fare = 10,
                              time_window = 10,
                              all_to_all = T,
                              shortest_path = F,
                              progress = T)
  )


ttm <- expanded_travel_time_matrix(r5r_core,
                                   origins = points,
                                   destinations = points,
                                   mode = c("WALK", "TRANSIT"),
                                   departure_datetime = departure_datetime,
                                   max_walk_dist = 1000,
                                   max_trip_duration = 120,
                                   # fare_structure = fare_structure,
                                   # max_fare = 10,
                                   time_window = 1,
                                   progress = T)

mapview::mapview(det2, zcol = "mode")
mapview::mapview(dplyr::filter(det2, option == 36), zcol = "mode")
mapview::mapview(dplyr::filter(det2,
                               from_id == "beira_rio_stadium",
                               to_id == "bus_central_station",
                               option == 2), zcol = "mode")

r5r::select_mode(c("WALK", "BICYCLE"), "WALK", style = "dit")
set_fare_structure(r5r_core, fare_structure)
r5r_core$setMaxFare(rJava::.jfloat(10.0))

r5r_core$dropFareCalculator()

library("tidyverse")

saveRDS(det, "det_v2.rds")
saveRDS(det2, "det_v1.rds")

library(sf)
library(tidyverse)

det2 %>%
  st_set_geometry(NULL) %>%
  group_by(option) %>%
  summarise(mode = paste(mode, collapse = "|"),
            routes = paste(route, collapse = "|"),
            total_fare = mean(total_fare))

library(mapview)

tn <- transit_network_to_sf(r5r_core)
mapview(points, xcol="lon",ycol="lat", crs=4326) + tn$routes


d <- r5r_core$getTransitServicesByDate("2019-05-13")
d <- r5r::java_to_dt(d)

# Convert a travel time surface to an isochrone.
surface_isochrone <- function (travel_time_surface, cutoffs) {
  checkmate::check_class(travel_time_surface, "travel_time_surface")

  nonzero_cutoffs = cutoffs[cutoffs>0]

  # get the iso, in web mercator pixel space
  bands = isoband::isobands(
    travel_time_surface@west:(travel_time_surface@west + travel_time_surface@width - 1),
    travel_time_surface@north:(travel_time_surface@north + travel_time_surface@height - 1),
    travel_time_surface@matrix,
    rep(0, length(nonzero_cutoffs)), # bands should all start at 0 minutes
    nonzero_cutoffs
  )

  # convert back to lat lon
  bands <- purrr::map(bands, function (b) {
    b$x <- webmercator_pixel_to_lon(b$x, travel_time_surface@zoom)
    b$y <- webmercator_pixel_to_lat(b$y, travel_time_surface@zoom)
    return(b)
  })
  class(bands) <- c("isobands", "iso")

  iso <- data.table::data.table(sf::st_sf(
    isochrone = nonzero_cutoffs,
    polygons = sf::st_make_valid(sf::st_sfc(isoband::iso_to_sfg(bands), crs=4326))
  ))

  iso <- iso[ order(-isochrone), ]
  return(iso)
}

#' Download `R5.jar`
#'
#' Downloads `R5.jar` and saves it locally, inside the package directory.
#'
#' @param version A string. The version of R5 to be downloaded. When `NULL`, it
#'        defaults to the latest version.
#' @param quiet A logical. Whether to show informative messages when downloading
#'   the file. Defaults to `FALSE`.
#' @param force_update A logical. Whether to overwrite a previously downloaded
#'   `R5.jar` in the local directory. Defaults to `FALSE`.
#' @param temp_dir A logical. Whether the file should be saved in a temporary
#'   directory. Defaults to `FALSE`.
#'
#' @return The path to the downloaded file.
#'
#' @family Build network
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' download_r5(temp_dir = TRUE)
#' @export
download_r5 <- function(version = NULL,
                        quiet = FALSE,
                        force_update = FALSE,
                        temp_dir = FALSE) {

  # R5 version
  if (is.null(version)) {version = r5r_env$r5_jar_version}


  # check inputs ----------------------------------------------------------

  checkmate::assert_logical(quiet)
  checkmate::assert_logical(force_update)
  checkmate::assert_logical(temp_dir)


  # set timeout options ---------------------------------------------------

  old_options <- options()
  on.exit(options(old_options), add = TRUE)

  options(timeout = max(600, getOption("timeout")))


  # download R5 jar -----------------------------------------------------

  if (!dir.exists(r5r_env$cache_dir)) {dir.create(r5r_env$cache_dir, recursive = TRUE)}

  file_url <- fileurl_from_metadata(version)
  filename <- basename(file_url)

  jar_file <- data.table::fifelse(
    temp_dir,
    file.path(tempdir(), filename),
    file.path( r5r_env$cache_dir , filename)
  )

  # check if the file exists and is not corrupted
  if (file.exists(jar_file) && file.info(jar_file)$size < r5r_env$r5_jar_size && isFALSE(force_update)) {
    stop(message("R5 Jar file is corrupted. To fix this problem, download it again with 'r5r::download_r5(force_update = TRUE)'"))
  }

  # check if the file exists, and returns its path if it does. otherwise,
  # download it from IPEA's server - if there's no internet connection "fail
  # gracefully" (i.e invisibly returns NULL and outputs a informative message)"
  if (file.exists(jar_file) && (force_update == FALSE)) {
    if (!quiet) message("Using cached R5 version from ", jar_file)
    return(jar_file)
  }

  # download JAR
  message("Downloading R5 jar file to ", jar_file)

  try(silent = TRUE,
      utils::download.file(
        url = file_url,
        destfile = jar_file,
        mode = "wb",
        # method = "curl",
        # extra = "--insecure",
        quiet = quiet
      )
  )

  # try(silent = TRUE,
  #     httr::GET(url=file_url,
  #               if(isFALSE(quiet)){ httr::progress()},
  #               httr::write_disk(jar_file, overwrite = TRUE),
  #               config = httr::config(ssl_verifypeer = FALSE))
  # )


# Halt function if download failed (file must exist and be larger than 60 MB)
if (!file.exists(jar_file) | file.info(jar_file)$size < r5r_env$r5_jar_size) {
  message('Internet connection not working properly.')
  return(invisible(NULL))
  }

  return(jar_file)

}

context("download_r5")

# skips tests on CRAN
testthat::skip_on_cran()

# expected behavior
test_that("download_r5 - expected behavior", {

  testthat::expect_vector(download_r5(force_update = TRUE, temp_dir = TRUE))
  testthat::expect_vector(download_r5(force_update = TRUE, temp_dir = FALSE))
  testthat::expect_vector(download_r5(force_update = FALSE, temp_dir = TRUE))
  testthat::expect_vector(download_r5(force_update = FALSE, quiet = TRUE))

  # testthat::expect_vector(download_r5())
  # file.remove(file.path(.libPaths()[1], "r5r", "jar", "r5r_v4.9.0.jar"))
  # testthat::expect_vector(download_r5(version='4.9.0'))

})

# Expected errors
test_that("download_r5 - expected errors", {

  testthat::expect_error( download_r5(version = "0") )
  testthat::expect_error(download_r5(force_update = 'a'))
  testthat::expect_error(download_r5(quiet = 'a'))
  testthat::expect_error(download_r5(temp_dir = 'a'))

  })

# Allocate memory, load R packages, set paths
options(java.parameters = "-Xmx4G")

library(r5r)
library(sf)
library(data.table)

library(dplyr)

# Locate data path
data_path = system.file("extdata/poa", package = "r5r")
gtfs = gtfstools::read_gtfs( file.path(data_path, 'poa_eptc.zip'))

# Select a specific route
routes_data = gtfs$routes
stops_data = gtfs$stops
stop_times_data = gtfs$stop_times
trips_data = gtfs$trips
calendar_dates_data = gtfs$calendar_dates
SELECTED_ROUTE_ID = routes_data$route_id[5]  # Result is route_id ="654"

# Select an available trip as an example
stop_sequence_for_trip = trips_data %>%
  filter(route_id == SELECTED_ROUTE_ID) %>%
  slice(1) %>%
  left_join(stop_times_data, by = "trip_id") %>%
  left_join(stops_data, by = "stop_id") %>%
  arrange(stop_sequence) %>%
  select(stop_sequence, stop_id, stop_name, stop_lat, stop_lon,
         arrival_time, departure_time)

# Query information for the selected trip
temp1_record_selected_trip = trips_data %>% filter(route_id == SELECTED_ROUTE_ID) %>%  slice(1)   # Result is trip_id = 654-1@1#1240
temp2_record_selected_trip_stops = stop_times_data[stop_times_data$trip_id == temp1_record_selected_trip$trip_id,]  # Operation time: 12:40:00 - 13:25:00

# Rename columns
names(stop_sequence_for_trip)[2] = 'id' # stop_id
names(stop_sequence_for_trip)[4] = 'lat' # stop_lat
names(stop_sequence_for_trip)[5] = 'lon' # stop_lon

# Select Stop No.1 → No.40 as the OD pair
A_origin = stop_sequence_for_trip[1,]
A_stop_id = A_origin$id
B_destination = stop_sequence_for_trip[40,]
B_stop_id = B_destination$id

r5r_network = build_network(data_path = data_path)

# Service schedule
calendar_date_654 = gtfs$calendar_dates[calendar_dates_data$service_id %in% temp1_record_selected_trip$service_id,]

# Set assumed departure time and search for feasible routes
T1 = as.POSIXct("2019-06-18 12:35:00",tz = 'America/Sao_Paulo')

# T1 = as.POSIXct(
#   "18-06-2019 12:35:00",
#   format = "%d-%m-%Y %H:%M:%S"
#   )

# Q1: Route 654 cannot be found when setting time between 12:35:00 - 12:39:00
# Q2: Why can Route 654 be found when setting time between 12:40:00 - 12:50:00? Why do we get different departure times?
# Known trip duration is 12:40:00 - 13:25:00
# Q3: Route 654 can be found when setting time between 12:30:00 - 12:34:00, but the departure time is sometimes unstable

A_B_trip = detailed_itineraries(r5r_network,
                                origins = A_origin,
                                destinations = B_destination,
                                mode = c("WALK", "TRANSIT"),
                                departure_datetime = T1,
                                max_walk_time = 120,
                                max_trip_duration = 120,
                                time_window = 5,
                                # suboptimal_minutes = 15,
                                shortest_path = T)   # Set to FALSE to find Route 654

A_B_trip
plot(A_B_trip)

# rafa ---------------------------------

library(gtfstools)
library(mapview)
library(sfheaders)


A_origin_sf <- sfheaders::sf_point(A_origin, x="lon", y="lat", keep = T)
B_destination_sf <- sfheaders::sf_point(B_destination, x="lon", y="lat", keep = T)

route_sf <- gtfstools::filter_by_route_id(gtfs, route_id = SELECTED_ROUTE_ID) |>
  gtfstools::convert_shapes_to_sf()


mapview(A_B_trip, zcol = "option" ) +
  mapview(route_sf, col.regions="orange") +
  mapview(A_origin_sf, col.regions="green") +
  mapview(B_destination_sf, col.regions="red")

#> Travel-time results are currently limited to 4 million destinations,
#> and 16 million total origin-destination pairs. Path detail results are currently
#> limited to 5000 destinations




link <- 'https://ftp.ibge.gov.br/Trabalho_e_Rendimento/Pesquisa_Nacional_por_Amostra_de_Domicilios_continua/Trimestral/Microdados/2023/PNADC_032023.zip'
file <- basename(link)


tic()
httr::GET(url = link,
          # httr::timeout(10),
          httr::progress(),
          httr::write_disk(file, overwrite = T))
toc()


tic()
link |>
  httr2::request() |>
  httr2::req_progress() |>
  httr2::req_perform(path = file)
toc()



# devtools::install_github("ipeaGIT/r5r", subdir = "r-package", force=T)
options(java.parameters = '-Xmx10G')
library(sf)
library(data.table)
library(magrittr)
library(roxygen2)
library(devtools)
library(usethis)
library(profvis)
library(dplyr)
library(mapview)
library(covr)
library(testthat)
library(ggplot2)
library(checkmate)
library(geobr)
library(tictoc)
library(mapview)
mapviewOptions(platform = 'leafgl')


# utils::remove.packages('r5r')
# devtools::install_github("ipeaGIT/r5r", subdir = "r-package", ref = 'detach_r5_codebase')
# library(r5r)



############## issue 281 - fare calculator not working with Sao Paulo example ------------------------------

# https://github.com/ipeaGIT/r5r/blob/8101578e5178333cceb969ab5efd040755c48afa/r-package/R/detailed_itineraries.R#L134



##################################################### POA
library(data.table)
library(r5r)
library(dplyr)

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path = data_path)

# load origin/destination points
points <- read.csv(file.path(data_path, "poa_hexgrid.csv"))[1:5,]

# load fare structure object
fare_structure_path <- system.file(
  "extdata/poa/fares/fares_poa.zip",
  package = "r5r"
)
fare_structure <- read_fare_structure(fare_structure_path)

departure_datetime <- as.POSIXct(
  "13-05-2019 14:00:00",
  format = "%d-%m-%Y %H:%M:%S"
)

pf_10 <- pareto_frontier(
  r5r_core,
  origins = points,
  destinations = points,
  mode = c("WALK", "TRANSIT"),
  time_window = 30,
  departure_datetime = departure_datetime,
  fare_structure = fare_structure,
  fare_cutoffs = c(4.5, 4.8, 9, 9.3, 9.6),
  progress = T,
  draws_per_minute = 10
)

pf_05 <- pareto_frontier(
  r5r_core,
  origins = points,
  destinations = points,
  mode = c("WALK", "TRANSIT"),
  time_window = 30,
  departure_datetime = departure_datetime,
  fare_structure = fare_structure,
  fare_cutoffs = c(4.5, 4.8, 9, 9.3, 9.6),
  progress = T,
  draws_per_minute = 5
)
stop_r5(r5r_core)

head(pf_10)
head(pf_05)


identical(pf_10$from_id, pf_05$from_id)
identical(pf_10$to_id, pf_05$to_id)
identical(pf_10$travel_time, pf_05$travel_time)

pf_10$travel_time - pf_05$travel_time

dplyr::all_equal(pf_10, pf_05)

pf_10[, id := paste(from_id, to_id)]
pf_05[, id := paste(from_id, to_id)]

df <- left_join(pf_10[,.(id, travel_time,monetary_cost)],
                pf_05[,.(id, travel_time,monetary_cost)], by=c('id', 'travel_time', 'monetary_cost'))

setDT(df)
df[, diff_time := travel_time.x - travel_time.y]
df[, diff_cost := monetary_cost.x - monetary_cost.y]




##################################################### SPO

library(r5r)

# build transport network
data_path <- system.file("extdata/spo", package = "r5r")
r5r_core <- setup_r5(data_path = data_path)

# load origin/destination points
points <- read.csv(file.path(data_path, "spo_hexgrid.csv"))[1:5,]

# load fare structure object
fare_structure_path <- system.file(
  "extdata/poa/fares/fares_poa.zip",
  package = "r5r"
)
fare_structure <- read_fare_structure(fare_structure_path)

departure_datetime <- as.POSIXct(
  "13-05-2019 14:00:00",
  format = "%d-%m-%Y %H:%M:%S"
)

detailed_itineraries(r5r_core = r5r_core,
                            origins = points,
                            destinations = points,
                     mode = c("WALK", "TRANSIT"),
                     departure_datetime = departure_datetime,
                            suboptimal_minutes = 8,
                            shortest_path = FALSE)


pf_10 <- pareto_frontier(
  r5r_core,
  origins = points,
  destinations = points,
  mode = c("WALK", "TRANSIT"),
  time_window = 30,
  departure_datetime = departure_datetime,
  fare_structure = fare_structure,
  fare_cutoffs = c(4.5, 4.8, 9, 9.3, 9.6),
  progress = T,
  draws_per_minute = 10
)

pf_05 <- pareto_frontier(
  r5r_core,
  origins = points,
  destinations = points,
  mode = c("WALK", "TRANSIT"),
  time_window = 30,
  departure_datetime = departure_datetime,
  fare_structure = fare_structure,
  fare_cutoffs = c(4.5, 4.8, 9, 9.3, 9.6),
  progress = T,
  draws_per_minute = 5
)
stop_r5(r5r_core)

head(pf_10)
head(pf_05)


identical(pf_10$from_id, pf_05$from_id)
identical(pf_10$to_id, pf_05$to_id)
identical(pf_10$travel_time, pf_05$travel_time)

pf_10$travel_time - pf_05$travel_time

dplyr::all_equal(pf_10, pf_05)

pf_10[, id := paste(from_id, to_id)]
pf_05[, id := paste(from_id, to_id)]

df <- left_join(pf_10[,.(id, travel_time,monetary_cost)],
                pf_05[,.(id, travel_time,monetary_cost)], by=c('id', 'travel_time', 'monetary_cost'))

setDT(df)
df[, diff_time := travel_time.x - travel_time.y]
df[, diff_cost := monetary_cost.x - monetary_cost.y]










identical(small_draws$from_id, big_draws$from_id)
identical(small_draws$to_id, big_draws$to_id)
identical(small_draws$percentile, big_draws$percentile)
identical(small_draws$travel_time, big_draws$travel_time)
identical(small_draws$monetary_cost, big_draws$monetary_cost)

v$travel_time - big_draws$travel_time

dplyr::all_equal(small_draws, big_draws)

small_draws[, id := paste(from_id, to_id)]
big_draws[, id := paste(from_id, to_id)]

df <- left_join(small_draws[,.(id, travel_time,monetary_cost)],
                big_draws[,.(id, travel_time,monetary_cost)], by=c('id', 'travel_time', 'monetary_cost'))

setDT(df)
df[, diff_time := travel_time.x - travel_time.y]
df[, diff_cost := monetary_cost.x - monetary_cost.y]



############## get_all_od_combinations ------------------------------

library(r5r)
library(data.table)

points <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))

origins = destinations = points

# function
get_all_od_combinations <- function(origins, destinations){

        # cross join to get all possible id combinations
        base <- CJ(origins$id, destinations$id, unique = TRUE)

        # rename df
        setnames(base, 'V1', 'id_orig')
        setnames(base, 'V2', 'id_dest')

        # bring spatial coordinates from origin and destination
        base[origins, on=c('id_orig'='id'), c('lon_orig', 'lat_orig') := list(i.lon, i.lat)]
        base[destinations, on=c('id_dest'='id'), c('lon_dest', 'lat_dest') := list(i.lon, i.lat)]

        return(base)
        }

# example
origins <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))[1:800,]
destinations <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))[400:1200,]

df <- get_all_od_combinations(origins, destinations)





##### TESTS isochrone ------------------------

library(profvis)


# allocate RAM memory to Java
options(java.parameters = "-Xmx8G")

library(r5r)
library(ggplot2)
library(mapview)
mapviewOptions(platform = 'leafgl')

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path = data_path)

# load origin/point of interest
origin <- read.csv(file.path(data_path, "poa_hexgrid.csv"))[700,]

departure_datetime <- as.POSIXct("13-05-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S")


###### with no destinations input

iso <- isochrone(r5r_core,
                 origin = origin,
                 mode = c("WALK", "TRANSIT"),
                 departure_datetime = departure_datetime,
                 cutoffs = c(0, 15, 30, 45, 60)
                 #, sample_size = .8
                 )
plot(iso['isochrone'])
head(iso)


profvis({
  iso <- isochrone(r5r_core,
                   origin = origin,
                   mode = c("WALK", "TRANSIT"),
                   departure_datetime = departure_datetime,
                   cutoffs = c(0, 15, 30, 45, 60, 75, 90, 120)
  )
})

# streets <- r5r::street_network_to_sf(r5r_core)

ggplot() +
  #  geom_sf(data=streets$edges, color='gray', alpha=.5) +
  geom_sf(data=iso, aes(fill= isochrone), alpha=.5) +
  scale_fill_viridis_c(direction = -1)

mapview(iso, z = 'isochrone')


##### with polygons

# prep grid with destinations
dest_points <- read.csv(file.path(data_path, "poa_hexgrid.csv"))
grid <- h3jsr::cell_to_polygon(input = dest_points$id, simple = FALSE)
grid$id <- dest_points$id


iso2 <- isochrone(r5r_core,
                 origin = origin,
                 destinations = grid,
                 mode = c("transit"),
                 departure_datetime = departure_datetime,
                 cutoffs = seq(10, 100, 10)
                 )

head(iso2)

isocrhone

ggplot() +
  #  geom_sf(data=streets$edges, color='gray', alpha=.5) +
  geom_sf(data=iso2, aes(fill= isochrone), alpha=.5) +
  scale_fill_viridis_d(direction = -1)

mapview(iso2, z = 'isochrone')











##### set_road_speed ------------------------
# https://github.com/ipeaGIT/r5r/issues/187

library(jsonlite)
library(data.table)
library(r5r)

# build network
points <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))[c(1:100, 500:600),]
path <- system.file("extdata/poa", package = "r5r")

path <- 'R:/Dropbox/git/r5r/r-package/inst/extdata/poa'
r5r_core <- setup_r5(data_path = path, verbose = F, overwrite = T)

tt <- r5r::travel_time_matrix(r5r_core = r5r_core, origins = points,destinations = points, mode = 'car')
# fwrite(tt, 'tt_test.csv')

tt1 <- fread('tt_default.csv')

df=merge(tt, tt1, by=c('from_id','to_id'))
df[, d := travel_time.x - travel_time.y]
summary(df$d)






# cria json so com vars de input
# criar r5r_core, depois atualiza velocidades

# default config
speeds_js <- r5r_core$defaultBuildConfig()
cat(speeds_js)


a <- sub('"defaultSpeed" ://d+' ,'aaa', speeds_js)
cat(a)


# replace primary speed
speeds_js2 <- gsub('"primary" : 45', '"primary" : 111', speeds_js)


cat(speeds_js2)

speeds_df <- fromJSON(speeds_js, simplifyDataFrame = T)



x <- toJSON(speeds_df)
cat(x)





set_road_speed <- function( # path,
                            motorway = NULL
                           , motorway_link = NULL
                           , trunk = NULL
                           , trunk_link = NULL
                           , primary = NULL
                           , primary_link = NULL
                           , secondary = NULL
                           , secondary_link = NULL
                           , tertiary = NULL
                           , tertiary_link = NULL
                           , living_street = NULL
                           , pedestrian = NULL
                           , residential = NULL
                           , unclassified = NULL
                           , service = NULL
                           , track = NULL
                           , road = NULL
                           , defaultSpeed = NULL){
speeds_df <- data.frame(
                  'motorway' =      ifelse(is.null(motorway), NA, motorway)
                , 'motorway_link' = ifelse(is.null(motorway_link), NA, motorway_link)
                , 'trunk' =         ifelse(is.null(trunk), NA, trunk)
                , 'trunk_link' =    ifelse(is.null(trunk_link), NA, trunk_link)
                , 'primary' =       ifelse(is.null(primary), NA, primary)
                , 'primary_link' =  ifelse(is.null(primary_link), NA, primary_link)
                , 'secondary' =     ifelse(is.null(secondary), NA, secondary)
                , 'secondary_link' =ifelse(is.null(secondary_link), NA, secondary_link)
                , 'tertiary' =      ifelse(is.null(tertiary), NA, tertiary)
                , 'tertiary_link' = ifelse(is.null(tertiary_link), NA, tertiary_link)
                , 'living_street' = ifelse(is.null(living_street), NA, living_street)
                , 'pedestrian' =     ifelse(is.null(pedestrian), NA, pedestrian)
                , 'residential' =   ifelse(is.null(residential), NA, residential)
                , 'unclassified' =  ifelse(is.null(unclassified), NA, unclassified)
                , 'service' =       ifelse(is.null(service), NA, service)
                , 'track' =         ifelse(is.null(track), NA, track)
                , 'road' =          ifelse(is.null(road), NA, road)
                , 'defaultSpeed' =  ifelse(is.null(defaultSpeed), NA, defaultSpeed))

# remove columns with missing value
speeds_df <- speeds_df[ , apply(speeds_df, 2, function(x) !any(is.na(x)))]

return(speeds_df)
}

speeds_df <- set_road_speed(primary = 666,
                            secondary = 666,
                            trunk = 666,
                            road = 666,
                            defaultSpeed = 666)


# retrieve default speed
dflt_speed <- speeds_df$defaultSpeed
speeds_df$defaultSpeed <- NULL

my_list <- list( speeds = list ( units = 'km/h',
                                 values= list(speeds_df),
                                 defaultSpeed = dflt_speed
                                 ) )

my_list <- jsonlite::toJSON(my_list, pretty = TRUE)

# remove brackets
my_list <- gsub("\\[|\\]", '', my_list)


write(my_list,file = paste0(path, '/build-config.json'))


# save json
jsonlite::write_json(my_list, path = 'build-config.json')

66666

speeds_js <- toJSON(speeds_df, pretty=TRUE)
cat(speeds_js)

# export json to path file
jsonlite::write_json(speeds_js, path = paste0('build-configsssss.json'))





cat(a)

speeds_df <- as.data.frame(fromJSON(a))
x <- toJSON(speeds_df, pretty=TRUE)
cat(x)

a <- jsonlite::read_json('build-config.json')
jsonlite::write_json(cat(a), path = 'build-config2.json')

writeLines(a,con  = 'build-config3.json')










##### downloads ------------------------
library(ggplot2)
library(dlstats)
library(data.table)


x <- cran_stats(c('geobr', 'geocodebr', 'flightsbr', 'enderecobr'))

x <- cran_stats(c('r5r', 'otpr', 'opentripplanner', 'gtfsrouter','dodgr'))
#x <- cran_stats(c('geobr', 'aopdata', 'flightsbr'))

 if (!is.null(x)) {
         head(x)
         ggplot(x, aes(end, downloads, group=package, color=package)) +
                 geom_line() + geom_point(aes(shape=package))
 }

setDT(x)

x[, .(total = sum(downloads)) , by=package][order(total)]

x[ start > as.Date('2022-01-01'), .(total = sum(downloads)) , by=package][order(total)]

xx <- x[package=='r5r',]

ggplot() +
  geom_line(data=x, aes(x=end, y=downloads, color=package))


library(cranlogs)

a <- cran_downloads( package = c("r5r"), from = "2020-01-01", to = "last-day")

a
ggplot() +
  geom_line(data=a, aes(x=date, y=count, color=package))







library(ggplot2)
library(dlstats)
library(cranlogs)


dl <- cran_stats(c('r5r'))
cl <- cran_downloads( package = c("r5r", "opentripplanner"), from = "2020-01-01", to = "2023-02-01")

sum(dl$downloads)
sum(cl$count)




ggplot() +
 # geom_line(data=dl, aes(x=end, y=downloads), color='blue') +
  geom_line(data=cl, aes(x=date, y=count), color='red')


ggplot() +
  # geom_line(data=dl, aes(x=end, y=downloads), color='blue') +
  geom_line(data=cl, aes(x=date, y=count, color=package))

##### HEX sticker ------------------------

# load origin/destination points
 points <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))
 points_sf <- sfheaders::sf_multipoint(points, x='lon', y='lat', multipoint_id = 'id')
 points_sf2 <- sf::st_as_sf(points, coords = c("lon", "lat"))

 data_path <- system.file("extdata/poa", package = "r5r")
 points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))
 a <-  sf::st_as_sf(points, coords = c("lon", "lat"))



 box <- st_as_sfc( st_bbox(points_sf), crs=st_crs(points_sf) )
 box <- st_sf(box)
 st_crs(box) <- 4674

# get hex
 hex_ids <- h3jsr::polyfill(box, res = 7, simple = FALSE)

 # pass the h3 ids to return the hexagonal grid
 hex_grid <- unlist(hex_ids$h3_polyfillers) %>%
         h3jsr::h3_to_polygon(simple = FALSE) %>%
         rename(id_hex = h3_address) %>%
         st_sf()

 selected_points <- subset(points_sf,
                           id %in% c('89a90129d97ffff',
                                     '89a90129977ffff',
                                     '89a90128a83ffff'))

 origin <- subset(points_sf, id == '89a90129977ffff')
 destinations <- subset(points_sf, id %like% c('89a901299'))

#  selected_points2 <- subset(points_sf,
#                            id %like% c('89a901299'))
#
# mapview(hex_grid) + selected_points + selected_points2
#
#
#
# # 89a90129d97ffff >> para direita
# # 89a90129977ffff ?? pra baixo
# # 89a90128a83ffff ?? nordeste







# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path = data_path)


# input
mode = c('WALK', 'TRANSIT')
departure_time <- "14:00:00"
trip_date <- "2019-03-15"
max_street_time <- 30000L

df <- detailed_itineraries(r5r_core,
                    origins = origin,
                    destinations = destinations,
                    mode = 'WALK',
                    trip_date,
                    departure_time,
                    max_street_time)





# load origin/destination points
street_net <- street_network_to_sf(r5r_core)
mapview(street_net) + points_sf




 # plot
 ggplot() +
         geom_sf(data = df, color='gray1', alpha=1) +
         theme_void()



##### Coverage ------------------------
library(covr)
library(testthat)
# library(r5r)
Sys.setenv(NOT_CRAN = "true")


# each function separately
covr::function_coverage(fun=r5r::download_r5, test_file("tests/testthat/test-download_r5.R"))
covr::function_coverage(fun=r5r::setup_r5, test_file("tests/testthat/test-setup_r5.R"))
a <- covr::function_coverage(fun=r5r::travel_time_matrix, test_file("tests/testthat/test-travel_time_matrix.R"))
a <- covr::function_coverage(fun=r5r::isochrone, test_file("tests/testthat/test-isochrone.R"))
covr::function_coverage(fun=r5r::detailed_itineraries, test_file("tests/testthat/test-detailed_itineraries.R"))
a <- covr::function_coverage(fun=r5r::expanded_travel_time_matrix, test_file("tests/testthat/test-expanded_travel_time_matrix.R"))
a <- covr::function_coverage(fun=r5r::pareto_frontier, test_file("tests/testthat/test-pareto_frontier.R"))
a <- covr::function_coverage(fun=r5r::accessibility, test_file("tests/testthat/test-accessibility.R"))
a <- covr::function_coverage(fun=r5r::build_custom_network, test_file("tests/testthat/test-build_custom_network.R"))

a <- covr::function_coverage(fun=r5r:::set_verbose, test_file("tests/testthat/test-set_verbose.R"))


covr::function_coverage(fun=r5r::street_network_to_sf, test_file("tests/testthat/test-street_network_to_sf.R"))
covr::function_coverage(fun=r5r::transit_network_to_sf, test_file("tests/testthat/test-transit_network_to_sf.R"))


a <- covr::function_coverage(fun=r5r::r5r_cache, test_file("tests/testthat/test-z_r5r_cache.R"))


covr::function_coverage(fun=r5r::set_max_walk_distance, test_file("tests/testthat/test-utils.R"))
covr::function_coverage(fun=r5r::posix_to_string, test_file("tests/testthat/test-utils.R"))
covr::function_coverage(fun=r5r::assert_points_input, test_file("tests/testthat/test-utils.R"))

covr::function_coverage(fun=tobler_hiking, test_file("tests/testthat/test-elevation_utils.R"))
# covr::function_coverage(fun=apply_elevation, test_file("tests/testthat/test-elevation_utils.R"))


covr::function_coverage(fun=r5r::find_snap, test_file("tests/testthat/test-find_snap.R"))


covr::function_coverage(fun=r5r::select_mode, test_file("tests/testthat/test-utils.R"))
covr::function_coverage(fun=r5r::set_verbose, test_file("tests/testthat/test-utils.R"))
covr::function_coverage(fun=r5r::set_n_threads, test_file("tests/testthat/test-utils.R"))
covr::function_coverage(fun=r5r::set_speed, test_file("tests/testthat/test-utils.R"))

# nocov start

# nocov end

# the whole package
Sys.setenv(NOT_CRAN = "true")
r5r_cov <- covr::package_coverage(path = ".", type = "tests")
r5r_cov

covr::report()

zeroCov <- covr::zero_coverage(a)






##### Profiling function ------------------------
# p <-   profvis( update_newstoptimes("T2-1@1#2146") )
#
# p <-   profvis( b <- corefun("T2-1") )





# checks spelling
library(spelling)
devtools::spell_check(pkg = ".", vignettes = TRUE, use_wordlist = TRUE)

# Update documentation
devtools::document(pkg = ".")


# Write package manual.pdf
system("R CMD Rd2pdf --title=Package gtfs2gps --output=./gtfs2gps/manual.pdf")
# system("R CMD Rd2pdf gtfs2gps")


# find unicode errors
tools::showNonASCIIfile(file = './man/accessibility.Rd')
tools::showNonASCIIfile(file = './man/roxygen/templates/mcraptor_algorithm_section.R')
tools::showNonASCIIfile(file = './man/roxygen/templates/raptor_algorithm_section.R')
tools::showNonASCIIfile(file = './vignettes/references.json')

functions <- list.files(path = './R', all.files = T, recursive = T, full.names = T)
lapply(X=functions, FUN = tools::showNonASCIIfile)

docs <- list.files(path = './man', all.files = T, recursive = T, full.names = T)
lapply(X=docs, FUN = tools::showNonASCIIfile)


### CMD Check ----------------
# Check package errors
library(tictoc)
library(beepr)

# run only the tests
testthat::test_local()


# LOCAL
utils::remove.packages('r5r')
jar_dir <- tools::R_user_dir("r5r", which = "cache")
unlink(jar_dir, recursive = TRUE)

tictoc::tic()
Sys.setenv(NOT_CRAN = "true")
devtools::check(pkg = ".",  cran = FALSE, env_vars = c(NOT_CRAN = "true"))
tictoc::toc()
beepr::beep()


# CRAN
tictoc::tic()
Sys.setenv(NOT_CRAN = "false")
devtools::check(pkg = ".",  cran = TRUE, env_vars = c(NOT_CRAN = "false"))
tictoc::toc()
beepr::beep()


devtools::check_win_release(pkg = ".")

# devtools::check_win_oldrelease()
# devtools::check_win_devel()


beepr::beep()



tictoc::tic()
devtools::check(pkg = ".",  cran = TRUE, env_vars = c(NOT_CRAN = "false"))
tictoc::toc()


# extrachecks -----------------
#' https://github.com/JosiahParry/extrachecks
#' remotes::install_github("JosiahParry/extrachecks")

library(extrachecks)
extrachecks::extrachecks()



# submit to CRAN -----------------
usethis::use_cran_comments('teste 2222, , asdadsad')

urlchecker::url_check()
devtools::check(remote = TRUE, manual = TRUE)
devtools::check_win_release()
devtools::check_mac_release()
rhub::check_for_cran(show_status = FALSE)


devtools::submit_cran()


# build binary -----------------
system("R CMD build . --resave-data") # build tar.gz


mode_list
mode_list <- list('direct_modes' = "WALK",
                  'transit_mode' =  "",# "TRANSIT;TRAM;SUBWAY;RAIL;BUS;FERRY;CABLE_CAR;GONDOLA;FUNICULAR",
                  'access_mode'  = "WALK",
                  'egress_mode'  = "CAR")



### pkgdown: update website ----------------

# Run to build the website
pkgdown::build_site()



#### argentina buenos aires -----

options(java.parameters = "-Xmx20G")

library(r5r)
library(osmextract)

library(sf)
library(ggplot2)
library(here)
library(data.table)
library(magrittr)


# create subdirectories "data" and "img"
dir.create(here::here("data"))
        # dir.create(here::here("img"))


# get city boundaries
city <- 'Buenos Aires'
# city_code <- lookup_muni(name_muni = city)
# city_boundary <- read_municipality(code_muni = city_code$code_muni, simplified = F)


# define city center
city_center_df <- data.frame(id='center', lon=-43.182811, lat=-22.906906)
city_center_sf <- sfheaders::sfc_point(obj = city_center_df, x='lon', y='lat')
st_crs(city_center_sf) <- 4326

# get OSM data
city_match <- osmextract::oe_match("Buenos Aires", quiet = TRUE)

osmextract::oe_download(file_url = city_match$url,
                        file_size = city_match$file_size,
                        provider = 'a',
                        download_directory = here::here("data"), force_download = T)


# build routing network
r5r_core <- r5r::setup_r5(data_path = here::here("data"), verbose = FALSE)

# get street network as sf
street_network <- r5r::street_network_to_sf(r5r_core)

# drop network outside our buffer
edges_buff <- street_network$edges[buff, ] %>% st_intersection(., buff)
vertices_buff <- street_network$vertices[buff, ] %>% st_intersection(., buff)
city_boundary_buff <- st_intersection(city_boundary, buff)
plot(city_boundary_buff)

# add id to vertices
vertices_buff$id <- vertices_buff$index


# calculate travel times to city center
tt <- r5r::travel_time_matrix(r5r_core,
                              origins = vertices_buff,
                              destinations = city_center_df,
                              mode = 'walk')



library(tibble)
library(ggplot2)

df <- tribble(~option, ~modes,       ~time, ~cost,
              1,      'Walk',         50,   0,
              2,      'Bus',          35,   3,
              3,      'Bus + Bus',    29,   5,
              4,      'Subway',       20,   6,
              5,      'Bus + Subway', 15,   8)


# data.frame
df <- structure(list(option = c(1, 2, 3, 4, 5),
                     modes = c("Walk", "Bus","Bus + Bus", "Subway", "Bus + Subway"),
                     time = c(50, 35, 29, 20, 15),
                     cost = c(0, 3, 5, 6, 8)), class = "data.frame",
                row.names = c(NA, -5L))

# figure
ggplot(data=df, aes(x=cost, y=time, label = modes)) +
        geom_step(linetype = "dashed") +
        geom_point() +
        geom_text(color='gray30', hjust = -.2, nudge_x = 0.05, angle = 45) +
        # labs(y="Travel time in minutes", x="Travel cost in R$") +
        scale_x_continuous(name="Travel cost (R$)", breaks=seq(0,12,3)) +
        scale_y_continuous(name="Travel time (minutes)", breaks=seq(0,60,10)) +
        coord_cartesian(xlim = c(0,15), ylim = c(0, 60)) +
        theme_classic()

#' Calculate travel time matrix between origin destination pairs considering a
#' time of arrival
#'
#' Computation of travel time estimates between one or multiple origin
#' destination pairs considering a time of arrival. This function considers a
#' time of arrival set by the user. The function returns the travel time of the
#' trip with the latest departure time that arrives before the arrival time set
#' by the user. If you want to calculate travel times considering a departure
#' time, have a' look at the [travel_time_matrix()] function. This function is a
#' wrapper around [expanded_travel_time_matrix()]. On one hand, this means this
#' the output of this function has more columns (more info) compared the output
#' of [travel_time_matrix()]. On the other hand, this function can be very memory
#' intensive if the user allows for really long max trip duration.
#'
#' @inheritParams expanded_travel_time_matrix
#' @param arrival_datetime A POSIXct object.
#'
#' @return A `data.table` with travel time estimates (in minutes) and the
#'   routes used in each trip between origin and destination pairs, for each
#'   minute of the specified time window. Each set of origin, destination and
#'   departure minute can appear up to N times, where N is the number of Monte
#'   Carlo draws specified in the function arguments (please note that this
#'   only applies when the GTFS feeds that describe the transit network include
#'   a frequencies table, otherwise only a single draw is performed). A pair is
#'   completely absent from the final output if no trips could be completed in
#'   any of the minutes of the time window. If for a single pair trips could be
#'   completed in some of the minutes of the time window, but not for all of
#'   them, the minutes in which trips couldn't be completed will have `NA`
#'   travel time and routes used. If `output_dir` is not `NULL`, the function
#'   returns the path specified in that parameter, in which the `.csv` files
#'   containing the results are saved.
#'
#' @template transport_modes_section
#' @template lts_section
#' @template datetime_parsing_section
#' @template raptor_algorithm_section
#'
#' @family routing
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' # build transport network
#' data_path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(data_path )
#'
#' # load origin/destination points
#' points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))
#'
#' arrival_datetime <- as.POSIXct(
#'   "13-05-2019 14:00:00",
#'   format = "%d-%m-%Y %H:%M:%S"
#' )
#'
#' # by default only returns the total time between each pair in each minute of
#' # the specified time window
#' arrival_ttm <- arrival_travel_time_matrix(
#'   r5r_network,
#'   origins = points,
#'   destinations = points,
#'   mode = c("WALK", "TRANSIT"),
#'   arrival_datetime = arrival_datetime,
#'   max_trip_duration = 60
#' )
#'
#' head(arrival_ttm)
#'
#' # when breakdown = TRUE the output contains much more information
#' arrival_ttm2 <- arrival_travel_time_matrix(
#'   r5r_network,
#'   origins = points,
#'   destinations = points,
#'   mode = c("WALK", "TRANSIT"),
#'   arrival_datetime = arrival_datetime,
#'   max_trip_duration = 60,
#'   breakdown = TRUE
#' )
#'
#' head(arrival_ttm2)
#'
#' stop_r5(r5r_network)
#' @export
arrival_travel_time_matrix <- function(r5r_network,
                                       r5r_core = deprecated(),
                                       origins,
                                       destinations,
                                       mode = "WALK",
                                       mode_egress = "WALK",
                                       arrival_datetime = Sys.time(),
                                       breakdown = FALSE,
                                       max_walk_time = Inf,
                                       max_bike_time = Inf,
                                       max_car_time = Inf,
                                       max_trip_duration = 120L,
                                       walk_speed = 3.6,
                                       bike_speed = 12,
                                       max_rides = 3,
                                       max_lts = 2,
                                       new_carspeeds = NULL,
                                       carspeed_scale = 1,
                                       new_lts = NULL,
                                       draws_per_minute = 5L,
                                       n_threads = Inf,
                                       verbose = FALSE,
                                       progress = FALSE,
                                       output_dir = NULL) {

  # deprecating r5r_core --------------------------------------
  if (lifecycle::is_present(r5r_core)) {

    cli::cli_warn(c(
      "!" = "The `r5r_core` argument is deprecated as of r5r v2.3.0.",
      "i" = "Please use the `r5r_network` argument instead."
    ))

    r5r_network <- r5r_core
  }

  old_options <- options(datatable.optimize = Inf)
  on.exit(options(old_options), add = TRUE)
  checkmate::assert_number(n_threads, lower = 1)
  old_dt_threads <- data.table::getDTthreads()
  dt_threads <- ifelse(is.infinite(n_threads), 0, n_threads)
  data.table::setDTthreads(dt_threads)
  on.exit(data.table::setDTthreads(old_dt_threads), add = TRUE)

  # check inputs and set r5r options --------------------------------------

  checkmate::assert_class(r5r_network, "r5r_network")

  origins <- assign_points_input(origins, "origins")
  destinations <- assign_points_input(destinations, "destinations")
  mode_list <- assign_mode(mode, mode_egress)

  # calculate departure datetime
  departure_datetime <- arrival_datetime - as.difftime(max_trip_duration, units = "mins")
  departure <- assign_departure(departure_datetime)

  # check availability of transit services on the selected date
  if (mode_list$transit_mode %like% 'TRANSIT|TRAM|SUBWAY|RAIL|BUS|CABLE_CAR|GONDOLA|FUNICULAR') {
    check_transit_availability_on_date(r5r_network, departure_date = departure$date)
  }

  r5r_network <- r5r_network@jcore

  # in direct modes reverse origin/destination to take advantage of R5's One to Many algorithm
  data_path <- r5r_network$getDataPath()
  res <- reverse_if_direct_mode(origins, destinations, mode_list, data_path)
  origins <- res$origins
  destinations <- res$destinations


  max_walk_time <- assign_max_street_time(
    max_walk_time,
    walk_speed,
    max_trip_duration,
    "walk"
  )
  max_bike_time <- assign_max_street_time(
    max_bike_time,
    bike_speed,
    max_trip_duration,
    "bike"
  )
  max_car_time <- assign_max_street_time(
    max_car_time,
    8, # 8 km/h, R5's default.
    max_trip_duration,
    "car"
  )
  max_trip_duration <- assign_max_trip_duration(
    max_trip_duration,
    mode_list,
    max_walk_time,
    max_bike_time
  )

  set_time_window(r5r_network, max_trip_duration)
  set_monte_carlo_draws(r5r_network, draws_per_minute, max_trip_duration)
  set_speed(r5r_network, walk_speed, "walk")
  set_speed(r5r_network, bike_speed, "bike")
  set_max_rides(r5r_network, max_rides)
  set_max_lts(r5r_network, max_lts)
  set_n_threads(r5r_network, n_threads)
  set_verbose(r5r_network, verbose)
  set_progress(r5r_network, progress)
  set_output_dir(r5r_network, output_dir)
  set_expanded_travel_times(r5r_network, TRUE)
  r5r_network$setSearchType("ARRIVE_BY")
  set_breakdown(r5r_network, breakdown)
  set_fare_structure(r5r_network, NULL)

  # SCENARIOS -------------------------------------------
  set_new_congestion(r5r_network, new_carspeeds, carspeed_scale)
  set_new_lts(r5r_network, new_lts)


  # call r5r_network method and process result -------------------------------

  travel_times <- r5r_network$travelTimeMatrix(
    origins$id,
    origins$lat,
    origins$lon,
    destinations$id,
    destinations$lat,
    destinations$lon,
    mode_list$direct_modes,
    mode_list$transit_mode,
    mode_list$access_mode,
    mode_list$egress_mode,
    departure$date,
    departure$time,
    max_walk_time,
    max_bike_time,
    max_car_time,
    max_trip_duration
  )

  if (!verbose & progress) cat("Preparing final output...", file = stderr())

  travel_times <- java_to_dt(travel_times)

  # reverse order of origins destinations back
  travel_times <- reverse_back_if_direct_mode(travel_times, origins, destinations, mode_list, data_path)

  # replace travel-times of non-viable trips with NAs
  # if breakdown is TRUE, there are more columns in the output

  if (nrow(travel_times) > 0) {
    if (breakdown) {
      travel_times[
        total_time > max_trip_duration,
        `:=`(
          access_time = NA_integer_,
          wait_time = NA_integer_,
          ride_time = NA_integer_,
          transfer_time = NA_integer_,
          egress_time = NA_integer_,
          routes = NA_character_,
          n_rides = NA_integer_,
          total_time = NA_integer_
        )
      ]
    } else {
      travel_times[
        total_time > max_trip_duration,
        `:=`(routes = NA_character_, total_time = NA_integer_)
      ]
    }
  }

  if (!verbose & progress) cat(" DONE!\n", file = stderr())

  if (!is.null(output_dir)) return(output_dir)
  return(travel_times[])
}

context("Arrival travel time matrix function")

# skips tests on CRAN since they require a specific version of java
testthat::skip_on_cran()

# create testing function

default_tester <- function(r5r_network,
                           origins = points[1:10,],
                           destinations = points[1:10,],
                           mode = "TRANSIT",
                           arrival_datetime = as.POSIXct("13-05-2019 14:00:00",
                                                           format = "%d-%m-%Y %H:%M:%S"),
                           breakdown = FALSE,
                           max_walk_time = Inf,
                           max_bike_time = Inf,
                           max_trip_duration = 120L,
                           walk_speed = 3.6,
                           bike_speed = 12,
                           max_rides = 3,
                           n_threads = Inf,
                           verbose = FALSE,
                           progress=TRUE) {

  results <- arrival_travel_time_matrix(
    r5r_network,
    origins = origins,
    destinations = destinations,
    mode = mode,
    arrival_datetime = arrival_datetime,
    breakdown = breakdown,
    max_walk_time = max_walk_time,
    max_bike_time = max_bike_time,
    max_trip_duration = max_trip_duration,
    walk_speed = walk_speed,
    bike_speed = bike_speed,
    max_rides = max_rides,
    n_threads = n_threads,
    verbose = verbose
  )

  return(results)

}


# errors and warnings -----------------------------------------------------


test_that("adequately raises errors", {

  # error related to using object with wrong type as r5r_network
  expect_error(default_tester("r5r_network"))

  # error related to using wrong origins/destinations object type
  multipoint_origins      <- sf::st_cast(sf::st_as_sf(points[1:2,], coords = c("lon", "lat")), "MULTIPOINT")
  multipoint_destinations <- multipoint_origins
  list_origins      <- list(id = c("1", "2"), lat = c(-30.02756, -30.02329), long = c(-51.22781, -51.21886))
  list_destinations <- list_origins

  expect_error(default_tester(r5r_network, origins = multipoint_origins))
  expect_error(default_tester(r5r_network, destinations = multipoint_destinations))
  expect_error(default_tester(r5r_network, origins = list_origins))
  expect_error(default_tester(r5r_network, destinations = list_destinations))
  expect_error(default_tester(r5r_network, origins = "origins"))
  expect_error(default_tester(r5r_network, destinations = "destinations"))

  # error/warning related to using wrong origins/destinations column types
  origins <- destinations <- points[1:2, ]

  origins_char_lat   <- data.frame(id = origins$id, lat = as.character(origins$lat), lon = origins$lon)
  origins_char_lon   <- data.frame(id = origins$id, lat = origins$lat, lon = as.character(origins$lon))
  destinations_char_lat   <- data.frame(id = destinations$id, lat = as.character(destinations$lat), lon = destinations$lon)
  destinations_char_lon   <- data.frame(id = destinations$id, lat = destinations$lat, lon = as.character(destinations$lon))

  expect_error(default_tester(r5r_network, origins = origins_char_lat))
  expect_error(default_tester(r5r_network, origins = origins_char_lon))
  expect_error(default_tester(r5r_network, destinations = destinations_char_lat))
  expect_error(default_tester(r5r_network, destinations = destinations_char_lon))

  # error related to nonexistent mode
  expect_error(default_tester(r5r_network, mode = "pogoball"))

  # errors related to date formatting
  numeric_datetime <- as.numeric(as.POSIXct("13-05-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S"))

  expect_error(default_tester(r5r_network, arrival_datetime = "13-05-2019 14:00:00"))
  expect_error(default_tester(r5r_network, numeric_datetime))


  # error with breakdown
  expect_error(default_tester(r5r_network, breakdown ='test'))

  # errors related to max_walk_time
  expect_error(default_tester(r5r_network, max_walk_time = "1000"))
  expect_error(default_tester(r5r_network, max_walk_time = NULL))

  # errors related to max_bike_time
  expect_error(default_tester(r5r_network, max_bike_time = "1000"))
  expect_error(default_tester(r5r_network, max_bike_time = NULL))

  # error/warning related to max_street_time
  expect_error(default_tester(r5r_network, max_trip_duration = "120"))

  # error related to non-numeric walk_speed
  expect_error(default_tester(r5r_network, walk_speed = "3.6"))

  # error related to non-numeric bike_speed
  expect_error(default_tester(r5r_network, bike_speed = "12"))
})

test_that("adequately raises warnings - needs java", {

  # error/warning related to using wrong origins/destinations column types
  origins <- destinations <- points[1:2, ]

  origins_numeric_id <- data.frame(id = 1:2, lat = origins$lat, lon = origins$lon)
  destinations_numeric_id <- data.frame(id = 1:2, lat = destinations$lat, lon = destinations$lon)

  expect_warning(default_tester(r5r_network, origins = origins_numeric_id))
  expect_warning(default_tester(r5r_network, destinations = destinations_numeric_id))


})


# adequate behaviour ------------------------------------------------------


test_that("output is correct", {

  #  * output class ---------------------------------------------------------


  # expect results to be of class 'data.table', independently of the class of
  # 'origins'/'destinations'

  origins_sf <- destinations_sf <- sf::st_as_sf(
    points[1:10, ],
    coords = c("lon", "lat"),
    crs = 4326
  )

  result_df_input <- default_tester(r5r_network)
  result_sf_input <- default_tester(r5r_network, origins_sf, destinations_sf)

  expect_true(is(result_df_input, "data.table"))
  expect_true(is(result_sf_input, "data.table"))

  # expect each column to be of right class

  expect_true(typeof(result_df_input$from_id) == "character")
  expect_true(typeof(result_df_input$to_id) == "character")
  expect_true(typeof(result_df_input$total_time) == "double")


  #  * r5r options ----------------------------------------------------------

  result_sf_input <- default_tester(r5r_network, origins_sf, destinations_sf,
                                    verbose = FALSE, progress=TRUE)


  #  * arguments ------------------------------------------------------------


  # expect all travel times to be lower than max_trip_duration

  origins <- destinations <- points[1:10,]
  max_trip_duration <- 60L

  df <- default_tester(r5r_network, origins, destinations, max_trip_duration = max_trip_duration)
  max_duration <- data.table::setDT(df)[, max(total_time, na.rm=T)]

  expect_true(max_duration <= max_trip_duration)



  # expect number of columns to be larger when breakdown = TRUE
  df2 <- default_tester(r5r_network, breakdown =TRUE)
  df3 <- default_tester(r5r_network, breakdown =FALSE)
  expect_true(ncol(df2) > ncol(df3))

  expect_true(typeof(df2$routes) == "character")
  expect_true(typeof(df2$access_time ) == "double")

  # # expect Empty data.table for trips walking and cycling and by car
  #
  # origins <- destinations <- points[1:10,]
  #
  # df <- default_tester(r5r_network, origins = origins, destinations = destinations,
  #                      mode = "WALK")
  # expect_true(nrow(df) == 0)
  #
  # df <- default_tester(r5r_network, origins = origins, destinations = destinations,
  #                      mode = "BICYCLE")
  # expect_true(nrow(df) == 0)
  #
  # df <- default_tester(r5r_network, origins = origins, destinations = destinations,
  #                      mode = "CAR")
  # expect_true(nrow(df) == 0)


})


test_that("using transit outside the gtfs dates throws an error", {
  expect_error(
    tester(r5r_network,
           mode='transit',
           arrival_datetime = as.POSIXct("13-05-2025 14:00:00",
                                           format = "%d-%m-%Y %H:%M:%S")
    )
  )
})

#' Manage cached files from the r5r package
#'
#' @param list_files Logical. Whether to print a message with the address of r5r
#'        JAR files cached locally. Defaults to `TRUE`.
#' @param delete_file String. The file name (basename) of a JAR file cached
#'        locally that should be deleted. Defaults to `NULL`, so that no
#'        file is deleted. If `delete_file = "all"`, then all cached files are
#'        deleted.
#'
#' @return A message indicating which file exist and/or which ones have been
#'         deleted from local cache directory.
#' @export
#' @family Cache data
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' # download r5 JAR
#' r5r::download_r5()
#'
#' # list all files cached
#' r5r_cache(list_files = TRUE)
#'
#' # delete r5 JAR
#' r5r_cache(delete_file = 'r5-v7.0')
#'
r5r_cache <- function(list_files = TRUE,
                      delete_file = NULL){

  # check inputs
  checkmate::assert_logical(list_files)
  checkmate::assert_character(delete_file, null.ok = TRUE)

  # find / create local dir
  if (!dir.exists(r5r_env$cache_dir)) { dir.create(r5r_env$cache_dir, recursive=TRUE) }

  # list cached files
  files <- list.files(dirname(r5r_env$cache_dir), full.names = TRUE, recursive = TRUE)

  # if wants to delete file
  # delete_file = "r5-v7.0-all.jar"
  if (!is.null(delete_file)) {

    # IF file does not exist, print message
    if (!any(grepl(delete_file, files)) & delete_file != "all") {
      message(paste0("The file '", delete_file, "' is not cached."))
    }

    # IF file exists, delete file
    if (any(grepl(delete_file, files))) {
      f <- files[grepl(delete_file, files)]
      unlink(f, recursive = TRUE, force = TRUE)
      message(paste0("The file '", delete_file, "' has been removed."))
    }

    # Delete ALL file
    if (delete_file=='all') {

      # delete any files from censobr, current and old data releases
      dir_above <- dirname(r5r_env$cache_dir)
      unlink(dir_above, recursive = TRUE, force = TRUE)
      message(paste0("All files have been removed."))

    }
  }

  # list cached files
  files <- list.files(r5r_env$cache_dir, full.names = TRUE)

  # print file names
  if(isTRUE(list_files)){
    message('Files currently cached:')
    message(paste0(files, collapse = '\n'))
  }
}

#' Create a transport network used for routing in R5 (deprecated)
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `setup_r5()` was renamed to [`build_network()`] to create a more consistent
#' API. **`setup_r5()` is being deprecated** after *r5r* v2.3.0 and will be
#' **removed in a future release**. Please switch to [`build_network()`].
#'
#' @template verbose
#' @param data_path A string pointing to the directory where data inputs are
#'        stored and where the built `network.dat` will be saved.
#' @param temp_dir A logical. Whether the `network.dat` file should be saved to
#'        a temporary directory. Defaults to `FALSE`.
#' @template elevation
#' @param overwrite A logical. Whether to overwrite an existing `network.dat`
#'        or to use a cached file. Defaults to `FALSE` (i.e. use a cached
#'        network).
#'
#' @return A `r5r_network` object representing the built network to connect with
#'         `R5` routing engine.
#'
#' @template elevation_section
#'
#' @family Build network
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' # directory with street network and gtfs files
#' data_path <- system.file("extdata/poa", package = "r5r")
#'
#' # `setup_r5()` has been deprecated, please switch to `build_network()`
#' r5r_network <- build_network(data_path)
#' @export
setup_r5 <- function(data_path,
                     verbose = FALSE,
                     temp_dir = FALSE,
                     elevation = "TOBLER",
                     overwrite = FALSE) { # nocov start

  # Deprecation warning --------------------------------------------------------
  cli::cli_warn(c(
    "!" = "{.fn setup_r5} is being deprecated in *r5r* and will be removed in a future release.",
    "i" = "Please use {.fn r5r::build_network} instead."
  ))

  # lifecycle::deprecate_soft(
  #   when = "2.3.0",
  #   what = "setup_r5()",
  #   with = "r5r::build_network()"
  #   )

  # # Optional base-R signal for tools that watch .Deprecated()
  # .Deprecated("build_network", package = "r5r")


  # Pass through to the replacement -------------------------------------------

  r5r_network <- build_network(
    data_path,
    verbose = verbose,
    temp_dir = temp_dir,
    elevation = elevation,
    overwrite = overwrite
    )

  return(r5r_network)
} # nocov end

#' Find snapped locations of input points on street network
#'
#' Finds the snapped location of points on `R5` network. Snapping is an
#' important step of the routing process, which is when the origins and
#' destinations specified by the user are actually positioned on the network
#' created by `R5`. The snapping process in `R5` is composed of two rounds.
#' First, it tries to snap the points within a radius of 300 meters from
#' themselves. If the first round is unsuccessful, then `R5` expands the search
#' to the radius specified (by default 1.6km). If yet again it is unsuccessful,
#' then the unsnapped points won't be used during the routing process. The
#' snapped location of each point depends on the transport mode set by the user,
#' because some network edges are not available to specific modes (e.g. a
#' pedestrian-only street cannot be used to snap car trips).
#'
#' @template r5r_network
#' @template r5r_core
#' @param points Either a `POINT sf` object with WGS84 CRS, or a `data.frame`
#'        containing the columns `id`, `lon` and `lat`.
#' @param radius Numeric. The maximum radius in meters within which to snap.
#'        Defaults to 1600m.
#' @param mode A string. Which mode to consider when trying to snap the points
#'        to the network. Defaults to `WALK`, also allows `BICYCLE` and `CAR`.
#'
#' @return A `data.table` with the original points, their respective
#' snapped coordinates on the street network and the Euclidean distance (in
#' meters) between the original points and their snapped location. Points that
#' could not be snapped show `NA` coordinates and `found = FALSE`.
#'
#' @family network functions
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(data_path = path)
#' points <- read.csv(file.path(path, "poa_hexgrid.csv"))
#'
#' snap_df <- find_snap(
#'   r5r_network,
#'   points = points,
#'   radius = 2000,
#'   mode = "WALK"
#'   )
#'
#' stop_r5(r5r_network)
#' @export
find_snap <- function(r5r_network,
                      r5r_core = deprecated(),
                      points,
                      radius = 1600,
                      mode = "WALK"){

  # deprecating r5r_core --------------------------------------
  if (lifecycle::is_present(r5r_core)) {

    cli::cli_warn(c(
      "!" = "The `r5r_core` argument is deprecated as of r5r v2.3.0.",
      "i" = "Please use the `r5r_network` argument instead."
    ))

    r5r_network <- r5r_core
  }

  checkmate::assert_class(r5r_network, "r5r_network")
  r5r_network <- r5r_network@jcore

  checkmate::assert_numeric(radius, lower = 0, finite = TRUE, max.len = 1)
  mode_options <- c("WALK", "BICYCLE", "CAR")
  checkmate::assert(
    checkmate::check_string(mode),
    checkmate::check_names(mode, subset.of = mode_options),
    combine = "and"
  )

  points <- assign_points_input(points, "points")

  snap_df <- r5r_network$findSnapPoints(points$id,
                                     points$lat,
                                     points$lon,
                                     radius,
                                     mode)
  snap_df <- java_to_dt(snap_df)

  snap_df[found == FALSE, `:=`(snap_lat = NA, snap_lon = NA, distance = NA)]

  return(snap_df)
}

context("find_snap function")

testthat::skip_on_cran()


# errors and warnings -----------------------------------------------------


test_that("adequately raises errors", {

  # invalid mode
  expect_error( find_snap(r5r_network, points = points, mode = 'AAA') )

  # invalid r5r_network
  expect_error( find_snap(r5r_network='a', points = points, mode = 'WALK') )

  # invalid points
  expect_error( find_snap(r5r_network=r5r_network, points = 'a', mode = 'WALK') )

  # invalid radius
  expect_error( find_snap(r5r_network=r5r_network, points = 'a', radius = -90, mode = 'WALK') )

})


# adequate behavior ------------------------------------------------------


test_that("output is correct", {

  # expected behavior
  expect_s3_class( find_snap(r5r_network=r5r_network, points = points, mode = 'WALK'), 'data.table')

  # 6 points don't get snapped using the default radius after moving them
  points_snap <- points
  points_snap$lat <- points_snap$lat + 0.01
  expect_equal(
    sum(!find_snap(r5r_network, points=points_snap, mode = 'WALK')$found),
    6
  )

  # all points are snapped after increasing the radius
  expect_equal(
    sum(!find_snap(r5r_network, points=points_snap, mode = 'WALK', radius = 5000)$found),
    0
  )

})

#' @param new_carspeeds A `data.frame` specifying the new car speed for each OSM
#'        edge id. This table must contain columns \code{osm_id}, \code{max_speed}
#'        and \code{speed_type}. The `"speed_type"` column is of class character
#'        and it indicates whether the values in `"max_speed"` should be
#'        interpreted as percentages of original speeds (`"scale"`) or as absolute
#'        speeds (`"km/h"`). Alternatively, the `new_carspeeds` parameter can
#'        receive an `sf data.frame` with POLYGON geometry that indicates the new
#'        car speed for all the roads that fall within each polygon. In this case,
#'        the table must contain the columns \code{poly_id} with a unique id for
#'        each polygon, \code{scale} with the new speed scaling factors and
#'        \code{priority}, which is a number ranking which polygon should be
#'        considered in case of overlapping polygons. See more into in the
#'        `link to congestion vignette`.
#' @param carspeed_scale Numeric. The default car speed to use for road segments
#'        not specified in `new_carspeeds`. By default, it is `NULL` and the speeds
#'        of the unlisted roads are kept unchanged.
#' @param new_lts A `data.frame` specifying the new LTS levels for each OSM edge
#'        id. The table must contain columns \code{osm_id} and \code{lts}.
#'        Alternatively, the `new_lts` parameter can receive an `sf data.frame`
#'        with LINESTRING geometry. R5 will then find the nearest road for each
#'        LINESTRING and update its LTS value accordingly.

# library(tidyverse)
options(java.parameters = '-Xmx10G')
devtools::load_all(".")
library(r5r)
library(sf)
library(tidyverse)
library(mapview)

# initialize r5r
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path, verbose = FALSE, use_elevation = TRUE)

# get regular grid at resolution 8
  grid_df <- r5r_core$getGrid(8L, TRUE)
  grid_df <- java_to_dt(grid_df)
  # grid_df <- jdx::convertToR(grid_df)


# grid_df$geometry <- st_as_sfc(grid_df$geometry)
# grid_df <- st_as_sf(grid_df, crs = 4326)

grid_df %>%
  mapview(xcol="lon", ycol="lat", crs=4326)

grid_df %>% mutate(id = as.integer(id)) %>% mapview::mapview(alpha.regions = 0)
  mapview::mapview(snap_df, xcol="lon", ycol="lat")
  mapview::mapview(snap_df %>% filter(point_id=="4"), xcol="snap_lon", ycol="snap_lat")

r5r_core$silentMode()
system.time(
  snap_df <- find_snap(r5r_core, grid_df, "WALK")
)


mapview::mapview(snap_df, xcol="lon", ycol="lat", crs=4326)
mapview::mapview(snap_df %>% drop_na(), xcol="snap_lon", ycol="snap_lat", zcol="distance", crs=4326)

leafsync::sync(mv1, mv2)


original <- snap_df %>% select(id = point_id, lat, lon)
snapped <- snap_df %>% select(id = point_id, lat = snap_lat, lon = snap_lon)

system.time(
  ttm_orig <- travel_time_matrix(r5r_core, origins = original, destinations = original,
                                 verbose = FALSE)
)

ttm_snap <- travel_time_matrix(r5r_core, origins = snapped, destinations = snapped,
                          verbose = FALSE)

ttm_join <- left_join(ttm_orig, ttm_snap, by=c("fromId", "toId"),
                      suffix = c("_or", "_sn"))

street_net <- street_network_to_sf(r5r_core)

street_net$vertices %>% mapview()
street_net$edges %>% ggplot() + geom_sf()

hex <- read_csv(system.file("extdata/poa", "poa_hexgrid.csv", package = "r5r"))

hex %>%
  mapview(xcol="lon", ycol="lat", crs=4326)

min(hex$lat)

departure_datetime <- as.POSIXct("13-03-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S")

ttm <- travel_time_matrix(r5r_core, origins = grid_df[20000,], destinations = grid_df,
                          departure_datetime = departure_datetime,
                          mode = c("BICYCLE"), max_trip_duration = 30,
                          max_walk_dist = 800)

ttm %>% left_join(grid_df, by = c("toId"="id")) %>%
  mutate(travel_time = travel_time %/% 2) %>%
  st_as_sf(crs = 4326) %>%
  mapview(zcol="travel_time")
  ggplot() +
  geom_sf(aes(geometry=geometry, fill=travel_time), color=NA) +
  scale_fill_distiller(palette = "Spectral")

View()
grid_df


original <- dplyr::select(snap_df, id = point_id, lat, lon)
original$opportunities <- 1

system.time(
  ttm_orig <- accessibility(r5r_core, origins = original, destinations = original,
                            verbose = FALSE)
)

#' Set custom road speed for vehicles
#'
#' @description
#' By default, R5 considers that vehicles travel at the speed limit allowed on
#' each type of road, as informed on OpenStreetMap.pbf data. This `set_road_speed()`
#' function allows users to set a custom speed for different types of roads. The
#' user needs to run this function before building the network with `setup_r5()`,
#' and then `set_road_speed()` will save a `build-config.json` file to the data
#' path, which will be picked up by R5 when building the network.
#'
#' @param data_path character string, the directory where data inputs are stored
#'                  and where the built `network.dat` will be saved.
#' @param motorway numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param motorway_link numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param trunk numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param trunk_link numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param primary numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param primary_link numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param secondary numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param secondary_link numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param tertiary numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param tertiary_link numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param living_street numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param pedestrian numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param residential numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param unclassified numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param service numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param track numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param road numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#' @param defaultSpeed numeric. Speed in Km/h. Defaults to (`NULL`) use the speed recorded on OpenStreetMap.pdf data.
#'
#' @return The functions saves a `build-config.json` file to the data path.
#'
#' @family support functions
#' @examples if (interactive()) {
#' library(r5r)
#'
#'# Get data path
#'path <- system.file("extdata/spo", package = "r5r")
#'
#'# set road speeds
#'set_road_speed(data_path = path,
#'                primary = 30,
#'                secondary = 20,
#'                trunk = 60,
#'                defaultSpeed = 50)
#'
#' # build transport network
#' r5r_core <- setup_r5(data_path = data_path, temp_dir = TRUE)
#'
#' }
#' @export
set_road_speed <- function(data_path,
                           motorway = NULL
                         , motorway_link = NULL
                         , trunk = NULL
                         , trunk_link = NULL
                         , primary = NULL
                         , primary_link = NULL
                         , secondary = NULL
                         , secondary_link = NULL
                         , tertiary = NULL
                         , tertiary_link = NULL
                         , living_street = NULL
                         , pedestrian = NULL
                         , residential = NULL
                         , unclassified = NULL
                         , service = NULL
                         , track = NULL
                         , road = NULL
                         , defaultSpeed = NULL){

  # create data frame with speeds passed by user
  speeds_df <- data.frame(
      'motorway' =      ifelse(is.null(motorway), NA, motorway)
    , 'motorway_link' = ifelse(is.null(motorway_link), NA, motorway_link)
    , 'trunk' =         ifelse(is.null(trunk), NA, trunk)
    , 'trunk_link' =    ifelse(is.null(trunk_link), NA, trunk_link)
    , 'primary' =       ifelse(is.null(primary), NA, primary)
    , 'primary_link' =  ifelse(is.null(primary_link), NA, primary_link)
    , 'secondary' =     ifelse(is.null(secondary), NA, secondary)
    , 'secondary_link' =ifelse(is.null(secondary_link), NA, secondary_link)
    , 'tertiary' =      ifelse(is.null(tertiary), NA, tertiary)
    , 'tertiary_link' = ifelse(is.null(tertiary_link), NA, tertiary_link)
    , 'living_street' = ifelse(is.null(living_street), NA, living_street)
    , 'pedestrian' =     ifelse(is.null(pedestrian), NA, pedestrian)
    , 'residential' =   ifelse(is.null(residential), NA, residential)
    , 'unclassified' =  ifelse(is.null(unclassified), NA, unclassified)
    , 'service' =       ifelse(is.null(service), NA, service)
    , 'track' =         ifelse(is.null(track), NA, track)
    , 'road' =          ifelse(is.null(road), NA, road)
    , 'defaultSpeed' =  ifelse(is.null(defaultSpeed), NA, defaultSpeed))

  # remove columns with missing value
  speeds_df <- speeds_df[ , apply(speeds_df, 2, function(x) !any(is.na(x)))]

  # retrieve default speed
  dflt_speed <- speeds_df$defaultSpeed
  speeds_df$defaultSpeed <- NULL

  # Convert to list in the json structure we need
  my_list <- list( speeds = list ( units = 'km/h',
                                   values= list(speeds_df),
                                   defaultSpeed = dflt_speed ) )

  # Prettify (add indentation)
  my_list <- jsonlite::toJSON(my_list, pretty = TRUE)

  # remove brackets
  my_list <- gsub("\\[|\\]", '', my_list)

  # save json
  message(paste0('build-config.json file saved at ', paste0(path, '/build-config.json')))
  write(my_list,file = paste0(path, '/build-config.json'))

}

#### example -----------------

library(jsonlite)
library(data.table)
library(r5r)

# build network
points <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))[c(1:100, 500:600),]
path <- system.file("extdata/poa", package = "r5r")

# set road speeds
set_road_speed(data_path = path,
               primary = 666,
               secondary = 666,
               trunk = 666,
               road = 666,
               defaultSpeed = 666)

r5r_core <- setup_r5(data_path = path, verbose = F, overwrite = T)

tt <- r5r::travel_time_matrix(r5r_core = r5r_core, origins = points,destinations = points, mode = 'car')
# fwrite(tt, 'tt_test.csv')

#' @param time_window An integer. The time window in minutes for which `r5r`
#'   will calculate multiple travel time matrices departing each minute.
#'   Defaults to 10 minutes. By default, the function returns the result based
#'   on median travel times, but the user can set the `percentiles` parameter to
#'   extract more results. Please read the time window vignette for more details
#'   on its usage `vignette("time_window", package = "r5r")`

#' Set verbose argument
#'
#' Indicates whether R5 should output informative messages or not. Please note
#' that R5 error messages are still reported even when `verbose` is `FALSE`.
#'
#' @template r5r_network
#' @param verbose A logical, passed from the function above.
#'
#' @param verbose A logical, passed from function above.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_verbose <- function(r5r_network, verbose) {
  checkmate::assert_logical(verbose, len = 1, any.missing = FALSE)

  if (verbose) {
    r5r_network$verboseMode()
  } else {
    r5r_network$silentMode()
  }

  return(invisible(TRUE))
}


#' Set progress argument
#'
#' Indicates whether or not a progress counter must be printed during
#' computations. Applies to all routing functions.
#'
#' @template r5r_network
#' @param progress A logical, passed from the function above.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_progress <- function(r5r_network, progress) {
  checkmate::assert_logical(progress, len = 1, any.missing = FALSE)

  r5r_network$setProgress(progress)

  return(invisible(TRUE))
}


#' Set number of threads
#'
#' Sets the number of threads to be used by the r5r `.jar`.
#'
#' @template r5r_network
#' @param n_threads A number, passed from the function above.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_n_threads <- function(r5r_network, n_threads) {
  checkmate::assert_number(n_threads, lower = 1)

  if (is.infinite(n_threads)) {
    r5r_network$setNumberOfThreadsToMax()
  } else {
    n_threads <- as.integer(n_threads)
    r5r_network$setNumberOfThreads(n_threads)
  }

  return(invisible(TRUE))
}


#' Set max Level of Transit Stress (LTS)
#'
#' @template r5r_network
#' @param max_lts A number (between 1 and 4). The maximum level of traffic
#'   stress that cyclists will tolerate. A value of 1 means cyclists will only
#'   travel through the quietest streets, while a value of 4 indicates cyclists
#'   can travel through any road.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_max_lts <- function(r5r_network, max_lts) {
  checkmate::assert_number(max_lts)

  if (max_lts < 1 | max_lts > 4) {
    stop(
      max_lts,
      " is not a valid value for the maximum Level of Transit Stress (LTS).\n",
      "Please enter a value between 1 and 4."
    )
  }

  r5r_network$setMaxLevelTrafficStress(as.integer(max_lts))

  return(invisible(TRUE))
}


#' Set max number of rides
#'
#' Sets the maximum number of rides a trip can use in R5.
#'
#' @template r5r_network
#' @param max_rides A number. The max number of public transport rides allowed
#'   in the same trip. Passed from routing function.
#'
#' @return No return value, called for side effects.
#'
#' @family setting functions
#'
#' @keywords internal
set_max_rides <- function(r5r_network, max_rides) {
  checkmate::assert_number(max_rides, lower = 1, finite = TRUE)

  r5r_network$setMaxRides(as.integer(max_rides))

  return(invisible(TRUE))
}


#' Set walk and bike speed
#'
#' This function receives the walk and bike 'speed' inputs in Km/h from routing
#' functions above and converts them to meters per second, which is then used
#' to set these speed profiles in r5r JAR.
#'
#' @template r5r_network
#' @param speed A number representing the speed in km/h.
#' @param mode A string. Either `"bike"` or `"walk"`.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_speed <- function(r5r_network, speed, mode) {
  checkmate::assert(
    checkmate::check_string(mode),
    checkmate::check_names(mode, subset.of = c("bike", "walk")),
    combine = "and"
  )
  var_name <- paste0(mode, "_speed")
  checkmate::assert_number(speed, finite = TRUE, .var.name = var_name)
  if (speed <= 0) {
    stop(
      "Assertion on '", var_name, "' failed: Must have value greater than 0."
    )
  }

  speed <- speed * 5 / 18

  if (mode == "walk") {
    r5r_network$setWalkSpeed(speed)
  } else {
    r5r_network$setBikeSpeed(speed)
  }

  return(invisible(TRUE))
}


#' Set time window
#'
#' Sets the time window to be used by R5.
#'
#' @template r5r_network
#' @param time_window A number.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_time_window <- function(r5r_network, time_window) {
  checkmate::assert_number(time_window, lower = 1, finite = TRUE)

  time_window <- as.integer(time_window)

  r5r_network$setTimeWindowSize(time_window)

  return(invisible(TRUE))
}


#' Set percentiles
#'
#' Sets the percentiles to be used by R5.
#'
#' @template r5r_network
#' @param percentiles An integer vector of maximum length 5.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_percentiles <- function(r5r_network, percentiles) {
  checkmate::assert_numeric(
    percentiles,
    lower = 1,
    upper = 99,
    max.len = 5,
    unique = TRUE,
    any.missing = FALSE,
    finite = TRUE
  )

  percentiles <- as.integer(percentiles)

  r5r_network$setPercentiles(percentiles)

  return(invisible(TRUE))
}


#' Set number of Monte Carlo draws
#'
#' Sets the number of Monte Carlo draws to be used by R5.
#'
#' @template r5r_network
#' @param draws_per_minute A number.
#' @param time_window A number.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_monte_carlo_draws <- function(r5r_network, draws_per_minute, time_window) {
  # time_window is previously checked in set_time_window()
  checkmate::assert_number(draws_per_minute, lower = 1, finite = TRUE)

  draws <- time_window * draws_per_minute
  draws <- as.integer(draws)

  r5r_network$setNumberOfMonteCarloDraws(draws)

  return(invisible(TRUE))
}


#' Set the fare structure used when calculating transit fares
#'
#' Sets the fare structure used by our "generic" fare calculator. A value of
#' `NULL` is passed to `fare_structure` by the upstream routing and
#' accessibility functions when fares are not to be calculated.
#'
#' @template r5r_network
#' @template fare_structure
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_fare_structure <- function(r5r_network, fare_structure) {
  if (!is.null(fare_structure)) {
    assert_fare_structure(fare_structure)

    if (!("type" %in% names(fare_structure))) {
      # this is an R5R fare structure object
      if (fare_structure$fare_cap == Inf) {
        fare_structure$fare_cap <- -1
      }
      if (fare_structure$transfer_time_allowance == Inf) {
        fare_structure$transfer_time_allowance <- -1
      }
      if (fare_structure$max_discounted_transfers == Inf) {
        fare_structure$max_discounted_transfers <- -1
      }
    }

    fare_settings_json <- jsonlite::toJSON(fare_structure, auto_unbox = TRUE)
    json_string <- as.character(fare_settings_json)

    r5r_network$setFareCalculator(json_string)
  } else {
    r5r_network$dropFareCalculator()
  }

  return(invisible(TRUE))
}


#' Set max fare
#'
#' Sets the max fare allowed when calculating transit fares.
#'
#' @template r5r_network
#' @param max_fare A number.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_max_fare <- function(r5r_network, max_fare) {
  checkmate::assert_number(max_fare, lower = 0)

  # Inf values are not allowed in Java, so -1 is used to indicate when max_fare
  # is unconstrained

  if (!is.infinite(max_fare)) {
    r5r_network$setMaxFare(rJava::.jfloat(max_fare))
  } else {
    r5r_network$setMaxFare(rJava::.jfloat(-1.0))
  }

  return(invisible(TRUE))
}


#' Set output directory
#'
#' Sets whether r5r should save output to a specified directory.
#'
#' @template r5r_network
#' @param output_dir A path.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_output_dir <- function(r5r_network, output_dir) {
  checkmate::assert_string(output_dir, null.ok = TRUE)

  if (!is.null(output_dir)) {
    checkmate::assert_directory_exists(output_dir)
    r5r_network$setCsvOutput(output_dir)
  } else {
    r5r_network$setCsvOutput("")
  }

  return(invisible(TRUE))
}


#' Set cutoffs
#'
#' Sets the cutoffs used when calculating accessibility.
#'
#' @template r5r_network
#' @param cutoffs A numeric vector.
#' @param decay_function A string, the name of the decay function.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_cutoffs <- function(r5r_network, cutoffs, decay_function) {
  checkmate::assert_numeric(
    cutoffs,
    min.len = 1,
    max.len = 12,
    any.missing = FALSE,
    finite = TRUE,
    null.ok = TRUE
  )

  non_null_cutoffs <- c("step", "exponential", "linear", "logistic")
  if (!is.null(cutoffs) & decay_function == "fixed_exponential") {
    stop(
      "Assertion on cutoffs failed: must be NULL when decay_function ",
      "is ", decay_function, "."
    )
  } else if (is.null(cutoffs) & decay_function %in% non_null_cutoffs) {
    stop(
      "Assertion on cutoffs failed: must not be NULL when decay_function ",
      "is ", decay_function, "."
    )
  }

  # java does not accept NULL values, so if cutoffs is NULL we assign a
  # placeholder number to it (it's ignored in R5 anyway)

  if (is.null(cutoffs)) {
    cutoffs <- 0L
  } else {
    cutoffs <- as.integer(cutoffs)
  }

  r5r_network$setCutoffs(cutoffs)

  return(invisible(TRUE))
}


#' Set monetary cutoffs
#'
#' Sets the monetary cutoffs that should be considered when calculating the
#' Pareto frontier.
#'
#' @template r5r_network
#' @param fare_cutoffs A path.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_fare_cutoffs <- function(r5r_network, fare_cutoffs) {
  checkmate::assert_numeric(
    fare_cutoffs,
    lower = 0,
    any.missing = FALSE,
    min.len = 1,
    unique = TRUE
  )

  r5r_network$setFareCutoffs(rJava::.jfloat(fare_cutoffs))

  return(invisible(TRUE))
}


#' Set breakdown
#'
#' Sets whether travel time matrices should include detailed trip information or
#' not.
#'
#' @template r5r_network
#' @param breakdown A logical.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_breakdown <- function(r5r_network, breakdown) {
  checkmate::assert_logical(breakdown, any.missing = FALSE, len = 1)

  r5r_network$setTravelTimesBreakdown(breakdown)

  return(invisible(TRUE))
}


#' Set expanded travel times
#'
#' Sets whether travel time matrices should return results for each minute of
#' the specified time window.
#'
#' @template r5r_network
#' @param expanded A logical.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_expanded_travel_times <- function(r5r_network, expanded) {
  checkmate::assert_logical(expanded, any.missing = FALSE, len = 1)

  r5r_network$setExpandedTravelTimes(expanded)

  return(invisible(TRUE))
}


#' Set suboptimal minutes
#'
#' Sets the number of suboptimal minutes considered in [detailed_itineraries()]
#' routing. From R5 documentation: "This parameter compensates for the fact that
#' GTFS does not contain information about schedule deviation (lateness). The
#' min-max travel time range for some trains is zero, since the trips are
#' reported to always have the same timings in the schedule. Such an option
#' does not overlap (temporally) its alternatives, and is too easily eliminated
#' by an alternative that is only marginally better. We want to effectively
#' push the max travel time of alternatives out a bit to account for the fact
#' that they don't always run on schedule".
#'
#' @template r5r_network
#' @param suboptimal_minutes A number.
#' @template fare_structure
#' @param shortest_path A logical.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @family setting functions
#'
#' @keywords internal
set_suboptimal_minutes <- function(r5r_network,
                                   suboptimal_minutes,
                                   fare_structure,
                                   shortest_path) {
  checkmate::assert_number(suboptimal_minutes, lower = 0, finite = TRUE)

  if (!is.null(fare_structure) && suboptimal_minutes > 0) {
    stop(
      "Assertion on 'suboptimal_minutes' failed: Must be 0 when calculating ",
      "fares with detailed_itineraries()."
    )
  }

  if (shortest_path && suboptimal_minutes > 0) {
    stop(
      "Assertion on 'suboptimal_minutes' failed: Must be 0 when ",
      "'shortest_path' is TRUE."
    )
  }

  suboptimal_minutes <- as.integer(suboptimal_minutes)

  r5r_network$setSuboptimalMinutes(suboptimal_minutes)

  return(invisible(TRUE))
}


#' Reverse Origins and Destinations for Direct Modes
#'
#' Swaps the `origins` and `destinations` data frames if certain conditions are
#' met, specifically to optimize routing performance with R5's one-to-many
#' algorithm. The function reverses the direction of analysis when the transit
#' mode is empty and the direct modes are WALK or BICYCLE and when the number of
#' origin points is greater than the number of destination points.
#'
#' @param origins A data frame representing origin locations.
#' @param destinations A data frame representing destination locations.
#' @param mode_list A named list containing the routing modes.
#' @param data_path The data path used to build the network
#'
#' @return List origins and destinations unchanged or in swapped order
#'
#' @family setting functions
#'
#' @keywords internal
reverse_if_direct_mode <- function(origins, destinations, mode_list, data_path) {

  # In direct modes, reverse origin/destination to take advantage of R5's One to Many algorithm
  if (
    mode_list$transit_mode == "" &&
    mode_list$direct_modes %in% c("WALK") &&
    nrow(origins) > nrow(destinations) &&
    !exists_tiff(data_path) # only if no elevation data is present
    ) {
    temp <- origins
    origins <- destinations
    destinations <- temp
  }

  return(list(origins = origins, destinations = destinations))
}


#' Reverse BACK Origins and Destinations for Direct Modes
#'
#' Swaps BACK the `origins` and `destinations` data frames if they were
#' originally reversed by `reverse_if_direct_mode()`.
#'
#' @param travel_times A travel time matrix.
#' @param origins A data frame representing origin locations.
#' @param destinations A data frame representing destination locations.
#' @param mode_list A named list containing the routing modes.
#' @param data_path The data path used to build the network
#'
#' @return A df travel time matrix with origins and destinations unchanged or
#'         in swapped order
#'
#' @family setting functions
#'
#' @keywords internal
reverse_back_if_direct_mode <- function(travel_times, origins, destinations, mode_list, data_path){

  # In direct modes, reverse origin/destination to take advantage of R5's One to Many algorithm
  if (
    mode_list$transit_mode == "" &&
    mode_list$direct_modes %in% c("WALK") &&
    nrow(origins) < nrow(destinations) &&
    !exists_tiff(data_path) # only if no elevation data is present
  ) {
    # rename and redorder columns
    data.table::setnames(
      x = travel_times,
      old = c('from_id', 'to_id'),
      new = c('to_id', 'from_id')
    )

    travel_times <- data.table::setcolorder(travel_times, c('from_id', 'to_id'))
  }

    return(travel_times)

}




#' Set elevation
#'
#' Verifies whether elevation mode is correct.
#'
#' @param elevation Character.
#'
#' @return Character. Corretly formatted elevation.
#' @family setting functions
#'
#' @keywords internal
set_elevation <- function(elevation) {
  elevation <- toupper(elevation)
  valid_elev <- c("TOBLER", "MINETTI", "NONE")
  if (!elevation %in% valid_elev) {
    cli::cli_abort(c(
      "Invalid value for {.arg elevation}: {.val {elevation}}.",
      "x" = "Must be one of: {.val {valid_elev}}")
    )
  }

  elevation
}

#' Set car congestion
#'
#' Verifies if and which congestion mode to use and applies it.
#'
#' @template r5r_network
#' @param new_carspeeds A df or sf polygon.
#' @param carspeed_scale Numeric > 0.
#'
#' @return Invisibly returns `TRUE`.
#' @family setting functions
#'
#' @keywords internal
set_new_congestion <- function(r5r_network, new_carspeeds, carspeed_scale) {
  checkmate::assert_class(new_carspeeds, "data.frame", null.ok = T)
  checkmate::assert_numeric(carspeed_scale, lower = 0, finite = TRUE, null.ok = F)
  if (!is.null(new_carspeeds) || carspeed_scale != 1){
    cli::cli_inform("Modifying carspeeds...")
    cli::cli_warn("A scenario was used for this calculation. This may affect results for WALK or BICYCLE due to missing elevation. See issue #555.")

    if (inherits(new_carspeeds, "sf")) { # polygon mode
      geojson_path <- congestion_poly2geojson(new_carspeeds)
      errors <- r5r_network$applyCongestionPolygon(geojson_path,
                                               "scale",
                                               "priority",
                                               "poly_id",
                                               rJava::.jfloat(carspeed_scale))
    } else { # OSM mode or scale != 1
      speed_map <- dt_to_speed_map(new_carspeeds)
      absolute_mode <- !is.null(new_carspeeds) && (new_carspeeds$speed_type[1] == "km/h")
      errors <- r5r_network$applyCongestionOsm(speed_map,
                                     rJava::.jfloat(carspeed_scale),
                                     absolute_mode)
    }

    if (errors != "[]"){
      cli::cli_inform(c(
        "!" = "Encountered the following errors modifying carspeeds:",
        " " = errors ))
    }
  }
}


#' Set LTS level
#'
#' Verifies if and which LTS mode to use and applies it.
#'
#' @template r5r_network
#' @param new_lts A df or sf polygon.
#'
#' @return Invisibly returns `TRUE`.
#' @family setting functions
#'
#' @keywords internal
set_new_lts <- function(r5r_network, new_lts) {
  checkmate::assert_class(new_lts, "data.frame", null.ok = TRUE)
  if (!is.null(new_lts)){
    cli::cli_inform("Modifying LTS levels...")
    cli::cli_warn("A scenario was used for this calculation. This may affect results for WALK or BICYCLE due to missing elevation. See issue #555.")

    if (inherits(new_lts, "sf")) { # polygon mode
      shp_path <- lts_lines2shp(new_lts)
      errors <- r5r_network$applyLtsPolygon(shp_path)
    } else { # OSM mode
      lts_map <- dt_to_lts_map(new_lts)
      errors <- r5r_network$applyLtsOsm(lts_map)
    }
    if (errors != "[]"){
      cli::cli_inform(c(
        "!" = "Encountered the following errors modifying LTS:",
        " " = errors ))
    }
  }
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

tester <- function(progress) set_progress(r5r_network@jcore, progress)

test_that("input is correct", {
  expect_error(tester("TRUE"))
  expect_error(tester(c(TRUE, TRUE)))
  expect_error(tester(NA))
})

test_that("progress argument works in routing functions", {
  assert_function <- function(f) {
    expr <- if (identical(f, travel_time_matrix)) {
      "f(r5r_network, origins=pois[1:5], destinations=pois[1:5])"
    } else if (identical(f, expanded_travel_time_matrix)) {
      "f(r5r_network, origins=pois[1:5], destinations=pois[1:5])"
    } else if (identical(f, detailed_itineraries)) {
      "f(r5r_network, origins=pois[1:5], destinations=pois[5:1])"
    } else if (identical(f, pareto_frontier)) {
      "f(
        r5r_network,
        origins=pois[1:5],
        destinations=pois[1:5],
        departure_datetime = departure_datetime,
        fare_structure = fare_structure,
        fare_cutoffs = c(0, 5, 10),
        max_trip_duration = 60
      )"
    } else if (identical(f, accessibility)) {
      "f(
        r5r_network=r5r_network,
        origins = points[1:5],
        destinations = points[1:5],
        opportunities_colnames = \"jobs\" ,
        cutoffs = 60
      )"
    }

    progress_expr <- sub("\\)$", ", progress = TRUE\\)", expr)
    non_progress_expr <- sub("\\)$", ", progress = FALSE\\)", expr)

    progress_regex <- "\\d+ out of \\d+ origins processed\\."


    log_file <- file.path(r5r_network@jcore$getLogPath())
    # Clean log before test
    if (file.exists(log_file)) writeLines("", log_file)

    res <- eval(parse(text = progress_expr))
    # Wait a bit to ensure Java flushed the logs
    Sys.sleep(0.2)
    progress_messages <- readLines(log_file)
    expect_true(any(grepl(progress_regex, progress_messages)))

    # reset log again
    writeLines("", log_file)
    res <- eval(parse(text = non_progress_expr))
    # Wait a bit to ensure Java flushed the logs
    Sys.sleep(0.2)
    non_progress_messages <- readLines(log_file)
    expect_false(any(grepl(progress_regex, non_progress_messages)))
  }

  assert_function(travel_time_matrix)
  assert_function(expanded_travel_time_matrix)
  assert_function(detailed_itineraries)
  assert_function(pareto_frontier)
  assert_function(accessibility)
})

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

tester <- function(n_threads) set_n_threads(r5r_network@jcore, n_threads)

test_that("input is correct", {
  expect_error(tester("1"))
  expect_error(tester(c(1, 1)))
  expect_error(tester(0))
})

test_that("number of threads is set correctly", {
  expect_true(tester(1))
  expect_true(r5r_network@jcore$getNumberOfThreads() == 1)

  expect_true(tester(Inf))
  expect_true(r5r_network@jcore$getNumberOfThreads() == parallel::detectCores() - 1)
})

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

tester <- function(verbose) set_verbose(r5r_network@jcore, verbose)

test_that("input is correct", {
  expect_error(tester("TRUE"))
  expect_error(tester(c(TRUE, TRUE)))
  expect_error(tester(NA))
})

test_that("verbose argument works in routing functions", {
  assert_function <- function(f) {
    expr <- if (identical(f, travel_time_matrix)) {
      "f(r5r_network=r5r_network, origins=pois[1:5], destinations=pois[1:5])"
    } else if (identical(f, expanded_travel_time_matrix)) {
      "f(r5r_network=r5r_network, origins=pois[1:5], destinations=pois[1:5])"
    } else if (identical(f, detailed_itineraries)) {
      "f(r5r_network=r5r_network, origins=pois[1:5], destinations=pois[5:1])"
    } else if (identical(f, pareto_frontier)) {
      "f(
        r5r_network,
        origins=pois[1:5],
        destinations=pois[1:5],
        departure_datetime = departure_datetime,
        fare_structure = fare_structure,
        fare_cutoffs = c(0, 5, 10),
        max_trip_duration = 60
      )"
    } else if (identical(f, accessibility)) {
      "f(
        r5r_network,
        origins=points[1:5],
        destinations=points[1:5],
        opportunities_colnames = \"schools\",
        cutoffs = 60
      )"
    }

    verbose_expr <- sub("\\)$", ", verbose = TRUE\\)", expr)
    non_verbose_expr <- sub("\\)$", ", verbose = FALSE\\)", expr)

    info_regex <- "\\[\\s*(INFO|ERROR|WARN)\\s*\\]"

    log_file <- file.path(r5r_network@jcore$getLogPath())
    # Clean log before test
    if (file.exists(log_file)) writeLines("", log_file)

    res <- eval(parse(text = non_verbose_expr))
    # Wait a bit to ensure Java flushed the logs
    Sys.sleep(0.2)
    non_verbose_messages <- readLines(log_file)
    expect_false(any(grepl(info_regex, non_verbose_messages)))

    # reset log again
    writeLines("", log_file)
    res <- eval(parse(text = verbose_expr))
    # Wait a bit to ensure Java flushed the logs
    Sys.sleep(0.2)
    verbose_messages <- readLines(log_file)

    expect_true(any(grepl(info_regex, verbose_messages)))
  }

  assert_function(travel_time_matrix)
  assert_function(expanded_travel_time_matrix)
  assert_function(detailed_itineraries)
  assert_function(pareto_frontier)
  assert_function(accessibility)
})

#' r5r: Rapid Realistic Routing with 'R5'
#'
#' Rapid realistic routing on multimodal transport networks (walk, bike, public
#' transport and car) using `R5`, the Rapid Realistic Routing on Real-world and
#' Reimagined networks engine <https://github.com/conveyal/r5>. The package
#' allows users to generate detailed routing analysis or calculate travel time
#' matrices using seamless parallel computing on top of the R5 Java machine.
#' While `R5` is developed by Conveyal, the package `r5r` is independently
#' developed by a team at the Institute for Applied Economic Research (Ipea)
#' with contributions from collaborators. Apart from the documentation in this
#' package, users will find additional information on `R5` documentation at
#' <https://docs.conveyal.com/>. Although we try to keep new releases of `r5r`
#' in synchrony with `R5`, the development of `R5` follows Conveyal's
#' independent update process. Hence, users should confirm if the `R5` version
#' implied by the Conveyal user manual (see
#' <https://docs.conveyal.com/changelog>) corresponds with the `R5` version
#' that `r5r` depends on.
#'
#' @section Usage:
#' Please check the vignettes on the [website](https://ipeagit.github.io/r5r/).
#'
#' @docType package
#' @name r5r
#' @aliases r5r-package
#'
#' @importFrom data.table := %between% %chin% %like% .N
#' @importFrom lifecycle deprecated
#'
#' @keywords internal
"_PACKAGE"

if (getRversion() >= "2.15.1") {
  utils::globalVariables(
    c(
      "duration",
      "from_id",
      "to_id",
      "option",
      ".SD",
      "geometry",
      "route",
      "temp_duration",
      "temp_route",
      "route",
      "temp_sign",
      ".I",
      "segment_duration",
      "total_duration",
      "wait",
      "release_date",
      "con",
      "start_lon",
      "start_lat",
      "end_lon",
      "end_lat",
      "slope",
      "lat",
      "lon",
      "walk_multiplier",
      "bike_multiplier",
      "found",
      ".",
      "%>%",
      ":=",
      "%like%",
      "%chin%",
      "set",
      "travel_time",
      "id_orig",
      "lat_orig",
      "lon_orig",
      "id_dest",
      "lat_dest",
      "lon_dest",
      "i.lon",
      "i.lat",
      "total_time",
      "setting",
      "cutoff",
      'travel_time_p50',
      'id',
      'i.travel_time_p50',
      'i.isochrone',
      'edge_index',
      'count',
      'active_on_date'
    )
  )
}


## usethis namespace: end
NULL

# options(java.parameters = '-Xmx2G')

# library(r5r)
devtools::load_all(".")
library(data.table)
library(tidyverse)

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path = data_path, verbose = TRUE, overwrite = FALSE,
                     temp_dir = FALSE)

# load origin/destination points

departure_datetime <- as.POSIXct("13-05-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S")
points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))

r5r_core$setBenchmark(TRUE)


# r5r_core$setTravelTimesBreakdown(FALSE)
  ttm <- expanded_travel_time_matrix(r5r_core,
                            origins = points,
                            destinations = points,
                            departure_datetime = departure_datetime,
                            breakdown = T,
                            mode = c("WALK", "TRANSIT"),
                            max_trip_duration = 60,
                            max_walk_dist = 800,
                            time_window = 15,
                            verbose = FALSE,
                            progress = TRUE)


# r5r_core$setTravelTimesBreakdown(TRUE)
t_ttm_breakdown <- system.time(
  ttm_b_mean <- travel_time_matrix(r5r_core,
                                   origins = points,
                                   destinations = points,
                                   departure_datetime = departure_datetime,
                                   breakdown = TRUE,
                                   breakdown_stat = "mean",
                                   mode = c("WALK", "TRANSIT"),
                                   max_trip_duration = 60,
                                   max_walk_dist = 800,
                                   time_window = 30,
                                   # percentiles = c(25),
                                   percentiles = c(25, 50, 75),
                                   verbose = FALSE,
                                   progress = FALSE)
)

# t_ttm_breakdown <- system.time(
#   ttm_b_min <- travel_time_matrix(r5r_core,
#                               origins = points[1:100, ],
#                               destinations = points,
#                               departure_datetime = departure_datetime,
#                               breakdown = TRUE,
#                               breakdown_stat = "minimum",
#                               mode = c("WALK", "TRANSIT"),
#                               max_trip_duration = 60,
#                               max_walk_dist = 800,
#                               time_window = 30,
#                               # percentiles = c(25),
#                               percentiles = c(25, 50, 75),
#                               verbose = FALSE)
# )

View(ttm_b_mean)

# ttm_first <- read_csv("ttm_first_option.csv")
# ttm_b_mean %>% write_csv(file = "ttm_first_option.csv")
# t_ttm_normal
# t_ttm_breakdown

# rbind(
#   ttm_n %>% select(fromId, execution_time) %>% distinct() %>% mutate(method = "normal"),
#   ttm_b_mean %>% select(fromId, execution_time) %>% distinct() %>% mutate(method = "breakdown")
# ) %>%
#   ggplot(aes(execution_time)) + geom_histogram() + facet_wrap(~method, scales = "free")
#


# max_trip_duration = 60
#
# travel_times = data.table::copy(ttm)
# for(j in seq(from = 3, to = (length(percentiles) + 3))){
#   data.table::set(travel_times, i=which(travel_times[[j]]>max_trip_duration), j=j, value=NA_integer_)
# }
#
# names
# colnames <- names(travel_times)[startsWith(names(travel_times), "travel_time")]
# for(j in colnames){
#   data.table::set(travel_times, i=which(travel_times[j]>max_trip_duration), j=j, value=NA_integer_)
# }
#
# length(c(25, 50, 75))

#
# ttm_n$saveToCsv("teste_ttm.csv")
#
# s_ttm <- read_csv("teste_ttm.csv")
#
#

ttm_b_mean %>%
  select(fromId, execution_time) %>%
  distinct() %>%
  summarise(t = sum(execution_time) / 1000 / 12)
t_ttm_breakdown

setClass("travel_time_surface", slots=list(
  matrix="matrix",
  zoom="integer",
  north="integer",
  west="integer",
  height="integer",
  width="integer"
))

#' Compute travel time surfaces.
#'
#' @description A travel time surface is a raster grid (in the Web Mercator projection)
#' containing travel times from a specified point.
#'
#' @template r5r_network
#' @param origins Either a `POINT sf` object with WGS84 CRS, or a
#'        `data.frame` containing the columns `id`, `lon` and `lat`.
#' @param zoom Resolution of the travel time surface used to create isochrones,
#'        can be between 9 and 12. More detailed isochrones will result from
#'        larger numbers, at the expense of compute time. Specifically, a raster
#'        grid of travel times in the Web Mercator projection at this zoom level
#'        is created, and the isochrones are interpolated from this grid. For
#'        more information on how the grid cells are defined, see
#'        \href{https://docs.conveyal.com/analysis/methodology#zoom-levels}{the R5 documentation.}
#' @param mode A character vector. The transport modes allowed for access,
#'        transfer and vehicle legs of the trips. Defaults to `WALK`. Please see
#'        details for other options.
#' @param mode_egress A character vector. The transport mode used after egress
#'        from the last public transport. It can be either `WALK`, `BICYCLE` or
#'        `CAR`. Defaults to `WALK`. Ignored when public transport is not used.
#' @param departure_datetime A POSIXct object. Please note that the departure
#'        time only influences public transport legs. When working with public
#'        transport networks, please check the `calendar.txt` within your GTFS
#'        feeds for valid dates. Please see details for further information on
#'        how datetimes are parsed.
#' @param time_window An integer. The time window in minutes for which `r5r`
#'        will calculate multiple travel time matrices departing each minute.
#'        Defaults to 10 minutes. The function returns the result based on
#'        median travel times. Please read the time window vignette for more
#'        details on its usage `vignette("time_window", package = "r5r")`
#' @param max_walk_time An integer. The maximum walking time (in minutes) to
#'        access and egress the transit network, or to make transfers within the
#'        network. Defaults to no restrictions, as long as `max_trip_duration`
#'        is respected. The max time is considered separately for each leg (e.g.
#'        if you set `max_walk_time` to 15, you could potentially walk up to 15
#'        minutes to reach transit, and up to _another_ 15 minutes to reach the
#'        destination after leaving transit). Defaults to `Inf`, no limit.
#' @param max_bike_time An integer. The maximum cycling time (in minutes) to
#'        access and egress the transit network. Defaults to no restrictions, as
#'        long as `max_trip_duration` is respected. The max time is considered
#'        separately for each leg (e.g. if you set `max_bike_time` to 15 minutes,
#'        you could potentially cycle up to 15 minutes to reach transit, and up
#'        to _another_ 15 minutes to reach the destination after leaving
#'        transit). Defaults to `Inf`, no limit.
#' @param max_car_time An integer. The maximum driving time (in minutes) to
#'        access and egress the transit network. Defaults to no restrictions, as
#'        long as `max_trip_duration` is respected. The max time is considered
#'        separately for each leg (e.g. if you set `max_car_time` to 15 minutes,
#'        you could potentially drive up to 15 minutes to reach transit, and up
#'        to _another_ 15 minutes to reach the destination after leaving transit).
#'        Defaults to `Inf`, no limit.
#' @param max_trip_duration An integer. The maximum trip duration in minutes.
#'        Defaults to 120 minutes (2 hours).
#' @param walk_speed A numeric. Average walk speed in km/h. Defaults to 3.6 km/h.
#' @param bike_speed A numeric. Average cycling speed in km/h. Defaults to 12 km/h.
#' @param max_rides An integer. The maximum number of public transport rides
#'        allowed in the same trip. Defaults to 3.
#' @param max_lts An integer between 1 and 4. The maximum level of traffic
#'        stress that cyclists will tolerate. A value of 1 means cyclists will
#'        only travel through the quietest streets, while a value of 4 indicates
#'        cyclists can travel through any road. Defaults to 2. Please see
#'        details for more information.
#'
#' @template draws_per_minute
#' @param n_threads An integer. The number of threads to use when running the
#'        router in parallel. Defaults to use all available threads (`Inf`).
#' @param progress A logical. Whether to show a progress counter when running
#'        the router. Defaults to `FALSE`. Only works when `verbose` is set to
#'        `FALSE`, so the progress counter does not interfere with `R5`'s output
#'        messages. Setting `progress` to `TRUE` may impose a small penalty for
#'        computation efficiency, because the progress counter must be
#'        synchronized among all active threads.
#' @template fare_structure
#' @template max_fare
#' @template scenarios
#' @template verbose
#' @template percentiles
#'
#' @return A `"sf" "data.frame"` for each isochrone of each origin.
#'
#' @template transport_modes_section
#' @template lts_section
#' @template datetime_parsing_section
#' @template raptor_algorithm_section
#'
#'
#' @family support functions
#'
#' @keywords internal
travel_time_surface <- function(r5r_network,
                               origins,
                               zoom = 10,
                               mode = "WALK",
                               mode_egress = "WALK",
                               departure_datetime = Sys.time(),
                               time_window = 10L,
                               percentiles = 50L,
                               max_walk_time = Inf,
                               max_bike_time = Inf,
                               max_car_time = Inf,
                               max_trip_duration = 120L,
                               walk_speed = 3.6,
                               bike_speed = 12,
                               max_rides = 3,
                               max_lts = 2,
                               fare_structure = NULL,
                               max_fare = Inf,
                               new_carspeeds = NULL,
                               carspeed_scale = 1,
                               new_lts = NULL,
                               draws_per_minute = 5L,
                               n_threads = Inf,
                               verbose = FALSE,
                               progress = FALSE
                      ) {
  # check inputs ------------------------------------------------------------
  checkmate::assert_class(r5r_network, "r5r_network")
  # R5 only supports grids between zoom 9 and 12
  checkmate::assert_numeric(zoom, lower=9, upper=12, len=1)
  r5r_network <- r5r_network@jcore

  origins <- assign_points_input(origins, "origins")
  mode_list <- assign_mode(mode, mode_egress)
  departure <- assign_departure(departure_datetime)

  max_walk_time <- assign_max_street_time(
    max_walk_time,
    walk_speed,
    max_trip_duration,
    "walk"
  )
  max_bike_time <- assign_max_street_time(
    max_bike_time,
    bike_speed,
    max_trip_duration,
    "bike"
  )
  max_car_time <- assign_max_street_time(
    max_car_time,
    8, # 8 km/h, R5's default.
    max_trip_duration,
    "car"
  )
  max_trip_duration <- assign_max_trip_duration(
    max_trip_duration,
    mode_list,
    max_walk_time,
    max_bike_time
  )

  set_time_window(r5r_network, time_window)
  set_percentiles(r5r_network, percentiles)
  set_monte_carlo_draws(r5r_network, draws_per_minute, time_window)
  set_speed(r5r_network, walk_speed, "walk")
  set_speed(r5r_network, bike_speed, "bike")
  set_max_rides(r5r_network, max_rides)
  set_max_lts(r5r_network, max_lts)
  set_n_threads(r5r_network, n_threads)
  set_verbose(r5r_network, verbose)
  set_progress(r5r_network, progress)
  set_fare_structure(r5r_network, fare_structure)
  set_max_fare(r5r_network, max_fare)
  set_expanded_travel_times(r5r_network, FALSE)
  set_breakdown(r5r_network, FALSE)
  r5r_network$setSearchType("DEPART_FROM")

  # SCENARIOS -------------------------------------------
  set_new_congestion(r5r_network, new_carspeeds, carspeed_scale)
  set_new_lts(r5r_network, new_lts)

  surfaces <- rJava::.jcall(
    r5r_network,
    "[Lorg/ipea/r5r/RegularGridResult;",
    "travelTimeSurfaces",
    check=FALSE,
    origins$id, origins$lat, origins$lon,
    mode_list$direct_modes,
    mode_list$transit_mode,
    mode_list$access_mode,
    mode_list$egress_mode,
    departure$date,
    departure$time,
    max_walk_time,
    max_bike_time,
    max_car_time,
    max_trip_duration,
    as.integer(zoom)
  )
  ex <- rJava::.jgetEx(clear=TRUE)
  if (!is.null(ex)) {
    ex$printStackTrace()
    cli::cli_abort("Internal R5 error (see detailed message above)")
  }

  # convert from java objects to R travel_time_surface class
  return(lapply(surfaces, process_surfaces))
}

# process each percentile of a travel time surface
process_surfaces <- function (sfaces) {
  result = list()

  for (i in seq_along(sfaces$percentiles)) {
    # R has issues with nonconsecutive ints being used as list indices, so use strings
    result[[paste0("p", sfaces$percentiles[i])]] = process_surface(
      sfaces$values[i,],
      sfaces$zoom,
      sfaces$west,
      sfaces$north,
      sfaces$width,
      sfaces$height
      )
  }

  return(result)
}

# process a single surface into an R object
process_surface <- function (sface, zoom, west, north, width, height) {
  methods::new("travel_time_surface",
    matrix=matrix(as.double(sface), height, width, byrow = TRUE),
    zoom = zoom,
    width = width,
    height = height,
    north = north,
    west = west
  )
}

#' Calculate travel time and monetary cost Pareto frontier
#'
#' Fast computation of travel time and monetary cost Pareto frontier between
#' origin and destination pairs.
#'
#' @template r5r_core
#' @template common_arguments
#' @template time_window_related_args
#' @template fare_structure
#' @template max_fare
#' @template verbose
#'
#' @return A `data.table` with the travel time and monetary cost Pareto
#' frontier between the specified origins and destinations. An additional
#' column identifying the travel time percentile is present if more than one
#' value was passed to `percentiles`. Origin and destination pairs whose trips
#' couldn't be completed within the maximum travel time using less money than
#' the specified monetary cutoffs are not returned in the `data.table`. If
#' `output_dir` is not `NULL`, the function returns the path specified in that
#' parameter, in which the `.csv` files containing the results are saved.
#'
#' @template transport_modes_section
#' @template lts_section
#' @template datetime_parsing_section
#' @template mcraptor_algorithm_section
#'
#' @family routing
#'
#' @keywords internal
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' # build transport network
#' data_path <- system.file("extdata/poa", package = "r5r")
#' r5r_core <- setup_r5(data_path = data_path)
#'
#' # load origin/destination points
#' points <- read.csv(file.path(data_path, "poa_hexgrid.csv"))[1:5,]
#'
#' # load fare structure object
#' fare_structure_path <- system.file(
#'   "extdata/poa/fares/fares_poa.zip",
#'   package = "r5r"
#' )
#' fare_structure <- read_fare_structure(fare_structure_path)
#'
#' departure_datetime <- as.POSIXct(
#'   "13-05-2019 14:00:00",
#'   format = "%d-%m-%Y %H:%M:%S"
#' )
#'
#' pf <- pareto_frontier(
#'   r5r_core,
#'   origins = points,
#'   destinations = points,
#'   mode = c("WALK", "TRANSIT"),
#'   departure_datetime = departure_datetime,
#'   fare_structure = fare_structure,
#'   monetary_cost_cutoffs = c(4.5, 4.8, 9, 9.3, 9.6)
#' )
#' head(pf)
#'
#' stop_r5(r5r_core)
#' @export
pareto_itineraries <- function(r5r_core,
                            origins,
                            destinations,
                            mode = c("WALK", "TRANSIT"),
                            mode_egress = "WALK",
                            departure_datetime = Sys.time(),
                            time_window = 1L,
                            percentiles = 50L,
                            max_walk_dist = Inf,
                            max_bike_dist = Inf,
                            max_trip_duration = 120L,
                            fare_structure = NULL,
                            max_fare = Inf,
                            walk_speed = 3.6,
                            bike_speed = 12,
                            max_rides = 3,
                            max_lts = 2,
                            draws_per_minute = 5L,
                            n_threads = Inf,
                            verbose = FALSE,
                            progress = FALSE,
                            output_dir = NULL) {

  old_options <- options(datatable.optimize = Inf)
  on.exit(options(old_options), add = TRUE)

  old_dt_threads <- data.table::getDTthreads()
  dt_threads <- ifelse(is.infinite(n_threads), 0, n_threads)
  data.table::setDTthreads(dt_threads)
  on.exit(data.table::setDTthreads(old_dt_threads), add = TRUE)


  # check inputs ------------------------------------------------------------

  # r5r_core
  checkmate::assert_class(r5r_core, "jobjRef")

  # modes
  mode_list <- assign_mode(mode, mode_egress)

  # departure time
  departure <- posix_to_string(departure_datetime)

  # max trip duration
  checkmate::assert_numeric(max_trip_duration, lower=1)
  max_trip_duration <- as.integer(max_trip_duration)

  # max_walking_distance, max_bike_distance, and max_street_time
  max_walk_time <- set_max_street_time(max_walk_dist,
                                       walk_speed,
                                       max_trip_duration)
  max_bike_time <- set_max_street_time(max_bike_dist,
                                       bike_speed,
                                       max_trip_duration)

  # origins and destinations
  origins      <- assert_points_input(origins, "origins")
  destinations <- assert_points_input(destinations, "destinations")

  checkmate::assert_subset("id", names(origins))
  checkmate::assert_subset("id", names(destinations))

  # time window
  checkmate::assert_numeric(time_window, lower=1)
  time_window <- as.integer(time_window)

  # montecarlo draws per minute
  draws <- time_window * draws_per_minute
  draws <- as.integer(draws)

  # percentiles
  if (length(percentiles) > 5) {
    stop("Maximum number of percentiles allowed is 5.")
  }
  percentiles <- percentiles[!is.na(percentiles)]
  checkmate::assert_numeric(percentiles)
  percentiles <- as.integer(percentiles)

  # set r5r_core options ----------------------------------------------------

  if (!is.null(output_dir)) r5r_core$setCsvOutput(output_dir)
  on.exit(r5r_core$setCsvOutput(""), add = TRUE)

  # time window
  r5r_core$setTimeWindowSize(time_window)
  r5r_core$setPercentiles(percentiles)
  r5r_core$setNumberOfMonteCarloDraws(draws)

  # set bike and walk speed
  set_speed(r5r_core, walk_speed, "walk")
  set_speed(r5r_core, bike_speed, "bike")

  # set max transfers
  set_max_rides(r5r_core, max_rides)

  # set max lts (level of traffic stress)
  set_max_lts(r5r_core, max_lts)

  # set number of threads to be used by r5 and data.table
  set_n_threads(r5r_core, n_threads)

  # set verbose
  set_verbose(r5r_core, verbose)

  # set progress
  set_progress(r5r_core, progress)

  # fare structure
  set_fare_structure(r5r_core, fare_structure)

  # set max fare
  # Inf and NULL values are not allowed in Java,
  # so -1 is used to indicate max_fare is unconstrained
  if (max_fare != Inf) {
    r5r_core$setMaxFare(rJava::.jfloat(max_fare))
  } else {
    r5r_core$setMaxFare(rJava::.jfloat(-1.0))
  }

  # call r5r_core method ----------------------------------------------------

  travel_times <- r5r_core$paretoItineraries(origins$id,
                                            origins$lat,
                                            origins$lon,
                                            destinations$id,
                                            destinations$lat,
                                            destinations$lon,
                                            mode_list$direct_modes,
                                            mode_list$transit_mode,
                                            mode_list$access_mode,
                                            mode_list$egress_mode,
                                            departure$date,
                                            departure$time,
                                            max_walk_time,
                                            max_bike_time,
                                            max_trip_duration)


  # process results ---------------------------------------------------------

  # convert travel_times from java object to data.table
  if (!verbose & progress) { cat("Preparing final output...") }

  travel_times <- java_to_dt(travel_times)

  if (!verbose & progress) { cat(" DONE!\n") }

  if (!is.null(output_dir)) return(output_dir)
  return(travel_times)
}

# options(java.parameters = '-Xmx2G')


# load stuff --------------------------------------------------------------

# library(r5r)
devtools::load_all(".")
library(data.table)
library(tidyverse)
library(sf)
library(mapview)

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path = data_path, verbose = FALSE, elevation = "none")

f <- setup_fare_structure(r5r_core, 5)
# load points
points <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))

# load fare structure
fare_structure_path <- system.file("extdata/poa/fares/fares_poa.zip", package = "r5r")
fare_structure <- read_fare_structure(fare_structure_path)

diff_frontiers_df <- read_csv(here::here("diff_frontiers.csv"))



# check differences -------------------------------------------------------


# find itineraries --------------------------------------------------------

plot_pareto <- function(df) {
  o <- df$from_id[1]
  d <- df$to_id[1]

  orig <- points  %>% filter(id == o)
  dest <-  points %>% filter(id == d)

  r5r_core$setTravelAllowance(TRUE)
  full_frontiers_with_ta_df <- pareto_frontier(r5r_core,
                                               origins = orig,
                                               destinations = dest,
                                               departure_datetime = as.POSIXct("13-05-2019 14:00:00",
                                                                               format = "%d-%m-%Y %H:%M:%S"),
                                               mode = c("WALK", "TRANSIT"),
                                               max_trip_duration = 180,
                                               fare_structure = fare_structure,
                                               monetary_cost_cutoffs = c(1, 4.5, 4.8, 7.20, 8.37, 11.4, 12.57),
                                               max_rides = 5,
                                               progress = TRUE
  )

  r5r_core$setTravelAllowance(FALSE)
  full_frontiers_no_ta_df <- pareto_frontier(r5r_core,
                                             origins = orig,
                                             destinations = dest,
                                             departure_datetime = as.POSIXct("13-05-2019 14:00:00",
                                                                             format = "%d-%m-%Y %H:%M:%S"),
                                             mode = c("WALK", "TRANSIT"),
                                             max_trip_duration = 180,
                                             fare_structure = fare_structure,
                                             monetary_cost_cutoffs = c(1, 4.5, 4.8, 7.20, 8.37, 11.4, 12.57),
                                             max_rides = 5,
                                             progress = TRUE)

  full_frontiers_with_ta_df$scenario <- "with_ta"
  full_frontiers_no_ta_df$scenario <- "no_ta"

  full_frontiers <- rbind(full_frontiers_no_ta_df, full_frontiers_with_ta_df)

  p <- full_frontiers %>%
    ggplot(aes(x=monetary_cost, y=travel_time, color=scenario)) +
    geom_step(position=position_dodge(width=0.2)) +
    geom_point(position=position_dodge(width=0.2))

  return(p)
}

plots <- lapply(1:16, function(i) plot_pareto(diff_frontiers_df[i, ]))

library(patchwork)

wrap_plots(plots, ncol=4) + plot_layout(guides = "collect")


plot_pareto(diff_frontiers_df[1,])
plot_pareto(diff_frontiers_df[9,])
plot_pareto(diff_frontiers_df[15,])







system.time(
  frontiers_df <- pareto_itineraries(r5r_core,
                                     # origins = poi[10, ],
                                     # destinations = poi[12, ],
                                     # origins = poi[c(10, 10), ],
                                     # destinations = poi[c(9, 12), ],
                                     origins = orig,
                                     destinations = dest,
                                     departure_datetime = as.POSIXct("13-05-2019 14:00:00",
                                                                     format = "%d-%m-%Y %H:%M:%S"),
                                     time_window = 30,
                                     mode = c("WALK", "TRANSIT"),
                                     max_trip_duration = 90,
                                     max_walk_dist = 1000,
                                     fare_structure = fare_structure,
                                     max_fare = 10.0,
                                     max_rides = 5,
                                     progress = TRUE)
)

#' Java object to data.table
#'
#' @description Converts a Java object returned by r5r_network to an R `data.table`
#'
#' @param obj A Java Object reference
#'
#' @return An R data.table
#' @family java support functions
#'
#' @keywords internal
java_to_dt <- function(obj) {

  # check input
  if(class(obj)[1] != "jobjRef"){
    stop("Input must be an object of class 'jobjRef'")}

  # get column names from Java table
  columns <- obj$getColumnNames()

  # get the contents of each column in a vector, and return them in a list
  dt <- lapply(columns, function(column_name) {
    # check column data type, so we can call the appropriate Java function
    column_type <- obj$getColumnType(column_name)

    if (column_type == "String") { v <- obj$getStringColumn(column_name) }
    if (column_type == "Integer") { v <- obj$getIntegerColumn(column_name) }
    if (column_type == "Long") { v <- obj$getLongColumn(column_name) }
    if (column_type == "Double") { v <- obj$getDoubleColumn(column_name) }
    if (column_type == "Boolean") { v <- obj$getBooleanColumn(column_name) }
    return(v)
  })

  # convert list of vectors to a data.table, and rename columns accordingly
  data.table::setDT(dt)
  data.table::setnames(dt, new = columns)
}

#' data.table to speedMap
#'
#' @description Converts a `data.frame` with road OSM id's and respective speeds
#'              to a Java Map<Long, Float> for use by r5r_network.
#'
#' @param dt data.frame/data.table. Table specifying the
#'        speed modifications. The table must contain columns \code{osm_id} and
#'        \code{max_speed}.
#' @return A speedMap (Java HashMap<Long, Float>)
#' @family java support functions
#' @keywords internal
dt_to_speed_map <- function(dt) {
  if (is.null(dt)){
    return (rJava::.jnew("java/util/HashMap"))
  }

  checkmate::assert_names(names(dt), must.include = c("osm_id", "max_speed", "speed_type"))
  checkmate::assert_numeric(dt$osm_id, any.missing = FALSE, all.missing = FALSE)
  checkmate::assert_numeric(dt$max_speed, any.missing = FALSE, all.missing = FALSE)
  checkmate::assert_true(length(unique(dt$speed_type)) == 1 && dt$speed_type[1] %in% c("km/h", "scale"))

  # Create new HashMap<long, float>
  map_builder <- rJava::.jnew("org.ipea.r5r.Utils.RMapBuilder")
  speed_map <- map_builder$buildSpeedMap(paste(as.character(dt$osm_id), collapse = ","),
                                        paste(as.character(dt$max_speed), collapse = ","))

  return(speed_map)
}


#' Determine the Java version installed locally
#'
#' @return The number of the Java version
#' @family java support functions
#' @keywords internal
get_java_version <- function(){
  ver <- rJava::.jcall("java.lang.System", "S", "getProperty", "java.version")
  ver <- as.numeric(gsub("\\..*", "", ver))
  return(ver)
}


#' data.table to ltsMap
#'
#' @description Converts a `data.frame` with road OSM id's and respective LTS
#'              levels a Java Map<Long, Integer> for use by r5r_network.
#'
#' @param dt data.frame/data.table. Table specifying the
#'        LTS levels. The table must contain columns \code{osm_id} and
#'        \code{lts}.
#' @return A speedMap (Java HashMap<Long, Integer>)
#' @family java support functions
#' @keywords internal
dt_to_lts_map <- function(dt) {
  checkmate::assert_names(names(dt), must.include = c("osm_id", "lts"))
  checkmate::assert_numeric(dt$osm_id, any.missing = FALSE, all.missing = FALSE)
  checkmate::assert_integer(dt$lts, any.missing = FALSE, all.missing = FALSE)

  # Create new HashMap<Long, Integer>
  map_builder <- rJava::.jnew("org.ipea.r5r.Utils.RMapBuilder")
  lts_map <- map_builder$buildLtsMap(paste(as.character(dt$osm_id), collapse = ","),
                                         paste(as.character(dt$lts), collapse = ","))

  return(lts_map)
}

# library("r5r")
devtools::load_all(".")
library("data.table")
library("tidyverse")
library("sf")
library("gtfstools")

data_path <- here::here("tests_marcus/kinshasa")
r5r_core <- r5r::setup_r5(data_path, verbose = FALSE)



# build fare calculator ---------------------------------------------------

fare_calculator <- r5r::setup_fare_calculator(r5r_core, 100)


  # adjust the fare values based on the fuel_price_change. the weight of the
  # fuel price on the final fare depends on the agency that operates each mode.
  # round the price to keep all fares as integers

  gtfs <- gtfstools::read_gtfs(here::here("tests_marcus/kinshasa", "Time.zip"))

  # assign the fare_type based on the fare_id of each route
  # we use a relational vector to get the fare id based on the route id

  fare_id_per_route <- gtfs$fare_rules$fare_id
  names(fare_id_per_route) <- gtfs$fare_rules$route_id

  fare_calculator$fares_per_route[
    ,
    fare_type := paste0(fare_type, "_FAREID_", fare_id_per_route[route_id])
  ]

  # update the fare values based on the values listed on fare_attributes

  new_fare_per_mode <- data.table::data.table(
    mode = paste0("BUS_FAREID_", gtfs$fare_attributes$fare_id),
    unlimited_transfers = FALSE,
    allow_same_route_transfer = FALSE,
    use_route_fare = FALSE,
    fare = gtfs$fare_attributes$price
  )
  fare_calculator$fares_per_mode <- new_fare_per_mode

  # remove any discounted transfers from the fare calculator

  fare_calculator$fares_per_transfer <- data.table::data.table()

  # update routes_info to display the correct fares - won't change anything,
  # because the values will be taken from fare_per_mode, but just makes
  # everything clearer

  fare_calculator$fares_per_route[
    fare_calculator$fares_per_mode,
    on = c(fare_type = "mode"),
    route_fare := i.fare
  ]


# test fare calculator ----------------------------------------------------

  # get centroids coordinates

  grid_path <- here::here("tests_marcus/kinshasa", "Uber_H3_8.csv")
  grid_data <- data.table::fread(grid_path)

  sample = 100
  if (sample) grid_data <- grid_data[sample(1:nrow(grid_data), sample)]



  max_rides <- 5
  max_fare <- 2000
  monetary_cost_cutoffs <- c(500, 1000, 1500, 2000, 5000)
  # monetary_cost_cutoffs <- generate_cost_cutoffs(
  #   fare_calculator,
  #   "kinshasa",
  #   max_fare = 2000,
  #   max_rides = 3
  # )

  departure <- as.POSIXct("10-05-2021 07:00:00", format = "%d-%m-%Y %H:%M:%S")

  mapview::mapview(grid_data, xcol="lon", ycol="lat", crs=4326)
  # a <- r5r_core$getTransitServicesByDate("2021-05-10")
  # b <- java_to_dt(a)

  # sn <- street_network_to_sf(r5r_core)
  # mapview::mapview(sn$edges)

  fare_calculator$debug_settings$output_file <- here::here("tests_marcus/kinshasa/kins_debug.csv")

  frontier <- r5r::pareto_frontier(
    r5r_core,
    origins = grid_data,
    destinations = grid_data,
    mode = c("WALK", "TRANSIT"),
    departure_datetime = departure,
    time_window = 1,
    max_walk_dist = 2500,
    max_trip_duration = 180,
    max_rides = max_rides,
    fare_calculator_settings = fare_calculator,
    monetary_cost_cutoffs = monetary_cost_cutoffs,
    verbose = FALSE
  )

  ttm2 <- r5r::travel_time_matrix(
    r5r_core,
    origins = grid_data,
    destinations = grid_data,
    mode = c("WALK", "TRANSIT"),
    departure_datetime = departure,
    time_window = 1,
    max_walk_dist = 2500,
    max_trip_duration = 180,
    max_rides = max_rides,
    fare_calculator_settings = fare_calculator,
    max_fare = 1000,
    verbose = FALSE
  )

View(fare_calculator$fares_per_mode)
View(fare_calculator$fares_per_route)
View(fare_calculator$fares_per_transfer)



fare_json <- r5r_core$getFareStructure()
clipr::write_clip(fare_json)


ttm_c <- full_join(ttm1, ttm2, by = c("from_id", "to_id")) %>%
  mutate(diff = travel_time.x - travel_time.y)

dbg <- read_csv(here::here("tests_marcus/kinshasa/kins_debug.csv"))

#' Calculate access to opportunities
#'
#' Fast computation of access to opportunities given a selected decay function.
#'
#' @template r5r_network
#' @template r5r_core
#' @template common_arguments
#' @template time_window_related_args
#' @template draws_per_minute
#' @template fare_structure
#' @template max_fare
#' @template scenarios
#' @template verbose
#' @param opportunities_colnames A character vector. The names of the columns
#'   in the `destinations` input that tells the number of opportunities in each
#'   location. Several different column names can be passed, in which case the
#'   accessibility to each kind of opportunity will be calculated.
#' @template percentiles
#' @param decay_function A string. Which decay function to use when calculating
#'   accessibility. One of `step`, `exponential`, `fixed_exponential`, `linear`
#'   or `logistic`. Defaults to `step`, which is equivalent to a cumulative
#'   opportunities measure. Please see the details to understand how each
#'   alternative works and how they relate to the `cutoffs` and `decay_value`
#'   parameters.
#' @param cutoffs A numeric vector (maximum length of 12). This parameter has
#'   different effects for each decay function: it indicates the cutoff times
#'   in minutes when calculating cumulative opportunities accessibility with
#'   the `step` function, the median (or inflection point) of the decay curves
#'   in the `logistic` and `linear` functions, and the half-life in the
#'   `exponential` function. It has no effect when using the
#'   `fixed_exponential` function.
#' @param decay_value A number. Extra parameter to be passed to the selected
#'   `decay_function`. Has no effects when `decay_function` is either `step` or
#'   `exponential`.
#'
#' @return A `data.table` with accessibility estimates for all origin points.
#'   This `data.table` contain columns listing the origin id, the type of
#'   opportunities to which accessibility was calculated, the travel time
#'   percentile considered in the accessibility estimate and the specified
#'   cutoff values (except in when `decay_function` is `fixed_exponential`, in
#'   which case the `cutoff` parameter is not used). If `output_dir` is not
#'   `NULL`, the function returns the path specified in that parameter, in
#'   which the `.csv` files containing the results are saved.
#'
#' @template decay_functions_section
#' @template transport_modes_section
#' @template lts_section
#' @template datetime_parsing_section
#' @template raptor_algorithm_section
#'
#' @family accessibility
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' data_path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(data_path)
#' points <- read.csv(file.path(data_path, "poa_hexgrid.csv"))[1:500, ]
#'
#' departure_datetime <- as.POSIXct(
#'   "13-05-2019 14:00:00",
#'   format = "%d-%m-%Y %H:%M:%S"
#' )
#'
#' access <- accessibility(
#'   r5r_network ,
#'   origins = points,
#'   destinations = points,
#'   opportunities_colnames = "schools",
#'   mode = "WALK",
#'   departure_datetime = departure_datetime,
#'   decay_function = "step",
#'   cutoffs = 30,
#'   max_trip_duration = 30
#' )
#' head(access)
#'
#' # using a different decay function
#' access <- accessibility(
#'   r5r_network ,
#'   origins = points,
#'   destinations = points,
#'   opportunities_colnames = "schools",
#'   mode = "WALK",
#'   departure_datetime = departure_datetime,
#'   decay_function = "logistic",
#'   cutoffs = 30,
#'   decay_value = 1,
#'   max_trip_duration = 30
#' )
#' head(access)
#'
#' # using several cutoff values
#' access <- accessibility(
#'   r5r_network ,
#'   origins = points,
#'   destinations = points,
#'   opportunities_colnames = "schools",
#'   mode = "WALK",
#'   departure_datetime = departure_datetime,
#'   decay_function = "step",
#'   cutoffs = c(15, 30),
#'   max_trip_duration = 30
#' )
#' head(access)
#'
#' # calculating access to different types of opportunities
#' access <- accessibility(
#'   r5r_network ,
#'   origins = points,
#'   destinations = points,
#'   opportunities_colnames = c("schools", "healthcare"),
#'   mode = "WALK",
#'   departure_datetime = departure_datetime,
#'   decay_function = "step",
#'   cutoffs = 30,
#'   max_trip_duration = 30
#' )
#' head(access)
#'
#' stop_r5(r5r_network )
#'
#' @export
accessibility <- function(r5r_network,
                          r5r_core = deprecated(),
                          origins,
                          destinations,
                          opportunities_colnames = "opportunities",
                          mode = "WALK",
                          mode_egress = "WALK",
                          departure_datetime = Sys.time(),
                          time_window = 10L,
                          percentiles = 50L,
                          decay_function = "step",
                          cutoffs = NULL,
                          decay_value = NULL,
                          max_walk_time = Inf,
                          max_bike_time = Inf,
                          max_car_time = Inf,
                          max_trip_duration = 120L,
                          walk_speed = 3.6,
                          bike_speed = 12,
                          max_rides = 3,
                          max_lts = 2,
                          fare_structure = NULL,
                          max_fare = Inf,
                          new_carspeeds = NULL,
                          carspeed_scale = 1,
                          new_lts = NULL,
                          draws_per_minute = 5L,
                          n_threads = Inf,
                          verbose = FALSE,
                          progress = FALSE,
                          output_dir = NULL) {

  # deprecating r5r_core --------------------------------------
  if (lifecycle::is_present(r5r_core)) {

    cli::cli_warn(c(
      "!" = "The `r5r_core` argument is deprecated as of r5r v2.3.0.",
      "i" = "Please use the `r5r_network` argument instead."
    ))

    r5r_network <- r5r_core
  }

  old_options <- options(datatable.optimize = Inf)
  on.exit(options(old_options), add = TRUE)

  old_dt_threads <- data.table::getDTthreads()
  dt_threads <- ifelse(is.infinite(n_threads), 0, n_threads)
  data.table::setDTthreads(dt_threads)
  on.exit(data.table::setDTthreads(old_dt_threads), add = TRUE)

  # check inputs and set r5r options --------------------------------------

  checkmate::assert_class(r5r_network, "r5r_network")
  origins <- assign_points_input(origins, "origins")
  destinations <- assign_points_input(destinations, "destinations")
  opportunities <- assign_opportunities(destinations, opportunities_colnames)
  mode_list <- assign_mode(mode, mode_egress)
  departure <- assign_departure(departure_datetime)

  # check availability of transit services on the selected date
  if (mode_list$transit_mode %like% 'TRANSIT|TRAM|SUBWAY|RAIL|BUS|CABLE_CAR|GONDOLA|FUNICULAR') {
    check_transit_availability_on_date(r5r_network, departure_date = departure$date)
  }

  r5r_network <- r5r_network@jcore

  # cap trip duration with cutoffs
  set_cutoffs(r5r_network, cutoffs, decay_function)
  checkmate::assert_number(max_trip_duration, lower = 1, finite = TRUE)

  if(!is.null(cutoffs)){
    max_trip_duration <- ifelse(max_trip_duration > max(cutoffs), max(cutoffs), max_trip_duration)

    if(max_trip_duration < max(cutoffs)){stop("'max_trip_duration' cannot be shorter than 'max(cutoffs)'")}
  }

  max_walk_time <- assign_max_street_time(
    max_walk_time,
    walk_speed,
    max_trip_duration,
    "walk"
  )
  max_bike_time <- assign_max_street_time(
    max_bike_time,
    bike_speed,
    max_trip_duration,
    "bike"
  )
  max_car_time <- assign_max_street_time(
    max_car_time,
    8, # 8 km/h, R5's default.
    max_trip_duration,
    "car"
  )
  max_trip_duration <- assign_max_trip_duration(
    max_trip_duration,
    mode_list,
    max_walk_time,
    max_bike_time
  )


  decay_list <- assign_decay_function(decay_function, decay_value)

  set_time_window(r5r_network, time_window)
  set_percentiles(r5r_network, percentiles)
  set_monte_carlo_draws(r5r_network, draws_per_minute, time_window)
  set_speed(r5r_network, walk_speed, "walk")
  set_speed(r5r_network, bike_speed, "bike")
  set_max_rides(r5r_network, max_rides)
  set_max_lts(r5r_network, max_lts)
  set_n_threads(r5r_network, n_threads)
  set_verbose(r5r_network, verbose)
  set_progress(r5r_network, progress)
  set_fare_structure(r5r_network, fare_structure)
  set_max_fare(r5r_network, max_fare)
  set_output_dir(r5r_network, output_dir)

  # SCENARIOS -------------------------------------------
  set_new_congestion(r5r_network, new_carspeeds, carspeed_scale)
  set_new_lts(r5r_network, new_lts)


  # call r5r_network method and process results ------------------------------

  # wrap r5r_network inputs in arrays (this helps to simplify the Java code)

  from_id_arr <- rJava::.jarray(origins$id)
  from_lat_arr <- rJava::.jarray(origins$lat)
  from_lon_arr <- rJava::.jarray(origins$lon)

  to_id_arr <- rJava::.jarray(destinations$id)
  to_lat_arr <- rJava::.jarray(destinations$lat)
  to_lon_arr <- rJava::.jarray(destinations$lon)

  opportunities_names <- rJava::.jarray(opportunities_colnames)
  opportunities_values <- rJava::.jarray(opportunities, "[I")

  accessibility <- r5r_network$accessibility(
    from_id_arr,
    from_lat_arr,
    from_lon_arr,
    to_id_arr,
    to_lat_arr,
    to_lon_arr,
    opportunities_names,
    opportunities_values,
    decay_list$fun,
    decay_list$value,
    mode_list$direct_modes,
    mode_list$transit_mode,
    mode_list$access_mode,
    mode_list$egress_mode,
    departure$date,
    departure$time,
    max_walk_time,
    max_bike_time,
    max_car_time,
    max_trip_duration
  )

  if (!verbose & progress) cat("Preparing final output...", file = stderr())

  accessibility <- java_to_dt(accessibility)

  if (decay_function == "fixed_exponential") accessibility[, cutoff := NULL]

  if (!verbose & progress) cat(" DONE!\n", file = stderr())

  if (!is.null(output_dir)) return(output_dir)
  return(accessibility[])
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

departure_datetime <- as.POSIXct(
  "13-05-2019 14:00:00",
  format = "%d-%m-%Y %H:%M:%S"
)

tester <- function(r5r_network = get("r5r_network", envir = parent.frame()),
                   origins = points[1:10, ],
                   destinations = points[1:10, ],
                   opportunities_colname = "schools",
                   mode = "WALK",
                   mode_egress = "WALK",
                   departure_datetime = Sys.time(),
                   time_window = 1L,
                   percentiles = 50L,
                   decay_function = "step",
                   cutoffs = 10,
                   decay_value = NULL,
                   fare_structure = NULL,
                   max_fare = Inf,
                   max_walk_time = Inf,
                   max_bike_time = Inf,
                   max_trip_duration = 120L,
                   walk_speed = 3.6,
                   bike_speed = 12,
                   max_rides = 3,
                   max_lts = 2,
                   draws_per_minute = 5L,
                   n_threads = Inf,
                   verbose = FALSE,
                   progress = FALSE,
                   output_dir = NULL) {
  accessibility(
    r5r_network = r5r_network,
    origins = origins,
    destinations = destinations,
    opportunities_colname = opportunities_colname,
    mode = mode,
    mode_egress = mode_egress,
    departure_datetime = departure_datetime,
    time_window = time_window,
    percentiles = percentiles,
    decay_function = decay_function,
    cutoffs = cutoffs,
    decay_value = decay_value,
    fare_structure = fare_structure,
    max_fare = max_fare,
    max_walk_time = max_walk_time,
    max_bike_time = max_bike_time,
    max_trip_duration = max_trip_duration,
    walk_speed = walk_speed,
    bike_speed = bike_speed,
    max_rides = max_rides,
    max_lts = max_lts,
    draws_per_minute = draws_per_minute,
    n_threads = n_threads,
    verbose = verbose,
    progress = progress,
    output_dir = output_dir
  )
}


# errors and warnings -----------------------------------------------------


test_that("adequately raises errors", {

  # error related to using object with wrong type as r5r_network
  expect_error(tester("r5r_network"))

  # error related to using wrong origins/destinations object type
  multipoint_origins      <- sf::st_cast(sf::st_as_sf(points[1:2,], coords = c("lon", "lat")), "MULTIPOINT")
  multipoint_destinations <- multipoint_origins
  list_origins      <- list(id = c("1", "2"), lat = c(-30.02756, -30.02329), long = c(-51.22781, -51.21886))
  list_destinations <- list_origins

  expect_error(tester(origins = multipoint_origins))
  expect_error(tester(destinations = multipoint_destinations))
  expect_error(tester(origins = list_origins))
  expect_error(tester(destinations = list_destinations))
  expect_error(tester(origins = "origins"))
  expect_error(tester(destinations = "destinations"))

  # error/warning related to using wrong origins/destinations column types
  origins <- destinations <- points[1:2, ]

  origins_char_lat   <- data.frame(id = origins$id, lat = as.character(origins$lat), lon = origins$lon)
  origins_char_lon   <- data.frame(id = origins$id, lat = origins$lat, lon = as.character(origins$lon))
  destinations_char_lat   <- data.frame(id = destinations$id, lat = as.character(destinations$lat), lon = destinations$lon)
  destinations_char_lon   <- data.frame(id = destinations$id, lat = destinations$lat, lon = as.character(destinations$lon))

  expect_error(tester(origins = origins_char_lat))
  expect_error(tester(origins = origins_char_lon))
  expect_error(tester(destinations = destinations_char_lat))
  expect_error(tester(destinations = destinations_char_lon))

  # error related to nonexistent mode
  expect_error(tester(mode = "pogoball"))

  # errors related to date formatting
  numeric_datetime <- as.numeric(as.POSIXct("13-05-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S"))

  expect_error(tester(departure_datetime = "13-05-2019 14:00:00"))
  expect_error(tester(numeric_datetime))



  # errors related to max_trip_duration
  expect_error(tester(cutoffs = 10, max_trip_duration = 5))
  expect_error(tester(max_trip_duration = "1000"))
  expect_error(tester(max_trip_duration = NULL))

  # errors related to max_walk_time
  expect_error(tester(max_walk_time = "1000"))
  expect_error(tester(max_walk_time = NULL))

  # errors related to max_bike_time
  expect_error(tester(max_bike_time = "1000"))
  expect_error(tester(max_bike_time = NULL))


  # error related to non-numeric walk_speed
  expect_error(tester(walk_speed = "3.6"))

  # error related to non-numeric bike_speed
  expect_error(tester(bike_speed = "12"))

  # error related to too many or invalid percentiles
  expect_error(tester(percentiles = .3))
  expect_error(tester(percentiles = 1:6))

  # decay_function
  expect_error(tester(decay_function = "fixed_exponential"))
  expect_error(tester(decay_function = "bananas"))
  expect_error(tester(opportunities_colname = "bananas"))
  expect_error(tester(cutoffs = "bananas"))
  expect_error(tester(decay_value = "bananas"))
  expect_error(
    tester(decay_function = "fixed_exponential", decay_value = 0.5, cutoffs = 30)
  )

})




# adequate behavior ------------------------------------------------------


test_that("output is correct", {

  # decay functions
  expect_s3_class(tester(decay_function = "step", cutoffs = 30), "data.table")
  expect_s3_class(
    tester(decay_function = "exponential", cutoffs = 30),
    "data.table"
  )
  expect_s3_class(
    tester(decay_function = "linear", cutoffs = 30, decay_value = 1),
    "data.table"
  )
  expect_s3_class(
    tester(decay_function = "logistic", cutoffs = 30, decay_value = 1),
    "data.table"
  )
  expect_s3_class(
    tester(decay_function = "fixed_exponential", decay_value = 0.5, cutoffs = NULL),
    "data.table"
  )

  #  * output class ---------------------------------------------------------


  # expect results to be of class 'data.table', independently of the class of
  # 'origins'/'destinations'

  origins_sf <- destinations_sf <- sf::st_as_sf(
    points[1:10, ],
    coords = c("lon", "lat"),
    crs = 4326
  )

  result_df_input <- tester(cutoffs = 30)
  result_sf_input <- tester(
    origins = origins_sf,
    destinations = destinations_sf,
    cutoffs = 30
  )

  expect_s3_class(result_df_input, "data.table")
  expect_s3_class(result_sf_input, "data.table")

  # expect each column to be of right class

  expect_true(typeof(result_df_input$id) == "character")
  expect_true(typeof(result_df_input$accessibility) == "double")


  #  * r5r options ----------------------------------------------------------


  # access to multiple opportunities
  one_opport <- tester(cutoffs = 30)
  two_opport <- tester(
    opportunities_colname = c("schools", "healthcare"),
    cutoffs = 30
  )
  expect_true( nrow(two_opport) > nrow(one_opport))
  expect_true(is(two_opport, "data.table"))
})


test_that("using transit outside the gtfs dates throws an error", {
  expect_error(
    tester(r5r_network,
           mode='transit',
           departure_datetime = as.POSIXct("13-05-2025 14:00:00",
                                           format = "%d-%m-%Y %H:%M:%S")
    )
  )
})

options(java.parameters = "-Xmx2G")
library(r5r)
library(ggplot2)

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path = data_path)

# load origin/point of interest
points <- read.csv(file.path(data_path, "poa_hexgrid.csv"))
origin_1 <- points[936,]

departure_datetime <- as.POSIXct(
  "13-05-2019 14:00:00",
  format = "%d-%m-%Y %H:%M:%S"
)


iso_poly <- isochrone(r5r_core,
                       origins = origin_1,
                       mode = "walk",
                       polygon_output = T,
                       departure_datetime = departure_datetime,
                       cutoffs = seq(0, 100, 10)
)

ggplot() +
  geom_sf(data=iso_poly, aes(fill=factor(isochrone))) +
  scale_fill_manual(values = colors, name='Isochrone(min.)') +
  geom_point(data=origin_1, aes(x=lon, y=lat), color='red')+
  theme_void()

# estimate line-based isochrone from origin_1
iso_lines <- isochrone(r5r_core,
                      origins = origin_1,
                      mode = "walk",
                      polygon_output = F,
                      departure_datetime = departure_datetime,
                      cutoffs = seq(0, 100, 10)
)
head(iso_lines)


#### plot
colors <- c('#ffe0a5','#ffcb69','#ffa600','#ff7c43','#f95d6a',
            '#d45087','#a05195','#665191','#2f4b7c','#003f5c')


# lines
ggplot() +
  geom_sf(data=iso_lines, aes(color=factor(isochrone))) +
  scale_color_manual(values = colors, name='Isochrone(min.)') +
  geom_point(data=origin_1, aes(x=lon, y=lat), color='red')+
  theme_void()

#' Estimate isochrones from a given location
#'
#' @description Fast computation of isochrones from a given location. The
#' function can return either polygon-based or line-based isochrones.
#' Polygon-based isochrones are generated as concave polygons based on the
#' travel times from the trip origin to all nodes in the transport network.
#' Meanwhile, line-based isochronesare based on travel times from each origin
#' to the centroids of all segments in the transport network.
#'
#' @template r5r_network
#' @template r5r_core
#' @param origins Either a `POINT sf` object with WGS84 CRS, or a
#'        `data.frame` containing the columns `id`, `lon` and `lat`.
#' @param cutoffs numeric vector. Number of minutes to define the time span of
#'        each Isochrone. Defaults to `c(0, 15, 30)`.
#' @param zoom Resolution of the travel time surface used to create isochrones,
#'        can be between `9` and `12.` The default is `10` (which uses cells of
#'        153 meters at the Equator). More detailed isochrones will result from
#'        larger numbers, at the expense of compute time. Specifically, a raster
#'        grid of travel times in the Web Mercator projection at this zoom level
#'        is created, and the isochrones are interpolated from this grid. For
#'        more information on how the grid cells are defined, see
#'        \href{https://docs.conveyal.com/analysis/methodology#zoom-levels}{the R5 documentation.}
#' @param mode A character vector. The transport modes allowed for access,
#'        transfer and vehicle legs of the trips. Defaults to `WALK`. Please see
#'        details for other options.
#' @param mode_egress A character vector. The transport mode used after egress
#'        from the last public transport. It can be either `WALK`, `BICYCLE` or
#'        `CAR`. Defaults to `WALK`. Ignored when public transport is not used.
#' @param departure_datetime A POSIXct object. Please note that the departure
#'        time only influences public transport legs. When working with public
#'        transport networks, please check the `calendar.txt` within your GTFS
#'        feeds for valid dates. Please see details for further information on
#'        how datetimes are parsed.
#' @param polygon_output A Logical. If `TRUE`, the function outputs
#'        polygon-based isochrones (the default) based on travel times from each
#'        origin to a sample of a random  sample nodes in the transport network
#'        (see parameter `sample_size`). If `FALSE`, the function outputs
#'        line-based isochrones based on travel times from each origin to the
#'        centroids of all segments in the transport network.
#' @param time_window An integer. The time window in minutes for which `r5r`
#'        will calculate multiple travel time matrices departing each minute.
#'        Defaults to 10 minutes. The function returns the result based on
#'        median travel times. Please read the time window vignette for more
#'        details on its usage `vignette("time_window", package = "r5r")`
#' @param max_walk_time An integer. The maximum walking time (in minutes) to
#'        access and egress the transit network, or to make transfers within the
#'        network. Defaults to no restrictions, as long as `max_trip_duration`
#'        is respected. The max time is considered separately for each leg (e.g.
#'        if you set `max_walk_time` to 15, you could potentially walk up to 15
#'        minutes to reach transit, and up to _another_ 15 minutes to reach the
#'        destination after leaving transit). Defaults to `Inf`, no limit.
#' @param max_bike_time An integer. The maximum cycling time (in minutes) to
#'        access and egress the transit network. Defaults to no restrictions, as
#'        long as `max_trip_duration` is respected. The max time is considered
#'        separately for each leg (e.g. if you set `max_bike_time` to 15 minutes,
#'        you could potentially cycle up to 15 minutes to reach transit, and up
#'        to _another_ 15 minutes to reach the destination after leaving
#'        transit). Defaults to `Inf`, no limit.
#' @param max_car_time An integer. The maximum driving time (in minutes) to
#'        access and egress the transit network. Defaults to no restrictions, as
#'        long as `max_trip_duration` is respected. The max time is considered
#'        separately for each leg (e.g. if you set `max_car_time` to 15 minutes,
#'        you could potentially drive up to 15 minutes to reach transit, and up
#'        to _another_ 15 minutes to reach the destination after leaving transit).
#'        Defaults to `Inf`, no limit.
#' @param max_trip_duration An integer. The maximum trip duration in minutes.
#'        Defaults to 120 minutes (2 hours).
#' @param walk_speed A numeric. Average walk speed in km/h. Defaults to 3.6 km/h.
#' @param bike_speed A numeric. Average cycling speed in km/h. Defaults to 12 km/h.
#' @param max_rides An integer. The maximum number of public transport rides
#'        allowed in the same trip. Defaults to 3.
#' @param max_lts An integer between 1 and 4. The maximum level of traffic
#'        stress that cyclists will tolerate. A value of 1 means cyclists will
#'        only travel through the quietest streets, while a value of 4 indicates
#'        cyclists can travel through any road. Defaults to 2. Please see
#'        details for more information.
#' @template percentiles
#' @template draws_per_minute
#' @param n_threads An integer. The number of threads to use when running the
#'        router in parallel. Defaults to use all available threads (`Inf`).
#' @param progress A logical. Whether to show a progress counter when running
#'        the router. Defaults to `FALSE`. Only works when `verbose` is set to
#'        `FALSE`, so the progress counter does not interfere with `R5`'s output
#'        messages. Setting `progress` to `TRUE` may impose a small penalty for
#'        computation efficiency, because the progress counter must be
#'        synchronized among all active threads.
#' @template verbose
#' @param sample_size deprecated, no longer has any effect.
#'
#' @return A `"sf" "data.frame"` for each isochrone of each origin.
#'
#' @template transport_modes_section
#' @template lts_section
#' @template datetime_parsing_section
#' @template raptor_algorithm_section
#'
#' @family Isochrone
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' options(java.parameters = "-Xmx2G")
#' library(r5r)
#' library(ggplot2)
#'
#' # build transport network
#' data_path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(data_path = data_path)
#'
#' # load origin/point of interest
#' points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))
#' origin <- points[2,]
#'
#' departure_datetime <- as.POSIXct(
#'  "13-05-2019 14:00:00",
#'  format = "%d-%m-%Y %H:%M:%S"
#' )
#'
#' # estimate polygon-based isochrone from origin
#' iso_poly <- isochrone(
#'   r5r_network,
#'   origins = origin,
#'   mode = "walk",
#'   polygon_output = TRUE,
#'   departure_datetime = departure_datetime,
#'   cutoffs = seq(0, 120, 30)
#'   )
#'
#' head(iso_poly)
#'
#'
#' # estimate line-based isochrone from origin
#' iso_lines <- isochrone(
#'   r5r_network,
#'   origins = origin,
#'   mode = "walk",
#'   polygon_output = FALSE,
#'   departure_datetime = departure_datetime,
#'   cutoffs = seq(0, 100, 25)
#'   )
#'
#' head(iso_lines)
#'
#'
#' # plot colors
#' colors <- c('#ffe0a5','#ffcb69','#ffa600','#ff7c43','#f95d6a',
#'             '#d45087','#a05195','#665191','#2f4b7c','#003f5c')
#'
#' # polygons
#' ggplot() +
#'   geom_sf(data=iso_poly, aes(fill=factor(isochrone))) +
#'   scale_fill_manual(values = colors) +
#'   theme_minimal()
#'
#' # lines
#' ggplot() +
#'   geom_sf(data=iso_lines, aes(color=factor(isochrone))) +
#'   scale_color_manual(values = colors) +
#'   theme_minimal()
#'
#' stop_r5(r5r_network)
#'
#' @export
isochrone <- function(r5r_network,
                      r5r_core = deprecated(),
                      origins,
                      mode = "transit",
                      mode_egress = "walk",
                      cutoffs = c(0, 15, 30),
                      zoom = 10,
                      departure_datetime = Sys.time(),
                      polygon_output = TRUE,
                      time_window = 10L,
                      max_walk_time = Inf,
                      max_bike_time = Inf,
                      max_car_time = Inf,
                      max_trip_duration = 120L,
                      walk_speed = 3.6,
                      bike_speed = 12,
                      max_rides = 3,
                      max_lts = 2,
                      draws_per_minute = 5L,
                      # TODO this is unused for line-based isos
                      percentiles = NULL,
                      n_threads = Inf,
                      verbose = FALSE,
                      progress = TRUE,
                      # no longer used
                      sample_size = deprecated()
                      ){


  # deprecating r5r_core --------------------------------------
  if (lifecycle::is_present(r5r_core)) {

    cli::cli_warn(c(
      "!" = "The `r5r_core` argument is deprecated as of r5r v2.3.0.",
      "i" = "Please use the `r5r_network` argument instead."
    ))

    r5r_network <- r5r_core
  }

  # sample size no longer used
  if (lifecycle::is_present(sample_size)) {
    cli::cli_warn(c(
      "!" = "The `sample_size` argument is no longer used and has no effect."
    ))
  }

# check inputs ------------------------------------------------------------
  checkmate::assert_class(r5r_network, "r5r_network")

  # check cutoffs
  checkmate::assert_numeric(cutoffs, lower = 0)
  checkmate::assert_logical(polygon_output)

  # max cutoff is used as max_trip_duration
  max_trip_duration = as.integer(max(cutoffs))

  # sort cutoffs and include 0
  if (min(cutoffs) > 0) {cutoffs <- sort(c(0, cutoffs))}

  ## whether polygon- or line-based isochrones
  if (isTRUE(polygon_output)) {
    if (checkmate::test_null(percentiles)) {
      percentiles = 50L
    }

    # R5 only supports grids between zoom 9 and 12
    checkmate::assert_numeric(zoom, lower=9, upper=12, len=1)

    # create the travel time surfaces
    surfaces = travel_time_surface(r5r_network = r5r_network,
                              origins = origins,
                              mode = mode,
                              mode_egress = mode_egress,
                              departure_datetime = departure_datetime,
                              time_window = time_window,
                              percentiles = percentiles,
                              max_walk_time = max_walk_time,
                              max_bike_time = max_bike_time,
                              max_car_time = max_car_time,
                              max_trip_duration = max_trip_duration,
                              walk_speed = walk_speed,
                              bike_speed = bike_speed,
                              max_rides = max_rides,
                              max_lts = max_lts,
                              draws_per_minute = draws_per_minute,
                              n_threads = n_threads,
                              verbose = verbose,
                              zoom = zoom
    )

    # convert surfaces to isochrones
    # this uses a few named functions to traverse down the nested list
    # (percentiles inside origins)
    isos <- purrr::list_rbind(purrr::map2(surfaces, origins$id, percentiles_to_isodt, cutoffs))
    isos <- isos[, c("id", "isochrone", "percentile", "polygons")]
    return(sf::st_as_sf(isos, sf_column_name="polygons"))
  }

  if(isFALSE(polygon_output)){

    if (!checkmate::test_null(percentiles)) {
      cli::cli_abort("Percentiles not supported for line-based isochrones")
    }

    network_e <- r5r::street_network_to_sf(r5r_network)$edges

    destinations <- sf::st_centroid(network_e)
    }

  # rename id col
  names(destinations)[1] <- 'id'
  destinations$id <- as.character(destinations$id)


    # estimate travel time matrix
    ttm <- travel_time_matrix(r5r_network = r5r_network,
                              origins = origins,
                              destinations = destinations,
                              mode = mode,
                              mode_egress = mode_egress,
                              departure_datetime = departure_datetime,
                              time_window = time_window,
                              # percentiles = percentiles,
                              max_walk_time = max_walk_time,
                              max_bike_time = max_bike_time,
                              max_car_time = max_car_time,
                              max_trip_duration = max_trip_duration,
                              walk_speed = walk_speed,
                              bike_speed = bike_speed,
                              max_rides = max_rides,
                              max_lts = max_lts,
                              draws_per_minute = draws_per_minute,
                              n_threads = n_threads,
                              verbose = verbose,
                              progress = progress
                              )

    # ignore travel times equal to 0
    ttm <- ttm[travel_time_p50>0, ]

    # aggregate travel-times
    # ttm[, isochrone_interval := cut(x=travel_time_p50, breaks=cutoffs)]
    ttm[, isochrone := cut(x=travel_time_p50, breaks=cutoffs, labels=F)]
    ttm[, isochrone := cutoffs[cutoffs>0][isochrone]]


    # line-based isochrones
      prep_iso_lines <- function(orig){ # orig = '89a90128107ffff'

        temp_ttm <- subset(ttm, from_id == orig)

        # join ttm results to destinations
        temp_iso <- subset(network_e, edge_index %in% temp_ttm$to_id)
        data.table::setDT(temp_iso)[, edge_index := as.character(edge_index)]
        temp_iso[temp_ttm, on=c('edge_index' ='to_id'), c('travel_time_p50', 'isochrone') := list(i.travel_time_p50, i.isochrone)]
       # temp_iso <- sf::st_as_sf(temp_iso)

      temp_iso <- temp_iso[order(-isochrone, -travel_time_p50)]
      data.table::setcolorder(temp_iso, c('edge_index', 'osm_id', 'isochrone', 'travel_time_p50'))
      # plot(temp_iso)
      return(temp_iso)
    }


    # get the isocrhone from each origin
    iso_list <- lapply(X = unique(origins$id), FUN = prep_iso_lines)

    # put output together
    iso <- data.table::rbindlist(iso_list)
    iso <- sf::st_sf(iso)
    iso <- subset(iso, isochrone < Inf)

    # remove data.table from class
    class(iso) <- c("sf", "data.frame")
    return(iso)
  }


# Helper function to handle multiple percentiles
percentiles_to_isodt <- function (surfaceList, origin_id, cutoffs) {
  iso <- purrr::imap(surfaceList, percentile_to_isodt, cutoffs)
  iso <- purrr::list_rbind(iso)
  iso$id <- origin_id
  return(iso)
}

percentile_to_isodt <- function(surface, percentile, cutoffs) {
  iso <- surface_isochrone(surface, cutoffs)
  iso$percentile <- percentile
  return(iso)
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

departure_datetime <- as.POSIXct(
  "13-05-2019 14:00:00",
  format = "%d-%m-%Y %H:%M:%S"
)

tester <- function(r5r_network = get("r5r_network", envir = parent.frame()),
                   origins = pois[2,],
                   cutoffs = c(0, 30),
                   zoom = 10,
                   mode = "WALK",
                   mode_egress = "WALK",
                   departure_datetime = Sys.time(),
                   polygon_output = TRUE,
                   max_walk_time = Inf,
                   max_bike_time = Inf,
                   max_trip_duration = 120L,
                   walk_speed = 3.6,
                   bike_speed = 12,
                   max_rides = 3,
                   max_lts = 2,
                   n_threads = Inf,
                   verbose = FALSE,
                   progress = FALSE
                   ) {
  isochrone(
    r5r_network,
    origins = origins,
    cutoffs = cutoffs,
    zoom = zoom,
    mode = mode,
    mode_egress = mode_egress,
    departure_datetime = departure_datetime,
    polygon_output = polygon_output,
    max_walk_time = max_walk_time,
    max_bike_time = max_bike_time,
    max_trip_duration = max_trip_duration,
    walk_speed = walk_speed,
    bike_speed = bike_speed,
    max_rides = max_rides,
    max_lts = max_lts,
    n_threads = n_threads,
    verbose = verbose,
    progress = progress
    )
}

# tests -------------------------------------------------------------------

test_that("errors due to incorrect input types - origins and destinations", {
  multipoint_origins <- sf::st_cast(
    sf::st_as_sf(pois, coords = c("lon", "lat")),
    "MULTIPOINT"
  )
  list_destinations <- list_origins <- unclass(pois)

  expect_error(tester(origins = multipoint_origins))
  expect_error(tester(origins = list_origins))
  expect_error(tester(origins = "origins"))

  # wrong columns types

  pois_char_lat <- pois
  pois_char_lat$lat <- as.character(pois$lat)
  pois_char_lon <- pois
  pois_char_lon$lon <- as.character(pois$lon)

  expect_error(tester(origins = pois_char_lat))
  expect_error(tester(origins = pois_char_lon))
})

test_that("errors due to incorrect input types - other inputs", {
  # mode and mode_egress are tested in assign_mode() tests

  expect_error(tester(unclass(r5r_network)))

  expect_error(tester(departure_datetime = unclass(departure_datetime)))
  expect_error(tester(departure_datetime = rep(departure_datetime, 2)))

  expect_error(tester(cutoffs = "50"))
  expect_error(tester(cutoffs = -5))
  expect_error(tester(polygon_output = 'banana'))
  expect_error(tester(zoom=8))
  expect_error(tester(zoom=13))
  expect_error(tester(zoom="banana"))
})


# test polygon-based output
test_that("output is an sf with correct columns", {
  iso <- tester()
  expect_s3_class(iso, "sf")
  expect_identical(names(iso), c("id", "isochrone", "percentile", "polygons"))
  expect_type(iso$id, "character")
  expect_type(iso$isochrone, "double")
  expect_type(iso$polygons, "list")
  print(class(iso$polygons[1]))
  expect_true("sfc_POLYGON" %in% class(iso$polygons[1]))

  # more cutoffs means more rows

  iso2 <- tester(cutoffs = c(30, 50))
  expect_true(
    nrow(iso2) > nrow(iso)
  )
})


# test line-based output
test_that("output is an sf with correct columns", {
  iso <- tester(polygon_output = FALSE)
  expect_s3_class(iso, "sf")

  expect_type(iso$edge_index, "character")
  expect_type(iso$isochrone, "double")
  expect_true("sfc_LINESTRING" %in% class(iso$geometry[1]))
})

#' Get GTFS eventual errors encountered in network building
#'
#' This returns a data frame of GTFS errors R5 encountered when building the
#' network. You can call this with the network itself as the main parameter. If
#' network build fails, you won't have a network object, so you can also call
#' this with the `data_path` to where the network is stored.
#'
#' @param r5r_network the R5R network object, or a path to the location where
#'     the network is stored (useful if network build failed).
#'
#' @return A `data.frame`
#'
#' @family support functions
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' # directory with street network and gtfs files
#' data_path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(data_path)
#'
#' get_gtfs_errors(r5r_network)
#'
#' @export
get_gtfs_errors <- function (r5r_network) {
    checkmate::assert_multi_class(r5r_network, c("r5r_network", "character"))

    if (inherits(r5r_network, "r5r_network")) {
      r5r_network = r5r_network@jcore$getDataPath()
    }

    err_file = file.path(r5r_network, "gtfs_errors.csv")

    if (!file.exists(err_file)) {
      cli::cli_abort(c(
        x = "Error file {.path err_file} does not exist",
        i = "Perhaps the network has not been built or was built by an older version of r5r"
      ))
    }

    return(data.table::fread(err_file, header=T))
}

#' @param draws_per_minute An integer. The number of Monte Carlo draws to
#'   perform per time window minute when calculating travel time matrices and
#'   when estimating accessibility. Defaults to 5. This would mean 300 draws in
#'   a 60-minute time window, for example. This parameter only affects the
#'   results when the GTFS feeds contain a `frequencies.txt` table. If the GTFS
#'   feed does not have a frequency table, r5r still allows for multiple runs
#'   over the set `time_window` but in a deterministic way.

testthat::skip_on_cran()

options(java.parameters = "-Xmx2G")

if (Sys.getenv("NOT_CRAN") != "false") {

  # porto alegre
  data_path <- system.file("extdata/poa", package = "r5r")
  r5r_network <- r5r::build_network(data_path, verbose = FALSE)
  points <- data.table::fread(file.path(data_path, "poa_hexgrid.csv"))
  pois <- data.table::fread(file.path(data_path, "poa_points_of_interest.csv"))
  departure_datetime <- as.POSIXct(
    "13-05-2019 14:00:00",
    format = "%d-%m-%Y %H:%M:%S"
  )
  fare_structure <- r5r::read_fare_structure(
    system.file("extdata/poa/fares/fares_poa.zip", package = "r5r")
  )

  # sao paulo
  spo_path <- system.file("extdata/spo", package = "r5r")
  spo_network <- r5r::build_network(spo_path, verbose = FALSE)
  spo_points <- data.table::fread(file.path(spo_path, "spo_hexgrid.csv"))
  spo_points[, opportunities := 1]
  spo_fare_struc <- r5r::setup_fare_structure(r5r_network = spo_network, base_fare = 5)
  spo_fare_struc$fares_per_transfer <- data.table::data.table(NULL)
}

# options(java.parameters = '-Xmx16384m')
# options(java.parameters = c("-XX:+UseConcMarkSweepGC", "-Xmx16384m"))

# library(r5r)
devtools::load_all(".")
# library(ggplot2)
# library(data.table)
library(tidyverse)
# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path = data_path, verbose = FALSE, overwrite = FALSE)

# load origin/destination points

departure_datetime <- as.POSIXct("13-05-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S")

poi <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))
points <- read.csv(file.path(data_path, "poa_hexgrid.csv"))
dest <- points

calculate_access <- function(fares) {

  access_df <- map_df(fares, function(f) {
    f <- as.integer(f)

    r5r_core$setMaxFare(f, "porto-alegre")

    access <- accessibility(r5r_core,
                            origins = points,
                            destinations = dest,
                            departure_datetime = departure_datetime,
                            opportunities_colname = "schools",
                            mode = c("WALK", "TRANSIT"),
                            cutoffs = c(30, 60),
                            max_trip_duration = 60,
                            time_window = 1,
                            percentiles = c(5, 50, 95),
                            verbose = FALSE)

    access$max_fare <- f

    return(access)
  })

  return(access_df)
}

access_df <- calculate_access(c(480, 720)) %>%
  # access_df <- calculate_access(c(240, 480, 720, 960, -1)) %>%
  left_join(points, by = c("from_id" = "id"))

access_df %>%
  ggplot(aes(x=lon, y=lat, color= accessibility)) +
  geom_point() +
  coord_map() +
  scale_color_distiller(palette = "Spectral") +
  facet_wrap(~max_fare)

access_df %>%
  pivot_wider(names_from = max_fare, values_from = accessibility, names_prefix = "fare_") %>%
  View()


r5r_core$setMaxFare(-1L, "porto-alegre")
r5r_core$setMaxFare(200L, "porto-alegre")
r5r_core$setMaxFare(480L, "porto-alegre")
r5r_core$setMaxFare(1000L, "porto-alegre")

system.time(
  access <- accessibility(r5r_core,
                        origins = points,
                        destinations = dest,
                        departure_datetime = departure_datetime,
                        opportunities_colname = "schools",
                        decay_function = "logistic",
                        decay_value = 15,
                        mode = c("WALK", "TRANSIT"),
                        cutoffs = c(60),
                        max_trip_duration = 60,
                        verbose = FALSE)
)

access %>% left_join(points, by = c("id" = "id")) %>%
  ggplot(aes(x=lon, y=lat, color= accessibility)) +
  geom_point() +
  coord_map() +
  scale_color_distiller(palette = "Spectral") +
  facet_wrap(~cutoff)


system.time(
  ttm <- travel_time_matrix(r5r_core, origins = points,
                            destinations = dest,
                            mode = c("WALK", "TRANSIT"),
                            breakdown = FALSE,
                            departure_datetime = departure_datetime,
                            max_trip_duration = 60,
                            max_walk_dist = 800,
                            time_window = 30,
                            percentiles = c(25, 50, 75),
                            verbose = FALSE,
                            progress = TRUE)
)

calculate_ttm <- function(fare) {
  r5r_core$setMaxFare(fare, "porto-alegre")

  ttm <- travel_time_matrix(r5r_core, origins = points,
                            destinations = dest,
                            mode = c("WALK", "TRANSIT"),
                            breakdown = FALSE,
                            departure_datetime = departure_datetime,
                            max_trip_duration = 60,
                            max_walk_dist = 800,
                            time_window = 1,
                            percentiles = c(50),
                            verbose = FALSE,
                            progress = TRUE)
  ttm$max_fare <- fare

  return(ttm)
}

ttm_max = calculate_ttm(-1L)
ttm_200 = calculate_ttm(200L)
ttm_480 = calculate_ttm(480L)
ttm_1000 = calculate_ttm(1000L)

ttm <- rbind(ttm_max, ttm_200, ttm_480, ttm_1000)

access_df <- ttm %>%
  group_by(fromId, max_fare) %>%
  summarise(access = n(), .groups = "drop") %>%
  left_join(points, by = c("fromId" = "id"))

access_df %>%
  ggplot(aes(x=lon, y=lat, color= access)) +
  geom_point() +
  coord_map() +
  scale_color_distiller(palette = "Spectral") +
  facet_wrap(~max_fare)


# Detailed Itineraries ----------------------------------------------------

r5r_core$setMaxFare(240L, "porto-alegre")
r5r_core$setMaxFare(480L, "porto-alegre")
r5r_core$setMaxFare(720L, "porto-alegre")

origins <- poi
destinations <- poi

mode = c("WALK", "BUS")
max_walk_dist <- 10000


system.time(
  df <- detailed_itineraries(r5r_core,
                             origins = origins[2,],
                             destinations = destinations[3,],
                             departure_datetime = departure_datetime,
                             max_walk_dist = max_walk_dist,
                             mode = mode,
                             shortest_path = F,
                             n_threads= Inf,
                             verbose = F,
                             progress=T)
)



# Pareto ------------------------------------------------------------------

pareto_df <- pareto_frontier(r5r_core,
                             origins = poi[1:2,],
                             destinations = poi[3:4,],
                             mode = c("WALK", "TRANSIT"),
                             departure_datetime = departure_datetime,
                             monetary_cost_cutoffs = seq(0, 1000, 100),
                             fare_calculator = "porto-alegre",
                             max_trip_duration = 60,
                             max_walk_dist = 8000,
                             time_window = 30,
                             percentiles = c(5, 50, 95),
                             max_rides = 5,
                             verbose = FALSE,
                             progress = TRUE)

pareto_df$monetary_cost <- pareto_df$monetary_cost / 100
pareto_df$monetary_cost_upper <- pareto_df$monetary_cost_upper / 100
View(pareto_df)

pareto_df %>%
  mutate(percentile = factor(percentile)) %>%
  pivot_longer(cols=starts_with("monetary"), names_to = "mon", values_to="cost") %>%
  ggplot(aes(x=cost, y=travel_time, color=percentile, group=percentile)) +
  geom_step() +
  # geom_path() +
  scale_color_brewer(palette = "Set1") +
  scale_x_continuous(breaks = 1:10) +
  facet_grid(from_id~to_id)

r5r_core$setMaxFare(10L, "rio-de-janeiro")
r5r_core$verboseMode()


## accessibility decay

library(r5r)
library(dplyr)
library(tidyr)
library(ggplot2)

data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path = data_path, verbose = FALSE, overwrite = FALSE)


decay_step <- r5r_core$testDecay("STEP", 0.0)
decay_exp <- r5r_core$testDecay("EXPONENTIAL", 0.0)
decay_fixed_exp <- r5r_core$testDecay("FIXED_EXPONENTIAL", 0)
decay_linear <- r5r_core$testDecay("LINEAR", 10.0)
decay_logistic <- r5r_core$testDecay("LOGISTIC", 10.0)

decays_df <- data.frame(seconds = 1:3600,
                        step = decay_step,
                        exponential = decay_exp,
                        fixed_exponential = decay_fixed_exp,
                        linear = decay_linear,
                        logistic = decay_logistic)

decays_df <- pivot_longer(decays_df, cols = 2:6, names_to = "decay_function", values_to = "decay")

ggplot(decays_df, aes(x=seconds, y=decay, color=decay_function)) +
  geom_point() +
  geom_vline(xintercept = 1800) +
  facet_wrap(~decay_function) +
  theme(legend.position = "none")



# LTS ---------------------------------------------------------------------

street_net <- street_network_to_sf(r5r_core)
View(street_net$vertices)
View(street_net$edges |> head(1000))

mapview::mapview(street_net$edges, zcol = "car_speed")

speeds <- unique(street_net$edges$car_speed) |> sort()

street_net$edges$car_speed_f <- factor(street_net$edges$car_speed,
                                       levels = speeds,
                                       labels = scales::comma(speeds, accuracy = 0.01))

street_net$edges$car_speed_f <- cut(street_net$edges$car_speed, include.lowest = TRUE,
                                       breaks = seq(0, 80, 20))

street_net$edges$street_class <- factor(street_net$edges$street_class,
                                        levels = c("MOTORWAY", "PRIMARY", "SECONDARY", "TERTIARY", "OTHER"))

library(viridis)
street_net$edges |>
  arrange(car_speed) |>
  ggplot() +
  geom_sf(aes(color=car_speed_f)) +
  coord_sf(datum = NA) +
  scale_color_discrete(type = heat.colors(4, rev=T)) +
  # scale_color_viridis_d(direction = -1) +
  theme_minimal() +
  labs(color = "car speed")


street_net$edges |>
  arrange(desc(street_class)) |>
  filter(street_class != "OTHER") |>
  ggplot() +
  geom_sf(data = filter(street_net$edges, street_class == "OTHER"), color = "grey90", size = 0.5) +
  geom_sf(aes(color=street_class)) +
  coord_sf(datum = NA) +
  scale_color_manual(values = c("#E41A1C", "#377EB8", "#4DAF4A", "#984EA3", "grey90"), drop = FALSE) +
  theme_minimal() +
  labs(color = "class")

RColorBrewer::brewer.pal(5, "Set1")

options(java.parameters = '-Xmx16384m')
# options(java.parameters = c("-XX:+UseConcMarkSweepGC", "-Xmx16384m"))

devtools::load_all(".")
# library(ggplot2)
library(data.table)
library(tidyverse)

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path = data_path, verbose = FALSE, overwrite = FALSE)

# load origin/destination points

departure_datetime <- as.POSIXct("13-05-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S")
points <- fread(file.path(data_path, "poa_hexgrid.csv"))
fares <- read_fare_structure(file.path(data_path, "fares/fares_poa.zip"))


dir.create(here::here("csv"))

# r5r_core$setCsvOutput(here::here("csv"))

tictoc::tic()
normal_ttm <- travel_time_matrix(r5r_core, origins = points, #[id == "89a9012a3cfffff",],
                                 destinations = points, #[id == "89a901284a3ffff",],
                                 mode = c("WALK", "TRANSIT"),
                                 departure_datetime = departure_datetime,
                                 max_trip_duration = 60,
                                 max_walk_time = 15,
                                 time_window = 30,
                                 percentiles = c(1, 25, 50, 75, 99),
                                 verbose = FALSE,
                                 progress = TRUE,
                                 fare_structure = fares,
                                 max_fare = 8)
tictoc::toc()

write_csv(normal_ttm, here::here('csv', 'ttm_original.csv'))
# fixed equal - 658.054 sec elapsed
# fixed diff  - 703.872 sec elapsed
# original    - 702.352 sec elapsed


ttm_or <- read_csv(here::here('csv', 'ttm_original.csv'))
ttm_fx <- read_csv(here::here('csv', 'ttm_fixed_equal.csv'))

ttm_or2 <-  ttm_or |> pivot_longer(cols = starts_with('travel_time'), names_to = 'percentile', values_to = 'travel_time_or')
ttm_fx2 <-  ttm_fx |> pivot_longer(cols = starts_with('travel_time'), names_to = 'percentile', values_to = 'travel_time_fx')

ttm_j <- left_join(ttm_or2, ttm_fx2) |>
  mutate(eq = travel_time_or == travel_time_fx)

# test R5
options(java.parameters = "-Xmx16G")
# source('./R/fun/setup.R')
# source("./R/fun/selecionar_data_gtfs.R")


# devtools::install_github("ipeaGIT/r5r", subdir = "r-package")
# library(r5r)
devtools::load_all(".")
library(data.table)

# get folders
# folders_graphs <- list.dirs("../../otp/graphs")
# folders_graphs <- folders_graphs[-1]
# folders_graphs <- folders_graphs[-1]
# folders_graphs <- folders_graphs[folders_graphs %like% paste0(munis_df_2019[modo == "todos"]$abrev_muni, collapse = "|")]
# folders_graphs <- folders_graphs[folders_graphs %like% "2019"]

folder_graphs <- "../../poa_data/"
folder_points <- "../../poa_data/"
folder_year <- "2019"
folder_muni <- "poa"


# selecionar_data_gtfs("poa", 2019)

# r5 setup - bho ---------------------------------------
setup <- setup_r5(data_path = folder_graphs, verbose = FALSE)

# points
points_file <- paste0(folder_points, "points_", folder_muni, "_09_", folder_year, ".csv")
points <- fread(points_file)
colnames(points) <- c("id", "lon", "lat")



# poa: 128.531 sec elapsed - 10 threads
# poa: 224.594 sec elapsed -  4 threads
tictoc::tic(folder_muni)
df_n <- travel_time_matrix(setup,
                           origin = points,
                           destination = points,
                           time_window = 120,
                           percentiles = c(5, 25, 50, 75, 95),
                           breakdown = FALSE,
                           max_walk_dist = 1000,
                           max_trip_duration = 120,
                           departure_datetime = as.POSIXct("2019-06-19 07:00:00"),
                           mode = c("WALK", "TRANSIT"),
                           n_threads = 4,
                           verbose = FALSE,
                           progress = TRUE)
tictoc::toc()

setup$setCsvOutput(here::here("tests_marcus/data"))
# ttmatrix bho - 222.32 secs
tictoc::tic(folder_muni)
df_b <- travel_time_matrix(setup,
                           origin = points,
                           destination = points,
                           time_window = 120,
                           percentiles = c(5, 25, 50, 75, 95),
                           breakdown = TRUE,
                           max_walk_dist = 1000,
                           max_trip_duration = 120,
                           departure_datetime = as.POSIXct("2019-06-19 07:00:00"),
                           mode = c("WALK", "TRANSIT"),
                           n_threads = 4,
                           verbose = FALSE,
                           progress = TRUE)
tictoc::toc()

library("tidyverse")

fare_schema_df <- read_csv("../java-r5rcore/src/main/resources/fares/rio/fare_schema.csv")
routes_info_df <- read_csv("../java-r5rcore/src/main/resources/fares/rio/routes_info.csv")


modes_prices_df <- routes_info_df %>%
  select(type, price) %>%
  distinct()

modes_prices_df %>% write_csv("../java-r5rcore/src/main/resources/fares/rio/price_per_mode.csv")

library(hexSticker) # https://github.com/GuangchuangYu/hexSticker
library(ggplot2)
library(sf)
library(r5r)
library(sysfonts)
library(data.table)

# add special text font
sysfonts::font_add_google(name = "Roboto", family = "Roboto")

### setup ------------------------

# load origin/destination points
points <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))
points_sf <- sfheaders::sf_point(points, x='lon', y='lat', keep = T)
sf::st_crs(points_sf) <- 4326


origin <- subset(points_sf, id == '89a90129977ffff')
destinations <- subset(points_sf, id %like% c('89a901299'))

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_core <- setup_r5(data_path = data_path)

# routing
df <- detailed_itineraries(r5r_core,
                           origins = origin,
                           destinations = destinations,
                           mode = 'WALK',
                           departure_datetime = as.POSIXct("13-03-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S"),
                           max_trip_duration = 300)


# st_crs(destinations) <- st_crs(df)


### network plot  ------------------------

# plot results
net <- ggplot() +
  geom_sf(data = df, color='gray95', alpha=.2) +
  # geom_sf(data=destinations,  color='gray95', size=1) +
  # geom_sf(data=destinations,  color='navyblue', size=.6) +
  # scale_x_continuous(limits = c(-51.20560, -51.18052 )) +
  # scale_y_continuous(limits = c(-30.02239, -30.0002 )) +
  theme_void() +
  theme(panel.grid.major=element_line(colour="transparent"))


### save sticker  ------------------------

# big
sticker(net,

        # package name
        package= expression( italic(paste("R"^5,"R"))),  p_size=10, p_y = 1.5, p_color = "gray95", p_family="Roboto",

        # ggplot image size and position
        s_x=1, s_y=.85, s_width=1.4, s_height=1.4,

        # blue hexagon
        h_fill="#0d8bb1", h_color="white", h_size=1.3,

        ## blackhexagon
        # h_fill="gray20", h_color="gray80", h_size=1.3,

        # url
        url = "github.com/ipeaGIT/r5r", u_color= "gray95", u_family = "Roboto", u_size = 1.8,

        # save output name and resolution
        filename="./man/figures/r5r_biagaa.png", dpi=300 #
)

#' @section Routing algorithm:
#'
#' The [detailed_itineraries()] and [pareto_frontier()] functions use an
#' `R5`-specific extension to the McRAPTOR routing algorithm. The
#' implementation used in `detailed_itineraries()` allows the router to find
#' paths that are optimal and less than optimal in terms of travel time, with
#' some heuristics around multiple access modes, riding the same patterns, etc.
#' The specific extension to McRAPTOR to do suboptimal path routing is not
#' documented yet, but a detailed description of base McRAPTOR can be found in
#' Delling et al (2015). The implementation used in `pareto_frontier()`, on the
#' other hand, returns only the fastest trip within a given monetary cutoff,
#' ignoring slower trips that cost the same. A detailed discussion on the
#' algorithm can be found in Conway and Stewart (2019).
#'
#' - Delling, D., Pajor, T., & Werneck, R. F. (2015). Round-based public
#' transit routing. Transportation Science, 49(3), 591-604.
#' \doi{10.1287/trsc.2014.0534}
#'
#' - Conway, M. W., & Stewart, A. F. (2019). Getting Charlie off the MTA: a
#' multiobjective optimization method to account for cost constraints in public
#' transit accessibility metrics. International Journal of Geographical
#' Information Science, 33(9), 1759-1787. \doi{10.1080/13658816.2019.1605075}

#' Check transit service availability by date
#'
#' @description
#' This function checks the number and proportion of public transport services
#' from the GTFS feeds in a `r5r_network` that are active on specified dates. This
#' is useful to verify that the selected departure dates for routing analysis are
#' valid and have adequate service levels. When routing with public transport, it
#' is crucial to use a departure date where services are operational, as indicated
#' in the GTFS `calendar.txt` file.
#'
#' @details
#' You can specify the dates to check in two ways:
#' \itemize{
#'   \item Using the `dates` argument to provide a vector of specific dates.
#'   \item Using the `start_date` and `end_date` arguments to provide a continuous date range.
#' }
#' You must use one of these two methods, but not both in the same function call.
#'
#' @param r5r_network A routable transport network created with `build_network()`.
#' @param r5r_core The `r5r_core` argument is deprecated as of r5r v2.3.0. Please use the `r5r_network` argument instead.
#' @param dates A vector of specific dates to be checked. Can be character strings
#'        in #'   "YYYY-MM-DD" format, or objects of class `Date`. This argument
#'        cannot be used with `start_date` or `end_date`.
#' @param start_date The start date for a continuous date range. Must be a single
#'        character string in "YYYY-MM-DD" format or a `Date` object. Must be used
#'        with `end_date`.
#' @param end_date The end date for a continuous date range. Must be a single
#'        character string in "YYYY-MM-DD" format or a `Date` object. Must be used
#'        with `start_date`.
#' @return A `data.table` with four columns: `date`, `total_services`,
#'         `active_services`, and `pct_active` (the proportion of active services).
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#' data_path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(data_path)
#'
#' # Example 1: Check a vector of specific dates
#' # Let's check a regular weekday and a Sunday, where service may differ.
#' dates_to_check <- c("2019-05-13", "2019-05-19")
#' availability1 <- check_transit_availability(r5r_network, dates = dates_to_check)
#' availability1
#'
#' # Example 2: Check a continuous date range using start_date and end_date
#' availability2 <- check_transit_availability(
#'   r5r_network,
#'   start_date = "2019-01-01",
#'   end_date = "2019-12-31"
#' )
#' availability2[121:124,]
#'
#' # plot availability over the year
#' library(ggplot2)
#' ggplot(availability2, aes(x = date, y = pct_active)) +
#'   geom_line() +
#'   geom_point() +
#'   theme_classic(base_size = 16)
#'
#' stop_r5(r5r_network)
#' @export
check_transit_availability <- function(r5r_network,
                                       r5r_core = deprecated(),
                                       dates = NULL,
                                       start_date = NULL,
                                       end_date = NULL
                                       ) {
  # deprecating r5r_core --------------------------------------
  if (lifecycle::is_present(r5r_core)) {
    cli::cli_warn(c(
      "!" = "The `r5r_core` argument is deprecated as of r5r v2.3.0.",
      "i" = "Please use the `r5r_network` argument instead."
    ))
    r5r_network <- r5r_core
  }

  # Check inputs
  checkmate::assert_class(r5r_network, "r5r_network")
  jcore <- r5r_network@jcore

  # Argument validation for date inputs (CONSOLIDATED)
  is_valid_dates_vector <- !is.null(dates) &&
    is.null(start_date) &&
    is.null(end_date)
  is_valid_date_range <- is.null(dates) &&
    !is.null(start_date) &&
    !is.null(end_date)

  if (!is_valid_dates_vector && !is_valid_date_range) {
    cli::cli_abort(
      c(
        "Incorrect date arguments provided.",
        "i" = "Please specify dates using one of the following methods:",
        "*" = "Use the {.arg dates} argument to provide a vector of specific dates.",
        "*" = "Use both {.arg start_date} and {.arg end_date} to provide a continuous date range.",
        "x" = "You cannot mix these methods or provide an incomplete date range."
      )
    )
  }

  # Helper function to parse and validate date inputs
  parse_date_input <- function(date_input, arg_name) {
    # Pass Date objects through directly
    if (inherits(date_input, "Date")) {
      return(date_input)
    }

    if (!is.character(date_input)) {
      cli::cli_abort(
        "{.arg {arg_name}} must be a vector of character strings or Date objects."
      )
    }

    # Use regex to strictly check for "YYYY-MM-DD" format
    is_iso_format <- grepl("^\\d{4}-\\d{2}-\\d{2}$", date_input)
    if (any(!is_iso_format)) {
      cli::cli_abort(c(
        "x" = "Invalid date format found in {.arg {arg_name}}.",
        "i" = "Please use the strict {.val 'YYYY-MM-DD'} format for all date strings."
      ))
    }

    # Use a tryCatch block to convert potential errors from as.Date() into NAs
    parsed_dates_list <- lapply(date_input, function(d) {
      tryCatch(
        {
          as.Date(d)
        },
        error = function(e) {
          # If as.Date fails, return NA instead of throwing an error
          as.Date(NA)
        }
      )
    })
    parsed_dates <- do.call("c", parsed_dates_list)

    # Final check for NAs, which now correctly indicate logically impossible dates
    if (anyNA(parsed_dates)) {
      cli::cli_abort(c(
        "x" = "Could not parse all values in {.arg {arg_name}}.",
        "i" = "One or more dates are logically invalid (e.g., '2025-02-29')."
      ))
    }

    return(parsed_dates)
  }

  # Prepare the list of dates to check
  if (!is.null(dates)) {
    dates_as_date <- parse_date_input(dates, "dates")
    dates_formatted <- format(dates_as_date, "%Y-%m-%d")
  } else {
    start_d <- parse_date_input(start_date, "start_date")
    end_d <- parse_date_input(end_date, "end_date")

    if (length(start_d) > 1 || length(end_d) > 1) {
      cli::cli_abort(
        "{.arg start_date} and {.arg end_date} must each be a single date."
      )
    }
    if (start_d > end_d) {
      cli::cli_abort(
        "{.arg start_date} must be before or the same as {.arg end_date}."
      )
    }

    date_sequence <- seq(from = start_d, to = end_d, by = "day")
    dates_formatted <- format(date_sequence, "%Y-%m-%d")
  }

  # Function to process a single date by querying the Java object
  process_single_date <- function(date_str) {
    services_java <- jcore$getTransitServicesByDate(date_str)
    services_dt <- java_to_dt(services_java)

    if (nrow(services_dt) == 0) {
      return(data.table::data.table(
        date = as.Date(date_str),
        total_services = 0L,
        active_services = 0L,
        pct_active = 0.0
      ))
    }

    total_s <- nrow(services_dt)
    active_s <- sum(services_dt$active_on_date, na.rm = TRUE)
    pct_s <- if (total_s > 0) active_s / total_s else 0.0

    return(data.table::data.table(
      date = as.Date(date_str),
      total_services = total_s,
      active_services = active_s,
      pct_active = pct_s
    ))
  }

  # Apply function to all dates and bind results into a single data.table
  results_list <- lapply(dates_formatted, process_single_date)
  final_dt <- data.table::rbindlist(results_list)

  return(final_dt)
}

# tests/testthat/test-check_transit_availability.R

# Individual skip is still good practice, though setup.R has a global one.
testthat::skip_on_cran()

context("check_transit_availability")

# --- Expected Behavior ---

test_that("check_transit_availability works with a vector of character dates", {
  dates_char <- c("2019-05-13", "2019-05-19") # Weekday and Sunday
  result <- check_transit_availability(r5r_network, dates = dates_char)

  # Check output type and structure
  expect_s3_class(result, "data.table")
  expect_equal(nrow(result), 2)
  expect_named(
    result,
    c("date", "total_services", "active_services", "pct_active")
  )
  expect_s3_class(result$date, "Date")

  # Check values
  expect_equal(result$active_services, c(116L, 1L))
})

test_that("check_transit_availability works with a vector of Date objects", {
  dates_obj <- as.Date(c("2019-05-13", "2019-05-19"))
  result <- check_transit_availability(r5r_network, dates = dates_obj)

  # Check structure and values
  expect_s3_class(result, "data.table")
  expect_equal(nrow(result), 2)
  expect_equal(result$date, dates_obj)
  expect_equal(result$active_services, c(116L, 1L))
})

test_that("check_transit_availability works with a start_date/end_date range", {
  start <- "2019-05-18" # Saturday
  end <- "2019-05-20" # Monday
  result <- check_transit_availability(
    r5r_network,
    start_date = start,
    end_date = end
  )

  # Check structure and values
  expect_s3_class(result, "data.table")
  expect_equal(nrow(result), 3)
  expect_equal(result$date, seq(as.Date(start), as.Date(end), by = "day"))
  # Values for Sat, Sun, Mon
  expect_equal(result$active_services, c(1L, 1L, 116L))
})

test_that("check_transit_availability returns 0 active services for dates outside the GTFS calendar", {
  result <- check_transit_availability(r5r_network, dates = "2025-01-01")

  expect_equal(result$active_services, 0L)
  expect_equal(result$pct_active, 0.0)
})

# --- Expected Errors ---

test_that("check_transit_availability throws errors for invalid arguments", {
  consolidated_error_msg <- "Incorrect date arguments provided."

  # Test cases for consolidated date argument logic
  expect_error(check_transit_availability(r5r_network), consolidated_error_msg)
  expect_error(
    check_transit_availability(
      r5r_network,
      dates = "2019-01-01",
      start_date = "2019-01-01"
    ),
    consolidated_error_msg
  )
  expect_error(
    check_transit_availability(r5r_network, start_date = "2019-01-01"),
    consolidated_error_msg
  )

  # --- More robust tests for date FORMAT and VALIDITY ---

  # Test for invalid format (slashes instead of dashes)
  expect_error(
    check_transit_availability(r5r_network, dates = "2019/05/13"),
    "Invalid date format found"
  )

  # Test for invalid format (day-month-year)
  expect_error(
    check_transit_availability(r5r_network, dates = "13-05-2019"),
    "Invalid date format found"
  )

  # Test for a logically impossible date (correct format but invalid value)
  # This test will now pass reliably.
  expect_error(
    check_transit_availability(r5r_network, dates = "2025-02-29"), # 2025 is not a leap year
    "logically invalid"
  )

  # Test for start date after end date
  expect_error(
    check_transit_availability(
      r5r_network,
      start_date = "2019-05-20",
      end_date = "2019-05-18"
    ),
    "`start_date` must be before or the same as `end_date`."
  )
})

options(java.parameters = '-Xmx16G')
# library(r5r)
devtools::load_all(".")
library(data.table)

data_path <- system.file("extdata/spo", package = "r5r")
r5r_core <- setup_r5(data_path = data_path)

# load origin/destination points
points <- fread(file.path(data_path, "spo_hexgrid.csv"))

# r5r_core$setCsvOutput(here::here("csv"))

# Calculate 3 TTMs, one for each minute of the TTM with the time_window setting
system.time(
  ttm <- travel_time_matrix(r5r_core,
                            origins = points[id == "89a8100c603ffff",], #[id == "89a901288c3ffff",],
                            destinations = points[id == "89a8100c2b3ffff"], #[id == "89a90129953ffff",],
                            mode = c("WALK", "TRANSIT"),
                            max_trip_duration = 60,
                            time_window = 30,
                            percentiles = c(25, 50, 75),
                            departure_datetime = lubridate::mdy_hm("4/19/19 12:00pm"),
                            draws_per_minute = 5,
                            verbose = F,
                            progress = T,
                            breakdown = T)
  )

ttm[from_id != to_id,] |>head(100) |> View()


ttm %>%
  arrange(departure_time) %>%
  write_csv("ttm_breakdown_full.csv")

#' Check and convert origin and destination inputs
#'
#' @param df Either a `data.frame` or a `POINT sf`.
#' @param name Object name.
#'
#' @return A `data.frame` with columns `id`, `lon` and `lat`.
#'
#' @family assigning functions
#'
#' @keywords internal
assign_points_input <- function(df, name) {
  if (!inherits(df, "data.frame")) {
    stop("'", name, "' must be either a 'data.frame' or a 'POINT sf'.")
  }

  if ("sf" %in% class(df)) {
    if (
      as.character(sf::st_geometry_type(df, by_geometry = FALSE)) != "POINT"
    ) {
      stop("'", name, "' must be either a 'data.frame' or a 'POINT sf'.")
    }

    if (sf::st_crs(df) != sf::st_crs(4326)) {
      stop(
        "'", name, "' CRS must be WGS 84 (EPSG 4326). ",
        "Please use either sf::set_crs() to set it or ",
        "sf::st_transform() to reproject it."
      )
    }

    df <- sfheaders::sf_to_df(df, fill = TRUE)
    data.table::setDT(df)
    data.table::setnames(df, c("x", "y"), c("lon", "lat"))
  }

  checkmate::assert_names(
    names(df),
    must.include = c("id", "lat", "lon"),
    .var.name = name
  )
  checkmate::assert_numeric(df$lon, .var.name = paste0(name, "$lon"))
  checkmate::assert_numeric(df$lat, .var.name = paste0(name, "$lat"))

  if (!is.character(df$id)) {
    df$id <- as.character(df$id)
    warning("'", name, "$id' forcefully cast to character.")
  }

  return(df)
}


#' Check and select transport modes from user input
#'
#' Selects the transport modes used in the routing functions. Only one direct
#' and access/egress modes are allowed at a time.
#'
#' @param mode A character vector, passed from routing functions.
#' @param mode_egress A character vector, passed from routing functions.
#'
#' @return A list with the transport modes to be used in the routing.
#'
#' @family assigning functions
#'
#' @keywords internal
assign_mode <- function(mode, mode_egress, style) {
  dr_modes <- c("WALK", "BICYCLE", "CAR", "BICYCLE_RENT", "CAR_PARK")
  tr_modes <- c(
    "TRANSIT",
    "TRAM",
    "SUBWAY",
    "RAIL",
    "BUS",
    "FERRY",
    "CABLE_CAR",
    "GONDOLA",
    "FUNICULAR"
  )
  all_modes <- c(tr_modes, dr_modes)

  mode <- toupper(unique(mode))
  checkmate::assert(
    checkmate::check_character(mode, min.len = 1, any.missing = FALSE),
    checkmate::check_names(mode, subset.of = all_modes),
    combine = "and"
  )

  mode_egress <- toupper(unique(mode_egress))
  checkmate::assert_string(mode_egress)
  checkmate::assert_names(
    mode_egress,
    subset.of = setdiff(dr_modes, c("CAR_PARK", "BICYCLE_RENT"))
  )

  direct_modes <- mode[which(mode %in% dr_modes)]
  transit_modes <- mode[which(mode %in% tr_modes)]

  if (length(direct_modes) > 1) {
    stop(
      "Please use only 1 of {",
      paste0("'", direct_modes, "'", collapse = ","),
      "} when routing."
    )
  }

  if (any(c("CAR_PARK", "BICYCLE_RENT") %in% direct_modes)) {
    stop("CAR_PARK and BICYCLE_RENT are currently unsupported by r5r.")
  }

  access_mode <- direct_modes

  if (length(transit_modes) == 0) {
    transit_modes <- ""
    egress_mode <- ""
  } else {
    if ("TRANSIT" %in% transit_modes) transit_modes <- tr_modes

    # if only transit mode is passed, assume "WALK" as access_mode
    if (length(direct_modes) == 0) access_mode <- direct_modes <- "WALK"

    egress_mode <- mode_egress
  }

  mode_list <- list(
    direct_modes = paste0(direct_modes, collapse = ";"),
    transit_mode = paste0(transit_modes, collapse = ";"),
    access_mode = paste0(access_mode, collapse = ";"),
    egress_mode = paste0(egress_mode, collapse = ";")
  )

  return(mode_list)
}


#' Check and convert POSIXct objects to strings
#'
#' @param datetime An object of POSIXct class.
#'
#' @return A list with the `date` and `time` of the trip departure as
#'   characters.
#'
#' @family assigning functions
#'
#' @keywords internal
assign_departure <- function(datetime) {
  checkmate::assert_posixct(
    datetime,
    len = 1,
    any.missing = FALSE,
    .var.name = "departure_datetime"
  )

  tz <- attr(datetime, "tzone")
  if (is.null(tz)) tz <- ""

  datetime_list <- list(
    date = strftime(datetime, format = "%Y-%m-%d", tz = tz),
    time = strftime(datetime, format = "%H:%M:%S", tz = tz)
  )

  return(datetime_list)
}


#' Assign max street time from walk/bike distance and speed
#'
#' Checks the time duration and speed inputs and converts them to distance.
#'
#' @param max_time A numeric of length 1. Maximum walking distance (in
#'   meters) for the whole trip. Passed from routing functions.
#' @param speed A numeric of length 1. Average walk speed in km/h.
#'   Defaults to 3.6 Km/h. Passed from routing functions.
#' @param max_trip_duration A numeric of length 1. Maximum trip duration in
#'   seconds. Defaults to 120 minutes (2 hours). Passed from routing functions.
#' @param mode A string. Either `"bike"` or `"walk"`.
#'
#' @return An `integer` representing the maximum number of minutes walking.
#'
#' @family assigning functions
#'
#' @keywords internal
assign_max_street_time <- function(max_time, speed, max_trip_duration, mode) {

  checkmate::assert(
    checkmate::check_string(mode),
    checkmate::check_names(mode, subset.of = c("bike", "walk", "car")),
    combine = "and"
  )

  checkmate::assert_number(
    max_time,
    .var.name = paste0("max_", mode, "_time"),
    lower = 1,
    finite = FALSE
  )

  checkmate::assert_number(
    speed,
    finite = TRUE,
    .var.name = paste0(mode, "_speed")
  )

  checkmate::assert_number(max_trip_duration, lower = 1, finite = TRUE)

  if (speed <= 0) {
    stop(
      "Assertion on '", mode, "_speed' failed: ",
      "Must have value greater than 0."
    )
  }

  if (is.infinite(max_time)){
    return(as.integer(max_trip_duration))
  }
  else {
    return(as.integer(max_time))
    }
}


#' Assign max trip duration
#'
#' Check and convert the max trip duration input.
#'
#' @param max_trip_duration A number.
#'
#' @return An `integer` representing the maximum trip duration in minutes.
#'
#' @family assigning functions
#'
#' @keywords internal
assign_max_trip_duration <- function(max_trip_duration,
                                     modes,
                                     max_walk_time,
                                     max_bike_time) {
  checkmate::assert_number(max_trip_duration, lower = 1, finite = TRUE)

  max_trip_duration <- as.integer(max_trip_duration)

  if (modes$transit_mode == "") {
    if (modes$direct_modes == "WALK" & max_walk_time < max_trip_duration) {
      max_trip_duration <- max_walk_time
    }
    if (modes$direct_modes == "BICYCLE" & max_bike_time < max_trip_duration) {
      max_trip_duration <- max_bike_time
    }
  }

  return(max_trip_duration)
}


#' Assign opportunities data
#'
#' Check and create an opportunities dataset.
#'
#' @param destinations Either a `data.frame` or a `POINT sf`.
#' @param opportunities_colnames A character vector with the names of the
#'   opportunities columns in `destinations`.
#'
#' @return A list of `Java-Array` objects.
#'
#' @family assigning functions
#'
#' @keywords internal
assign_opportunities <- function(destinations, opportunities_colnames) {
  checkmate::assert_character(
    opportunities_colnames,
    min.len = 1,
    unique = TRUE,
    any.missing = FALSE
  )
  checkmate::assert_names(
    names(destinations),
    must.include = opportunities_colnames,
    .var.name = "destinations"
  )

  opportunities_data <- lapply(
    opportunities_colnames,
    function(colname) {
      checkmate::assert_numeric(destinations[[colname]])

      opp_array <- as.integer(destinations[[colname]])
      opp_array <- rJava::.jarray(opp_array)

      opp_array
    }
  )

  return(opportunities_data)
}


#' Assign decay function and parameter values
#'
#' Checks and assigns decay function and values.
#'
#' @param decay_function A string, the name of the decay function.
#' @param decay_value A number, the value of decay parameter.
#'
#' @return A `list` with the validated decay function and parameter value.
#'
#' @family assigning functions
#'
#' @keywords internal
assign_decay_function <- function(decay_function, decay_value) {
  all_functions  <- c(
    "step",
    "exponential",
    "fixed_exponential",
    "linear",
    "logistic"
  )
  checkmate::assert(
    checkmate::check_string(decay_function),
    checkmate::check_names(decay_function, subset.of = all_functions),
    combine = "and"
  )
  checkmate::assert_number(decay_value, finite = TRUE, null.ok = TRUE)

  non_null_decay <- c("fixed_exponential", "linear", "logistic")
  if (!is.null(decay_value) & decay_function %in% c("step", "exponential")) {
    stop(
      "Assertion on decay_value failed: must be NULL when decay_function ",
      "is ", decay_function, "."
    )
  } else if (is.null(decay_value) & decay_function %in% non_null_decay) {
    stop(
      "Assertion on decay_value failed: must not be NULL when decay_function ",
      "is ", decay_function, "."
    )
  }

  if (decay_function == "fixed_exponential") {
    if (decay_value <= 0 | decay_value >= 1) {
      stop(
        "Assertion on decay_value failed: must be a number between 0 and 1 ",
        "(exclusive) when decay_function is fixed_exponential."
      )
    }
  }

  if (decay_function %in% c("logistic", "linear")) {
    if (decay_value < 1) {
      stop(
        "Assertion on decay_value failed: must be a number greater than or ",
        "equal to 1 when decay_function is ",
        decay_function, "."
      )
    }
  }

  decay_function <- toupper(decay_function)

  # java does not accept NULL values, so if decay_value is NULL we assign a
  # placeholder number to it (it's ignored in R5 anyway)

  if (is.null(decay_value)) {
    decay_value <- 0
  } else {
    decay_value <- as.double(decay_value)
  }

  decay_list <- list("fun" = decay_function, "value" = decay_value)

  return(decay_list)
}


#' Assign shortest path
#'
#' Check the shortest path input.
#'
#' @param shortest_path A logical.
#'
#' @return A logical.
#'
#' @family assigning functions
#'
#' @keywords internal
assign_shortest_path <- function(shortest_path) {
  checkmate::assert_logical(shortest_path, len = 1, any.missing = FALSE)

  return(shortest_path)
}


#' Assign drop geometry
#'
#' Check the drop geometry input.
#'
#' @param drop_geometry A logical.
#'
#' @return A logical.
#'
#' @family assigning functions
#'
#' @keywords internal
assign_drop_geometry <- function(drop_geometry) {
  checkmate::assert_logical(drop_geometry, len = 1, any.missing = FALSE)

  return(drop_geometry)
}


#' Assign osm_link_ids
#'
#' Check the osm_link_ids input.
#'
#' @param osm_link_ids A logical.
#' @param drop_geometry A logical.
#'
#' @return A logical.
#'
#' @family assigning functions
#'
#' @keywords internal
assign_osm_link_ids <- function(osm_link_ids, drop_geometry) {
  checkmate::assert_logical(osm_link_ids, len = 1, any.missing = FALSE)

  if (isTRUE(drop_geometry) & isTRUE(osm_link_ids)) {
    cli::cli_abort("To set osm_link_ids to `TRUE`, the parameter 'drop_geometry' must also be `FALSE`")
  }
  return(osm_link_ids)
}

tester <- function(mode = "WALK", mode_egress = "WALK") {
  assign_mode(mode, mode_egress)
}

test_that("expects correct modes", {
  expect_error(tester(character()))
  expect_error(tester("oie"))
  expect_error(tester(NA))
  expect_error(tester(mode_egress = c("WALK", "CAR"), style = "ttm"))
  expect_error(tester(mode_egress = "CAR_PARK"))
})

test_that("raises error when receives either CAR_PARK or BICYCLE_RENT", {
  expect_error(tester("CAR_PARK"))
  expect_error(tester("BICYCLE_RENT"))
})

test_that("raises errors when it cannot disambiguate direct modes", {
  expect_error(tester(c("CAR", "WALK")))
  expect_error(tester(c("CAR", "WALK", "TRANSIT")))
})

test_that("mode_egress is ignored when direct modes are passed to mode", {
  expect_identical(tester("WALK", "WALK"), tester("WALK", "BICYCLE"))
  expect_identical(tester("CAR", "WALK"), tester("CAR", "BICYCLE"))
  expect_identical(
    tester("BICYCLE", "WALK"),
    tester("BICYCLE", "CAR")
  )

  expect_identical(
    tester("CAR", "WALK"),
    list(
      direct_modes = "CAR",
      transit_mode = "",
      access_mode = "CAR",
      egress_mode = ""
    )
  )
})

test_that("walk is set to access_mode when not listed in mode", {
  expect_identical(tester(c("BUS", "WALK"), "WALK"), tester("BUS", "WALK"))

  expect_identical(
    tester("BUS", "WALK"),
    list(
      direct_modes = "WALK",
      transit_mode = "BUS",
      access_mode = "WALK",
      egress_mode = "WALK"
    )
  )

  expect_identical(
    tester(c("BUS", "CAR"), "WALK"),
    list(
      direct_modes = "CAR",
      transit_mode = "BUS",
      access_mode = "CAR",
      egress_mode = "WALK"
    )
  )
})

test_that("all pt modes are passed to transit_modes when TRANSIT is present", {
  tr_modes <- c(
    "TRANSIT",
    "TRAM",
    "SUBWAY",
    "RAIL",
    "BUS",
    "FERRY",
    "CABLE_CAR",
    "GONDOLA",
    "FUNICULAR"
  )

  expect_identical(
    tester(c("TRANSIT", "WALK"), "WALK"),
    list(
      direct_modes = "WALK",
      transit_mode = paste0(tr_modes, collapse = ";"),
      access_mode = "WALK",
      egress_mode = "WALK"
    )
  )
})

testthat::skip_on_cran()

# assign_max_street_time -----------------------------------------------------


test_that("assign_max_street_time adequately raises warnings and errors", {

  expect_error(assign_max_walk_distance("1000", 3.6, 60L, "walk"))
  expect_error(assign_max_walk_distance(1000, "3.6", 60L, "walk"))
  expect_error(assign_max_walk_distance(0, 3.6, 60L, "walk"))
  expect_error(assign_max_walk_distance(-3, 3.6, 60L, "walk"))
  # expect_error(assign_max_walk_distance(3700, 3.6, "60L")) # should this fail though?

})

test_that("assign_max_street_time output is coherent", {

  expect_equal(assign_max_street_time(Inf, 3.6, 60L, "walk"), 60L)
  expect_equal(assign_max_street_time(30, 3.6, 60L, "walk"), 30L)
  expect_equal(assign_max_street_time(Inf, 3.6, 60L, "bike"), 60L)
  expect_equal(assign_max_street_time(30, 3.6, 60L, "bike"), 30L)

})


# assign_departure ---------------------------------------------------------


test_that("assign_departure adequately raises warnings and errors", {

  datetime <- as.POSIXct("13-03-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S")

  expect_error(assign_departure(as.character(datetime)))
  expect_error(assign_departure(as.integer(datetime)))

})

test_that("assign_departure output is coherent", {

  datetime <- as.POSIXct("13-03-2019 14:00:00", format = "%d-%m-%Y %H:%M:%S")
  datetime <- assign_departure(datetime)

  expect_equal(datetime$date, "2019-03-13")
  expect_equal(datetime$time, "14:00:00")

  datetime <- as.POSIXct("13-03-1919 2:00:00 pm", format = "%d-%m-%Y %I:%M:%S %p")
  datetime <- assign_departure(datetime)

  expect_equal(datetime$date, "1919-03-13")
  expect_equal(datetime$time, "14:00:00")

})


# assign_points_input -----------------------------------------------------

sf_points <- sf::st_as_sf(points, coords = c("lon", "lat"), crs = 4326)

test_that("assign_points_input adequately raises warnings and errors", {

  # object class

  multipoint_points <- sf::st_cast(
    sf::st_as_sf(points, coords = c("lon", "lat"), crs = 4326),
    "MULTIPOINT"
  )
  list_points <- setNames(
    lapply(names(points), function(i) points[[i]]),
    names(points)
  )

  expect_error(assign_points_input(as.matrix(points), "points"))
  expect_error(assign_points_input(list_points, "points"))
  expect_error(assign_points_input(multipoint_points, "points"))

  # object columns types

  points_numeric_id <- data.table::setDT(data.table::copy(points))[, id := 1:.N]
  points_char_lat <- data.table::setDT(data.table::copy(points))[
    ,
    lat := as.character(lat)
  ]
  points_char_lon <- data.table::setDT(data.table::copy(points))[
    ,
    lon := as.character(lon)
  ]

  expect_warning(assign_points_input(points_numeric_id, "points"))
  expect_error(assign_points_input(points_char_lat, "points"))
  expect_error(assign_points_input(points_char_lon, "points"))

  # object crs

  wrong_crs_sf <- sf::st_transform(sf_points, 4674)
  expect_error(assign_points_input(wrong_crs_sf, "points"))

})

test_that("assign_points_input output is coherent", {
  sf_points_output <- assign_points_input(sf_points, "points")
  df_points_output <- assign_points_input(points, "points")

  # correct output column types

  expect_type(sf_points_output$id, "character")
  expect_type(sf_points_output$lat, "double")
  expect_type(sf_points_output$lon, "double")

  expect_type(df_points_output$id, "character")
  expect_type(df_points_output$lat, "double")
  expect_type(df_points_output$lon, "double")

  # expect output columns to have the same value, irrespective of input class

  expect_equal(sf_points_output$id, points$id)
  expect_equal(sf_points_output$id, df_points_output$id)

  expect_equal(sf_points_output$lat, points$lat)
  expect_equal(sf_points_output$lat, df_points_output$lat)

  expect_equal(sf_points_output$lon, points$lon)
  expect_equal(sf_points_output$lon, df_points_output$lon)
})



  # assign_decay_function -----------------------------------------------------

  test_that("assign_decay_function adequately raises warnings and errors", {

    expect_error(assign_decay_function(decay_function='step', decay_value=4))
    expect_error(assign_decay_function(decay_function='step', decay_value='bananas'))
    expect_error(assign_decay_function(decay_function='bananas', decay_value=4))
    expect_error(assign_decay_function(decay_function= 444, decay_value=4))
    expect_error(assign_decay_function(decay_function= 'logistic', decay_value=0.4))
  })

  test_that("assign_decay_function expected behavior", {
    expect_equal(class(assign_decay_function(decay_function='step', decay_value=NULL)), 'list')
  })


# r5r_sitrep situation report --------------------------------------------------

  test_that("raises error if version is not a string", {
    expect_is(r5r::r5r_sitrep(), 'list')
  })



# fileurl_from_metadata --------------------------------------------------


test_that("raises error if version is not a string", {
  expect_error(fileurl_from_metadata(version = 6))
})

test_that("returns expected result", {
  expect_equal("https://github.com/ipeaGIT/r5/releases/download/v6.4/r5-v6.4-all.jar", fileurl_from_metadata("6.4.0"))
})

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

departure_datetime <- as.POSIXct(
  "13-05-2019 14:00:00",
  format = "%d-%m-%Y %H:%M:%S"
)

tester <- function(r5r_network = get("r5r_network", envir = parent.frame()),
                   origins = pois,
                   destinations = pois,
                   mode = "WALK",
                   mode_egress = "WALK",
                   departure_datetime = Sys.time(),
                   time_window = 1L,
                   percentiles = 50L,
                   fare_structure = NULL,
                   max_fare = Inf,
                   max_walk_time = Inf,
                   max_bike_time = Inf,
                   max_trip_duration = 120L,
                   walk_speed = 3.6,
                   bike_speed = 12,
                   max_rides = 3,
                   max_lts = 2,
                   new_carspeeds = NULL,
                   carspeed_scale = 1,
                   draws_per_minute = 5L,
                   n_threads = Inf,
                   verbose = FALSE,
                   progress = FALSE,
                   output_dir = NULL) {
  travel_time_matrix(
    r5r_network,
    origins = origins,
    destinations = destinations,
    mode = mode,
    mode_egress = mode_egress,
    departure_datetime = departure_datetime,
    time_window = time_window,
    percentiles = percentiles,
    fare_structure = fare_structure,
    max_fare = max_fare,
    max_walk_time = max_walk_time,
    max_bike_time = max_bike_time,
    max_trip_duration = max_trip_duration,
    walk_speed = walk_speed,
    bike_speed = bike_speed,
    max_rides = max_rides,
    max_lts = max_lts,
    new_carspeeds = new_carspeeds,
    carspeed_scale = carspeed_scale,
    draws_per_minute = draws_per_minute,
    n_threads = n_threads,
    verbose = verbose,
    progress = progress,
    output_dir = output_dir
  )
}

# tests -------------------------------------------------------------------

test_that("errors due to incorrect input types - origins and destinations", {
  multipoint_origins <- sf::st_cast(
    sf::st_as_sf(pois, coords = c("lon", "lat")),
    "MULTIPOINT"
  )
  multipoint_destinations <- multipoint_origins

  list_destinations <- list_origins <- unclass(pois)

  expect_error(tester(origins = multipoint_origins))
  expect_error(tester(destinations = multipoint_destinations))
  expect_error(tester(origins = list_origins))
  expect_error(tester(destinations = list_destinations))
  expect_error(tester(origins = "origins"))
  expect_error(tester(destinations = "destinations"))

  # wrong columns types

  pois_char_lat <- pois
  pois_char_lat$lat <- as.character(pois$lat)
  pois_char_lon <- pois
  pois_char_lon$lon <- as.character(pois$lon)

  expect_error(tester(origins = pois_char_lat))
  expect_error(tester(origins = pois_char_lon))
  expect_error(tester(destinations = pois_char_lat))
  expect_error(tester(destinations = pois_char_lon))
})

test_that("errors due to incorrect input types - other inputs", {
  # mode and mode_egress are tested in assign_mode() tests

  expect_error(tester(unclass(r5r_network)))

  expect_error(tester(departure_datetime = unclass(departure_datetime)))
  expect_error(tester(departure_datetime = rep(departure_datetime, 2)))

  expect_error(tester(percentiles = "50"))
  expect_error(tester(percentiles = 0))
  expect_error(tester(percentiles = 100))
  expect_error(tester(percentiles = c(50, 50)))
  expect_error(tester(percentiles = 1:6))
  expect_error(tester(percentiles = NA))

  # TODO: test fare_structure

  expect_error(tester(max_fare = "0"))
  expect_error(tester(max_fare = -1))
  expect_error(tester(max_fare = c(0, 20)))
  expect_error(tester(max_fare = NA))

  expect_error(tester(max_walk_time = "1000"))
  expect_error(tester(max_walk_time = NULL))
  expect_error(tester(max_walk_time = c(1000, 2000)))
  expect_error(tester(max_walk_time = 0))

  expect_error(tester(max_bike_time = "1000"))
  expect_error(tester(max_bike_time = NULL))
  expect_error(tester(max_bike_time = c(1000, 2000)))
  expect_error(tester(max_bike_time = 0))

  expect_error(tester(max_trip_duration = "120"))
  expect_error(tester(max_trip_duration = c(25, 30)))
  expect_error(tester(max_trip_duration = Inf))

  expect_error(tester(walk_speed = "3.6"))
  expect_error(tester(walk_speed = c(3.6, 5)))
  expect_error(tester(walk_speed = 0))

  expect_error(tester(bike_speed = "12"))
  expect_error(tester(bike_speed = c(12, 15)))
  expect_error(tester(bike_speed = 0))

  expect_error(tester(draws_per_minute = "1"))
  expect_error(tester(draws_per_minute = c(12, 15)))
  expect_error(tester(draws_per_minute = 0))
  expect_error(tester(draws_per_minute = Inf))

  expect_error(tester(output_dir = 1))
  expect_error(tester(output_dir = "non_existent_dir"))
})

test_that("raises errors when non-character ids are used in origs/dests", {
  origins <- destinations <- pois[1:2, ]

  origins_numeric_id <- origins
  origins_numeric_id$id <- 1:2
  destinations_numeric_id <- destinations
  destinations_numeric_id$id <- 1:2

  expect_warning(tester(origins = origins_numeric_id))
  expect_warning(tester(destinations = destinations_numeric_id))
})

test_that("output is a data.table with correct columns", {
  ttm <- tester()
  expect_s3_class(ttm, "data.table")
  expect_identical(names(ttm), c("from_id", "to_id", "travel_time_p50"))
  expect_type(ttm$from_id, "character")
  expect_type(ttm$to_id, "character")
  expect_type(ttm$travel_time_p50, "integer")

  # more percentiles means more columns

  ttm <- tester(percentiles = c(25, 50, 75))
  expect_identical(
    names(ttm),
    c("from_id", "to_id", paste0("travel_time_p", c(25, 50, 75)))
  )
  expect_type(ttm$from_id, "character")
  expect_type(ttm$to_id, "character")
  expect_type(ttm$travel_time_p25, "integer")
  expect_type(ttm$travel_time_p50, "integer")
  expect_type(ttm$travel_time_p75, "integer")
})

test_that("output is identical independent of origs/dests type", {
  origins_sf <- destinations_sf <- sf::st_as_sf(
    pois,
    coords = c("lon", "lat"),
    crs = 4326
  )

  ttm_df <- tester()
  ttm_sf <- tester(origins = origins_sf, destinations = destinations_sf)
  expect_identical(ttm_df, ttm_sf)
})


test_that("we get more results (and faster trips) when using faster modes", {
  ttm_walk_only <- tester(
    mode = "WALK",
    departure_datetime = departure_datetime
  )
  data.table::setnames(ttm_walk_only, "travel_time_p50", new = "only_walk_time")
  ttm_with_transit <- tester(
    mode = c("WALK", "TRANSIT"),
    departure_datetime = departure_datetime
  )

  expect_true(nrow(ttm_with_transit) > nrow(ttm_walk_only))

  ttm <- ttm_walk_only[
    ttm_with_transit,
    on = c("from_id", "to_id"),
    with_transit_time := i.travel_time_p50
  ]
  expect_true(all(ttm$only_walk_time >= ttm$with_transit_time))
})


test_that("higher percentiles travel times are slower", {
  ttm <- tester(
    mode = c("WALK", "TRANSIT"),
    departure_datetime = departure_datetime,
    percentiles = c(1, 50, 99),
    time_window = 20
  )
  expect_true(all(ttm$travel_time_p01 <= ttm$travel_time_p50))
  expect_true(all(ttm$travel_time_p50 <= ttm$travel_time_p99))
})

test_that("walk trips are shorter with higher walk speeds", {
  ttm_low_speed <- tester(mode = "WALK", walk_speed = 3.6)
  data.table::setnames(ttm_low_speed, "travel_time_p50", new = "low_speed_time")
  ttm_high_speed <- tester(mode = "WALK", walk_speed = 6)
  ttm <- ttm_low_speed[
    ttm_high_speed,
    on = c("from_id", "to_id"),
    high_speed_time := i.travel_time_p50
  ]
  expect_true(all(ttm$low_speed_time >= ttm$high_speed_time))
})



test_that("bike trips are shorter with higher bike speeds", {
  ttm_low_speed <- tester(mode = "BICYCLE", bike_speed = 12)
  data.table::setnames(ttm_low_speed, "travel_time_p50", new = "low_speed_time")
  ttm_high_speed <- tester(mode = "BICYCLE", bike_speed = 20)
  ttm <- ttm_low_speed[
    ttm_high_speed,
    on = c("from_id", "to_id"),
    high_speed_time := i.travel_time_p50
  ]
  expect_true(all(ttm$low_speed_time >= ttm$high_speed_time))
})

test_that("all travel times are lower than max_trip_duration", {
  max_trip_duration <- 60L
  ttm <- tester(max_trip_duration = max_trip_duration, percentiles = 99)
  expect_true(all(ttm$travel_time_p99 <= max_trip_duration))
})

test_that("all od pairs are unique", {
  ttm <- tester()
  ttm <- ttm[, .N, keyby = .(from_id, to_id)]
  expect_equal(unique(ttm$N), 1)
})

test_that("output is saved to dir and function returns path with output_dir", {
  tmpdir <- tempfile("ttm_output")
  dir.create(tmpdir)

  ttm_output_dir <- tester(output_dir = tmpdir)
  expect_equal(normalizePath(ttm_output_dir), normalizePath(tmpdir))

  ttm_from_files <- lapply(
    list.files(ttm_output_dir, full.names = TRUE),
    data.table::fread
  )
  ttm_from_files <- data.table::rbindlist(ttm_from_files)
  ttm_from_files <- ttm_from_files[order(from_id, to_id)]

  ttm_normal <- tester()
  ttm_normal <- ttm_normal[order(from_id, to_id)]

  expect_identical(ttm_normal, ttm_from_files)
})

test_that("returns ttm even if last call saved to dir", {
  tmpdir <- tempfile("ttm_output")
  dir.create(tmpdir)
  ttm_output_dir <- tester(output_dir = tmpdir)
  ttm_normal <- tester()
  expect_s3_class(ttm_normal, "data.table")
})


test_that("using transit outside the gtfs dates throws an error", {
  expect_error(
    tester(r5r_network,
           mode='transit',
           departure_datetime = as.POSIXct("13-05-2025 14:00:00",
                                           format = "%d-%m-%Y %H:%M:%S")
    )
  )
})

#' Expand origin-destination pairs
#'
#' This function is used in [detailed_itineraries()] to update the `origins` and
#' `destinations` datasets.
#'
#' @param origins Passed by [detailed_itineraries()].
#' @param destinations Passed by [detailed_itineraries()].
#' @param all_to_all Passed by [detailed_itineraries()].
#'
#' @keywords internal
expand_od_pairs <- function(origins, destinations, all_to_all) {
  checkmate::assert_logical(all_to_all, len = 1, any.missing = FALSE)

  n_origs <- nrow(origins)
  n_dests <- nrow(destinations)

  if (all_to_all || n_origs == 1 || n_dests == 1) {
    origins <- origins[rep(1:n_origs, each = n_dests), ]
    destinations <- destinations[rep(1:n_dests, times = n_origs), ]

    if (!all_to_all && (n_origs > 1 || n_dests > 1)) {
      if (n_origs == 1) {
        message("'origins' was expanded to match the number of destinations.")
      } else {
        message("'destinations' was expanded to match the number of origins.")
      }
    }
  } else if (n_origs != n_dests) {
    stop(
      "When 'all_to_all' is FALSE, 'origins' and 'destinations' must either ",
      "have the same number of rows or one of them must have only one row."
    )
  }

  result <- list(origins = origins, destinations = destinations)
  return(result)
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

testthat::skip_on_cran()

test_that("all_to_all is of correct type", {
  expect_error(expand_od_pairs(pois, pois, all_to_all = "FALSE"))
  expect_error(expand_od_pairs(pois, pois, all_to_all = NA))
  expect_error(expand_od_pairs(pois, pois, all_to_all = c(FALSE, FALSE)))
})

test_that("raises errors when !all_to_all and nrow(origs/dests) is diff", {
  origins <- points[1:3]
  destinations <- points[6:7]

  expect_error(expand_od_pairs(origins, destinations, all_to_all = FALSE))
})

test_that("works correctly when nrow(origins/destinations) == 1", {
  origins <- as.data.frame(pois)[1, ]
  destinations <- as.data.frame(pois)
  expect_message(
    result <- expand_od_pairs(origins, destinations, all_to_all = FALSE)
  )
  expect_true(nrow(result$origins) == nrow(result$destinations))
  expect_true(all(result$origins$id == origins$id))
  expect_identical(result$destinations, destinations)

  origins <- as.data.frame(pois)
  destinations <- as.data.frame(pois)[1, ]
  expect_message(
    result <- expand_od_pairs(origins, destinations, all_to_all = FALSE)
  )
  expect_true(nrow(result$origins) == nrow(result$destinations))
  expect_true(all(result$destinations$id == destinations$id))
  expect_identical(result$origins, origins)

  origins <- as.data.frame(pois)[1, ]
  destinations <- as.data.frame(pois)[1, ]
  expect_silent(
    result <- expand_od_pairs(origins, destinations, all_to_all = TRUE)
  )
  expect_true(nrow(result$origins) == nrow(result$destinations))
  expect_identical(result$origins, origins)
  expect_identical(result$destinations, destinations)
})

test_that("works correctly when all_to_all = TRUE", {
  origins <- as.data.frame(pois)[1:3, ]
  destinations <- as.data.frame(pois)[4:5, ]

  expect_silent(
    result <- expand_od_pairs(origins, destinations, all_to_all = TRUE)
  )
  expect_true(nrow(result$origins) == nrow(result$destinations))
  expect_identical(result$origins$id, origins$id[rep(1:3, each = 2)])
  expect_identical(result$destinations$id, destinations$id[rep(1:2, times = 3)])
})

test_that("works with sf objects", {
  origins_sf <- destinations_sf <- sf::st_as_sf(
    pois,
    coords = c("lon", "lat"),
    crs = 4326
  )
  origins_sf <- origins_sf[1:3, ]
  destinations_sf <- destinations_sf[4:5, ]

  expect_silent(
    result <- expand_od_pairs(origins_sf, destinations_sf, all_to_all = TRUE)
  )
  expect_true(nrow(result$origins) == nrow(result$destinations))
  expect_identical(result$origins$id, origins_sf$id[rep(1:3, each = 2)])
  expect_identical(
    result$destinations$id,
    destinations_sf$id[rep(1:2, times = 3)]
  )
})

test_that("doesn't do anything when nrow > 1, equal and all_to_all is FALSE", {
  origins <- pois
  destinations <- pois
  expect_silent(
    result <- expand_od_pairs(origins, destinations, all_to_all = FALSE)
  )
  expect_true(nrow(result$origins) == nrow(result$destinations))
  expect_identical(result$origins, origins)
  expect_identical(result$destinations, destinations)
})

test_that("is silent when nrow(dests & origs) == 1", {
  origins <- pois[1]
  destinations <- pois[1]
  expect_silent(
    result <- expand_od_pairs(origins, destinations, all_to_all = FALSE)
  )
})

#' Calculate travel time matrix between origin destination pairs considering a
#' departure time
#'
#' Fast computation of travel time estimates between one or multiple origin
#' destination pairs. This function considers a departure time set by the user.
#' If you want to calculate travel times considering a time of arrival, have a
#' look at the [arrival_travel_time_matrix()] function.
#'
#' @template r5r_network
#' @template r5r_core
#' @template common_arguments
#' @template time_window_related_args
#' @template draws_per_minute
#' @template fare_structure
#' @template max_fare
#' @template scenarios
#' @template verbose
#' @param percentiles An integer vector (max length of 5). Specifies the
#'   percentile to use when returning travel time estimates within the given
#'   time window. For example, if the 25th travel time percentile between A and
#'   B is 15 minutes, 25% of all trips taken between these points within the
#'   specified time window are shorter than 15 minutes. Defaults to 50,
#'   returning the median travel time. If a vector with length bigger than 1 is
#'   passed, the output contains an additional column for each percentile
#'   specifying the percentile travel time estimate. each estimate. Due to
#'   upstream restrictions, only 5 percentiles can be specified at a time. For
#'   more details, please see R5 documentation at
#'   <https://docs.conveyal.com/analysis/methodology#accounting-for-variability>.
#'
#' @return A `data.table` with travel time estimates (in minutes) between
#'   origin and destination pairs. Pairs whose trips couldn't be completed
#'   within the maximum travel time and/or whose origin is too far from the
#'   street network are not returned in the `data.table`. If `output_dir` is
#'   not `NULL`, the function returns the path specified in that parameter, in
#'   which the `.csv` files containing the results are saved.
#'
#' @template transport_modes_section
#' @template lts_section
#' @template datetime_parsing_section
#' @template raptor_algorithm_section
#'
#' @family routing
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' # build transport network
#' data_path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(data_path)
#'
#' # load origin/destination points
#' points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))
#'
#' departure_datetime <- as.POSIXct(
#'   "13-05-2019 14:00:00",
#'   format = "%d-%m-%Y %H:%M:%S"
#' )
#'
#' ttm <- travel_time_matrix(
#'   r5r_network,
#'   origins = points,
#'   destinations = points,
#'   mode = c("WALK", "TRANSIT"),
#'   departure_datetime = departure_datetime,
#'   max_trip_duration = 60
#' )
#' head(ttm)
#'
#' # using a larger time window
#' ttm <- travel_time_matrix(
#'   r5r_network,
#'   origins = points,
#'   destinations = points,
#'   mode = c("WALK", "TRANSIT"),
#'   departure_datetime = departure_datetime,
#'   time_window = 30,
#'   max_trip_duration = 60
#' )
#' head(ttm)
#'
#' # selecting different percentiles
#' ttm <- travel_time_matrix(
#'   r5r_network,
#'   origins = points,
#'   destinations = points,
#'   mode = c("WALK", "TRANSIT"),
#'   departure_datetime = departure_datetime,
#'   time_window = 30,
#'   percentiles = c(25, 50, 75),
#'   max_trip_duration = 60
#' )
#' head(ttm)
#'
#' # use a fare structure and set a max fare to take monetary constraints into
#' # account
#' fare_structure <- read_fare_structure(
#'   file.path(data_path, "fares/fares_poa.zip")
#' )
#' ttm <- travel_time_matrix(
#'   r5r_network,
#'   origins = points,
#'   destinations = points,
#'   mode = c("WALK", "TRANSIT"),
#'   departure_datetime = departure_datetime,
#'   fare_structure = fare_structure,
#'   max_fare = 5,
#'   max_trip_duration = 60,
#' )
#' head(ttm)
#'
#' stop_r5(r5r_network)
#'
#' @export
travel_time_matrix <- function(r5r_network,
                               r5r_core = deprecated(),
                               origins,
                               destinations,
                               mode = "WALK",
                               mode_egress = "WALK",
                               departure_datetime = Sys.time(),
                               time_window = 10L,
                               percentiles = 50L,
                               max_walk_time = Inf,
                               max_bike_time = Inf,
                               max_car_time = Inf,
                               max_trip_duration = 120L,
                               walk_speed = 3.6,
                               bike_speed = 12,
                               max_rides = 3,
                               max_lts = 2,
                               fare_structure = NULL,
                               max_fare = Inf,
                               new_carspeeds = NULL,
                               carspeed_scale = 1,
                               new_lts = NULL,
                               draws_per_minute = 5L,
                               n_threads = Inf,
                               verbose = FALSE,
                               progress = FALSE,
                               output_dir = NULL) {

  # deprecating r5r_core --------------------------------------
  if (lifecycle::is_present(r5r_core)) {

    cli::cli_warn(c(
      "!" = "The `r5r_core` argument is deprecated as of r5r v2.3.0.",
      "i" = "Please use the `r5r_network` argument instead."
    ))

    r5r_network <- r5r_core
  }

  old_options <- options(datatable.optimize = Inf)
  on.exit(options(old_options), add = TRUE)

  old_dt_threads <- data.table::getDTthreads()
  dt_threads <- ifelse(is.infinite(n_threads), 0, n_threads)
  data.table::setDTthreads(dt_threads)
  on.exit(data.table::setDTthreads(old_dt_threads), add = TRUE)

  # check inputs and set r5r options --------------------------------------

  origins <- assign_points_input(origins, "origins")
  destinations <- assign_points_input(destinations, "destinations")
  mode_list <- assign_mode(mode, mode_egress)
  departure <- assign_departure(departure_datetime)

  # check availability of transit services on the selected date
  if (mode_list$transit_mode %like% 'TRANSIT|TRAM|SUBWAY|RAIL|BUS|CABLE_CAR|GONDOLA|FUNICULAR') {
    check_transit_availability_on_date(r5r_network, departure_date = departure$date)
  }

  checkmate::assert_class(r5r_network, "r5r_network")
  r5r_network <- r5r_network@jcore

  # in direct modes reverse origin/destination to take advantage of R5's One to Many algorithm
  data_path <- r5r_network$getDataPath()
  res <- reverse_if_direct_mode(origins, destinations, mode_list, data_path)
  origins <- res$origins
  destinations <- res$destinations


  max_walk_time <- assign_max_street_time(
    max_walk_time,
    walk_speed,
    max_trip_duration,
    "walk"
  )
  max_bike_time <- assign_max_street_time(
    max_bike_time,
    bike_speed,
    max_trip_duration,
    "bike"
  )
  max_car_time <- assign_max_street_time(
    max_car_time,
    8, # 8 km/h, R5's default.
    max_trip_duration,
    "car"
  )
  max_trip_duration <- assign_max_trip_duration(
    max_trip_duration,
    mode_list,
    max_walk_time,
    max_bike_time
  )

  set_time_window(r5r_network, time_window)
  set_percentiles(r5r_network, percentiles)
  set_monte_carlo_draws(r5r_network, draws_per_minute, time_window)
  set_speed(r5r_network, walk_speed, "walk")
  set_speed(r5r_network, bike_speed, "bike")
  set_max_rides(r5r_network, max_rides)
  set_max_lts(r5r_network, max_lts)
  set_n_threads(r5r_network, n_threads)
  set_verbose(r5r_network, verbose)
  set_progress(r5r_network, progress)
  set_fare_structure(r5r_network, fare_structure)
  set_max_fare(r5r_network, max_fare)
  set_output_dir(r5r_network, output_dir)
  set_expanded_travel_times(r5r_network, FALSE)
  set_breakdown(r5r_network, FALSE)
  r5r_network$setSearchType("DEPART_FROM")

  # SCENARIOS -------------------------------------------
  set_new_congestion(r5r_network, new_carspeeds, carspeed_scale)
  set_new_lts(r5r_network, new_lts)


  # call r5r_network method and process result -------------------------------

  travel_times <- r5r_network$travelTimeMatrix(
    origins$id,
    origins$lat,
    origins$lon,
    destinations$id,
    destinations$lat,
    destinations$lon,
    mode_list$direct_modes,
    mode_list$transit_mode,
    mode_list$access_mode,
    mode_list$egress_mode,
    departure$date,
    departure$time,
    max_walk_time,
    max_bike_time,
    max_car_time,
    max_trip_duration
  )

  if (!verbose & progress) cat("Preparing final output...", file = stderr())

  travel_times <- java_to_dt(travel_times)

  # reverse order of origins destinations back
  travel_times <- reverse_back_if_direct_mode(travel_times, origins, destinations, mode_list, data_path)


  if (nrow(travel_times) > 0) {
    # replace travel-times of nonviable trips with NAs.
    # the first column with travel time information is column 3, because
    # columns 1 and 2 contain the ids of OD point.
    # the percentiles parameter indicates how many travel times columns we'll
    # have
    for (j in seq(from = 3, to = (length(percentiles) + 2))) {
      data.table::set(
        travel_times,
        i = which(travel_times[[j]] > max_trip_duration),
        j = j,
        value = NA_integer_
      )
    }
  }

  if (!verbose & progress) cat(" DONE!\n", file = stderr())

  if (!is.null(output_dir)) return(output_dir)
  return(travel_times[])
}

#' Extract transit network in sf format
#'
#' Extracts the transit network in `sf` format from a routable transport network
#' built with [build_network()]).
#'
#' @template r5r_network
#' @template r5r_core
#'
#' @return A list with two components of a transit network in `sf` format:
#' route shapes (`LINESTRING`) and transit stops (`POINT`). The same
#' `route_id`/`short_name` might appear with different geometries. This occurs
#' when the same route is associated to more than one `shape_id`s in the GTFS
#' feed used to create the transit network. Some transit stops might be
#' returned with geometry `POINT EMPTY` (i.e. missing spatial coordinates).
#' This may occur when a transit stop is not snapped to the road network,
#' possibly because the GTFS feed used to create the transit network covers an
#' area larger than the `.osm.pbf` input data.
#'
#' @family network functions
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' # build transport network
#' path <- system.file("extdata/poa", package = "r5r")
#' r5r_network <- build_network(path)
#'
#' # extract transit network from r5r_network
#' transit_net <- transit_network_to_sf(r5r_network)
#'
#' stop_r5(r5r_network)
#' @export
transit_network_to_sf <- function(r5r_network,
                                  r5r_core = deprecated()) {

  # deprecating r5r_core --------------------------------------
  if (lifecycle::is_present(r5r_core)) {

    cli::cli_warn(c(
      "!" = "The `r5r_core` argument is deprecated as of r5r v2.3.0.",
      "i" = "Please use the `r5r_network` argument instead."
    ))

    r5r_network <- r5r_core
  }

  checkmate::assert_class(r5r_network, "r5r_network")

  network <- r5r_network@jcore$getTransitNetwork()

  # Convert edges to SF (linestring)
  routes_df <- java_to_dt(network$get(0L))
  routes_df[, geometry := sf::st_as_sfc(geometry)]
  routes_sf <- sf::st_sf(routes_df, crs = 4326) # WGS 84

  if (any(!sf::st_is_valid(routes_sf))) {
    routes_sf <- sf::st_make_valid(routes_sf)
  }

  routes_sf <- routes_sf[!sf::st_is_empty(routes_sf), ] # removing empty geometries

  # Convert stops to SF (point)
  stops_df <- java_to_dt(network$get(1L))
  stops_df[
    ,
    `:=`(
      lat = data.table::fifelse(lat == -1, NA_real_, lat),
      lon = data.table::fifelse(lon == -1, NA_real_, lon)
    )
  ]
  stops_sf <- sfheaders::sf_point(stops_df, x = "lon", y = "lat", keep = TRUE)
  sf::st_crs(stops_sf) <- 4326 # WGS 84

  transit_network <- list(stops = stops_sf, routes = routes_sf)

  return(transit_network)
}

context("transit_network_to_sf")

testthat::skip_on_cran()

# expected behavior
test_that("transit_network_to_sf - expected behavior", {

  testthat::expect_type( transit_network_to_sf(r5r_network), 'list')
})



# Expected errors
test_that("transit_network_to_sf - expected errors", {

  # invalid input
  testthat::expect_error( transit_network_to_sf('a') )
})

#' Generate an r5r situation report to help debug errors
#'
#' @description
#' The function reports a list with the following information:
#'
#' - The package version of `{r5r}` in use.
#' - The installed version of `R5.jar`.
#' - The Java version in use.
#' - The amount of memory set to Java through the `java.parameters` option.
#' - The user's Session Info.
#'
#' @return A `list` with information of the versions of the r5r package, Java
#' and R5 Jar in use, the memory set to Java and user's Session Info.
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' r5r_sitrep()
#'
#' @export
r5r_sitrep <- function() {
  r5r_package_version <- utils::packageVersion("r5r")

  jar_dir <- r5r_env$cache_dir
  jar_dir_files <- list.files(jar_dir)
  jar_dir_files_full_names <- list.files(jar_dir, full.names = TRUE)

  r5r_jar <- jar_dir_files[grepl("r5r_\\d_\\d_\\d.*\\.jar", jar_dir_files)]

  r5r_jar_version <- r5r_env$r5_jar_version
  r5r_jar_path <- jar_dir_files_full_names[
    grepl("\\/r5r_\\d_\\d_\\d*\\.jar", jar_dir_files_full_names)
  ]

  r5_jar <- jar_dir_files[grepl("r5-v\\d\\.\\d.*\\.jar", jar_dir_files)]
  r5_jar_version <- substr(r5_jar, 5, 7)
  r5_jar_path <- jar_dir_files_full_names[
    grepl("\\/r5-v\\d\\.\\d.*\\.jar", jar_dir_files_full_names)
  ]

  rJava::.jinit()
  java_version <- rJava::.jcall(
    "java.lang.System",
    "S",
    "getProperty",
    "java.version"
  )

  set_memory <- getOption("java.parameters")

  output_list <- list(r5r_package_version,
                      r5_jar_version,
                      java_version,
                      # r5_jar_path,
                      # r5r_jar_path,
                      set_memory,
                     # jar_dir_files_full_names,
                      utils::sessionInfo())

  names(output_list) <- c('r5r_package_version',
                          'r5_jar_version',
                          'java_version',
                         # 'r5_jar_path',
                         # 'r5r_jar_path',
                          'set_memory',
                         # 'jar_dir_files_full_names',
                          'session_info'
                          )


  return(output_list)
}

#' Build a transport network used for routing in R5
#'
#' Builds a multimodal transport network used for routing in `R5`, combining
#' multiple data inputs present in the directory where the network should be
#' saved to. The directory must contain only one street network file (in
#' `.osm.pbf` format). It may optionally contain one or more public transport
#' GTFS feeds (in `.zip` format), when used for public transport routing, and a
#' `.tif` file describing the elevation profile of the study area. If there is
#' more than one GTFS feed in the directory, all feeds are automatically merged.
#' If there is already a `'network.dat'` file in the directory, the function will
#' simply read it and load it to memory (unless specified not to do so).
#'
#' @template verbose
#' @param data_path A string pointing to the directory where data inputs are
#'        stored and where the built `network.dat` will be saved.
#' @param temp_dir A logical. Whether the `network.dat` file should be saved to
#'        a temporary directory. Defaults to `FALSE`.
#' @template elevation
#' @param overwrite A logical. Whether to overwrite an existing `network.dat`
#'        or to use a cached file. Defaults to `FALSE` (i.e. use a cached
#'        network).
#'
#' @return A `r5r_network` object representing the built network to connect with
#'         `R5` routing engine.
#'
#' @template elevation_section
#'
#' @family Build network
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(r5r)
#'
#' # directory with street network and gtfs files
#' data_path <- system.file("extdata/poa", package = "r5r")
#'
#' r5r_network <- build_network(data_path)
#' @export
build_network <- function(data_path,
                          verbose = FALSE,
                          temp_dir = FALSE,
                          elevation = "TOBLER",
                          overwrite = FALSE) {

  # check inputs ------------------------------------------------------------

  checkmate::assert_directory_exists(data_path)
  checkmate::assert_logical(verbose)
  checkmate::assert_logical(temp_dir)
  checkmate::assert_character(elevation)
  checkmate::assert_logical(overwrite)

  elevation <- set_elevation(elevation)

  # expand data_path to full path, as required by rJava api call
  data_path <- path.expand(data_path)

  # check Java version installed locally and init java
  start_r5r_java(data_path = data_path, temp_dir = temp_dir, verbose = verbose)

  # check if data_path has osm.pbf, .tif gtfs data, or a network.dat file
  any_network <- length(grep("network.dat", list.files(data_path))) > 0
  any_pbf  <- length(grep(".pbf", list.files(data_path))) > 0
  any_gtfs <- length(grep(".zip", list.files(data_path))) > 0
  any_tif <- length(grep(".tif", list.files(data_path))) > 0

  # stop if there is no input data
  if (!(any_pbf | any_network)){
    cli::cli_abort("An OSM {.file .pbf} file is required in {.path {data_path}} to build a network.")
  }

  # use no elevation model if there is no raster.tif input data
  if (!(any_tif)) {
    elevation <- 'NONE'
    cli::cli_inform(c(
      i = "No raster {.file .tif} files found in {.path {data_path}}. Using elevation = {.val 'NONE'}."
    ))
  }

  # check if data_path already has a network.dat file
  dat_file <- file.path(data_path, "network.dat")

  if (checkmate::test_file_exists(dat_file) && !overwrite) {
    r5r_network <- rJava::.jnew("org.ipea.r5r.R5RCore", data_path, verbose, elevation)

    cli::cli_inform(c(
      i = "Using cached network from {.path {dat_file}}."
    ))

  } else {
    # check if the user has permission to write to the data directory. if not,
    # R5 won't be able to create the required files and will fail with a
    # not-that-enlightening error
    error_if_no_write_permission(data_path)

    # stop r5 in case it is already running
    suppressMessages( r5r::stop_r5() )

    # clean up any files that might have been created by previous r5r usage
    # if the files do not exist 'file.remove()' will raise a warning, which is
    # suppressed here
    mapdb_files <- list.files(data_path, full.names = TRUE)
    mapdb_files <- mapdb_files[grepl("\\.mapdb", mapdb_files)]
    suppressWarnings(
      invisible(file.remove(dat_file, mapdb_files))
    )

    # build new r5r_network
    r5r_network <- rJava::.jnew("org.ipea.r5r.R5RCore", data_path, verbose, elevation, check=F)
    ex = rJava::.jgetEx(clear=TRUE)
    if (!is.null(ex)) {
      msg <- rJava::.jcall(ex, "S", "toString")
      if (grepl("Geographic extent of street layer", msg)) {
        cli::cli_abort(
          "Geographic extent of street layer exceeds limit of {.val 975,000} {.unit km^2}."
        )
      } else {
        ex$printStackTrace()
        return(NULL)
      }
    }

    # display a message if there is a PBF file but no GTFS data
    if (any_pbf && !any_gtfs) {
      cli::cli_inform(c(
        i = "No public transport data ({.file GTFS}) found in {.path {data_path}}. Graph will be built with the street network only."
      ))
    }

    errors = java_to_dt(r5r_network$gtfsErrors)
    errfile = file.path(data_path, "gtfs_errors.csv")

    # always write error file even when empty, so that if you fix errors the error file gets overwritten on rebuild
    data.table::fwrite(errors, errfile)

    if (any(errors$priority == "HIGH")) {
      cli::cli_abort(c(
        "x"="High priority GTFS errors found; network build failed.",
        "i" = "Use {.fn r5r::get_gtfs_errors} to see them, or review {.file {errfile}}."

      ))
    } else if (nrow(errors) > 0) {
        cli::cli_inform(c(
          "!" = "{nrow(errors)} errors found in GTFS.",
          "i" = "Use {.fn r5r::get_gtfs_errors} to see them, or review {.file {errfile}}."
        ))
    }

    cli::cli_inform(c(
      v = "Finished building network at {.path {data_path}}"
    ))
  }

  return(wrap_r5r_network(r5r_network))
}

error_if_no_write_permission <- function(data_path) {
  write_permission <- file.access(data_path, mode = 2)

  normalized_path <- normalizePath(data_path)

  if (write_permission == -1) {
    cli::cli_abort(
      c(
        "Permission to write to {.path {normalized_path}} denied.",
        i = paste0(
          "{.pkg r5r} needs write privilege to create the network files. ",
          "Please make sure you have this privilege in the provided directory."
        )
      ),
      class = "dir_permission_denied",
      call = rlang::caller_env()
    )
  }

  return(invisible(TRUE))
}

---
title: 'Trade-offs between travel time and monetary cost'
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
abstract: "This vignette shows how to use the `pareto_frontier()` function to examine the trade-offs between travel time and monetary cost in travel time matrices in r5r."
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Trade-offs between travel time and monetary cost} 
  %\VignetteEngine{knitr::rmarkdown} 
  \usepackage[utf8]{inputenc}
bibliography: references.json
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)
```

# 1. Introduction

In most cases, transport routing models find either the fastest or the lowest-cost
routes that connect places in a given transport network. Sometimes, though, we 
might want a more sophisticated analysis that considers both the time and monetary 
costs that public transport passengers have to face. The problem here is that simultaneously accounting for both time and monetary costs is a major challenge 
for routing models because of the trade-offs between the objectives of minimizing 
trip duration and cost [@conway2019getting].

To address this problem, `r5r` has a function called `pareto_frontier()`, which calculates the most efficient route possibilities between origin destination pairs considering multiple combinations of travel time and monetary costs. This vignette 
uses a reproducible example to demonstrate how to use `pareto_frontier()` and 
interpret its results.



## 2. What the `pareto_frontier` means.

Imagine a hypothetical journey from A to B. There are multiple route alternatives 
between this origin and destination with varying combinations of travel time and 
cost (figure below).

* Walking from A to B would be the **cheapest** option but it would take 50 minutes.
* The **fastest** option would be to take a bus to a subway station and transfer to the subway. This option would only take 15 minutes, but it would cost $8.
* There are some intermediary alternatives, such as taking:
    * a single bus, $3 for 35 min.
    * two buses with one transfer, $5 for 29 min.
    * taking the subway after walking to the station , $6 for 20 min.

This figure illustrates the Pareto frontier of alternative routes from A to B. In 
other words, it shows the most optimal set of route alternatives between A and B. 
There are certainly other route options, but there is no other option that is both 
faster and cheaper at the same time.

```{r, echo = FALSE, fig.width=7, fig.height=4}
library(ggplot2)

# data.frame
df <- structure(list(option = c(1, 2, 3, 4, 5), 
                     modes = c("Walk", "Bus","Bus + Bus", "Subway", "Bus + Subway"), 
                     time = c(50, 35, 29, 20, 15), 
                     cost = c(0, 3, 5, 6, 8)), class = "data.frame", 
                row.names = c(NA, -5L))

# figure
ggplot(data=df, aes(x=cost, y=time, label = modes)) + 
  geom_step(linetype = "dashed") +
  geom_point() +
  geom_text(color='gray30', hjust = -.2, nudge_x = 0.05, angle = 45) +
  labs(title='Pareto frontier of alternative routes from A to B', subtitle = 'Hypotetical example') +
  scale_x_continuous(name="Travel cost (BRL)", breaks=seq(0,12,3)) +
  scale_y_continuous(name="Travel time (minutes)", breaks=seq(0,60,10)) +
  coord_cartesian(xlim = c(0,14), ylim = c(0, 60)) +
  theme_classic()
```

<br>

This kind of abstraction allows us to have a better grasp of the trade-offs between travel time and monetary cost passengers face when using public transport. It also 
allows us to calculate cumulative-opportunity accessibility metrics with cutoffs 
for both time and cost (e.g. the number of jobs reachable from a given origin with 
limits of 40 minutes and $5) (ref paper by Matt and Anson).

Let's see a couple concrete examples showing how `r5r` can calculate the Pareto 
frontier for multiple origins.



## 3. Demonstration of `pareto_frontier()`.

### 3.1 Build routable transport network with `build_network()`

First, let's build the network and create the routing inputs. In this example 
we'll be using the a sample data set for the city of Porto Alegre (Brazil) 
included in `r5r`.

```{r, message = FALSE}
# increase Java memory
options(java.parameters = "-Xmx2G")

# load libraries
library(r5r)
library(data.table)
library(ggplot2)
library(dplyr)

# build a routable transport network with r5r
data_path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path)

# routing inputs
mode <- c('walk', 'transit')
max_trip_duration <- 90 # minutes

# load origin/destination points of interest
points <- fread(file.path(data_path, "poa_points_of_interest.csv"))

```


### 3.2 Set up the fare structure

Now we need to set what are the fare rules of our public transport system. These 
rules will be used by `R5` to calculate the monetary cost of alternative routes. 
In the case of Porto Alegre, the fare rules are as follows:
* Each bus ticket costs R$ 4.80.
*  Riding a second bus adds `$` 2.40 to the total cost. Subsequent bus rides cost the full ticket price of $ 4.80.
* Each train ticket costs $ 4.50. Once a passenger enters a train station, she can take an unlimited amount of train trips as long as she doesn’t leave a station.
* The integrated fare between bus and train has a 10% discount, which totals $ 8.37.

We can create `list` object with these fare rules with the support of the `setup_fare_structure()` function as shown in the code below. A detailed explanation of how to use the fare structure of `5r5` can be found in [(this other vignette)](https://ipeagit.github.io/r5r/articles/fare_structure.html).


```{r}
# create basic fare structure
fare_structure <- setup_fare_structure(
  r5r_network, 
  base_fare = 4.8,
  by = "MODE"
  )

# update the cost of bus and train fares
fare_structure$fares_per_type[, fare := fcase(type == "BUS", 4.80,
                                             type == "RAIL", 4.50)]

# update the cost of tranfers
fare_structure$fares_per_transfer[, fare := fcase(first_leg == "BUS" & second_leg == "BUS", 7.2,
                                                 first_leg != second_leg, 8.37)]

# update transfer_time_allowance to 60 minutes
fare_structure$transfer_time_allowance <- 60

fare_structure$fares_per_type[type == "RAIL", unlimited_transfers := TRUE]
fare_structure$fares_per_type[type == "RAIL", allow_same_route_transfer := TRUE]

```

For convenience, we can save these fare rules as a `zip` file and load again for 
a future application.


```{r}
# save fare rules to temp file
temp_fares <- tempfile(pattern = "fares_poa", fileext = ".zip")
r5r::write_fare_structure(fare_structure, file_path = temp_fares)


fare_structure <- r5r::read_fare_structure(file.path(data_path, "fares/fares_poa.zip"))
```


### 3.3 Calculating a `pareto_frontier()`.

In this example, we calculate the Pareto frontier from all origins to all 
destinations considering multiple cutoffs of monetary costs:

* $1, which would only allow for walking trips
* $4.5, which would only allow for rail trips
* $4.8, which would allow for a single bus trip
* $7.20, which would allow for bus + bus
* $8.37, which would allow for walking walking + bus + rail

```{r}
departure_datetime <- as.POSIXct("13-05-2019 14:00:00", 
                                 format = "%d-%m-%Y %H:%M:%S")

prtf <- pareto_frontier(
  r5r_network,
  origins = points,
  destinations = points,
  mode = c("WALK", "TRANSIT"),
  departure_datetime = departure_datetime,
  fare_structure = fare_structure,
  fare_cutoffs = c(1, 4.5, 4.8, 7.20, 8.37),
  progress = TRUE
  )

head(prtf)
```


For the sake of illustration, let's check the optimum route alternatives from 
the Farrapos train station to (a) the Praia de Belas shopping mall and (b) the 
Moinhos hospital. An optimum route alternative means that one cannot make a faster trip without 
increasing costs, and one cannot make a cheaper trip without increasing travel time.


```{r, echo = FALSE, fig.width=7, fig.height=4, warning=FALSE}
# select origin and destinations
pf2 <- dplyr::filter(prtf, to_id == 'farrapos_station'  &
                       from_id %in% c('moinhos_de_vento_hospital', 
                                      'praia_de_belas_shopping_center'))

# recode modes
pf2[, modes := fcase(monetary_cost == 1,   'Walk',
                     monetary_cost == 4.5, 'Train',
                     monetary_cost == 4.8, 'Bus',
                     monetary_cost == 7.2, 'Bus + Bus',
                     monetary_cost == 8.37, 'Bus + Train')]

# plot
ggplot(data=pf2, aes(x=monetary_cost, y=travel_time, color=from_id, label = modes)) + 
  geom_step(linetype = "dashed") +
  geom_point() +
  geom_text(color='gray30', hjust = -.2, nudge_x = 0.05, angle = 45) +
  labs(title='Pareto frontier of alternative routes from Farrapos station to:', 
       subtitle = 'Praia de Belas shopping mall and Moinhos hospital',
       color='Destination') +
  scale_x_continuous(name="Travel cost ($)", breaks=seq(0,12,2)) +
  scale_y_continuous(name="Travel time (minutes)", breaks=seq(0,120,20)) +
  coord_cartesian(xlim = c(0,12), ylim = c(0, 120)) +
  theme_classic() + theme(legend.position=c(.2,0.8))
```

### Cleaning up after usage

`r5r` objects are still allocated to any amount of memory previously set after they are done with their calculations. In order to remove an existing `r5r` object and reallocate the memory it had been using, we use the `stop_r5` function followed by a call to Java's garbage collector, as follows:

```{r, message = FALSE}
r5r::stop_r5(r5r_network)
rJava::.jgc(R.gc = TRUE)
```

If you have any suggestions or want to report an error, please visit [the package GitHub page](https://github.com/ipeaGIT/r5r).


## References

---
title: 'Accessibility'
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
abstract: "This vignette shows how to calculate and visualize accessibility in R using the `r5r` package."
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Accessibility} 
  %\VignetteEngine{knitr::rmarkdown} 
  \usepackage[utf8]{inputenc}
bibliography: references.json
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)
```

# 1. Introduction 

Accessibility indicators measure the ease with which opportunities, such as jobs,
can be reached by a traveler from a particular location [@levinson2020manual]. This vignette shows how to calculate and visualize accessibility in R using the [`r5r` package](https://ipeagit.github.io/r5r/index.html) using a reproducible example. In this example, we will be using a sample data set for the city of Porto Alegre (Brazil) included in `r5r`. 

There are two ways to calculate / visualize accessibility using `r5r`. The quick and easy option is using the `r5r::accessibility()` function. The other alternative requires one to first calculate a travel time matrix, and then to use the [{accessibility} package](https://ipea.github.io/accessibility/). This is a more flexible options because the `{accessibility}` package provides a wider range of options of accessibility metrics. We will cover both approaches in this vignette. 

Before we start, we need to increase Java memory + load a few libraries, and to build routable transport network.


# 2. Build routable transport network with `build_network()`

### Increase Java memory and load libraries

First, we need to increase the memory available to Java and load the packages used in this vignette. Please note we allocate RAM memory to Java *before* loading our libraries.

```{r, message = FALSE}
options(java.parameters = "-Xmx2G")

library(r5r)
library(accessibility)
library(sf)
library(data.table)
library(ggplot2)
library(interp)
library(h3jsr)
library(dplyr)
```
To build a routable transport network with `r5r`, the user needs to call `build_network()` with the path to the directory where OpenStreetMap and GTFS data are stored.

```{r, message = FALSE}
# system.file returns the directory with example data inside the r5r package
# set data path to directory containing your own data if not running this example
data_path <- system.file("extdata/poa", package = "r5r")

r5r_network <- build_network(data_path)
```


# 3. Accessibility: quick and easy approach

There are different types of accessibility metrics. One of the simplest ones is the cumulative-opportunity metric, which counts the number of opportunities accessible from each location considering a maximum travel time cutoff. This is what we'll be calculating in this vignette using the parameter `decay_function = "step"`.

In this example, we will be calculating the number of schools and public healthcare facilities accessible by public transport within a travel time of up to 20 minutes. The sample data provided contains information on the spatial distribution of schools in Porto Alegre in the `points$schools` column, and healthcare facilities in the `points$healthcare` column. 

With the code below we compute the number of schools and healthcare accessible considering median of multiple travel time estimates departing every minute over a 60-minute time window, between 2pm and 3pm. The `accessibility()` function can calculate access to multiple opportunities in a single call, which is much more efficient and convenient than producing a travel time matrix of the study area and manually computing accessibility.


```{r, message = FALSE}
# read all points in the city
points <- fread(file.path(data_path, "poa_hexgrid.csv"))

# routing inputs
mode <- c("WALK", "TRANSIT")
max_walk_time <- 30      # in minutes
travel_time_cutoff <- 20 # in minutes
time_window <- 60        # in minutes
departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
                                 format = "%d-%m-%Y %H:%M:%S")

# calculate accessibility
access1 <- r5r::accessibility(
  r5r_network,
  origins = points,
  destinations = points,
  mode = mode,
  opportunities_colnames = c("schools", "healthcare"),
  decay_function = "step",
  cutoffs = travel_time_cutoff,
  departure_datetime = departure_datetime,
  max_walk_time = max_walk_time,
  time_window = time_window,
  progress = FALSE
  )

head(access1)
```


Mind you that the `r5r::accessibility()` also allow users to calculate gravity-based accessibility metrics, which can be calculated by setting the `decay_function` to one of the following: `"exponential"` `"fixed_exponential"`, `"linear"` or `"logistic"`. Nonetheless, there are several other types of accessibility metrics not implemented in R5, including floating catchment area metrics, travel cost to closest N opportunities, time interval based cumulative opportunity, etc. This is where the `{accessibility}` package comes in.


# 4. Accessibility: flexible approach

The `{accessibility}` package provides a much more flexible approach to calculate accessibility estimates. A key input here is a [travel time matrix](https://ipeagit.github.io/r5r/articles/travel_time_matrix.html), which we calculate using `r5r`:

```{r, message = FALSE}
# calculate travel time matrix
ttm <- r5r::travel_time_matrix(
  r5r_network,
  origins = points,
  destinations = points,
  mode = mode,
  departure_datetime = departure_datetime,
  max_walk_time = max_walk_time,
  time_window = time_window,
  progress = FALSE
  )

head(ttm)
```
Now to calculate a traditional cumulative opportunity metric like we did above, we just need to call the `accessibility::cumulative_cutoff()` function, and pass our travel time matrix and land use data as input:

```{r, message = FALSE}
# calculate accessibility
access_edu <- accessibility::cumulative_cutoff(
  travel_matrix = ttm, 
  land_use_data = points,
  opportunity = 'schools',
  travel_cost = 'travel_time_p50',
  cutoff = 20
  )

access_health <- accessibility::cumulative_cutoff(
  travel_matrix = ttm, 
  land_use_data = points,
  opportunity = 'healthcare',
  travel_cost = 'travel_time_p50',
  cutoff = 20
  )

head(access_edu)
head(access_health)
```


# 5. Map Accessibility

The final step is mapping the accessibility results calculated earlier. We can use at least two different approaches to map our accessibility estimates.

## 5.1 Choropleth maps

The first approach is to use choropleth maps. In our example, each point of reference is the centroid of a H3 hexagonal grid at a fine spatial resolution. In this case, we basically need to retrieve the polygons of the spatial grid, and merge it with our accessibility estimates.

```{r, message = FALSE, out.width='100%'}
# retrieve polygons of H3 spatial grid
grid <- h3jsr::cell_to_polygon(points$id, simple = FALSE)

# merge accessibility estimates
access_sf <- left_join(grid, access1, by = c('h3_address'='id'))

# plot
ggplot() +
  geom_sf(data = access_sf, aes(fill = accessibility), color= NA) +
  scale_fill_viridis_c(direction = -1, option = 'B') +
  labs(fill = "Number of\nfacilities within\n20 minutes") +
  theme_minimal() +
  theme(axis.title = element_blank()) +
  facet_wrap(~opportunity) +
  theme_void()

```


## 5.2 Spatial interpolation

An alternative approach is to use our accessibility estimates for each reference point and do some spatial interpolation so we can have a smoother spatial distribution. The code below demonstrates how to do that, producing a prettier map.

```{r, message = FALSE, out.width='100%'}
# interpolate estimates to get spatially smooth result
access_schools <- access1 %>% 
  filter(opportunity == "schools") %>%
  inner_join(points, by='id') %>%
  with(interp::interp(lon, lat, accessibility)) %>%
  with(cbind(acc=as.vector(z),  # Column-major order
             x=rep(x, times=length(y)),
             y=rep(y, each=length(x)))) %>% as.data.frame() %>% na.omit() %>%
  mutate(opportunity = "schools")

access_health <- access1 %>% 
  filter(opportunity == "healthcare") %>%
  inner_join(points, by='id') %>%
  with(interp::interp(lon, lat, accessibility)) %>%
  with(cbind(acc=as.vector(z),  # Column-major order
             x=rep(x, times=length(y)),
             y=rep(y, each=length(x)))) %>% as.data.frame() %>% na.omit() %>%
  mutate(opportunity = "healthcare")

access.interp <- rbind(access_schools, access_health)

# find results' bounding box to crop the map
bb_x <- c(min(access.interp$x), max(access.interp$x))
bb_y <- c(min(access.interp$y), max(access.interp$y))

# extract OSM network, to plot over map
street_net <- street_network_to_sf(r5r_network)

# plot
ggplot(na.omit(access.interp)) +
  geom_sf(data = street_net$edges, color = "gray55", size=0.01, alpha = 0.7) +
  geom_contour_filled(aes(x=x, y=y, z=acc), alpha=.7) +
  scale_fill_viridis_d(direction = -1, option = 'B') +
  scale_x_continuous(expand=c(0,0)) +
  scale_y_continuous(expand=c(0,0)) +
  coord_sf(xlim = bb_x, ylim = bb_y, datum = NA) + 
  labs(fill = "Number of\nfacilities within\n20 minutes") +
  theme_void() +
  facet_wrap(~opportunity)
```


### Cleaning up after usage

`r5r` objects are still allocated to any amount of memory previously set after they are done with their calculations. In order to remove an existing `r5r` object and reallocate the memory it had been using, we use the `stop_r5` function followed by a call to Java's garbage collector, as follows:

```{r, message = FALSE}
r5r::stop_r5(r5r_network)
rJava::.jgc(R.gc = TRUE)
```

If you have any suggestions or want to report an error, please visit [the package GitHub page](https://github.com/ipeaGIT/r5r).

## References

---
title: 'Using the time_window parameter'
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
abstract: "This vignette shows how to use and interpret the `time_window` parameter in r5r."
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Using the time_window parameter} 
  %\VignetteEngine{knitr::rmarkdown} 
  \usepackage[utf8]{inputenc}
bibliography: references.json
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

```

## 1. Introduction 

### The problem
To calculate the travel time from A to B, or to calculate the accessibility level at a given location, one has to select a departure time. However, travel time and accessibility estimates can differ significantly at different departure times because of how public transport service levels vary across the day [@stepniak2019impact]. Even a small difference, say leaving at `10:00am` or `10:04am` might importantly change travel time and accessibility estimates depending on when a person departs relative to when a public transport vehicle arrives, and how well transfers are coordinated given a service timetable. This is a very common issue related to the modifiable temporal unit problem (MTUP) [@pereiramtup2019; @levinson2020manual].

This problem gets even more complicated when public transport GTFS feeds have a `frequencies.txt` table. In these cases, we cannot know the exact departure time of vehicles, what creates greater uncertainty for our travel time and accessibility estimates [@conway2018accounting; @stewart2022half].


### The solution

A common strategy to overcome this problem is to calculate travel times and accessibilities at multiple departure times sampled over a time window, and then take the average or median value. Now you may ask, but how many departure times should I use? You might also be thinking  that doing multiple repeated routing analysis can be cumbersome and take a lot of time. Right?

Here is where `r5r` comes in. Both the `travel_time_matrix()`and `accessibility()` functions have a parameter called `time_window`. When this parameter is set, R<sup>5</sup> will automatically compute multiple travel times / accessibility estimates considering multiple departures per minute within the `time_window` selected by the user.

This vignette shows a reproducible example to explain how one can use the `time_window` and interpret the results.


## 2. How the `time_window` works and how to interpret the results.

As mentioned above, when `time_window` is set, R<sup>5</sup> computes multiple travel times / accessibility estimates starting at the specified `departure_datetime` and within the `time_window` selected by the user.  By default, `r5r` will generate one estimate per minute. Nonetheless, users can set a number to the `draws_per_minute` parameter that will change the number of Monte Carlo draws to perform per time window minute. The default value of `draws_per_minute` is 5, which mean 300 draws in a 60 minutes time window, for example. For a detailed discussion on the effect of number of draws on result stability, see Stewart et al [-@stewart2022half].

In this case, there isn't a single estimate of travel time / accessibility, but a distribution of several estimates that reflect the travel time / accessibility uncertainties in the specified time window. To get our heads around so many estimates, we can use the `percentiles` parameter to specify the percentiles of the distribution we are interested in. For example, if we select the 25th travel time percentile and the results show that the travel time estimate between A and B is 15 minutes, this means that 25% of all trips taken between these points within the specified time window are shorter than 15 minutes.


Let's see a couple concrete examples now.

## 3. Demonstration of `time_window`.

### 3.1 Build routable transport network with `build_network()`

First, let's build the network and create the routing inputs. In this example we'll be using the a sample data set for the city of São Paulo (Brazil) included in `r5r`.


```{r, message = FALSE}
# increase Java memory
options(java.parameters = "-Xmx2G")

# load libraries
library(r5r)
library(sf)
library(data.table)
library(ggplot2)
library(dplyr)

# build a routable transport network with r5r
data_path <- system.file("extdata/spo", package = "r5r")
r5r_network <- build_network(data_path)

# routing inputs
mode <- c('walk', 'transit')
max_walk_time <- 30 # minutes
max_trip_duration <- 90 # minutes

# load origin/destination points
points <- fread(file.path(data_path, "spo_hexgrid.csv"))

# departure datetime
departure_datetime = as.POSIXct("13-05-2019 14:00:00", 
                                format = "%d-%m-%Y %H:%M:%S")
```

ps. Please keep in mind that the Monte Carlo draws in `time_window` only affects the results when the GTFS feeds contain a `frequencies.txt` table. If the GTFS feed does not have a frequency table, r5r still allow for multiple runs over the set `time_window` in a deterministic way.


### 3.2 Accessibility with `time_window`.

In this example we calculate the number of schools accessible from each location within a 60-minute time window departing between 2pm and 3pm. In this example we'll be using a cumulative accessibility metric `decay_function = "step"` with a max time threshold of 45 minutes `cutoffs = 45`.

```{r, message = FALSE}
# estimate accessibility
acc <- r5r::accessibility(
  r5r_network,   
  origins = points,
  destinations = points, 
  opportunities_colnames = 'schools',
  mode = mode,
  max_walk_time = max_walk_time,
  decay_function = "step",
  cutoffs = 45,
  departure_datetime = departure_datetime,
  progress = FALSE,
  time_window = 60,
  percentiles = c(10, 20, 50, 70, 80)
  )

head(acc, n = 10)

```

This output is in long format, so the first 5 rows show the result for the same origin. In this case, we see that in only 10% of the trips departing from that origin between 2pm and 3pm a person would be able to access up to 111 schools. Meanwhile, 50% of the times she would only access 79 schools. By contrast, the accessibility from the other origin shown in the output above is 0, meaning there are no schools accessible from that location given the max travel time of 45 minutes.

We can use a plot like the one below to visualize this uncertainty in how accessibility levels might vary between 2pm and 3pm depending on the departure time within that 60-minute time window.

```{r}
# summarize
df <- acc[, .(min_acc = min(accessibility),
              median = accessibility[which(percentile == 50)],
              max_acc = max(accessibility)), by = id]

# plot
ggplot(data=df) +
  geom_linerange(color='gray', alpha=.5, aes(x = reorder(id, median) , 
                      y=median, ymin=min_acc, ymax=max_acc)) +
  geom_point(color='#0570b0', size=.5, aes(x = reorder(id, median), y=median)) +
  labs(y='N. of schools accessible\nby public transport', x='Origins sorted by accessibility',
       title="Accessibility uncertainty between 2pm and 3pm",
       subtitle = 'Upper limit 10% and lower limit 80% of the times') +
  theme_classic() +
  theme(axis.text.x=element_blank(),
        axis.ticks.x=element_blank())
```



### 3.3 Travel time matrix with `time_window`.

Now let's calculate all-to-all travel time estimates within a 60-minute time window departing between 2pm and 3pm and see how the output looks like.

```{r, message = FALSE}
# estimate travel time matrix
ttm <- travel_time_matrix(
  r5r_network,   
  origins = points,
  destinations = points,    
  mode = mode,
  max_walk_time = max_walk_time,
  max_trip_duration = max_trip_duration,
  departure_datetime = departure_datetime,
  progress = TRUE,
  time_window = 60,
  percentiles = c(10, 20, 50, 70, 80)
  )

head(ttm, n = 10)

```

Now let's look at the 2nd row of the output above. This output tell us that only 10% of the trips between 2pm and 3pm for that origin-destination pair took 39 minutes or less. Meanwhile, 50% of those trips took up tp 45 minutes and 80% of them were 48-minute long or shorter.

The last row in the result above has a few `NA`s. This tell us that at least 50% of all simulated trips between 2pm and 3pm for that origin-destination pair could not be completed because they took longer than the `max_trip_duration` we have set (90 minutes).



### 3.4 Expanded travel time matrix with `time_window`.

Finally, we can also use the `time_window` in the `expanded_travel_time_matrix()` function. In this case, though, when the user sets a `time_window` value, the `expanded_travel_time_matrix()` will return the fastest route alternative departing each minute within the specified time window. Please note this function can be very memory intensive for large data sets and time windows.

```{r}
ettm <- r5r::expanded_travel_time_matrix(
  r5r_network,
  origins = points[1:30,],
  destinations = points[31:61,],    
  mode = mode,
  max_walk_time = max_walk_time,
  max_trip_duration = max_trip_duration,
  departure_datetime = departure_datetime,
  progress = FALSE,
  time_window = 20
  )

head(ettm, n = 10)

```

### 3.5 Detailed itineraries with `time_window`.

In the `detailed_itineraries()` function, the number of Monte Carlo draws per minute is hardcoded to 1. This means that the function simulates only one departure per minute within the `time_window`. So if you set a `time_window` of 10 minutes, it would simulate 10 departures, one in each minute. This is largely because the `time_window` behaves slightly differently here.

See, functions like `travel_time_matrix()` or `accessibility()`, for example, return estimates of travel times or accessibility. In these cases, when we use the `time_window` parameter, these functions output selected percentiles of those values generated based on the distribution of all estimates that result from the multiple trip simulations.

The `detailed_itineraries()`, on the other hand, does not return travel times or accessibility estimates. It returns alternatives of trip journeys. In this case, when we use the `time_window` parameter, the function will return the optimal trip itinerary found within the time window. It can also return the optimal route along with multiple sub-optimal journey alternatives found within the time window if the the user sets `shortest_path = FALSE`.

*obs.* Mind you that `detailed_itineraries()` cannot be computed for public transport trips if the the network uses a frequencies-based GTFS feed. In these cases, we suggest using `gtfstools::frequencies_to_stop_times()` to create a suitable feed.



### Cleaning up after usage

`r5r` objects are still allocated to any amount of memory previously set after they are done with their calculations. In order to remove an existing `r5r` object and reallocate the memory it had been using, we use the `stop_r5` function followed by a call to Java's garbage collector, as follows:

```{r, message = FALSE}
r5r::stop_r5(r5r_network)
rJava::.jgc(R.gc = TRUE)
```

If you have any suggestions or want to report an error, please visit [the package GitHub page](https://github.com/ipeaGIT/r5r).

## References

---
title: 'Using custom OSM car speeds and LTS'
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
abstract: "This vignette shows how to calculate travel times and accessibility with custom OSM car speeds and Level of Traffic Stress, which can be used to simulate different scenarios of traffic congestion and road closures, or interventions in cycling infrastructure."
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Using custom OSM car speeds and LTS} 
  %\VignetteEngine{knitr::rmarkdown} 
  \usepackage[utf8]{inputenc}
bibliography: references.json
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

```

# 1. Introduction

By default, routing by car in `R5` considers that vehicles travel at the legal speed limit in each OSM road edge. This is commonly referred to as a "free flow scenario", without congestion. However, the average speed of car trips is different (usually slower) than the legal speed limit in most real case scenarios due to traffic conditions and driving behavior. Similarly, `R5` considers road speeds, road hierarchy and the quality of cycling infrastructure to infer the Level of Traffic Stress (LTS) of each road segment in the transportation network.

This vignette shows how you can calculate travel times and accessibility using custom OSM car speeds or LTS, which can be used to simulate different scenarios of traffic congestion and road closures, or interventions in cycling infrastructure.

In short, OSM car speeds and LTS values can be changed using two different strategies: (1) using a  `data.frame` to apply changes to individual road segments, or (2) using a spatial `sf` to apply changes to certain roads within / touching the spatial simple feature. Let's see how this works with a few examples using the a sample data set for the city of Porto Alegre (Brazil) included in `r5r`. First, let's load a few libraries and build our rotatable transportation network.

```{r, message=FALSE}
# increase Java memory
options(java.parameters = "-Xmx2G")

# load libraries
library(r5r)
library(dplyr)
library(data.table)
library(ggplot2)

# data path
data_path <- system.file("extdata/poa", package = "r5r")

# build network
r5r_network <- r5r::build_network(
  data_path = data_path,
  verbose = FALSE
  )
```


# 2. Changing car speeds

All of the routing and accessibility functions in {r5r} now have the following parameters that allow one to use custom OSM car speeds.

- `new_carspeeds`: here, users must pass either a `data.frame` that indicates the 
new car speed for each OSM edge id, OR an `sf data.frame` polygon that indicates 
the new car speed for all the roads that fall within each polygon.
- `carspeed_scale`: this parameter allows one to set the default car speed for all 
of the road segments not specified in `new_carspeeds`. By default, `carspeed_scale = NULL` and the speeds of the unlisted roads are kept unchanged.

Let's see a couple examples.

## 2.1 Changing car speeds by OSM edge

In this first example, we will pass the new car speeds using a sample `data.frame` that comes with the package. Mind you that this data frame must contain the columns `"osm_id"`, `"max_speed"` and `"speed_type"`. The `"speed_type"` column is of class character where all values must be either `"scale"` or `"km/h"`, indicating whether the values in `"max_speed"` should be interpreted as percentages of original speeds (`"scale"`) or as absolute speeds (`"km/h"`). Like this:

```{r, message=FALSE}
# read data.frame with new car speeds
edge_speed_factors <- read.csv(
  file.path(data_path, "poa_osm_congestion.csv")
  )

head(edge_speed_factors)
```

In this example, the values of the `"max_speed"` column are all set to `0.5` and `speed_type == "scale"`, which means that the driving speed of those OSM edges listed in the `data.frame` will be at 50% of the original speed in the OSM data.

Now you can simply run any routing or accessibility function passing the parameters `new_carspeeds` and `carspeed_scale`:

```{r, message=FALSE}
# origins and destination points
points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))

# travel time matrix
ttm_congestion <- r5r::travel_time_matrix(
  r5r_network = r5r_network,
  origins = points,
  destinations = points,
  mode = 'car',
  departure_datetime = Sys.time(),
  max_trip_duration = 30,
  new_carspeeds = edge_speed_factors,
  carspeed_scale = 0.8
)
```
Obs. Mind you that, even though we have set the speed factors to `0.5`, travel times might not become twice as long. This is because of how travel times by car are also affected by intersections, and how changes in the road speeds might also affect the route and hence the trip distance itself.  

Now let's dive into a more realistic examples.

### 2.1.1 Setting different congestion levels by road hierarchy

In this example, we'll set different speed factors for roads of different hierarchy levels. We can assume for example that congestion levels tend to be more intense in roads of higher hierarchy. We can do this in two simple steps.

First we need to do read the OSM data from our `.pbf` file, and to filter the OSM edges with the road types we want.


```{r, message=FALSE}
# path to OSM pbf
pbf_path <- paste0(data_path, "/poa_osm.pbf")
  
# read layer of lines from pbf
roads <- sf::st_read(
  pbf_path, 
  layer = 'lines', 
  quiet = TRUE
  )

# Filter only road types of interest
rt <- c("motorway", "primary", "secondary", "tertiary") 

roads <- roads |>
  select(osm_id, highway) |>
  filter(highway %in% rt)

head(roads)
```

Here's how the road network looks like.

```{r}
# map
plot(roads["highway"])

```

Now we only need to add a new column `"max_speed"` with values conditioned on the road type, and make sure the `osm_id` is of class `numeric`. We also need to include the column `speed_type` to make it clear that max speeds should be interpreted as a `"scale"` factor. The `data.frame` looks like this:

```{r}
new_edge_speeds <- roads |>
  mutate( 
    osm_id = as.numeric(osm_id),
    max_speed = case_when(
      highway == "motorway"  ~ 0.75,
      highway == "primary"   ~ 0.8,
      highway == "secondary" ~ 0.85,
      highway == "tertiary"  ~ 0.9)) |>
  sf::st_drop_geometry()

new_edge_speeds$speed_type <- "scale"

head(new_edge_speeds)

```

That's it. Now we can calculate travel times with the modified OSM car speeds.

```{r, message = FALSE, warning=FALSE}
# travel time matrix
ttm_congestion <- r5r::travel_time_matrix(
  r5r_network = r5r_network,
  origins = points,
  destinations = points,
  mode = 'car',
  departure_datetime = Sys.time(),
  max_trip_duration = 30,
  new_carspeeds = new_edge_speeds
  )

```

### 2.1.2 Applying the same speed factor to all roads

In this example, we'll simulate as if the speed limits of all roads were changed to 40 Km/h. To do this, we simply edit the `new_edge_speeds` table we created in our previous example to assign a 40 Km/h to each and every OSM id.


```{r, message = FALSE}
# edit table with custom speeds to 40 km/h
new_edge_speeds40 <- new_edge_speeds |>
  mutate(max_speed = 40,
         speed_type = "km/h")
  
# travel time matrix
ttm_congestion <- r5r::travel_time_matrix(
  r5r_network = r5r_network,
  origins = points,
  destinations = points,
  mode = 'car',
  departure_datetime = Sys.time(),
  max_trip_duration = 30,
  new_carspeeds = new_edge_speeds40
  )
```

### Extra tip: 

- **Road closure**: one can simulate a road closure by setting the `"max_speed"` value to `0`. This can be quite handy for studies that try to measure the resilience of transport systems to network disruptions.


## 2.2 Changing car speeds with a spatial polygon

If you do not want to set the speed factor for each individual OSM road edge, you 
can use one or more spatial polygons to set the new car speeds of all the 
roads within those polygons. In this example with the sample data from {r5r}, we have two polygons in the city of Porto Alegre. The first one covers the extended city center, and the second polygon covers a few important roads that connect two major avenues in the city.

```{r, message = FALSE}
# read sf with congestion polygons
congestion_poly <- readRDS(file.path(data_path, "poa_poly_congestion.rds"))

# preview
mapview::mapview(congestion_poly, zcol="scale")
```

Mind you that this `sf data.frame` must have a few mandatory columns:

- `"poly_id"`: a unique id for each polygon
- `"scale"`: the speed scaling factor for each polygon. Notice that this parameter only works with relative speed as percentages of original speeds. It *does not* work with absolute speeds in km/h.
- `"priority"`: a number ranking which polygon should be considered in case of overlapping polygons.

```{r, message = FALSE}
head(congestion_poly)
```
In this example, we are simulating a higher congestion level of the roads in the city center, which would be running at 70% of the legal speed limit, and a slightly better performance for the roads in the second polygon, running at 80% of the speed limit. Finally, we can set `carspeed_scale = 0.95` to simulate that all the other roads in the city would be running at 95%.

```{r, message = FALSE}
ttm_congestion <- r5r::travel_time_matrix(
  r5r_network = r5r_network,
  origins = points,
  destinations = points,
  mode = 'car',
  departure_datetime = Sys.time(),
  max_trip_duration = 30,
  new_carspeeds = congestion_poly,
  carspeed_scale = 0.95
  )
```

And that's it!



# 3. Changing cycling LTS values

There are two approaches to changing LTS, which are fairly similar to how we can change road speeds. This can be done either with a `data.frame` indicating the new LTS of each individual OSM edge id, or an `sf data.frame`. The key difference is that the sf object needs to be of type LINESTRING. R5 will then find the nearest road for each line and update its LTS value accordingly.

## 3.1 Changing LTS by OSM edge

Now let's simulate the local government has implemented protected cycle lanes or  tracks across a few designated roads in our transport network. Mind you that the
`data.frame` must contain the column `"lts"` indicating the new LTS value.

```{r, message=FALSE}
# read data.frame with new lts
edge_lts <- read.csv(
  file.path(data_path, "poa_osm_lts.csv")
  )

head(edge_lts)
```
Now we simply need to pass this `data.frame` with the new LTS values to the parameter `new_lts` in any routing/accessibility function:

```{r, message = FALSE}
ttm_new_lts <- r5r::travel_time_matrix(
  r5r_network = r5r_network,
  origins = points,
  destinations = points,
  mode = 'bicycle',
  departure_datetime = Sys.time(),
  max_trip_duration = 30,
  new_lts = edge_lts
  )
```

## 3.2. Changing LTS with a spatial polygon

Alternatively, we can pass a `sf` linestring of the roads where we want to 
simulate a some cycling network intervetion. Here we simulate that local authorities would have build dedicated lanes along all secondary roads in the city.

```{r, message = FALSE}
# read sf with congestion polygons
lts_lines <- readRDS(file.path(data_path, "poa_ls_lts.rds"))

# preview
mapview::mapview(lts_lines, zcol="lts")
```

Now we only need to pass this `sf` with the new LTS values to the `new_lts` parameter in any routing/accessibility function:

```{r, message = FALSE}
ttm_new_lts <- r5r::travel_time_matrix(
  r5r_network = r5r_network,
  origins = points,
  destinations = points,
  mode = 'bicycle',
  departure_datetime = Sys.time(),
  max_trip_duration = 30,
  new_lts = lts_lines
  )
```

# Cleaning up after usage

`r5r` objects are still allocated to any amount of memory previously set after they are done with their calculations. In order to remove an existing `r5r` object and reallocate the memory it had been using, we use the `stop_r5` function followed by a call to Java's garbage collector, as follows:

```{r, message = FALSE}
# stop an specific r5r network
r5r::stop_r5(r5r_network)

# or stop all r5r networks at once
r5r::stop_r5()
rJava::.jgc(R.gc = TRUE)
```


If you have any suggestions or want to report an error, please visit [the package GitHub page](https://github.com/ipeaGIT/r5r).

---
title: 'FAQ - Frequently Asked Questions'
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
abstract: "A short list of responses and clarifications about Frequently Asked Questions from r5r users."
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{FAQ - Frequently Asked Questions} 
  %\VignetteEngine{knitr::rmarkdown} 
  \usepackage[utf8]{inputenc}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

```

# 1. Why do some trips from/to the same ID have travel times larger than zero?

<details><summary>click to expand</summary>

> This can happen when the input point is distant to any routable road segment. In this case, R5 will snap the point to the closest road segment that can be traversed with the selected transport mode, and it will consider that the person would walk in Euclidean distance from the input point to the snapped location. So R5 is accounting for this walking time "as the crow flies" in the routing.

</details>



# 2. Is it possible to run `r5r` with custom modifications to street nework data?

<details><summary>click to expand</summary>

> Yes, all routing and accessibility functions in `r5r` have new parameters `new_carspeeds`, `carspeed_scale` and `new_lts` which allow one to use custom car speeds and LTS levels for cycling. These parameters provide convient and efficient ways to build different scenarios of traffic congestion, road closure and interventions in cycling infrastructure. [See this vignette](https://ipeagit.github.io/r5r/articles/scenarios.html). For other changes to the OSM network (e.g. including a new road link), you would need to edit the OpenStreetMap `.pbf` file direclty before using it in `r5r`. To to this, you can edit the `.pbf` file with [JOSM (https://wiki.openstreetmap.org/wiki/JOSM)](https://wiki.openstreetmap.org/wiki/JOSM). Mind you that the the OpenStreetMap tags can be changed but they cannot be removed from the data.

</details>




# 3. Why are the output results of `time_travel_matrix()` and `detailed_itineraries()` different?
  
<details><summary>click to expand</summary>

> The functions `time_travel_matrix()` (and `expanded_time_travel_matrix`) and `detailed_itineraries()` use different routing algorithms, as explained in the documentation of these functions. As such, we advise not to use or combine the output of these functions.

</details>




# 4. What does the ERROR "Geographic extent of street layer exceeds limit" mean? and what to do about it?

<details><summary>click to expand</summary>

> If you an error message says something like: *"Geographic extent of street layer (5315196 km2) exceeds limit of 975000 km2"*. This means the your study area is too large. Unfortunately, this is a limit hardcoded upstream in R5 so we cannot change it in r5r. The advised **solution** here would be to reduce the extent of the `OpenStreetMap.pbf` to a smaller area of interest. One can do this using the [Osmosis](https://wiki.openstreetmap.org/wiki/Osmosis) software. The code below illustrates how to do this by calling Osmosis from within R. You need to donwload the Osmosis program to your computer. The latest releases can be downloaded from [here](https://github.com/openstreetmap/osmosis/releases). Next, you can use this step-by-step illustrated in the example below:

```
# get the bounding box of your study area
study_area_polygon <- geobr::read_state(code_state = "ES")
area_bbox <- sf::st_bbox(study_area_polygon)

# input: the path to osmosis and the large pbf file in your local computer
osmosis_path <- "./osmosis_dir/bin/osmosis.bat"
large_pbf_path <- "./project_dir/large.pbf"

# path where you want to save the smaller .pbf file
smaller_pbf <-"./project_dir/smaller.pbf"

# prepare call to osmosis
osmosis_cmd <- sprintf("%s --read-pbf %s --bounding-box left=%s bottom=%s right=%s top=%s --write-pbf %s",
                       osmosis_path, large_pbf_path, 
                       area_bbox@xmin, area_bbox@ymin, area_bbox@xmax, area_bbox@ymax,
                       smaller_pbf)

# call to osmosis
shell(osmosis_cmd, translate = TRUE)
```
</details>



# 5. Is it possible to use custom car speed data with r5r?

<details><summary>click to expand</summary>

> By default, R5 considers the max speed limit of each road as set in OpenStreetMap data. Unfortunately, there is currently no easy way to change the car speeds of road segments from within R. Although you could probably do that by editing the osm.pbf file using other programs. See issue [#289](https://github.com/ipeaGIT/r5r/issues/289).

</details>




# 6. Why do I get identical results by public transport and walking?

<details><summary>click to expand</summary>

> If your travel time / accessibility results are identical for public transport and walking, it is probably because R5 did not detect that the public transport network is available for your requested trips. This is likely to occur in case you set a `departure_datetime` outside the calendar of public transport operations in your GTFS data. Check the `calendar.txt` file in your gtfs.zip feed. Alternatively, this could occur because there is no public transport trip option that would be faster than walking for the given origin-destination pair you queried.

</details>

---
title: 'Isochrones'
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
abstract: "This vignette shows how to calculate and visualize isochrones in R using the `r5r` package."
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Isochrones} 
  %\VignetteEngine{knitr::rmarkdown} 
  \usepackage[utf8]{inputenc}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)
```

# 1. Introduction

An isochrone map shows how far one can travel from a given place within a certain amount of time. In other other words, it shows all the areas reachable from that place within a maximum travel time. This vignette shows how to calculate and visualize isochrones in R using the [`r5r` package](https://ipeagit.github.io/r5r/index.html) using a reproducible example. In this example, we will be using a sample data set for the city of Porto Alegre (Brazil) included in `r5r`. Our aim here is to calculate several isochrones departing from the central bus station given different travel time thresholds. 

The `r5r::isochrone()` function allows you to build both polygon- and line-based 
isochrones. We will cover both approaches in this vignette, where we will be 
calculating isochrones by public transport from the central bus station in Porto Alegre.


Before we start, we need to increase Java memory + load a few libraries, and to  
build routable  transport network.


***Warning:*** If you want to calculate how many opportunities (e.g. jobs, or schools or hospitals) are located inside each isochrone, we strongly recommend you NOT to use the `isochrone()` function. You will find much more efficient ways to do this in the [Accessibility vignette](https://ipeagit.github.io/r5r/articles/accessibility.html).


# 2. Build routable transport network with `build_network()`

### Increase Java memory and load libraries

First, we need to increase the memory available to Java and load the packages used in this vignette. Please note we allocate RAM memory to Java *before* loading our libraries.

```{r, message = FALSE}
options(java.parameters = "-Xmx2G")

library(r5r)
library(sf)
library(data.table)
library(ggplot2)
```


To build a routable transport network with `r5r`, the user needs to call `build_network()` with the path to the directory where OpenStreetMap and GTFS data are stored.

```{r, message = FALSE}
# system.file returns the directory with example data inside the r5r package
# set data path to directory containing your own data if not running this example
data_path <- system.file("extdata/poa", package = "r5r")

r5r_network <- build_network(data_path)
```

# 3. Calculating and visualizing isochrones 


## 3.1 Polygon-based isochrones 

The most common approach here is to create polygon-based isochrones. To do this,
you need to pass the arguments `polygon_output = TRUE` and choose the `zoom` level.
The polygon-based isochrone in {r5r} are built on top of a regular grid based on 
[Web Mercator pixels](https://docs.conveyal.com/analysis/methodology#spatial-resolution). 
The zoom level can be controlled with the `zoom` parameter; higher zooms lead to 
more detailed isochrones at the expense of computational time. The default is 10 
(which uses cells of 153m at the Equator). In large networks, high zooms may not 
be possible and will give an error.

With the code below, `r5r` determines the isochrones considering the median travel 
time of multiple travel time estimates calculated departing every minute over a 
60-minute time window, between 2pm and 4pm. 

```{r, message = FALSE}
# read all points in the city
points <- fread(file.path(data_path, "poa_hexgrid.csv"))

# subset point with the geolocation of the central bus station
central_bus_stn <- points[291,]

# isochrone intervals
time_intervals <- seq(0, 100, 10)

# routing inputs
mode <- c("WALK", "TRANSIT")
max_walk_time <- 30      # in minutes
max_trip_duration <- 90  # in minutes
time_window <- 60        # in minutes
departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
                                 format = "%d-%m-%Y %H:%M:%S")

# calculate travel time matrix
iso1 <- r5r::isochrone(
  r5r_network,
  origins = central_bus_stn,
  mode = mode,
  polygon_output = TRUE, 
  cutoffs = time_intervals,
  departure_datetime = departure_datetime,
  max_walk_time = max_walk_time,
  max_trip_duration = max_trip_duration,
  time_window = time_window,
  progress = FALSE,
  zoom = 10
  )

```
As you can see, the `isochrone()` functions works very similarly to the `travel_time_matrix()` function. However, instead of returning a table with travel time estimates, it returns a `POLYGON  "sf" "data.frame"` for each isochrone of each origin when you set `polygon_output = TRUE`.

```{r, message = FALSE}
head(iso1)
```

Now it becomes super simple to visualize our isochrones on a map:

```{r, message = FALSE}
# extract OSM network
street_net <- street_network_to_sf(r5r_network)
main_roads <- subset(street_net$edges, street_class %like% 'PRIMARY|SECONDARY')
  
colors <- c('#ffe0a5','#ffcb69','#ffa600','#ff7c43','#f95d6a',
            '#d45087','#a05195','#665191','#2f4b7c','#003f5c')

ggplot() +
  geom_sf(data = iso1, aes(fill=factor(isochrone)), color = NA, alpha = .7) +
  geom_sf(data = main_roads, color = "gray55", size=0.01, alpha = 0.2) +
  geom_point(data = central_bus_stn, aes(x=lon, y=lat, color='Central bus\nstation')) +
  scale_fill_manual(values = rev(colors) ) +
  scale_color_manual(values=c('Central bus\nstation'='black')) +
  labs(fill = "Travel time\n(in minutes)", color='') +
  theme_minimal() +
  theme(axis.title = element_blank())
```
## 3.1 Line-based isochrones 

Alternatively, you can build line-based isochrones by simply passing
`polygon_output = FALSE` to the `isochrone()` function. Note that you do not need 
the `zoom` parameter here, and that the output is `LINESTRING  "sf" "data.frame"`.


```{r, message = FALSE}
# calculate travel time matrix
iso2 <- r5r::isochrone(
  r5r_network,
  origins = central_bus_stn,
  mode = mode,
  polygon_output = FALSE, 
  cutoffs = time_intervals,
  departure_datetime = departure_datetime,
  max_walk_time = max_walk_time,
  max_trip_duration = max_trip_duration,
  time_window = time_window,
  progress = FALSE
  )

head(iso2)

```


Now it becomes super simple to visualize our isochrones on a map:

```{r, message = FALSE}
ggplot() +
  geom_sf(data = iso2, aes(color=factor(isochrone)), alpha = .7) +
  scale_color_manual(values = rev(colors) ) +
  geom_point(data = central_bus_stn, aes(x=lon, y=lat), color='black') +
  labs(color = "Travel time\n(in minutes)", color='sadasd') +
  theme_minimal() +
  theme(axis.title = element_blank())
```


### Cleaning up after usage

`r5r` objects are still allocated to any amount of memory previously set after they are done with their calculations. In order to remove an existing `r5r` object and reallocate the memory it had been using, we use the `stop_r5` function followed by a call to Java's garbage collector, as follows:

```{r, message = FALSE}
r5r::stop_r5(r5r_network)
rJava::.jgc(R.gc = TRUE)
```

If you have any suggestions or want to report an error, please visit [the package GitHub page](https://github.com/ipeaGIT/r5r).

---
title: 'Accounting for monetary costs'
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
abstract: "This vignette shows how to configure and use custom fare rules in order to account for monetary travel costs when generating travel time matrices and accessibility estimates with the `r5r` package."
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Accounting for monetary costs} 
  %\VignetteEngine{knitr::rmarkdown} 
  \usepackage[utf8]{inputenc}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)
```

# 1. Introduction

Considering the monetary costs of public transport trips in the calculation of 
travel time matrices and accessibility estimates is a major challenge faced by 
researchers and planning practitioners. Each public transport system can have 
its own set of rules for calculating fares, with varying levels of complexity. 
Moreover, there are important trade offs between travel time and monetary costs 
across multiple trip alternatives and are currently not captured by any 
multimodal routing engine, except for R5.

R5 has native capabilities and an open architecture for creating and including 
fare structures in routing models, making it possible to estimate travel time 
matrices and accessibility estimates simultaneously considering different 
combinations of time and monetary cost cutoffs. The main challenge, however, is 
that a specific fare structure for each city needs to be programmed in Java and 
tightly integrated into R5, making this functionality out of reach for those who 
do not know how to code in Java (i.e. most of us!).

To help tackle this challenge, `r5r` has a simple generic rule-based fare 
structure that can be configured via a predefined set of properties and rules 
that can be set directly from R or using external tools such as text editors and
spreadsheets. This approach currently available in `r5r` is able to account for 
the monetary costs of public transport systems that follow a simple set of fare 
rules according to which the cost of a journey depends on combinations of modes 
(see details below).

**This vignette shows the features of `r5r`s fare structure. It also uses a 
reproducible example to demonstrate how to configure the fare structure to 
account for monetary costs when generating travel time matrices and accessibility 
estimates with `r5r`**. 



## 1.1 Details

A common feature among many public transport services is the possibility of 
discounted transfers, when passengers can use a single ticket for a trip composed 
of multiple rides sometimes combining different transport modes. Such trips usually 
come with a discount in the second or subsequent fares, as well as a limit on the 
number of discounted transfers the user can make and/or a time limit for using that 
discount. This is the type of fare structure currently covered by `r5r`.

We acknowledge that there are several types fare rules that vary from one public 
transport system to another. According to these rules, the cost of a journey can 
differ, for example, depending on: different costs for each trip leg, transport 
mode or route; distance- or zone-based fares; different fares for types of riders 
(e.g elderly people or students) or time of the day (e.g. peak and off-peak hours); 
among many others rules. As such, taking all of these possible rules into 
consideration when calculating the monetary cost of multimodal can be quite 
difficult. `r5r` currently does not cover these more complex fare rules.

The fare calculator currently available in `r5r` is not intended to be a robust 
solution that can take into consideration all public transport systems and their 
specific fare rules. That would be a Herculean task. The features included in 
`r5r`'s fare calculator are inspired by our empirical observations of Brazilian 
public transport systems, and is meant to be used mainly in the [Access to Opportunities project](https://www.ipea.gov.br/acessooportunidades/en/). Everyone else is 
welcome to use it, if the current features suit their needs.

obs. The GTFS format has some features for specifying public transport fares, but
those features are quite limited and are not enough for adequately representing 
many use cases. A new version of that specification is currently being developed 
[Fares V2](https://github.com/google/transit/issues/252), but it may take some 
time for it to be approved and for transport agencies actually start providing 
GTFS feeds with full fare information.

<br>

# 2. Reprex: the public transport system of Porto Alegre

In this vignette, we will be using the sample data set for the city of Porto 
Alegre (Brazil) included in `r5r`.  Before we start, we need to increase the memory available to Java and load the packages used in this vignette

```{r, message = FALSE}
options(java.parameters = "-Xmx2G")

library(r5r)
library(sf)
library(data.table)
library(ggplot2)
library(patchwork)
library(dplyr)
library(h3jsr)
```

Porto Alegre has a relatively straightforward public transport system, where the vast majority of the population that rely on transit ride buses. The city also has a metropolitan rail service that connects the city center to the neighboring northbound municipalities. That system can be seen in the map below.

```{r}
# setup and load Porto Alegre multimodal network into memory

# system.file returns the directory with example data inside the r5r package
# set data path to directory containing your own data if not using the examples
data_path <- system.file("extdata/poa", package = "r5r")

r5r_network <- build_network(data_path)

# load transit network as an SF
transit_network <- transit_network_to_sf(r5r_network)

# map
ggplot() +
  geom_sf(data=transit_network$routes, aes(color=mode)) +
  theme_void()
```

According to the fare rules in Porto Alegre, as in most Brazilian cities, the 
cost a a journey depends on a combination of number of subsequent trips and/or 
transport modes. In the case of Porto Alegre, the fare rules are as follows:

-   Each bus ticket costs R\$ 4.80.
-   Riding a second bus adds R\$ 2.40 to the total cost. Subsequent bus rides cost the full ticket price of R\$ 4.80.
-   Each train ticket costs R\$ 4.50. Once a passenger enters a train station, she can take an unlimited amount of train trips as long as she doesn't leave a station.
-   The integrated fare between bus and train has a 10% discount, which totals R\$ 8.37.

In the following sections, we will demonstrate how to implement those rules within r5r's fare calculator.

<br>

# 3. Setting up the fare structure

There are three support functions in `r5r` to help users configure the fare structure:

-   `setup_fare_structure()` analyses the study area's GTFS and builds a 'skeleton' fare structure structure with the parameters that need to be set;
-   `write_fare_structure()` and `read_fare_structure()` allow saving the current fare structure settings to disk, and reading them back into memory. The settings are saved as standard `.csv` files inside a zipped folder. These files can be edited outside the R session using external text editors and spreadsheet software, for user's convenience.

First, we need to call `setup_fare_structure()`, providing three parameters: the
current `r5r_network` object, a `base_fare` used to populate the fare structure, and
the `by` parameters that identifies what is the main property of the route that 
defines the different fares.

In the example below, the `base_fare` is the standard bus ticket price of R\$ 4.80. 
We are also stating that `by = "MODE"`, so that each transport mode has its own 
fares and integration rules. Users can also create a fare structure where fare 
rules of routes differ by `"AGENCY_ID"` or `"AGENCY_NAME"`, or simply set `by = "GENERIC"` when the entire system follows the same rules.

```{r}
fare_structure <- setup_fare_structure(r5r_network, 
                                       base_fare = 4.8,
                                       by = "MODE")
```


Now let's check the contents of the `fare_structure` object. We can see below that it is simply a `list` with a few properties and data.frames.

```{r}
head(fare_structure, n=7)

```

## 3.1 Global Properties

Let's configure the global properties first, which are the ones that are applied 
to the entire system. 

### `max_discounted_transfers`
Note that `max_discounted_transfers` is set to 1 by default. This means that the 
passenger gets a fare discount in the first transfer between buses, but she would 
pay the full fare price in subsequent transfers. 

### `transfer_time_allowance`
By default, `transfer_time_allowance` is set to 120 minutes. We have to set it to 
60 minutes to fit our use case (passengers have 60 minutes to take the second bus 
on a discounted fare, otherwise a full fare is charged).

### `fare_cap`
Finally, the `fare_cap` setting indicates if there is a maximum value that can be 
charged in a trip, beyond which all subsequent rides are free of charge. In this 
example, we can leave `fare_cap` set to its default `Inf` value because this 
feature is not applicable to Porto Alegre.

Here is how we can check or update the values of these components:
```{r}
fare_structure$max_discounted_transfers
fare_structure$transfer_time_allowance <- 60 # update transfer_time_allowance
fare_structure$fare_cap
```

## 3.2 Configure fares by transport mode

To configure mode-, transfer-, and route-specific properties, we can use the three `data.frames` inside our `fare_structure` list. Let's configure the modes first. 
Below, we can see that the `fares_per_type` data.frame contains five columns:

-   `mode`: the transport mode to which rules on each row refer to;
-   `unlimited_transfers`: a logical value `TRUE` or `FALSE` that indicates if that transport mode allows unlimited transfers between trips of the same mode, such as a metro/subway system where the passenger pays a fare to access a station and then can use as many services as she wants as long as she doesn't exit the system;
-   `allow_same_route_transfer`: a logical value indicating if a discounted transfer can be done between vehicles of the same route;
-   `use_route_fare`: another logical value that indicates if each route will have its own fare, or if all routes in this mode will use the fare indicated in this table;
-   `fare`: the full fare price of this mode.

```{r}
fare_structure$fares_per_type
```

We need to do a few small changes in the `fares_per_type` table to accomodate the
fare rules of Porto Alegre. In the `"RAIL"` mode, we need to set `unlimited_transfers`
and `allow_same_route_transfer` to `TRUE`, and update `fare` to 4.50. In the `"BUS"` 
mode, we can let the `allow_same_route_transfer` set to its default `FALSE` value, 
because even though there is a discount for transfers between buses (which is set 
in the following section), that discount is not valid when transferring between 
buses within the same route (for example, from bus route T1 to another T1). We'll 
do those changes below, using `data.table` notation.

```{r}
fare_structure$fares_per_type[type == "RAIL", unlimited_transfers := TRUE]
fare_structure$fares_per_type[type == "RAIL", fare := 4.50]
fare_structure$fares_per_type[type == "RAIL", allow_same_route_transfer := TRUE]
```

Checking the results below, everything looks OK:

```{r}
fare_structure$fares_per_type
```

## 3.3 Configure fares by transfers

The fare rules for transfer are stored in the `fares_per_transfer` data.frame, 
which is shown below. Each row contains the fare prices for transfers between the
modes specified in `first_leg` and `second_leg` columns.

```{r}
fare_structure$fares_per_transfer
```

Let's update `fare_per_transfer` to account for the actual integration rules in 
Porto Alegre.

-   The fare for "BUS" to "BUS" integration is composed of 4.80 for the first leg plus 2.40 for the second leg, which equals to a total fare of 7.20.

```{r}
# conditional update fare value
fare_structure$fares_per_transfer[first_leg == "BUS" & second_leg == "BUS", fare := 7.2]
```

- Transfers between "BUS" and "RAIL" (in any direction) cost 8.37, once the 10% discount is applied. Let's make a final update in the data.frame to account for that.

```{r}
# conditional update fare value
fare_structure$fares_per_transfer[first_leg != second_leg, fare := 8.37]

# use fcase instead ?
fare_structure$fares_per_transfer[, fare := fcase(first_leg == "BUS" & second_leg == "BUS", 7.2,
                                                 first_leg != second_leg, 8.37)]

```

- Transfers between "RAIL" and "RAIL" are free and unlimited, which is already 
accounted for in the field `unlimited_transfers` of the `fare_per_mode` table. 
Thus, the equivalent row of the `fare_per_transfer` data.frame needs to be removed. 
If we leave the that row in `fare_per_transfer`, transfers between "RAIL" and "RAIL" 
will count to the global `max_discounted_transfers` allowance.

```{r}
# remove row
fare_structure$fares_per_transfer <- fare_structure$fares_per_transfer[!(first_leg == "RAIL" & second_leg == "RAIL")]

```


Once all changes are applied, the `fare_per_transfer` data.frame should look like 
this:

```{r}
fare_structure$fares_per_transfer
```

## 3.4 Routes configuration

The information on the fare price for each route is stored in the `fares_per_route` 
data.frame. Below, we can see a sample of the bus and train routes in Porto Alegre. 
In case there a few special routes (e.g. express services) with specific fares, 
these values can be updated in this `fares_per_route` data.frame.



```{r}
tail(fare_structure$fares_per_route)
```

Basic route information is taken directly from the GTFS data (agency, route id 
and names, mode, etc), but the `route_fare` and `fare_type` columns were added 
specifically for the `r5r` fare structure.

-   `route_fare`: is used to set a specific fare for each route. This field can 
be used to represent services that have many unique fares, such as metropolitan 
/ suburban trains and buses. This is used together with the `use_route_fare` 
column in the `fares_per_type` table: the `route_fare` field is only considered 
by the `r5r` fare structure when `use_route_fare` of that mode is set to `TRUE`.

-   `fare_type`: is used to link each route with information in the `fares_per_type` 
and `fares_per_transfer` tables. In this example, `fare_type` is always the same 
as `mode`, because that was what we chose in the `by` parameter when calling `setup_fare_structure` earlier (we could have chosen to discriminate fares by 
agency, for example).

We actually don't have any change do to in the `fares_per_route` table, in this 
example. It does not matter that the `route_fare` value is wrong for the "RAIL" 
lines, because we are using the fares set in `fares_per_type` and `fares_per_transfer` 
which we already set up correctly before.

Now that our `fare_structure` is complete, we can use it to calculate travel time 
matrices and accessibility while accounting for monetary cost cutoffs. Let's see 
how it's done in the next sections.


# 4. Calculating travel time and accessibiilty accounting for monetary costs

The `travel_time_matrix()` and `accessibility()` functions have two new parameters 
to account for monetary costs thresholds:

-   `fare_structure`: the settings object that we've been working on.
-   `max_fare`: the maximum total fare that can be used in the trip.


## 4.1 Travel time with monetary cost

The following example shows travel time differences when monetary costs are 
accounted for, using the `travel_time_matrix()` function.


```{r}
## load input data
points <- read.csv(system.file("extdata/poa/poa_hexgrid.csv", package = "r5r"))

# calculate travel times function
calculate_travel_times <- function(fare) {
  ttm_df <- travel_time_matrix(
    r5r_network,
    origins = points,
    destinations = points,
    mode = c("WALK", "TRANSIT"),
    departure_datetime = as.POSIXct(
      "13-05-2019 14:00:00",
      format = "%d-%m-%Y %H:%M:%S"
    ),
    time_window = 1,
    fare_structure = fare_structure,
    max_fare = fare,
    max_trip_duration = 40,
    max_walk_time = 20
  )

  return(ttm_df)
}


# calculate travel times, and combine results
ttm <- calculate_travel_times(fare = Inf)
ttm_500 <- calculate_travel_times(fare = 5)

# merge results
ttm[ttm_500, on = .(from_id, to_id), travel_time_500 := i.travel_time_p50]
ttm[, travel_time_unl := travel_time_p50]
ttm[, travel_time_p50 := NULL]
```

Below, we can see a sample of the travel time differences with and without monetary 
cost restriction. We can see that some trips are not affected at all (`travel_time_unl == travel_time_500`), some trips take a little longer to complete (`travel_time_500 > travel_time_unl`), and other trips cannot be completed at all (`travel_time_500 == NA`).

```{r}
tail(ttm, 10)
```

The plots below show the overall distribution of the travel time differences and 
unreachable destinations:

```{r}
# plot of overall travel time differences between limited and unlimited cost travel time matrices 
time_difference = ttm[!is.na(travel_time_500), .(count = .N), 
                      by = .(travel_time_unl, travel_time_500)]

p1 <- ggplot(time_difference, aes(y = travel_time_unl, x = travel_time_500)) +
  geom_point(size = 0.7) +
  coord_fixed() +
  scale_x_continuous(breaks = seq(0, 45, 5)) +
  scale_y_continuous(breaks = seq(0, 45, 5)) +
  theme_light() +
  theme(legend.position = "none") +
  labs(y = "travel time (minutes)\nunestricted monetary cost",
       x = "travel time (minutes)\nmonetary cost restricted to BRL 5.00"
       )

# plot of unreachable destinations when the monetary cost limit is too low
unreachable <- ttm[, .(count = .N), by = .(travel_time_unl, is.na(travel_time_500))]
unreachable[, perc := count / sum(count, na.rm = T), by = .(travel_time_unl)]
unreachable <- unreachable[is.na == TRUE]
unreachable <- na.omit(unreachable)

p2 <- ggplot(unreachable, aes(x=travel_time_unl, y=perc)) +
  geom_col() +
  coord_flip() +
  scale_x_continuous(breaks = seq(0, 45, 5)) +
  scale_y_continuous(limits = c(0, 1), breaks = seq(0, 1, 0.2), 
                     labels = paste0(seq(0, 100, 20), "%")) +
  theme_light() +
  labs(x = "travel time (minutes)\nwithout monetary cost restriction",
       y = "% of unreachable destinations\nconsidering a R$ 5.00 monetary cost limit")

# combine both plots using patchwork
p1 + p2 + plot_annotation(subtitle = "Comparing travel times with and without monetary cost restriction")

```



## 4.2 Calculating accessibility with monetary cost

Now, we can answer questions like "how many health care facilities one can access 
in 60 minutes using public transport, on a R\$5.00 budget?". We'll do that below, 
and compare the results the accessibility unconstrained by monetary costs: 

```{r}
# calculate accessibility function
calculate_accessibility <- function(fare, fare_string) {
  access_df <- accessibility(
    r5r_network,
    origins = points,
    destinations = points,
    mode = c("WALK", "TRANSIT"),
    departure_datetime = as.POSIXct(
      "13-05-2019 14:00:00",
      format = "%d-%m-%Y %H:%M:%S"),
    time_window = 1,
    opportunities_colname = "healthcare",
    cutoffs = 40,
    fare_structure = fare_structure,
    max_fare = fare,
    max_trip_duration = 40,
    max_walk_time = 20,
    progress = FALSE)

  access_df$max_fare <- fare_string

  return(access_df)
}

# calculate accessibility, combine results, and convert to SF
access_500 <- calculate_accessibility(fare=5, fare_string="R$ 5.00 budget")
access_unl <- calculate_accessibility(fare=Inf, fare_string="Unlimited budget")

access <- rbind(access_500, access_unl)

# bring geometry
access$geometry <- h3jsr::cell_to_polygon(access$id)
access <- st_as_sf(access)

```

Finally, we can plot the results and see how accessibility levels can differ 
quite substantially when we account for monetary costs. 

```{r}
# plot accessibility maps
ggplot(data = access) +
  geom_sf(aes(fill = accessibility), color=NA, size = 0.2) +
  scale_fill_distiller(palette = "Spectral") +
  facet_wrap(~max_fare) +
  labs(subtitle = "Effect of monetary cost on accessibility") +
  theme_minimal() +
  theme(legend.position = "bottom",
        axis.text = element_blank())

```


### Cleaning up after usage

`r5r` objects are still allocated to any amount of memory previously set after they are done with their calculations. In order to remove an existing `r5r` object and reallocate the memory it had been using, we use the `stop_r5` function followed by a call to Java's garbage collector, as follows:

```{r, message = FALSE}
r5r::stop_r5(r5r_network)
rJava::.jgc(R.gc = TRUE)
```

If you have any suggestions or want to report an error, please visit [the package GitHub page](https://github.com/ipeaGIT/r5r).

---
title: 'Intro to r5r: Rapid Realistic Routing with R5 in R'
author: "Rafael H. M. Pereira, Marcus Saraiva, Daniel Herszenhut, Carlos Kaue Braga"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
abstract: "`{r5r}` is an R package for rapid realistic routing on multimodal transport networks (walk, bike, public transport and car) using R<sup>5</sup>. The package allows users to generate detailed routing analysis or calculate travel time matrices using seamless parallel computing on top of the R<sup>5</sup> Java machine <https://github.com/conveyal/r5>"
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Intro to r5r: Rapid Realistic Routing with R5 in R} 
  %\VignetteEngine{knitr::rmarkdown} 
  \usepackage[utf8]{inputenc}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

## removes files previously created by 'build_network()'
#data_path <- system.file("extdata/poa", package = "r5r")
#existing_files <- list.files(data_path)
#files_to_keep <- c(
#  "poa_hexgrid.csv",
#  "poa_osm.pbf",
#  "poa_points_of_interest.csv",
#  "poa_eptc.zip",
#  "poa_trensurb.zip",
#  'fares'
#  )
#files_to_remove <- existing_files[! existing_files %in% files_to_keep]
#invisible(file.remove(file.path(data_path, files_to_remove)))
```

# 1. Introduction 

**r5r** is an [R package for rapid realistic routing on multimodal transport networks](https://github.com/ipeaGIT/r5r) (walk, bike, public transport and car). It provides a simple and friendly interface to R<sup>5</sup>, a really fast and open source Java-based routing engine developed separately by [Conveyal](https://www.conveyal.com/). R<sup>5</sup> stands for [Rapid Realistic Routing on Real-world and Reimagined networks](https://github.com/conveyal/r5). More details about **r5r** can be found on the [package webpage](https://ipeagit.github.io/r5r/index.html) or on this [paper](
https://doi.org/10.32866/001c.21262).

# 2. Installation

You can install `{r5r}` from CRAN, or the development version from github.

```{r, eval = FALSE}
# from CRAN
install.packages('r5r')

# dev version with latest features
devtools::install_github("ipeaGIT/r5r", subdir = "r-package")
```


Please bear in mind that you need to have *Java Development Kit (JDK) 21* installed 
on your computer to use `{r5r}`. No worries, you don't have to pay for it. There are
numerous open-source JDK implementations, and you only need to install one JDK. Here are a few options:

- [Adoptium/Eclipse Temurin](https://adoptium.net/) (our preferred option)
- [Amazon Corretto](https://aws.amazon.com/corretto/)
- [Oracle OpenJDK](https://jdk.java.net/21/). 

The easiest way to install JDK is using the new [{rJavaEnv}](https://www.ekotov.pro/rJavaEnv/) package in R:

```{r, eval = FALSE}
# install {rJavaEnv} from CRAN
install.packages("rJavaEnv")

# check version of Java currently installed (if any) 
rJavaEnv::java_check_version_rjava()

## if this is the first time you use {rJavaEnv}, you might need to run this code
## below to consent the installation of Java.
# rJavaEnv::rje_consent(provided = TRUE)

# install Java 21
rJavaEnv::java_quick_install(version = 21)

# check if Java was successfully installed
rJavaEnv::java_check_version_rjava()

```


# 3. Usage

First, we need to increase the memory available to Java. This has to be done **before** loading the `{r5r}` library because, by default, `R` allocates only 512MB of memory for Java processes, which is not enough for large queries using `{r5r}`. To increase available memory to 2GB, for example, we need to set the `java.parameters` option at the beginning of the script, as follows:

```{r, message = FALSE}
options(java.parameters = "-Xmx2G")

# By default, {r5r} uses all CPU cores available. If you want to limit the 
# number of CPUs to 4, for example, you can run:  
options(java.parameters = c("-Xmx2G", "-XX:ActiveProcessorCount=4"))
```

Note: It's very important to allocate enough memory before loading `{r5r}` or any other Java-based package, since `rJava` starts a Java Virtual Machine only once for each R session. It might be useful to restart your R session and execute the code above right after, if you notice that you haven't succeeded in your previous attempts. 

Then we can load the packages used in this vignette:

```{r, message = FALSE, warning = FALSE}
library(r5r)
library(sf)
library(data.table)
library(ggplot2)
```

The `{r5r}` package has seven **fundamental functions**:

1. `build_network()` to build a routable transport network;

2. `accessibility()` for fast computation of access to opportunities considering
a selected decay function;

3. `travel_time_matrix()` for fast computation of travel time estimates between origin/destination pairs considering departure time;

4. `arrival_travel_time_matrix()` for calculating travel time matrices between origin destination pairs considering a time of arrival. The output includes additional information such as the routes used and total time disaggregated by access, waiting, in-vehicle and transfer times.

5. `expanded_travel_time_matrix()` for calculating travel matrices between origin destination pairs with additional information such as routes used and total time disaggregated by access, waiting, in-vehicle and transfer times.

6. `detailed_itineraries()` to get detailed information on one or multiple alternative routes between origin/destination pairs.

7. `pareto_frontier()` for analyzing the trade-off between the travel time and monetary costs of multiple route alternatives between origin/destination pairs.

8. `isochrone()` to estimate the polygons of the areas that can be reached from an origin point at different travel time limits.

Most of these functions also allow users to account for monetary travel costs 
when generating travel time matrices and accessibility estimates. More info about
how to consider monetary costs can be found in [this vignette](https://ipeagit.github.io/r5r/articles/fare_structure.html).


The package also includes a few **support functions**.

1. `street_network_to_sf()` to extract OpenStreetMap network in sf format from a `network.dat` file.

2. `transit_network_to_sf()` to extract transit network in sf format from a `network.dat` file.

3. `find_snap()` to find snapped locations of input points on street network.

4. `r5r_sitrep()` to generate a situation report to help debug eventual errors.


## 3.1 Data requirements:

To use `{r5r}`, you will need:

- A road network data set from OpenStreetMap in `.pbf` format (*mandatory*)
- A public transport feed in `GTFS.zip` format (optional)
- A raster file of Digital Elevation Model data in `.tif` format (optional)

Here are a few places from where you can download these data sets:

- OpenStreetMap
  - [osmextract](https://docs.ropensci.org/osmextract/) R package
  - [geofabrik](https://download.geofabrik.de/) website
  - [hot export tool](https://export.hotosm.org/) website
  - [BBBike.org](https://extract.bbbike.org/) website

- GTFS
  - [tidytransit](https://r-transit.github.io/tidytransit/) R package
  - [transitland](https://www.transit.land/) website
  - [Mobility Database](https://mobilitydatabase.org/) website

- Elevation
  - [elevatr](https://github.com/USEPA/elevatr) R package
  - Nasa's SRTMGL1 website


Let's have a quick look at how `{r5r}` works using a sample data set.



# 4. Demonstration on sample data

## Data

To illustrate the functionalities of `{r5r}`, the package includes a small sample data for the city of Porto Alegre (Brazil). It includes seven files:

* An OpenStreetMap network: `poa_osm.pbf`
* Two public transport feeds: `poa_eptc.zip` and `poa_trensurb.zip`
* A raster elevation data: `poa_elevation.tif`
* A `poa_hexgrid.csv` file with spatial coordinates of a regular hexagonal grid covering the sample area, which can be used as origin/destination pairs in a travel time matrix calculation.
* A `poa_points_of_interest.csv` file containing the names and spatial coordinates of 15 places within Porto Alegre
* A `fares_poa.zip` file with the fare rules of the city's public transport system.


```{r}
data_path <- system.file("extdata/poa", package = "r5r")
list.files(data_path)
```

The points of interest data can be seen below.  In this example, we will be looking at transport alternatives between some of those places.

```{r}
poi <- fread(file.path(data_path, "poa_points_of_interest.csv"))
head(poi)
```

The data with origin destination pairs is shown below. In this example, we will be using  200 points randomly selected from this data set.

```{r}
points <- fread(file.path(data_path, "poa_hexgrid.csv"))

# sample points
sampled_rows <- sample(1:nrow(points), 200, replace=TRUE)
points <- points[ sampled_rows, ]
head(points)
```



## 4.1 Building routable transport network with `build_network()`

The first step is to build the multimodal transport network used for routing in R<sup>5</sup>. This is done with the `build_network` function. This function does two things: (1) downloads/updates a compiled JAR file of R<sup>5</sup> and stores it locally in the `{r5r}` package directory for future use; and (2) combines the osm.pbf and gtfs.zip data sets to build a routable network object.

```{r, message = FALSE}
# Indicate the path where OSM and GTFS data are stored
r5r_network <- build_network(data_path = data_path)
```


## 4.2 Accessibility analysis

The fastest way to calculate accessibility estimates is using the `accessibility()`
function. In this example, we calculate the number of schools and health care 
facilities accessible in less than 60 minutes by public transport and walking.
More details in this vignette on [Calculating and visualizing Accessibility](https://ipeagit.github.io/r5r/articles/accessibility.html).

```{r, message = FALSE}
# set departure datetime input
departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
                                 format = "%d-%m-%Y %H:%M:%S")

# calculate accessibility
access <- accessibility(
  r5r_network,
  origins = points,
  destinations = points,
  opportunities_colnames = c("schools", "healthcare"),
  mode = c("WALK", "TRANSIT"),
  departure_datetime = departure_datetime,
  decay_function = "step",
  cutoffs = 60
  )

head(access)
```



## 4.3 Routing analysis

For fast routing analysis, **r5r** currently has three core functions: 
`travel_time_matrix()`, `expanded_travel_time_matrix()` and `detailed_itineraries()`.


### Fast many to many travel time matrix

The `travel_time_matrix()` function is a really simple and fast function to 
compute travel time estimates between one or multiple origin/destination pairs. 
The origin/destination input can be either a spatial `sf POINT` object, or a 
`data.frame` containing the columns `id, lon, lat`. The function also receives 
as inputs the *max walking distance*, in meters, and the *max trip duration*, in
minutes. Resulting travel times are also output in minutes.

This function also allows users to very efficiently capture the travel time 
uncertainties inside a given time window considering multiple departure times. 
[More info on this vignette](https://ipeagit.github.io/r5r/articles/time_window.html). 


```{r, message = FALSE}
# set inputs
mode <- c("WALK", "TRANSIT")
max_walk_time <- 30 # minutes
max_trip_duration <- 120 # minutes
departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
                                 format = "%d-%m-%Y %H:%M:%S")

# calculate a travel time matrix
ttm <- travel_time_matrix(
  r5r_network,
  origins = poi,
  destinations = poi,
  mode = mode,
  departure_datetime = departure_datetime,
  max_walk_time = max_walk_time,
  max_trip_duration = max_trip_duration
  )

head(ttm)
```

```{r ttm head, echo=FALSE, message=FALSE, out.width='100%', eval = FALSE}
knitr::include_graphics("https://github.com/ipeaGIT/r5r/blob/master/r-package/inst/img/vig_output_ttm.png?raw=true")
```


### Expanded travel time matrix with minute-by-minute estimates

For those interested in more detailed outputs, the `expanded_travel_time_matrix()` 
works very similarly with `travel_time_matrix()` but it brings much more 
information. It estimates for each origin destination pair the routes used and 
total time disaggregated by access, waiting, in-vehicle and transfer times. 
Please note this  function can be very memory intensive for large data sets.


```{r, message = FALSE}
# calculate a travel time matrix
ettm <- expanded_travel_time_matrix(
  r5r_network,
  origins = poi,
  destinations = poi,
  mode = mode,
  departure_datetime = departure_datetime,
  breakdown = TRUE,
  max_walk_time = max_walk_time,
  max_trip_duration = max_trip_duration
  )

head(ettm)
```



### Detailed itineraries

Most routing packages only return the fastest route. A key advantage of the `detailed_itineraries()` function is that is allows for fast routing analysis 
while providing multiple alternative routes between origin destination pairs. 
The output also brings detailed information for each route alternative at the 
trip segment level, including the transport mode, waiting times, travel time and
distance of each trip segment. 

In this example below, we want to know some alternative routes between one origin/destination pair only.

```{r, message = FALSE}
# set inputs
origins <- poi[10,]
destinations <- poi[12,]
mode <- c("WALK", "TRANSIT")
max_walk_time <- 60 # minutes
departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
                                 format = "%d-%m-%Y %H:%M:%S")

# calculate detailed itineraries
det <- detailed_itineraries(
  r5r_network,
  origins = origins,
  destinations = destinations,
  mode = mode,
  departure_datetime = departure_datetime,
  max_walk_time = max_walk_time,
  shortest_path = FALSE
  )

head(det)
```
```{r detailed head, echo = FALSE, out.width='100%', message = FALSE, eval = FALSE}
knitr::include_graphics("https://github.com/ipeaGIT/r5r/blob/master/r-package/inst/img/vig_output_detailed.png?raw=true")
```

The output is a `data.frame sf` object, so we can easily visualize the results.

#### Visualize results

**Static visualization** with `ggplot2` package: To provide a geographic context 
for the visualization of the results in `ggplot2`, you can also use the `street_network_to_sf()` function to  extract the OSM street network used in the routing.

```{r, message = FALSE}
# extract OSM network
street_net <- r5r::street_network_to_sf(r5r_network)

# extract public transport network
transit_net <- r5r::transit_network_to_sf(r5r_network)

# plot
ggplot() +
  geom_sf(data = street_net$edges, color='gray85') +
  geom_sf(data = det, aes(color=mode)) +
  facet_wrap(.~option) + 
  theme_void()

```

```{r ggplot2 output, echo = FALSE, out.width='100%', message = FALSE, eval = FALSE}
knitr::include_graphics("https://github.com/ipeaGIT/r5r/blob/master/r-package/inst/img/vig_detailed_ggplot.png?raw=true")
```


### Cleaning up after usage

`{r5r}` objects are still allocated to any amount of memory previously set after they are done with their calculations. In order to remove an existing `{r5r}` object and reallocate the memory it had been using, we use the `stop_r5` function followed by a call to Java's garbage collector, as follows:

```{r, message = FALSE}
r5r::stop_r5(r5r_network)
rJava::.jgc(R.gc = TRUE)
```

```{r, eval = TRUE, include = FALSE, message = FALSE}
# clean cache (CRAN policy)
r5r::r5r_cache(delete_file = 'all')

```

If you have any suggestions or want to report an error, please visit [the package GitHub page](https://github.com/ipeaGIT/r5r).

---
title: 'Trip planning with detailed_itineraries()'
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
abstract: "This vignette shows how to do route planning using the `detailed_itineraries()` function in r5r."
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Trip planning with detailed_itineraries()} 
  %\VignetteEngine{knitr::rmarkdown} 
  \usepackage[utf8]{inputenc}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)
```


# 1. Introduction 

**r5r** has some extremely efficient functions to run multimodal routing and accessibility analysis. In general, though, these functions output only the essential information required by most transport planning applications and simulation models. Moreover, the algorithms behind these function return only the optimal route in terms of minimizing travel times and/or monetary costs. Sometimes, though, we would like to do more simple route planning analysis and extract more information for each route. Also, we might be interested in finding not only the fastest route but some other suboptimal route alternatives too.

This is where the `detailed_itineraries()` function comes in. This function outputs for each origin destination pair a detailed route plan with information per leg, meaning a route taken by a single mode such as a walk to the bus stop. In R5's documentation these legs are referred to as 'segments', a word more usually used to describe small sections on the transport network. Results contain information on the mode, waiting times, travel times and distances for each leg (or 'segment' in R5 documentation) and the trip geometry. Moreover, the `detailed_itineraries()` function can also return results for multiple route alternatives. Let's see how this function works using a reproducible example.
 

**obs.** We only recommend using `detailed_itineraries()` in case you are interested in finding suboptimal alternative routes and/or need the geometry information of the outputs. If you only want to have route information detailed by trip segments, then we would strongly encourage you to use the `expanded_travel_time_matrix()` function.

# 2. Build routable transport network with `build_network()`

First, let's load some libraries and build our multimodal transport network. In this example we'll be using the a sample data set for the city of Porto Alegre (Brazil) included in `r5r`.

```{r, message = FALSE}
# increase Java memory
options(java.parameters = "-Xmx2G")

# load libraries
library(r5r)
library(sf)
library(ggplot2)
library(data.table)

# build a routable transport network with r5r
data_path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path)

# routing inputs
mode <- c('walk', 'transit')
max_trip_duration <- 60 # minutes

# departure time
departure_datetime <- as.POSIXct("13-05-2019 14:00:00", 
                                 format = "%d-%m-%Y %H:%M:%S")

# load origin/destination points
poi <- fread(file.path(data_path, "poa_points_of_interest.csv"))

```


# 3. Detailed info by trip segment for multiple trip alternatives

In this example below, we want to know some alternative routes between a single origin/destination pair. To get multiple route alternatives, we need to set `shortest_path = FALSE`.

Note that in the example below we set `suboptimal_minutes = 8`. In this case, `r5r` will consider sub-optimal routes that arrive up to 8 minutes after the arrival of the optimal route.

```{r, message = FALSE}
# set inputs
origins <- poi[10,]
destinations <- poi[12,]
mode <- c("WALK", "TRANSIT")
max_walk_time <- 60
departure_datetime <- as.POSIXct("13-05-2019 14:00:00",
                                 format = "%d-%m-%Y %H:%M:%S")

# calculate detailed itineraries
det <- detailed_itineraries(
  r5r_network,
  origins = origins,
  destinations = destinations,
  mode = mode,
  departure_datetime = departure_datetime,
  max_walk_time = max_walk_time,
  suboptimal_minutes = 8,
  shortest_path = FALSE
  )

head(det)
```
```{r detailed head, echo = FALSE, out.width='100%', message = FALSE, eval = FALSE}
knitr::include_graphics("https://github.com/ipeaGIT/r5r/blob/master/r-package/inst/img/vig_output_detailed.png?raw=true")
```


The output is a `data.frame sf` object, so we can easily visualize the results.

## 3.1 Visualize results

**Static visualization** with `ggplot2` package: To provide a geographic context 
for the visualization of the results in `ggplot2`, you can also use the `street_network_to_sf((` function to  extract the OSM street network used in the routing.

```{r, message = FALSE}
# extract OSM network
street_net <- r5r::street_network_to_sf(r5r_network)

# extract public transport network
transit_net <- r5r::transit_network_to_sf(r5r_network)

# plot
fig <- ggplot() +
        geom_sf(data = street_net$edges, color='gray85') +
        geom_sf(data = subset(det, option <4), aes(color=mode)) +
        facet_wrap(.~option) + 
        theme_void()

fig
```
```{r, message = FALSE, eval = FALSE}
# SAVE image
ggsave(plot = fig, filename = 'inst/img/vig_detailed_ggplot.png', 
       height = 5, width = 15, units='cm', dpi=200)
```

```{r ggplot2 output, echo = FALSE, out.width='100%', message = FALSE, eval = FALSE}
knitr::include_graphics("https://github.com/ipeaGIT/r5r/blob/master/r-package/inst/img/vig_detailed_ggplot.png?raw=true")
```

# 4. A few options:

### 4.1 Combining orings and destinations

- By default, `detailed_itineraries()` will query routes between the 1st origin to the 1st destination, then the 2nd origin to the 2nd destination, and so on. If you would like to query routes between all origins to all destinations you can simply set `all_to_all = TRUE`.

### 4.2 Keep geometry data in the output

- Be default, `detailed_itineraries()` will return the spatial geometry of results. To prevent retrieving this information you can simply set `drop_geometry = TRUE`.



# 5. Hack for frequency-based GTFS feeds

Please note that the `detailed_itineraries()` functions does not run on frequency-based GTFS feeds. A simple hack to overcome this problem is to convert your GTFS data from frequencies to time tables. This can be easily done using the [`gtfstools` package](https://ipea.github.io/gtfstools/). Here is how:

```{r, message = FALSE, eval = FALSE}
library(gtfstools)

# location of your frequency-based GTFS
freq_gtfs_file <- system.file("extdata/spo/spo.zip", package = "r5r")

# read GTFS data
freq_gtfs <- gtfstools::read_gtfs(freq_gtfs_file)

# convert from frequencies to time tables
stop_times_gtfs <- gtfstools::frequencies_to_stop_times(freq_gtfs)

# save it as a new GTFS.zip file
gtfstools::write_gtfs(gtfs = stop_times_gtfs,
                      path = tempfile(pattern = 'stop_times_gtfs', fileext = '.zip'))


```

... and now you can use `r5r` on this `stop_times_gtfs.zip`.

### Cleaning up after usage

`r5r` objects are still allocated to any amount of memory previously set after they are done with their calculations. In order to remove an existing `r5r` object and reallocate the memory it had been using, we use the `stop_r5` function followed by a call to Java's garbage collector, as follows:

```{r, message = FALSE}
r5r::stop_r5(r5r_network)
rJava::.jgc(R.gc = TRUE)
```

If you have any suggestions or want to report an error, please visit [the package GitHub page](https://github.com/ipeaGIT/r5r).

---
title: 'Travel time matrices'
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
abstract: "This vignette shows how to use the travel_time_matrix() and expanded_travel_time_matrix() functions in r5r."
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Travel time matrices} 
  %\VignetteEngine{knitr::rmarkdown} 
  \usepackage[utf8]{inputenc}
bibliography: references.json
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

## removes files previously created by 'build_network()'
#data_path <- system.file("extdata/poa", package = "r5r")
#existing_files <- list.files(data_path)
#files_to_keep <- c(
#  "poa_hexgrid.csv",
#  "poa_osm.pbf",
#  "poa_points_of_interest.csv",
#  "poa_eptc.zip",
#  "poa_trensurb.zip",
#  'fares'
#  )
#files_to_remove <- existing_files[! existing_files %in% files_to_keep]
#invisible(file.remove(file.path(data_path, files_to_remove)))
```

# 1. Introduction 

Some of the most common tasks in transport planning and modeling involve require having good quality data with travel time estimates between origins and destinations. `R5` is incredibly fast in generating realistic door-to-door travel time estimates in multimodal transport systems.

The `r5r` package has two functions that allow users to leverage the computing power of `R5`:
- `travel_time_matrix()`
- `expanded_travel_time_matrix()`
- `arrival_travel_time_matrix()`

This vignette shows a reproducible example to explain how these two functions work and the differences between them.


# 2. Build routable transport network with `build_network()`

First, let's build the multimodal transport network we'll be using in this vignette. In this example we'll be using the a sample data set for the city of Porto Alegre (Brazil) included in `r5r`.


```{r, message = FALSE}
# increase Java memory
options(java.parameters = "-Xmx2G")

# load libraries
library(r5r)
library(data.table)
library(ggplot2)

# build a routable transport network with r5r
data_path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path)

# routing inputs
mode <- c('walk', 'transit')
max_trip_duration <- 60 # minutes

# departure time
departure_datetime <- as.POSIXct("13-05-2019 14:00:00", 
                                 format = "%d-%m-%Y %H:%M:%S")

# load origin/destination points
points <- fread(file.path(data_path, "poa_points_of_interest.csv"))

```


# 3. The `travel_time_matrix()` function

The `travel_time_matrix()` function provides a simple and really fast way to calculate the travel time between all possible origin destination pairs at a given departure time using a given transport mode.

The user can also customize many parameters such as:
- `max_trip_duration`: maximum trip duration
- `max_rides`: maximum number of transfer in the public transport system
- `max_walk_time` and `max_bike_time`: maximum walking or cycling time to and from public transport
- `walk_speed` and `bike_speed`: maximum walking or cycling speed
- `max_fare`: maximum monetary cost in public transport. [See this vignette](https://ipeagit.github.io/r5r/articles/fare_structure.html).


```{r, message = FALSE}
# estimate travel time matrix
ttm <- travel_time_matrix(
  r5r_network,   
  origins = points,
  destinations = points,    
  mode = mode,
  max_trip_duration = max_trip_duration,
  departure_datetime = departure_datetime
  )

head(ttm, n = 10)

```

Now remember that travel time estimates can vary significantly across the day because of variations in public transport service levels. In order to account for this, you might want to calculate multiple travel time matrices departing at different times.

This can be done very efficiently by using the `time_window` and `percentile` parameters in the `travel_time_matrix()` function. When these parameters are set, R<sup>5</sup> will automatically compute multiple travel times estimates considering multiple departures per minute within the `time_window` selected by the user. [More information about this functionality can found in this vignette](https://ipeagit.github.io/r5r/articles/time_window.html).


# 4. The `expanded_travel_time_matrix()` function

Sometimes, we want to know more than simply the total travel time from A to B. This is when the `expanded_travel_time_matrix()` function comes in. By default, the output of this function will also tell which public transport routes were taken between each origin destination pair.

Nonetheless, you may set the parameter `breakdown = TRUE` to gather much more info for each trip. In this case, `expanded_travel_time_matrix()` will tell the number of transfers used to complete each trip and their total access, waiting, in-vehicle and transfer times. Please note that setting `breakdown = TRUE` can make the function slower for large data sets.

*A general call to expanded_travel_time_matrix()*
```{r, message = FALSE}
ettm <- expanded_travel_time_matrix(
  r5r_network,   
  origins = points,
  destinations = points,    
  mode = mode,
  max_trip_duration = max_trip_duration,
  departure_datetime = departure_datetime
  )

head(ettm, n = 10)
```

*Calling expanded_travel_time_matrix() with `breakdown = TRUE`*

```{r, message = FALSE}
ettm2 <- expanded_travel_time_matrix(
  r5r_network,   
  origins = points,
  destinations = points,    
  mode = mode,
  max_trip_duration = max_trip_duration,
  departure_datetime = departure_datetime,
  breakdown = TRUE
  )

head(ettm2, n = 10)
```

You will notice in the documentation that the `expanded_travel_time_matrix()` also has a `time_window` parameter. In this case, though, when the user sets a `time_window` value, the `expanded_travel_time_matrix()` will return the fastest route alternative departing each minute within the specified time window. Please note this function can be very memory intensive for large data sets and time windows.

```{r, message = FALSE}
ettm_window <- expanded_travel_time_matrix(
  r5r_network,   
  origins = points,
  destinations = points,    
  mode = mode,
  max_trip_duration = max_trip_duration,
  departure_datetime = departure_datetime,
  breakdown = TRUE,
  time_window = 10
  )

ettm_window[15:25,]
```



# 4. The `arrival_travel_time_matrix()` function

Both of the functions `travel_time_matrix()` and `expanded_travel_time_matrix()` consider a **departure** time set by the user. In some cases, though, you might need to calculate travel times considering an **arrival time**. For such cases, you can use `arrival_travel_time_matrix()`. In this function, you need to set the latest arrival time desired and a maximum trip duration. The function returns the travel time of the trip with the latest departure time that arrives before the arrival time.

This function is useful when modeling user behavior in situations where arriving by a specific time is important, such as getting to work or school for a set start time (e.g., 9 a.m.). In these scenarios, it is often most convenient for a person to take the latest possible departure that still ensures arrival before their required start time, rather than choosing the trip with the shortest travel time but arriving much earlier and waiting unnecessarily at the destination.

Note that the output of this function includes more information with additional columns, like in the `expanded_travel_time_matrix()` function (you can also use `breakdown = TRUE`).

*A general call to arrival_travel_time_matrix()*
```{r, message = FALSE}

arrival_datetime <- as.POSIXct(
 "13-05-2019 14:00:00",
 format = "%d-%m-%Y %H:%M:%S"
)

arrival_ttm <- arrival_travel_time_matrix(
  r5r_network,
  origins = points,
  destinations = points,
  mode = c("WALK", "TRANSIT"),
  arrival_datetime = arrival_datetime,
  max_trip_duration = 60
)

head(arrival_ttm, n = 10)
```



### Cleaning up after usage

`r5r` objects are still allocated to any amount of memory previously set after they are done with their calculations. In order to remove an existing `r5r` object and reallocate the memory it had been using, we use the `stop_r5` function followed by a call to Java's garbage collector, as follows:

```{r, message = FALSE}
r5r::stop_r5(r5r_network)
rJava::.jgc(R.gc = TRUE)
```

```{r, eval = TRUE, include = FALSE, message = FALSE}
# clean cache (CRAN policy)
r5r::r5r_cache(delete_file = 'all')

```

If you have any suggestions or want to report an error, please visit [the package GitHub page](https://github.com/ipeaGIT/r5r).

## References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r5r_network.R
\docType{class}
\name{r5r_network-class}
\alias{r5r_network-class}
\title{Class to internally handle Java reference to R5RCore}
\description{
Class to internally handle Java reference to R5RCore
}
\seealso{
Other r5r_network: 
\code{\link{wrap_r5r_network}()}
}
\concept{r5r_network}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expand_od_pairs.R
\name{expand_od_pairs}
\alias{expand_od_pairs}
\title{Expand origin-destination pairs}
\usage{
expand_od_pairs(origins, destinations, all_to_all)
}
\arguments{
\item{origins}{Passed by \code{\link[=detailed_itineraries]{detailed_itineraries()}}.}

\item{destinations}{Passed by \code{\link[=detailed_itineraries]{detailed_itineraries()}}.}

\item{all_to_all}{Passed by \code{\link[=detailed_itineraries]{detailed_itineraries()}}.}
}
\description{
This function is used in \code{\link[=detailed_itineraries]{detailed_itineraries()}} to update the \code{origins} and
\code{destinations} datasets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/java_utils.R
\name{dt_to_lts_map}
\alias{dt_to_lts_map}
\title{data.table to ltsMap}
\usage{
dt_to_lts_map(dt)
}
\arguments{
\item{dt}{data.frame/data.table. Table specifying the
LTS levels. The table must contain columns \code{osm_id} and
\code{lts}.}
}
\value{
A speedMap (Java HashMap<Long, Integer>)
}
\description{
Converts a \code{data.frame} with road OSM id's and respective LTS
levels a Java Map<Long, Integer> for use by r5r_network.
}
\seealso{
Other java support functions: 
\code{\link{dt_to_speed_map}()},
\code{\link{get_java_version}()},
\code{\link{java_to_dt}()}
}
\concept{java support functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_r5.R
\name{download_r5}
\alias{download_r5}
\title{Download \code{R5.jar}}
\usage{
download_r5(
  version = NULL,
  quiet = FALSE,
  force_update = FALSE,
  temp_dir = FALSE
)
}
\arguments{
\item{version}{A string. The version of R5 to be downloaded. When \code{NULL}, it
defaults to the latest version.}

\item{quiet}{A logical. Whether to show informative messages when downloading
the file. Defaults to \code{FALSE}.}

\item{force_update}{A logical. Whether to overwrite a previously downloaded
\code{R5.jar} in the local directory. Defaults to \code{FALSE}.}

\item{temp_dir}{A logical. Whether the file should be saved in a temporary
directory. Defaults to \code{FALSE}.}
}
\value{
The path to the downloaded file.
}
\description{
Downloads \code{R5.jar} and saves it locally, inside the package directory.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)

download_r5(temp_dir = TRUE)
\dontshow{\}) # examplesIf}
}
\seealso{
Other Build network: 
\code{\link{build_network}()},
\code{\link{setup_r5}()}
}
\concept{Build network}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{exists_tiff}
\alias{exists_tiff}
\title{Check if there is an elevation data file in \code{.tif} format in a given data path}
\usage{
exists_tiff(data_path)
}
\arguments{
\item{data_path}{A string pointing to a directory.}
}
\value{
\code{TRUE} if there is a \code{.tif} file, \code{FALSE} otherwise
}
\description{
Check if there is an elevation data file in \code{.tif} format in a given data path
}
\seealso{
Other support functions: 
\code{\link{fileurl_from_metadata}()},
\code{\link{get_gtfs_errors}()},
\code{\link{start_r5r_java}()},
\code{\link{stop_r5}()},
\code{\link{tempdir_unique}()},
\code{\link{travel_time_surface}},
\code{\link{validate_bad_osm_ids}()}
}
\concept{support functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transit_network_to_sf.R
\name{transit_network_to_sf}
\alias{transit_network_to_sf}
\title{Extract transit network in sf format}
\usage{
transit_network_to_sf(r5r_network, r5r_core = deprecated())
}
\arguments{
\item{r5r_network}{A routable transport network created with \code{\link[=build_network]{build_network()}}.}

\item{r5r_core}{The \code{r5r_core} argument is deprecated as of r5r v2.3.0.
Please use the \code{r5r_network} argument instead.}
}
\value{
A list with two components of a transit network in \code{sf} format:
route shapes (\code{LINESTRING}) and transit stops (\code{POINT}). The same
\code{route_id}/\code{short_name} might appear with different geometries. This occurs
when the same route is associated to more than one \code{shape_id}s in the GTFS
feed used to create the transit network. Some transit stops might be
returned with geometry \verb{POINT EMPTY} (i.e. missing spatial coordinates).
This may occur when a transit stop is not snapped to the road network,
possibly because the GTFS feed used to create the transit network covers an
area larger than the \code{.osm.pbf} input data.
}
\description{
Extracts the transit network in \code{sf} format from a routable transport network
built with \code{\link[=build_network]{build_network()}}).
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)

# build transport network
path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(path)

# extract transit network from r5r_network
transit_net <- transit_network_to_sf(r5r_network)

stop_r5(r5r_network)
\dontshow{\}) # examplesIf}
}
\seealso{
Other network functions: 
\code{\link{find_snap}()},
\code{\link{street_network_bbox}()},
\code{\link{street_network_to_sf}()}
}
\concept{network functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pareto_frontier.R
\name{pareto_frontier}
\alias{pareto_frontier}
\title{Calculate travel time and monetary cost Pareto frontier}
\usage{
pareto_frontier(
  r5r_network,
  r5r_core = deprecated(),
  origins,
  destinations,
  mode = c("WALK", "TRANSIT"),
  mode_egress = "WALK",
  departure_datetime = Sys.time(),
  time_window = 10L,
  percentiles = 50L,
  max_walk_time = Inf,
  max_bike_time = Inf,
  max_car_time = Inf,
  max_trip_duration = 120L,
  fare_structure = NULL,
  fare_cutoffs = -1L,
  walk_speed = 3.6,
  bike_speed = 12,
  max_rides = 3,
  max_lts = 2,
  n_threads = Inf,
  verbose = FALSE,
  progress = FALSE,
  output_dir = NULL
)
}
\arguments{
\item{r5r_network}{A routable transport network created with \code{\link[=build_network]{build_network()}}.}

\item{r5r_core}{The \code{r5r_core} argument is deprecated as of r5r v2.3.0.
Please use the \code{r5r_network} argument instead.}

\item{origins, destinations}{Either a \verb{POINT sf} object with WGS84 CRS, or a
\code{data.frame} containing the columns \code{id}, \code{lon} and \code{lat}.}

\item{mode}{A character vector. The transport modes allowed for access,
transfer and vehicle legs of the trips. Defaults to \code{WALK}. Please see
details for other options.}

\item{mode_egress}{A character vector. The transport mode used after egress
from the last public transport. It can be either \code{WALK}, \code{BICYCLE} or
\code{CAR}. Defaults to \code{WALK}. Ignored when public transport is not used.}

\item{departure_datetime}{A POSIXct object. Please note that the departure
time only influences public transport legs. When working with public
transport networks, please check the \code{calendar.txt} within your GTFS feeds
for valid dates. Please see details for further information on how
datetimes are parsed.}

\item{time_window}{An integer. The time window in minutes for which \code{r5r}
will calculate multiple travel time matrices departing each minute.
Defaults to 10 minutes. By default, the function returns the result based
on median travel times, but the user can set the \code{percentiles} parameter to
extract more results. Please read the time window vignette for more details
on its usage \code{vignette("time_window", package = "r5r")}}

\item{percentiles}{An integer vector (max length of 5). Specifies the
percentile to use when returning travel time estimates within the given
time window. Please note that this parameter is applied to the travel time
estimates only (e.g. if the 25th percentile is specified, and the output
between A and B is 15 minutes and 10 dollars, 25\% of all trips cheaper
than 10 dollars taken between these points are shorter than 15 minutes).
Defaults to 50, returning the median travel time. If a vector with length
bigger than 1 is passed, the output contains an additional column that
specifies the percentile of each travel time and monetary cost
combination. Due to upstream restrictions, only 5 percentiles can be
specified at a time. For more details, please see R5 documentation at
\url{https://docs.conveyal.com/analysis/methodology#accounting-for-variability}.}

\item{max_walk_time}{An integer. The maximum walking time (in minutes) to
access and egress the transit network, to make transfers within the network
or to complete walk-only trips. Defaults to no restrictions (numeric value
of \code{Inf}), as long as \code{max_trip_duration} is respected. When routing
transit trips, the max time is considered separately for each leg (e.g. if
you set \code{max_walk_time} to 15, you could get trips with an up to 15 minutes
walk leg to reach transit and another up to 15 minutes walk leg to reach
the destination after leaving transit. In walk-only trips, whenever
\code{max_walk_time} differs from \code{max_trip_duration}, the lowest value is
considered.}

\item{max_bike_time}{An integer. The maximum cycling time (in minutes) to
access and egress the transit network, to make transfers within the network
or to complete bicycle-only trips. Defaults to no restrictions (numeric
value of \code{Inf}), as long as \code{max_trip_duration} is respected. When routing
transit trips, the max time is considered separately for each leg (e.g. if
you set \code{max_bike_time} to 15, you could get trips with an up to 15 minutes
cycle leg to reach transit and another up to 15 minutes cycle leg to reach
the destination after leaving transit. In bicycle-only trips, whenever
\code{max_bike_time} differs from \code{max_trip_duration}, the lowest value is
considered.}

\item{max_car_time}{An integer. The maximum driving time (in minutes) to
access and egress the transit network. Defaults to no restrictions, as long
as \code{max_trip_duration} is respected. The max time is considered separately
for each leg (e.g. if you set \code{max_car_time} to 15 minutes, you could
potentially drive up to 15 minutes to reach transit, and up to \emph{another} 15
minutes to reach the destination after leaving transit). Defaults to \code{Inf},
no limit.}

\item{max_trip_duration}{An integer. The maximum trip duration in minutes.
Defaults to 120 minutes (2 hours).}

\item{fare_structure}{A fare structure object, following the convention
set in \code{\link[=setup_fare_structure]{setup_fare_structure()}}. This object describes how transit fares
should be calculated. Please see the fare structure vignette to
understand how this object is structured:
\code{vignette("fare_structure", package = "r5r")}.}

\item{fare_cutoffs}{A numeric vector. The monetary cutoffs that
should be considered when calculating the Pareto frontier. Most of the
time you'll want this parameter to be the combination of all possible
fares listed in you \code{fare_structure}. Choosing a coarse distribution of
cutoffs may result in many different trips falling within the same cutoff.
For example, if you have two different routes in your GTFS, one costing $3
and the other costing $4, and you set this parameter to \code{5}, the output
will tell you the fastest trips that costed up to $5, but you won't be
able to identify which route was used to complete such trips. In this
case, it would be more beneficial to set the parameter as \code{c(3, 4)} (you
could also specify combinations of such values, such as 6, 7, 8 and so on,
because a transit user could hypothetically benefit from making transfers
between the available routes).}

\item{walk_speed}{A numeric. Average walk speed in km/h. Defaults to 3.6
km/h.}

\item{bike_speed}{A numeric. Average cycling speed in km/h. Defaults to 12
km/h.}

\item{max_rides}{An integer. The maximum number of public transport rides
allowed in the same trip. Defaults to 3.}

\item{max_lts}{An integer between 1 and 4. The maximum level of traffic
stress that cyclists will tolerate. A value of 1 means cyclists will only
travel through the quietest streets, while a value of 4 indicates cyclists
can travel through any road. Defaults to 2. Please see details for more
information.}

\item{n_threads}{An integer. The number of threads to use when running the
router in parallel. Defaults to use all available threads (Inf).}

\item{verbose}{A logical. Whether to show \code{R5} informative messages when
running the function. Defaults to \code{FALSE} (please note that in such case
\code{R5} error messages are still shown). Setting \code{verbose} to \code{TRUE} shows
detailed output, which can be useful for debugging issues not caught by
\code{r5r}.}

\item{progress}{A logical. Whether to show a progress counter when running
the router. Defaults to \code{FALSE}. Only works when \code{verbose} is set to
\code{FALSE}, so the progress counter does not interfere with \code{R5}'s output
messages. Setting \code{progress} to \code{TRUE} may impose a small penalty for
computation efficiency, because the progress counter must be synchronized
among all active threads.}

\item{output_dir}{Either \code{NULL} or a path to an existing directory. When not
\code{NULL} (the default), the function will write one \code{.csv} file with the
results for each origin in the specified directory. In such case, the
function returns the path specified in this parameter. This parameter is
particularly useful when running on memory-constrained settings because
writing the results directly to disk prevents \code{r5r} from loading them to
RAM memory.}
}
\value{
A \code{data.table} with the travel time and monetary cost Pareto frontier
between the specified origins and destinations. An additional column
identifying the travel time percentile is present if more than one value
was passed to \code{percentiles}. Origin and destination pairs whose trips
couldn't be completed within the maximum travel time using less money than
the specified monetary cutoffs are not returned in the \code{data.table}. If
\code{output_dir} is not \code{NULL}, the function returns the path specified in
that parameter, in which the \code{.csv} files containing the results are
saved.
}
\description{
Fast computation of travel time and monetary cost Pareto frontier between
origin and destination pairs.
}
\section{Transport modes}{


\code{R5} allows for multiple combinations of transport modes. The options
include:
\itemize{
\item \strong{Transit modes:} \code{TRAM}, \code{SUBWAY}, \code{RAIL}, \code{BUS}, \code{FERRY}, \code{CABLE_CAR},
\code{GONDOLA}, \code{FUNICULAR}. The option \code{TRANSIT} automatically considers all
public transport modes available.
\item \strong{Non transit modes:} \code{WALK}, \code{BICYCLE}, \code{CAR}, \code{BICYCLE_RENT},
\code{CAR_PARK}.
}
}

\section{Level of Traffic Stress (LTS)}{


When cycling is enabled in \code{R5} (by passing the value \code{BIKE} to either
\code{mode} or \code{mode_egress}), setting \code{max_lts} will allow cycling only on
streets with a given level of danger/stress. Setting \code{max_lts} to 1, for
example, will allow cycling only on separated bicycle infrastructure or
low-traffic streets and routing will revert to walking when traversing any
links with LTS exceeding 1. Setting \code{max_lts} to 3 will allow cycling on
links with LTS 1, 2 or 3. Routing also reverts to walking if the street
segment is tagged as non-bikable in OSM (e.g. a staircase), independently of
the specified max LTS.

The default methodology for assigning LTS values to network edges is based
on commonly tagged attributes of OSM ways. See more info about LTS in the
original documentation of R5 from Conveyal at
\url{https://docs.conveyal.com/learn-more/traffic-stress}. In summary:
\itemize{
\item \strong{LTS 1}: Tolerable for children. This includes low-speed, low-volume
streets, as well as those with separated bicycle facilities (such as
parking-protected lanes or cycle tracks).
\item \strong{LTS 2}: Tolerable for the mainstream adult population. This includes
streets where cyclists have dedicated lanes and only have to interact with
traffic at formal crossing.
\item \strong{LTS 3}: Tolerable for "enthused and confident" cyclists. This includes
streets which may involve close proximity to moderate- or high-speed
vehicular traffic.
\item \strong{LTS 4}: Tolerable only for "strong and fearless" cyclists. This
includes streets where cyclists are required to mix with moderate- to
high-speed vehicular traffic.
}

For advanced users, you can provide custom LTS values by adding a tag \verb{<key = "lts">} to the \code{osm.pbf} file.
}

\section{Datetime parsing}{


\code{r5r} ignores the timezone attribute of datetime objects when parsing dates
and times, using the study area's timezone instead. For example, let's say
you are running some calculations using Rio de Janeiro, Brazil, as your study
area. The datetime \code{as.POSIXct("13-05-2019 14:00:00", format = "\%d-\%m-\%Y \%H:\%M:\%S")} will be parsed as May 13th, 2019, 14:00h in
Rio's local time, as expected. But \code{as.POSIXct("13-05-2019 14:00:00", format = "\%d-\%m-\%Y \%H:\%M:\%S", tz = "Europe/Paris")} will also be parsed as
the exact same date and time in Rio's local time, perhaps surprisingly,
ignoring the timezone attribute.
}

\section{Routing algorithm}{


The \code{\link[=detailed_itineraries]{detailed_itineraries()}} and \code{\link[=pareto_frontier]{pareto_frontier()}} functions use an
\code{R5}-specific extension to the McRAPTOR routing algorithm. The
implementation used in \code{detailed_itineraries()} allows the router to find
paths that are optimal and less than optimal in terms of travel time, with
some heuristics around multiple access modes, riding the same patterns, etc.
The specific extension to McRAPTOR to do suboptimal path routing is not
documented yet, but a detailed description of base McRAPTOR can be found in
Delling et al (2015). The implementation used in \code{pareto_frontier()}, on the
other hand, returns only the fastest trip within a given monetary cutoff,
ignoring slower trips that cost the same. A detailed discussion on the
algorithm can be found in Conway and Stewart (2019).
\itemize{
\item Delling, D., Pajor, T., & Werneck, R. F. (2015). Round-based public
transit routing. Transportation Science, 49(3), 591-604.
\doi{10.1287/trsc.2014.0534}
\item Conway, M. W., & Stewart, A. F. (2019). Getting Charlie off the MTA: a
multiobjective optimization method to account for cost constraints in public
transit accessibility metrics. International Journal of Geographical
Information Science, 33(9), 1759-1787. \doi{10.1080/13658816.2019.1605075}
}
}

\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path = data_path)

# load origin/destination points
points <- read.csv(file.path(data_path, "poa_hexgrid.csv"))[1:5,]

# load fare structure object
fare_structure_path <- system.file(
  "extdata/poa/fares/fares_poa.zip",
  package = "r5r"
)
fare_structure <- read_fare_structure(fare_structure_path)

departure_datetime <- as.POSIXct(
  "13-05-2019 14:00:00",
  format = "\%d-\%m-\%Y \%H:\%M:\%S"
)

pf <- pareto_frontier(
  r5r_network,
  origins = points,
  destinations = points,
  mode = c("WALK", "TRANSIT"),
  departure_datetime = departure_datetime,
  fare_structure = fare_structure,
  fare_cutoffs = c(4.5, 4.8, 9, 9.3, 9.6)
)
head(pf)

stop_r5(r5r_network)
\dontshow{\}) # examplesIf}
}
\seealso{
Other routing: 
\code{\link{arrival_travel_time_matrix}()},
\code{\link{detailed_itineraries}()},
\code{\link{expanded_travel_time_matrix}()},
\code{\link{travel_time_matrix}()}
}
\concept{routing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_network.R
\name{build_network}
\alias{build_network}
\title{Build a transport network used for routing in R5}
\usage{
build_network(
  data_path,
  verbose = FALSE,
  temp_dir = FALSE,
  elevation = "TOBLER",
  overwrite = FALSE
)
}
\arguments{
\item{data_path}{A string pointing to the directory where data inputs are
stored and where the built \code{network.dat} will be saved.}

\item{verbose}{A logical. Whether to show \code{R5} informative messages when
running the function. Defaults to \code{FALSE} (please note that in such case
\code{R5} error messages are still shown). Setting \code{verbose} to \code{TRUE} shows
detailed output, which can be useful for debugging issues not caught by
\code{r5r}.}

\item{temp_dir}{A logical. Whether the \code{network.dat} file should be saved to
a temporary directory. Defaults to \code{FALSE}.}

\item{elevation}{A string. The name of the impedance function to be used to
calculate impedance for walking and cycling based on street slopes.
Available options include \code{TOBLER} (Default) and \code{MINETTI}, or \code{NONE}
to ignore elevation. R5 loads elevation data from \code{.tif} files saved
inside the \code{data_path} directory. Elevation raster must be in WGS 84
(EPSG:4326) coordinate reference system. See more info in the Details
section below.}

\item{overwrite}{A logical. Whether to overwrite an existing \code{network.dat}
or to use a cached file. Defaults to \code{FALSE} (i.e. use a cached
network).}
}
\value{
A \code{r5r_network} object representing the built network to connect with
\code{R5} routing engine.
}
\description{
Builds a multimodal transport network used for routing in \code{R5}, combining
multiple data inputs present in the directory where the network should be
saved to. The directory must contain only one street network file (in
\code{.osm.pbf} format). It may optionally contain one or more public transport
GTFS feeds (in \code{.zip} format), when used for public transport routing, and a
\code{.tif} file describing the elevation profile of the study area. If there is
more than one GTFS feed in the directory, all feeds are automatically merged.
If there is already a \code{'network.dat'} file in the directory, the function will
simply read it and load it to memory (unless specified not to do so).
}
\section{Elevation}{


More information about the \code{TOBLER} and \code{MINETTI} options to calculate the
effects of elevation on travel times can be found in the references below:
\itemize{
\item Campbell, M. J., et al (2019). Using crowdsourced fitness tracker data to
model the relationship between slope and travel rates. Applied geography, 106,
93-107. \doi{10.1016/j.apgeog.2019.03.008}.
\item Minetti, A. E., et al (2002). Energy cost of walking and running at extreme
uphill and downhill slopes. Journal of applied physiology. \doi{10.1152/japplphysiol.01177.2001}.
\item Tobler, W. (1993). Three presentations on geographical analysis and modeling:
Non-isotropic geographic modeling speculations on the geometry of geography
global spatial analysis. Technical Report. National center for geographic
information and analysis. 93 (1). \url{https://escholarship.org/uc/item/05r820mz}.
}
}

\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)

# directory with street network and gtfs files
data_path <- system.file("extdata/poa", package = "r5r")

r5r_network <- build_network(data_path)
\dontshow{\}) # examplesIf}
}
\seealso{
Other Build network: 
\code{\link{download_r5}()},
\code{\link{setup_r5}()}
}
\concept{Build network}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assign.R
\name{assign_max_trip_duration}
\alias{assign_max_trip_duration}
\title{Assign max trip duration}
\usage{
assign_max_trip_duration(
  max_trip_duration,
  modes,
  max_walk_time,
  max_bike_time
)
}
\arguments{
\item{max_trip_duration}{A number.}
}
\value{
An \code{integer} representing the maximum trip duration in minutes.
}
\description{
Check and convert the max trip duration input.
}
\seealso{
Other assigning functions: 
\code{\link{assign_decay_function}()},
\code{\link{assign_departure}()},
\code{\link{assign_drop_geometry}()},
\code{\link{assign_max_street_time}()},
\code{\link{assign_mode}()},
\code{\link{assign_opportunities}()},
\code{\link{assign_osm_link_ids}()},
\code{\link{assign_points_input}()},
\code{\link{assign_shortest_path}()}
}
\concept{assigning functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fare_structure.R
\name{read_fare_structure}
\alias{read_fare_structure}
\title{Read a fare structure object from a file}
\usage{
read_fare_structure(file_path, encoding = "UTF-8")
}
\arguments{
\item{file_path}{A path pointing to a fare structure with a \code{.zip}
extension.}

\item{encoding}{A string. Passed to \code{\link[data.table:fread]{data.table::fread()}}, defaults to
\code{"UTF-8"}. Other possible options are \code{"unknown"} and \code{"Latin-1"}. Please
note that this is not used to re-encode the input, but to enable handling
encoded strings in their native encoding.}
}
\value{
A fare structure object.
}
\description{
Read a fare structure object from a file
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
path <- system.file("extdata/poa/fares/fares_poa.zip", package = "r5r")
fare_structure <- read_fare_structure(path)
\dontshow{\}) # examplesIf}
}
\seealso{
Other fare structure: 
\code{\link{setup_fare_structure}()},
\code{\link{write_fare_structure}()}
}
\concept{fare structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_gtfs_errors.R
\name{get_gtfs_errors}
\alias{get_gtfs_errors}
\title{Get GTFS eventual errors encountered in network building}
\usage{
get_gtfs_errors(r5r_network)
}
\arguments{
\item{r5r_network}{the R5R network object, or a path to the location where
the network is stored (useful if network build failed).}
}
\value{
A \code{data.frame}
}
\description{
This returns a data frame of GTFS errors R5 encountered when building the
network. You can call this with the network itself as the main parameter. If
network build fails, you won't have a network object, so you can also call
this with the \code{data_path} to where the network is stored.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)

# directory with street network and gtfs files
data_path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path)

get_gtfs_errors(r5r_network)
\dontshow{\}) # examplesIf}
}
\seealso{
Other support functions: 
\code{\link{exists_tiff}()},
\code{\link{fileurl_from_metadata}()},
\code{\link{start_r5r_java}()},
\code{\link{stop_r5}()},
\code{\link{tempdir_unique}()},
\code{\link{travel_time_surface}},
\code{\link{validate_bad_osm_ids}()}
}
\concept{support functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assign.R
\name{assign_points_input}
\alias{assign_points_input}
\title{Check and convert origin and destination inputs}
\usage{
assign_points_input(df, name)
}
\arguments{
\item{df}{Either a \code{data.frame} or a \verb{POINT sf}.}

\item{name}{Object name.}
}
\value{
A \code{data.frame} with columns \code{id}, \code{lon} and \code{lat}.
}
\description{
Check and convert origin and destination inputs
}
\seealso{
Other assigning functions: 
\code{\link{assign_decay_function}()},
\code{\link{assign_departure}()},
\code{\link{assign_drop_geometry}()},
\code{\link{assign_max_street_time}()},
\code{\link{assign_max_trip_duration}()},
\code{\link{assign_mode}()},
\code{\link{assign_opportunities}()},
\code{\link{assign_osm_link_ids}()},
\code{\link{assign_shortest_path}()}
}
\concept{assigning functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r5r_cache.R
\name{r5r_cache}
\alias{r5r_cache}
\title{Manage cached files from the r5r package}
\usage{
r5r_cache(list_files = TRUE, delete_file = NULL)
}
\arguments{
\item{list_files}{Logical. Whether to print a message with the address of r5r
JAR files cached locally. Defaults to \code{TRUE}.}

\item{delete_file}{String. The file name (basename) of a JAR file cached
locally that should be deleted. Defaults to \code{NULL}, so that no
file is deleted. If \code{delete_file = "all"}, then all cached files are
deleted.}
}
\value{
A message indicating which file exist and/or which ones have been
deleted from local cache directory.
}
\description{
Manage cached files from the r5r package
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
# download r5 JAR
r5r::download_r5()

# list all files cached
r5r_cache(list_files = TRUE)

# delete r5 JAR
r5r_cache(delete_file = 'r5-v7.0')
\dontshow{\}) # examplesIf}
}
\concept{Cache data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r5r-package.R
\docType{package}
\name{r5r}
\alias{r5r}
\alias{r5r-package}
\title{r5r: Rapid Realistic Routing with 'R5'}
\description{
Rapid realistic routing on multimodal transport networks (walk, bike, public
transport and car) using \code{R5}, the Rapid Realistic Routing on Real-world and
Reimagined networks engine \url{https://github.com/conveyal/r5}. The package
allows users to generate detailed routing analysis or calculate travel time
matrices using seamless parallel computing on top of the R5 Java machine.
While \code{R5} is developed by Conveyal, the package \code{r5r} is independently
developed by a team at the Institute for Applied Economic Research (Ipea)
with contributions from collaborators. Apart from the documentation in this
package, users will find additional information on \code{R5} documentation at
\url{https://docs.conveyal.com/}. Although we try to keep new releases of \code{r5r}
in synchrony with \code{R5}, the development of \code{R5} follows Conveyal's
independent update process. Hence, users should confirm if the \code{R5} version
implied by the Conveyal user manual (see
\url{https://docs.conveyal.com/changelog}) corresponds with the \code{R5} version
that \code{r5r} depends on.
}
\section{Usage}{

Please check the vignettes on the \href{https://ipeagit.github.io/r5r/}{website}.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/ipeaGIT/r5r}
  \item \url{https://ipeagit.github.io/r5r/}
  \item Report bugs at \url{https://github.com/ipeaGIT/r5r/issues}
}

}
\author{
\strong{Maintainer}: Rafael H. M. Pereira \email{rafa.pereira.br@gmail.com} (\href{https://orcid.org/0000-0003-2125-7465}{ORCID})

Authors:
\itemize{
  \item Marcus Saraiva \email{marcus.saraiva@gmail.com} (\href{https://orcid.org/0000-0001-6218-2338}{ORCID})
  \item Daniel Herszenhut \email{dhersz@gmail.com} (\href{https://orcid.org/0000-0001-8066-1105}{ORCID})
  \item Alex Magnus \email{alex.magnus@mail.utoronto.ca}
  \item Matthew Wigginton Bhagat-Conway \email{mwbc@unc.edu} (\href{https://orcid.org/0000-0002-1210-2982}{ORCID})
}

Other contributors:
\itemize{
  \item Carlos Kaue Vieira Braga \email{kaue@kauebraga.dev} (\href{https://orcid.org/0000-0002-6104-7297}{ORCID}) [contributor]
  \item Luyu Liu \email{luyuliu@ufl.edu} (\href{https://orcid.org/0000-0002-6684-5570}{ORCID}) [contributor]
  \item Egor Kotov \email{kotov.egor@gmail.com} (\href{https://orcid.org/0000-0001-6690-5345}{ORCID}) [contributor]
  \item Daniel Snow \email{dan@sno.ws} [contributor]
  \item Ipea - Institute for Applied Economic Research [copyright holder, funder]
  \item Department of Geography & Planning, University of Toronto [funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set.R
\name{reverse_back_if_direct_mode}
\alias{reverse_back_if_direct_mode}
\title{Reverse BACK Origins and Destinations for Direct Modes}
\usage{
reverse_back_if_direct_mode(
  travel_times,
  origins,
  destinations,
  mode_list,
  data_path
)
}
\arguments{
\item{travel_times}{A travel time matrix.}

\item{origins}{A data frame representing origin locations.}

\item{destinations}{A data frame representing destination locations.}

\item{mode_list}{A named list containing the routing modes.}

\item{data_path}{The data path used to build the network}
}
\value{
A df travel time matrix with origins and destinations unchanged or
in swapped order
}
\description{
Swaps BACK the \code{origins} and \code{destinations} data frames if they were
originally reversed by \code{reverse_if_direct_mode()}.
}
\seealso{
Other setting functions: 
\code{\link{reverse_if_direct_mode}()},
\code{\link{set_breakdown}()},
\code{\link{set_cutoffs}()},
\code{\link{set_elevation}()},
\code{\link{set_expanded_travel_times}()},
\code{\link{set_fare_cutoffs}()},
\code{\link{set_fare_structure}()},
\code{\link{set_max_fare}()},
\code{\link{set_max_lts}()},
\code{\link{set_max_rides}()},
\code{\link{set_monte_carlo_draws}()},
\code{\link{set_n_threads}()},
\code{\link{set_new_congestion}()},
\code{\link{set_new_lts}()},
\code{\link{set_output_dir}()},
\code{\link{set_percentiles}()},
\code{\link{set_progress}()},
\code{\link{set_speed}()},
\code{\link{set_suboptimal_minutes}()},
\code{\link{set_time_window}()},
\code{\link{set_verbose}()}
}
\concept{setting functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{start_r5r_java}
\alias{start_r5r_java}
\title{Initialize Java and Check Version}
\usage{
start_r5r_java(data_path, temp_dir = FALSE, verbose = FALSE)
}
\arguments{
\item{data_path}{A character string. The directory where the log file should be saved.}

\item{temp_dir}{A logical. Whether the jar file should be saved in a temporary
directory. Defaults to \code{FALSE}.}

\item{verbose}{A logical. Whether to show informative messages. Defaults to \code{FALSE}.}
}
\value{
No return value. The function will stop execution with an error if Java 21 is not found.
}
\description{
Sets up Java logging for r5r and ensures Java SE Development Kit 21 is installed.
}
\details{
This function initializes the Java Virtual Machine (JVM) with a log path for r5r, and checks that
the installed Java version is 21. If not, it stops with an informative error message and download links.
}
\seealso{
Other support functions: 
\code{\link{exists_tiff}()},
\code{\link{fileurl_from_metadata}()},
\code{\link{get_gtfs_errors}()},
\code{\link{stop_r5}()},
\code{\link{tempdir_unique}()},
\code{\link{travel_time_surface}},
\code{\link{validate_bad_osm_ids}()}
}
\concept{support functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup_r5.R
\name{setup_r5}
\alias{setup_r5}
\title{Create a transport network used for routing in R5 (deprecated)}
\usage{
setup_r5(
  data_path,
  verbose = FALSE,
  temp_dir = FALSE,
  elevation = "TOBLER",
  overwrite = FALSE
)
}
\arguments{
\item{data_path}{A string pointing to the directory where data inputs are
stored and where the built \code{network.dat} will be saved.}

\item{verbose}{A logical. Whether to show \code{R5} informative messages when
running the function. Defaults to \code{FALSE} (please note that in such case
\code{R5} error messages are still shown). Setting \code{verbose} to \code{TRUE} shows
detailed output, which can be useful for debugging issues not caught by
\code{r5r}.}

\item{temp_dir}{A logical. Whether the \code{network.dat} file should be saved to
a temporary directory. Defaults to \code{FALSE}.}

\item{elevation}{A string. The name of the impedance function to be used to
calculate impedance for walking and cycling based on street slopes.
Available options include \code{TOBLER} (Default) and \code{MINETTI}, or \code{NONE}
to ignore elevation. R5 loads elevation data from \code{.tif} files saved
inside the \code{data_path} directory. Elevation raster must be in WGS 84
(EPSG:4326) coordinate reference system. See more info in the Details
section below.}

\item{overwrite}{A logical. Whether to overwrite an existing \code{network.dat}
or to use a cached file. Defaults to \code{FALSE} (i.e. use a cached
network).}
}
\value{
A \code{r5r_network} object representing the built network to connect with
\code{R5} routing engine.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{setup_r5()} was renamed to \code{\link[=build_network]{build_network()}} to create a more consistent
API. \strong{\code{setup_r5()} is being deprecated} after \emph{r5r} v2.3.0 and will be
\strong{removed in a future release}. Please switch to \code{\link[=build_network]{build_network()}}.
}
\section{Elevation}{


More information about the \code{TOBLER} and \code{MINETTI} options to calculate the
effects of elevation on travel times can be found in the references below:
\itemize{
\item Campbell, M. J., et al (2019). Using crowdsourced fitness tracker data to
model the relationship between slope and travel rates. Applied geography, 106,
93-107. \doi{10.1016/j.apgeog.2019.03.008}.
\item Minetti, A. E., et al (2002). Energy cost of walking and running at extreme
uphill and downhill slopes. Journal of applied physiology. \doi{10.1152/japplphysiol.01177.2001}.
\item Tobler, W. (1993). Three presentations on geographical analysis and modeling:
Non-isotropic geographic modeling speculations on the geometry of geography
global spatial analysis. Technical Report. National center for geographic
information and analysis. 93 (1). \url{https://escholarship.org/uc/item/05r820mz}.
}
}

\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)

# directory with street network and gtfs files
data_path <- system.file("extdata/poa", package = "r5r")

# `setup_r5()` has been deprecated, please switch to `build_network()`
r5r_network <- build_network(data_path)
\dontshow{\}) # examplesIf}
}
\seealso{
Other Build network: 
\code{\link{build_network}()},
\code{\link{download_r5}()}
}
\concept{Build network}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fare_structure.R
\name{write_fare_structure}
\alias{write_fare_structure}
\title{Write a fare structure object to disk}
\usage{
write_fare_structure(fare_structure, file_path)
}
\arguments{
\item{fare_structure}{A fare structure object, following the convention
set in \code{\link[=setup_fare_structure]{setup_fare_structure()}}. This object describes how transit fares
should be calculated. Please see the fare structure vignette to
understand how this object is structured:
\code{vignette("fare_structure", package = "r5r")}.}

\item{file_path}{A path to a \code{.zip} file. Where the fare structure should be
written to.}
}
\value{
The path passed to \code{file_path}, invisibly.
}
\description{
Writes a fare structure object do disk. Fare structure is saved as a
collection of \code{.csv} files inside a \code{.zip} file.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)

data_path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path)

fare_structure <- setup_fare_structure(r5r_network, base_fare = 5)

tmpfile <- tempfile("sample_fare_structure", fileext = ".zip")
write_fare_structure(fare_structure, tmpfile)
\dontshow{\}) # examplesIf}
}
\seealso{
Other fare structure: 
\code{\link{read_fare_structure}()},
\code{\link{setup_fare_structure}()}
}
\concept{fare structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fare_structure.R
\name{setup_fare_structure}
\alias{setup_fare_structure}
\title{Setup a fare structure to calculate the monetary costs of trips}
\usage{
setup_fare_structure(
  r5r_network,
  r5r_core = deprecated(),
  base_fare,
  by = "MODE",
  debug_path = NULL,
  debug_info = NULL
)
}
\arguments{
\item{r5r_network}{A routable transport network created with \code{\link[=build_network]{build_network()}}.}

\item{r5r_core}{The \code{r5r_core} argument is deprecated as of r5r v2.3.0.
Please use the \code{r5r_network} argument instead.}

\item{base_fare}{A numeric. A base value used to populate the fare
structure.}

\item{by}{A string. Describes how \code{fare_type}s (a classification we created
to assign fares to different routes) are distributed among routes.
Possible values are \code{MODE}, \code{AGENCY} and \code{GENERIC}. \code{MODE} is used when
the mode is what determines the price of a route (e.g. if all the buses of
a given city cost $5). \code{AGENCY} is used when the agency that operates each
route is what determines its price (i.e. when two different routes/modes
operated by a single agency cost the same; note that you can also use
\code{AGENCY_NAME}, if the agency_ids listed in your GTFS cannot be easily
interpreted). \code{GENERIC} is used when all the routes cost the same. Please
note that this classification can later be edited to better suit your
needs (when, for example, two types of buses cost the same, but one offers
discounts after riding the subway and the other one doesn't), but this
parameter may save you some work.}

\item{debug_path}{Either a path to a \code{.csv} file or \code{NULL}. When \code{NULL} (the
default), fare debugging capabilities are disabled - i.e. there's no way
to check if the fare calculation is correct. When a path is provided,
\code{r5r} saves different itineraries and their respective fares to the
specified file. How each itinerary is described is controlled by
\code{debug_info}.}

\item{debug_info}{Either a string (when \code{debug_path} is a path) or \code{NULL}
(the default). Doesn't have any effect if \code{debug_path} is \code{NULL}. When a
string, accepts the values \code{MODE}, \code{ROUTE} and \code{MODE_ROUTE}. These values
dictates how itinerary information is written to the output. Let's suppose
we have an itinerary composed by two transit legs: first a subway leg
whose route_id is 001, and then a bus legs whose route_id is 007. If
\code{debug_info} is \code{MODE}, then this itinerary will be described as
\code{SUBWAY|BUS}. If \code{ROUTE}, as \code{001|007}. If \code{MODE_ROUTE}, as \verb{SUBWAY 001|BUS 007}. Please note that the final debug information will contain
not only the itineraries that were in fact used in the itineraries
returned in \code{\link[=travel_time_matrix]{travel_time_matrix()}}, \code{\link[=accessibility]{accessibility()}} and
\code{\link[=pareto_frontier]{pareto_frontier()}}, but all the itineraries that \code{R5} checked when
calculating the routes. This imposes a performance penalty when tracking
debug information (but has the positive effect of returning a larger
sample of itineraries, which might help finding some implementation issues
on the fare structure).}
}
\value{
A fare structure object.
}
\description{
Creates a basic fare structure that describes how transit fares should be
calculated in \code{\link[=travel_time_matrix]{travel_time_matrix()}}, \code{\link[=expanded_travel_time_matrix]{expanded_travel_time_matrix()}},
\code{\link[=accessibility]{accessibility()}} and \code{\link[=pareto_frontier]{pareto_frontier()}}. This fare structure can be
manually edited and adjusted to the existing rules in your study area, as
long as they stick to some basic premises. Please see the \href{../doc/fare_structure.html}{fare-structure vignette} for more information.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)

data_path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path)

fare_structure <- setup_fare_structure(r5r_network, base_fare = 5)

# to debug fare calculation
fare_structure <- setup_fare_structure(
  r5r_network,
  base_fare = 5,
  debug_path = "fare_debug.csv",
  debug_info = "MODE"
)

fare_structure$debug_settings

# debugging can be manually turned off by setting output_file to ""
fare_structure$debug_settings <- ""
\dontshow{\}) # examplesIf}
}
\seealso{
Other fare structure: 
\code{\link{read_fare_structure}()},
\code{\link{write_fare_structure}()}
}
\concept{fare structure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set.R
\name{set_speed}
\alias{set_speed}
\title{Set walk and bike speed}
\usage{
set_speed(r5r_network, speed, mode)
}
\arguments{
\item{r5r_network}{A routable transport network created with \code{\link[=build_network]{build_network()}}.}

\item{speed}{A number representing the speed in km/h.}

\item{mode}{A string. Either \code{"bike"} or \code{"walk"}.}
}
\value{
Invisibly returns \code{TRUE}.
}
\description{
This function receives the walk and bike 'speed' inputs in Km/h from routing
functions above and converts them to meters per second, which is then used
to set these speed profiles in r5r JAR.
}
\seealso{
Other setting functions: 
\code{\link{reverse_back_if_direct_mode}()},
\code{\link{reverse_if_direct_mode}()},
\code{\link{set_breakdown}()},
\code{\link{set_cutoffs}()},
\code{\link{set_elevation}()},
\code{\link{set_expanded_travel_times}()},
\code{\link{set_fare_cutoffs}()},
\code{\link{set_fare_structure}()},
\code{\link{set_max_fare}()},
\code{\link{set_max_lts}()},
\code{\link{set_max_rides}()},
\code{\link{set_monte_carlo_draws}()},
\code{\link{set_n_threads}()},
\code{\link{set_new_congestion}()},
\code{\link{set_new_lts}()},
\code{\link{set_output_dir}()},
\code{\link{set_percentiles}()},
\code{\link{set_progress}()},
\code{\link{set_suboptimal_minutes}()},
\code{\link{set_time_window}()},
\code{\link{set_verbose}()}
}
\concept{setting functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set.R
\name{set_suboptimal_minutes}
\alias{set_suboptimal_minutes}
\title{Set suboptimal minutes}
\usage{
set_suboptimal_minutes(
  r5r_network,
  suboptimal_minutes,
  fare_structure,
  shortest_path
)
}
\arguments{
\item{r5r_network}{A routable transport network created with \code{\link[=build_network]{build_network()}}.}

\item{suboptimal_minutes}{A number.}

\item{fare_structure}{A fare structure object, following the convention
set in \code{\link[=setup_fare_structure]{setup_fare_structure()}}. This object describes how transit fares
should be calculated. Please see the fare structure vignette to
understand how this object is structured:
\code{vignette("fare_structure", package = "r5r")}.}

\item{shortest_path}{A logical.}
}
\value{
Invisibly returns \code{TRUE}.
}
\description{
Sets the number of suboptimal minutes considered in \code{\link[=detailed_itineraries]{detailed_itineraries()}}
routing. From R5 documentation: "This parameter compensates for the fact that
GTFS does not contain information about schedule deviation (lateness). The
min-max travel time range for some trains is zero, since the trips are
reported to always have the same timings in the schedule. Such an option
does not overlap (temporally) its alternatives, and is too easily eliminated
by an alternative that is only marginally better. We want to effectively
push the max travel time of alternatives out a bit to account for the fact
that they don't always run on schedule".
}
\seealso{
Other setting functions: 
\code{\link{reverse_back_if_direct_mode}()},
\code{\link{reverse_if_direct_mode}()},
\code{\link{set_breakdown}()},
\code{\link{set_cutoffs}()},
\code{\link{set_elevation}()},
\code{\link{set_expanded_travel_times}()},
\code{\link{set_fare_cutoffs}()},
\code{\link{set_fare_structure}()},
\code{\link{set_max_fare}()},
\code{\link{set_max_lts}()},
\code{\link{set_max_rides}()},
\code{\link{set_monte_carlo_draws}()},
\code{\link{set_n_threads}()},
\code{\link{set_new_congestion}()},
\code{\link{set_new_lts}()},
\code{\link{set_output_dir}()},
\code{\link{set_percentiles}()},
\code{\link{set_progress}()},
\code{\link{set_speed}()},
\code{\link{set_time_window}()},
\code{\link{set_verbose}()}
}
\concept{setting functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/detailed_itineraries.R
\name{detailed_itineraries}
\alias{detailed_itineraries}
\title{Detailed itineraries between origin-destination pairs}
\usage{
detailed_itineraries(
  r5r_network,
  r5r_core = deprecated(),
  origins,
  destinations,
  mode = "WALK",
  mode_egress = "WALK",
  departure_datetime = Sys.time(),
  time_window = 10L,
  suboptimal_minutes = 0L,
  max_walk_time = Inf,
  max_bike_time = Inf,
  max_car_time = Inf,
  max_trip_duration = 120L,
  walk_speed = 3.6,
  bike_speed = 12,
  max_rides = 3,
  max_lts = 2,
  shortest_path = TRUE,
  all_to_all = FALSE,
  fare_structure = NULL,
  max_fare = Inf,
  new_carspeeds = NULL,
  carspeed_scale = 1,
  new_lts = NULL,
  n_threads = Inf,
  verbose = FALSE,
  progress = FALSE,
  drop_geometry = FALSE,
  osm_link_ids = FALSE,
  output_dir = NULL
)
}
\arguments{
\item{r5r_network}{A routable transport network created with \code{\link[=build_network]{build_network()}}.}

\item{r5r_core}{The \code{r5r_core} argument is deprecated as of r5r v2.3.0.
Please use the \code{r5r_network} argument instead.}

\item{origins, destinations}{Either a \verb{POINT sf} object with WGS84 CRS, or a
\code{data.frame} containing the columns \code{id}, \code{lon} and \code{lat}.}

\item{mode}{A character vector. The transport modes allowed for access,
transfer and vehicle legs of the trips. Defaults to \code{WALK}. Please see
details for other options.}

\item{mode_egress}{A character vector. The transport mode used after egress
from the last public transport. It can be either \code{WALK}, \code{BICYCLE} or
\code{CAR}. Defaults to \code{WALK}. Ignored when public transport is not used.}

\item{departure_datetime}{A POSIXct object. Please note that the departure
time only influences public transport legs. When working with public
transport networks, please check the \code{calendar.txt} within your GTFS feeds
for valid dates. Please see details for further information on how
datetimes are parsed.}

\item{time_window}{An integer. The time window in minutes for which \code{r5r}
will calculate multiple itineraries departing each minute. Defaults to 10
minutes. If the same sequence of routes appear in different minutes of the
time window, only the fastest of them will be kept in the output. This
happens because the result is not aggregated by percentile, as opposed to
other routing functions in the package. Because of that, the output may
contain trips departing after the specified \code{departure_datetime}, but
still within the time window. Please read the time window vignette for
more details on how this argument affects the results of each routing
function: \code{vignette("time_window", package = "r5r")}.}

\item{suboptimal_minutes}{A number. The difference in minutes that each
non-optimal RAPTOR branch can have from the optimal branch without being
disregarded by the routing algorithm. If, for example, users set
\code{suboptimal_minutes = 10}, the routing algorithm will consider sub-optimal
routes that arrive up to 10 minutes after the arrival of the optimal one.
This argument emulates the real-life behaviour that makes people want to
take a path that is technically not optimal in terms of travel time, for
example, for some practical reasons (e.g. mode preference, safety, etc).
In practice, the higher this value, the more itineraries will be returned
in the final result.}

\item{max_walk_time}{An integer. The maximum walking time (in minutes) to
access and egress the transit network, to make transfers within the network
or to complete walk-only trips. Defaults to no restrictions (numeric value
of \code{Inf}), as long as \code{max_trip_duration} is respected. When routing
transit trips, the max time is considered separately for each leg (e.g. if
you set \code{max_walk_time} to 15, you could get trips with an up to 15 minutes
walk leg to reach transit and another up to 15 minutes walk leg to reach
the destination after leaving transit. In walk-only trips, whenever
\code{max_walk_time} differs from \code{max_trip_duration}, the lowest value is
considered.}

\item{max_bike_time}{An integer. The maximum cycling time (in minutes) to
access and egress the transit network, to make transfers within the network
or to complete bicycle-only trips. Defaults to no restrictions (numeric
value of \code{Inf}), as long as \code{max_trip_duration} is respected. When routing
transit trips, the max time is considered separately for each leg (e.g. if
you set \code{max_bike_time} to 15, you could get trips with an up to 15 minutes
cycle leg to reach transit and another up to 15 minutes cycle leg to reach
the destination after leaving transit. In bicycle-only trips, whenever
\code{max_bike_time} differs from \code{max_trip_duration}, the lowest value is
considered.}

\item{max_car_time}{An integer. The maximum driving time (in minutes) to
access and egress the transit network. Defaults to no restrictions, as long
as \code{max_trip_duration} is respected. The max time is considered separately
for each leg (e.g. if you set \code{max_car_time} to 15 minutes, you could
potentially drive up to 15 minutes to reach transit, and up to \emph{another} 15
minutes to reach the destination after leaving transit). Defaults to \code{Inf},
no limit.}

\item{max_trip_duration}{An integer. The maximum trip duration in minutes.
Defaults to 120 minutes (2 hours).}

\item{walk_speed}{A numeric. Average walk speed in km/h. Defaults to 3.6
km/h.}

\item{bike_speed}{A numeric. Average cycling speed in km/h. Defaults to 12
km/h.}

\item{max_rides}{An integer. The maximum number of public transport rides
allowed in the same trip. Defaults to 3.}

\item{max_lts}{An integer between 1 and 4. The maximum level of traffic
stress that cyclists will tolerate. A value of 1 means cyclists will only
travel through the quietest streets, while a value of 4 indicates cyclists
can travel through any road. Defaults to 2. Please see details for more
information.}

\item{shortest_path}{A logical. Whether the function should only return the
fastest itinerary between each origin and destination pair (the default)
or multiple alternatives.}

\item{all_to_all}{A logical. Whether to query routes between the 1st origin
to the 1st destination, then the 2nd origin to the 2nd destination, and so
on (\code{FALSE}, the default) or to query routes between all origins to all
destinations (\code{TRUE}).}

\item{fare_structure}{A fare structure object, following the convention
set in \code{\link[=setup_fare_structure]{setup_fare_structure()}}. This object describes how transit fares
should be calculated. Please see the fare structure vignette to
understand how this object is structured:
\code{vignette("fare_structure", package = "r5r")}.}

\item{max_fare}{A number. The maximum value that trips can cost when
calculating the fastest journey between each origin and destination pair.}

\item{new_carspeeds}{A \code{data.frame} specifying the new car speed for each OSM
edge id. This table must contain columns \code{osm_id}, \code{max_speed}
and \code{speed_type}. The \code{"speed_type"} column is of class character
and it indicates whether the values in \code{"max_speed"} should be
interpreted as percentages of original speeds (\code{"scale"}) or as absolute
speeds (\code{"km/h"}). Alternatively, the \code{new_carspeeds} parameter can
receive an \verb{sf data.frame} with POLYGON geometry that indicates the new
car speed for all the roads that fall within each polygon. In this case,
the table must contain the columns \code{poly_id} with a unique id for
each polygon, \code{scale} with the new speed scaling factors and
\code{priority}, which is a number ranking which polygon should be
considered in case of overlapping polygons. See more into in the
\verb{link to congestion vignette}.}

\item{carspeed_scale}{Numeric. The default car speed to use for road segments
not specified in \code{new_carspeeds}. By default, it is \code{NULL} and the speeds
of the unlisted roads are kept unchanged.}

\item{new_lts}{A \code{data.frame} specifying the new LTS levels for each OSM edge
id. The table must contain columns \code{osm_id} and \code{lts}.
Alternatively, the \code{new_lts} parameter can receive an \verb{sf data.frame}
with LINESTRING geometry. R5 will then find the nearest road for each
LINESTRING and update its LTS value accordingly.}

\item{n_threads}{An integer. The number of threads to use when running the
router in parallel. Defaults to use all available threads (Inf).}

\item{verbose}{A logical. Whether to show \code{R5} informative messages when
running the function. Defaults to \code{FALSE} (please note that in such case
\code{R5} error messages are still shown). Setting \code{verbose} to \code{TRUE} shows
detailed output, which can be useful for debugging issues not caught by
\code{r5r}.}

\item{progress}{A logical. Whether to show a progress counter when running
the router. Defaults to \code{FALSE}. Only works when \code{verbose} is set to
\code{FALSE}, so the progress counter does not interfere with \code{R5}'s output
messages. Setting \code{progress} to \code{TRUE} may impose a small penalty for
computation efficiency, because the progress counter must be synchronized
among all active threads.}

\item{drop_geometry}{A logical. Whether the output should include the
geometry of each trip leg or not. The default value of \code{FALSE} keeps the
geometry column in the result.}

\item{osm_link_ids}{A logical. Whether the output should include additional
columns: \code{osm_id_list} for the OSM ids of the road
segments used along the trip geometry, and \code{board_stop_id} and
\code{alight_stop_id} for the OSM ids of transit boarding and alighting stops.
Defaults to \code{FALSE}.

Keep in mind that \code{osm_id_list} will contain an id even if the route only
uses a small stretch of the road (e.g. 5m of a 600m street segment). For
more precision, use \code{edge_id_list}, which returns the exact internal edge
segments used in the trip. You can inspect these edge ids and their
associated properties, including OSM ids, with
\code{\link[=street_network_to_sf]{street_network_to_sf()}}.}

\item{output_dir}{Either \code{NULL} or a path to an existing directory. When not
\code{NULL} (the default), the function will write one \code{.csv} file with the
results for each origin in the specified directory. In such case, the
function returns the path specified in this parameter. This parameter is
particularly useful when running on memory-constrained settings because
writing the results directly to disk prevents \code{r5r} from loading them to
RAM memory.}
}
\value{
When \code{drop_geometry} is \code{FALSE}, the function outputs a \verb{LINESTRING sf} with detailed information on the itineraries between the specified
origins and destinations. When \code{TRUE}, the output is a \code{data.table}. All
distances are in meters and travel times are in minutes. If \code{output_dir}
is not \code{NULL}, the function returns the path specified in that parameter,
in which the \code{.csv} files containing the results are saved.
}
\description{
Returns detailed trip information between origin-destination pairs. The
output includes the waiting and moving time in each trip leg, as well as some
info such as the distance traveled, the routes used and the geometry of each
leg. Please note that this function was originally conceptualized as a trip
planning functionality, similar to other commercial and non-commercial APIs
and apps (e.g. Moovit, Google's Directions API, OpenTripPlanning's
PlannerResource API). Thus, it consumes much more time and memory than the
other (more analytical) routing functions included in the package.
}
\section{Transport modes}{


\code{R5} allows for multiple combinations of transport modes. The options
include:
\itemize{
\item \strong{Transit modes:} \code{TRAM}, \code{SUBWAY}, \code{RAIL}, \code{BUS}, \code{FERRY}, \code{CABLE_CAR},
\code{GONDOLA}, \code{FUNICULAR}. The option \code{TRANSIT} automatically considers all
public transport modes available.
\item \strong{Non transit modes:} \code{WALK}, \code{BICYCLE}, \code{CAR}, \code{BICYCLE_RENT},
\code{CAR_PARK}.
}
}

\section{Level of Traffic Stress (LTS)}{


When cycling is enabled in \code{R5} (by passing the value \code{BIKE} to either
\code{mode} or \code{mode_egress}), setting \code{max_lts} will allow cycling only on
streets with a given level of danger/stress. Setting \code{max_lts} to 1, for
example, will allow cycling only on separated bicycle infrastructure or
low-traffic streets and routing will revert to walking when traversing any
links with LTS exceeding 1. Setting \code{max_lts} to 3 will allow cycling on
links with LTS 1, 2 or 3. Routing also reverts to walking if the street
segment is tagged as non-bikable in OSM (e.g. a staircase), independently of
the specified max LTS.

The default methodology for assigning LTS values to network edges is based
on commonly tagged attributes of OSM ways. See more info about LTS in the
original documentation of R5 from Conveyal at
\url{https://docs.conveyal.com/learn-more/traffic-stress}. In summary:
\itemize{
\item \strong{LTS 1}: Tolerable for children. This includes low-speed, low-volume
streets, as well as those with separated bicycle facilities (such as
parking-protected lanes or cycle tracks).
\item \strong{LTS 2}: Tolerable for the mainstream adult population. This includes
streets where cyclists have dedicated lanes and only have to interact with
traffic at formal crossing.
\item \strong{LTS 3}: Tolerable for "enthused and confident" cyclists. This includes
streets which may involve close proximity to moderate- or high-speed
vehicular traffic.
\item \strong{LTS 4}: Tolerable only for "strong and fearless" cyclists. This
includes streets where cyclists are required to mix with moderate- to
high-speed vehicular traffic.
}

For advanced users, you can provide custom LTS values by adding a tag \verb{<key = "lts">} to the \code{osm.pbf} file.
}

\section{Datetime parsing}{


\code{r5r} ignores the timezone attribute of datetime objects when parsing dates
and times, using the study area's timezone instead. For example, let's say
you are running some calculations using Rio de Janeiro, Brazil, as your study
area. The datetime \code{as.POSIXct("13-05-2019 14:00:00", format = "\%d-\%m-\%Y \%H:\%M:\%S")} will be parsed as May 13th, 2019, 14:00h in
Rio's local time, as expected. But \code{as.POSIXct("13-05-2019 14:00:00", format = "\%d-\%m-\%Y \%H:\%M:\%S", tz = "Europe/Paris")} will also be parsed as
the exact same date and time in Rio's local time, perhaps surprisingly,
ignoring the timezone attribute.
}

\section{Routing algorithm}{


The \code{\link[=detailed_itineraries]{detailed_itineraries()}} and \code{\link[=pareto_frontier]{pareto_frontier()}} functions use an
\code{R5}-specific extension to the McRAPTOR routing algorithm. The
implementation used in \code{detailed_itineraries()} allows the router to find
paths that are optimal and less than optimal in terms of travel time, with
some heuristics around multiple access modes, riding the same patterns, etc.
The specific extension to McRAPTOR to do suboptimal path routing is not
documented yet, but a detailed description of base McRAPTOR can be found in
Delling et al (2015). The implementation used in \code{pareto_frontier()}, on the
other hand, returns only the fastest trip within a given monetary cutoff,
ignoring slower trips that cost the same. A detailed discussion on the
algorithm can be found in Conway and Stewart (2019).
\itemize{
\item Delling, D., Pajor, T., & Werneck, R. F. (2015). Round-based public
transit routing. Transportation Science, 49(3), 591-604.
\doi{10.1287/trsc.2014.0534}
\item Conway, M. W., & Stewart, A. F. (2019). Getting Charlie off the MTA: a
multiobjective optimization method to account for cost constraints in public
transit accessibility metrics. International Journal of Geographical
Information Science, 33(9), 1759-1787. \doi{10.1080/13658816.2019.1605075}
}
}

\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path)

# load origin/destination points
points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))

# inputs
departure_datetime <- as.POSIXct(
  "13-05-2019 14:00:00",
  format = "\%d-\%m-\%Y \%H:\%M:\%S"
)

det <- detailed_itineraries(
  r5r_network,
  origins = points[10,],
  destinations = points[12,],
  mode = c("WALK", "TRANSIT"),
  departure_datetime = departure_datetime,
  max_trip_duration = 60
)
head(det)

stop_r5(r5r_network)
\dontshow{\}) # examplesIf}
}
\seealso{
Other routing: 
\code{\link{arrival_travel_time_matrix}()},
\code{\link{expanded_travel_time_matrix}()},
\code{\link{pareto_frontier}()},
\code{\link{travel_time_matrix}()}
}
\concept{routing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assign.R
\name{assign_opportunities}
\alias{assign_opportunities}
\title{Assign opportunities data}
\usage{
assign_opportunities(destinations, opportunities_colnames)
}
\arguments{
\item{destinations}{Either a \code{data.frame} or a \verb{POINT sf}.}

\item{opportunities_colnames}{A character vector with the names of the
opportunities columns in \code{destinations}.}
}
\value{
A list of \code{Java-Array} objects.
}
\description{
Check and create an opportunities dataset.
}
\seealso{
Other assigning functions: 
\code{\link{assign_decay_function}()},
\code{\link{assign_departure}()},
\code{\link{assign_drop_geometry}()},
\code{\link{assign_max_street_time}()},
\code{\link{assign_max_trip_duration}()},
\code{\link{assign_mode}()},
\code{\link{assign_osm_link_ids}()},
\code{\link{assign_points_input}()},
\code{\link{assign_shortest_path}()}
}
\concept{assigning functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set.R
\name{reverse_if_direct_mode}
\alias{reverse_if_direct_mode}
\title{Reverse Origins and Destinations for Direct Modes}
\usage{
reverse_if_direct_mode(origins, destinations, mode_list, data_path)
}
\arguments{
\item{origins}{A data frame representing origin locations.}

\item{destinations}{A data frame representing destination locations.}

\item{mode_list}{A named list containing the routing modes.}

\item{data_path}{The data path used to build the network}
}
\value{
List origins and destinations unchanged or in swapped order
}
\description{
Swaps the \code{origins} and \code{destinations} data frames if certain conditions are
met, specifically to optimize routing performance with R5's one-to-many
algorithm. The function reverses the direction of analysis when the transit
mode is empty and the direct modes are WALK or BICYCLE and when the number of
origin points is greater than the number of destination points.
}
\seealso{
Other setting functions: 
\code{\link{reverse_back_if_direct_mode}()},
\code{\link{set_breakdown}()},
\code{\link{set_cutoffs}()},
\code{\link{set_elevation}()},
\code{\link{set_expanded_travel_times}()},
\code{\link{set_fare_cutoffs}()},
\code{\link{set_fare_structure}()},
\code{\link{set_max_fare}()},
\code{\link{set_max_lts}()},
\code{\link{set_max_rides}()},
\code{\link{set_monte_carlo_draws}()},
\code{\link{set_n_threads}()},
\code{\link{set_new_congestion}()},
\code{\link{set_new_lts}()},
\code{\link{set_output_dir}()},
\code{\link{set_percentiles}()},
\code{\link{set_progress}()},
\code{\link{set_speed}()},
\code{\link{set_suboptimal_minutes}()},
\code{\link{set_time_window}()},
\code{\link{set_verbose}()}
}
\concept{setting functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/java_utils.R
\name{java_to_dt}
\alias{java_to_dt}
\title{Java object to data.table}
\usage{
java_to_dt(obj)
}
\arguments{
\item{obj}{A Java Object reference}
}
\value{
An R data.table
}
\description{
Converts a Java object returned by r5r_network to an R \code{data.table}
}
\seealso{
Other java support functions: 
\code{\link{dt_to_lts_map}()},
\code{\link{dt_to_speed_map}()},
\code{\link{get_java_version}()}
}
\concept{java support functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stop_r5.R
\name{stop_r5}
\alias{stop_r5}
\title{Stop running r5r network}
\usage{
stop_r5(...)
}
\arguments{
\item{...}{\code{r5r_network} objects currently running. By default, if no
r5r network is supplied all running networks are stopped.}
}
\value{
No return value, called for side effects.
}
\description{
Stops running r5r network
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)

path <- system.file("extdata/poa", package = "r5r")

r5r_network <- build_network(path)

stop_r5(r5r_network)
\dontshow{\}) # examplesIf}
}
\seealso{
Other support functions: 
\code{\link{exists_tiff}()},
\code{\link{fileurl_from_metadata}()},
\code{\link{get_gtfs_errors}()},
\code{\link{start_r5r_java}()},
\code{\link{tempdir_unique}()},
\code{\link{travel_time_surface}},
\code{\link{validate_bad_osm_ids}()}
}
\concept{support functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{congestion_poly2geojson}
\alias{congestion_poly2geojson}
\title{Save speeds polygon to .geojson temporary file}
\usage{
congestion_poly2geojson(new_speeds_poly)
}
\arguments{
\item{new_speeds_poly}{An sf polygon}
}
\value{
The path to a \code{.geojson} saved as a temporary file.
}
\description{
Support function that checks the input of speeds polygon passed to
\code{build_custom_network()} and saves it to a \code{.geojson} temporary file.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}

# read polygons with new speeds
congestion_poly <- readRDS(
  system.file("extdata/poa/poa_poly_congestion.rds", package = "r5r")
  )

geojson_path <- r5r:::congestion_poly2geojson(
  new_speeds_poly = congestion_poly
  )
\dontshow{\}) # examplesIf}
}
\seealso{
Other Support functions: 
\code{\link{lts_lines2shp}()}
}
\concept{Support functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{lts_lines2shp}
\alias{lts_lines2shp}
\title{Save LTS lines to shapefile temporary file}
\usage{
lts_lines2shp(new_lts_lines)
}
\arguments{
\item{new_lts_lines}{An sf LINESTRING or MULTILINESTRING}
}
\value{
The path to a \code{.shp} saved as a temporary file.
}
\description{
Support function that checks the input of LTS lines passed to
and saves it to a \code{.shp} temporary file.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}

# read lines with new speeds
new_lts_lines <- readRDS(
  system.file("extdata/poa/poa_ls_lts.rds", package = "r5r")
  )

shp_path <- r5r:::lts_lines2shp(
  new_lts_lines = new_lts_lines
  )
\dontshow{\}) # examplesIf}
}
\seealso{
Other Support functions: 
\code{\link{congestion_poly2geojson}()}
}
\concept{Support functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assign.R
\name{assign_departure}
\alias{assign_departure}
\title{Check and convert POSIXct objects to strings}
\usage{
assign_departure(datetime)
}
\arguments{
\item{datetime}{An object of POSIXct class.}
}
\value{
A list with the \code{date} and \code{time} of the trip departure as
characters.
}
\description{
Check and convert POSIXct objects to strings
}
\seealso{
Other assigning functions: 
\code{\link{assign_decay_function}()},
\code{\link{assign_drop_geometry}()},
\code{\link{assign_max_street_time}()},
\code{\link{assign_max_trip_duration}()},
\code{\link{assign_mode}()},
\code{\link{assign_opportunities}()},
\code{\link{assign_osm_link_ids}()},
\code{\link{assign_points_input}()},
\code{\link{assign_shortest_path}()}
}
\concept{assigning functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assign.R
\name{assign_mode}
\alias{assign_mode}
\title{Check and select transport modes from user input}
\usage{
assign_mode(mode, mode_egress, style)
}
\arguments{
\item{mode}{A character vector, passed from routing functions.}

\item{mode_egress}{A character vector, passed from routing functions.}
}
\value{
A list with the transport modes to be used in the routing.
}
\description{
Selects the transport modes used in the routing functions. Only one direct
and access/egress modes are allowed at a time.
}
\seealso{
Other assigning functions: 
\code{\link{assign_decay_function}()},
\code{\link{assign_departure}()},
\code{\link{assign_drop_geometry}()},
\code{\link{assign_max_street_time}()},
\code{\link{assign_max_trip_duration}()},
\code{\link{assign_opportunities}()},
\code{\link{assign_osm_link_ids}()},
\code{\link{assign_points_input}()},
\code{\link{assign_shortest_path}()}
}
\concept{assigning functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assign.R
\name{assign_max_street_time}
\alias{assign_max_street_time}
\title{Assign max street time from walk/bike distance and speed}
\usage{
assign_max_street_time(max_time, speed, max_trip_duration, mode)
}
\arguments{
\item{max_time}{A numeric of length 1. Maximum walking distance (in
meters) for the whole trip. Passed from routing functions.}

\item{speed}{A numeric of length 1. Average walk speed in km/h.
Defaults to 3.6 Km/h. Passed from routing functions.}

\item{max_trip_duration}{A numeric of length 1. Maximum trip duration in
seconds. Defaults to 120 minutes (2 hours). Passed from routing functions.}

\item{mode}{A string. Either \code{"bike"} or \code{"walk"}.}
}
\value{
An \code{integer} representing the maximum number of minutes walking.
}
\description{
Checks the time duration and speed inputs and converts them to distance.
}
\seealso{
Other assigning functions: 
\code{\link{assign_decay_function}()},
\code{\link{assign_departure}()},
\code{\link{assign_drop_geometry}()},
\code{\link{assign_max_trip_duration}()},
\code{\link{assign_mode}()},
\code{\link{assign_opportunities}()},
\code{\link{assign_osm_link_ids}()},
\code{\link{assign_points_input}()},
\code{\link{assign_shortest_path}()}
}
\concept{assigning functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/java_utils.R
\name{get_java_version}
\alias{get_java_version}
\title{Determine the Java version installed locally}
\usage{
get_java_version()
}
\value{
The number of the Java version
}
\description{
Determine the Java version installed locally
}
\seealso{
Other java support functions: 
\code{\link{dt_to_lts_map}()},
\code{\link{dt_to_speed_map}()},
\code{\link{java_to_dt}()}
}
\concept{java support functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{tempdir_unique}
\alias{tempdir_unique}
\title{Return a temporary directory path that is unique with every call}
\usage{
tempdir_unique()
}
\value{
Path. Returns the path of the created temporary directory
}
\description{
This is different from the built in tempdir() in that it does not return the same directory within a given runtime. Always returns a unique directory
}
\seealso{
Other support functions: 
\code{\link{exists_tiff}()},
\code{\link{fileurl_from_metadata}()},
\code{\link{get_gtfs_errors}()},
\code{\link{start_r5r_java}()},
\code{\link{stop_r5}()},
\code{\link{travel_time_surface}},
\code{\link{validate_bad_osm_ids}()}
}
\concept{support functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/street_network_bbox.R
\name{street_network_bbox}
\alias{street_network_bbox}
\title{Extract the geographic bounding box of the transport network}
\usage{
street_network_bbox(
  r5r_network,
  output = c("polygon", "bbox", "vector"),
  r5r_core = deprecated()
)
}
\arguments{
\item{r5r_network}{A routable transport network created with \code{\link[=build_network]{build_network()}}.}

\item{output}{A character string specifying the desired output format. One of
\code{"polygon"} (the default), \code{"bbox"}, or \code{"vector"}.}

\item{r5r_core}{The \code{r5r_core} argument is deprecated as of r5r v2.3.0.
Please use the \code{r5r_network} argument instead.}
}
\value{
By default (\code{output = "polygon"}), an \code{sf} object with a single \code{POLYGON}
geometry. If \code{output = "bbox"}, an \code{sf} \code{bbox} object. If \code{output = "vector"},
a named numeric vector with \code{xmin}, \code{ymin}, \code{xmax}, \code{ymax} coordinates.
All outputs use the WGS84 coordinate reference system (EPSG: 4326).
}
\description{
Extracts the geographic bounding box of the street network layer from a
routable transport network built with \code{\link[=build_network]{build_network()}}). It is a fast and
memory-efficient alternative to \code{sf::st_bbox(street_network_to_sf(r5r_net))}.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)
library(sf)

data_path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path)

# Get the network's bounding box as an sf polygon (default)
poly <- street_network_bbox(r5r_network, output = "polygon")
plot(poly)

# Get an sf bbox object (order is xmin, ymin, xmax, ymax)
box <- street_network_bbox(r5r_network , output = "bbox")
box

# Get a simple named vector (order now also xmin, ymin, xmax, ymax)
vec <- street_network_bbox(r5r_network , output = "vector")
vec

stop_r5(r5r_network)
\dontshow{\}) # examplesIf}
}
\seealso{
Other network functions: 
\code{\link{find_snap}()},
\code{\link{street_network_to_sf}()},
\code{\link{transit_network_to_sf}()}
}
\concept{network functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{fileurl_from_metadata}
\alias{fileurl_from_metadata}
\title{Get most recent JAR file url from metadata}
\usage{
fileurl_from_metadata(version = NULL)
}
\arguments{
\item{version}{A string. The version of R5 to be downloaded. When \code{NULL}, it
defaults to the latest version.}
}
\value{
A url a string.
}
\description{
Returns the most recent JAR file url from metadata, depending on the version.
}
\seealso{
Other support functions: 
\code{\link{exists_tiff}()},
\code{\link{get_gtfs_errors}()},
\code{\link{start_r5r_java}()},
\code{\link{stop_r5}()},
\code{\link{tempdir_unique}()},
\code{\link{travel_time_surface}},
\code{\link{validate_bad_osm_ids}()}
}
\concept{support functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r5r_network.R
\name{wrap_r5r_network}
\alias{wrap_r5r_network}
\title{Constructor for r5r_network object}
\usage{
wrap_r5r_network(jcore)
}
\arguments{
\item{jcore}{A \code{jobjRef} Java object reference to R5RCore.}
}
\value{
\code{r5r_network}
}
\description{
Wraps a Java R5RCore as an r5r_network.
}
\seealso{
Other r5r_network: 
\code{\link{r5r_network-class}}
}
\concept{r5r_network}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{validate_bad_osm_ids}
\alias{validate_bad_osm_ids}
\title{Validate OSM IDs returned from Java backend and print warnings}
\usage{
validate_bad_osm_ids(bad_ids_string)
}
\arguments{
\item{bad_ids_string}{Character. A string formatted as a Java array (e.g., \code{"[id1, id2]"}).}
}
\value{
Warning if necessary.
}
\description{
Parses a Java-style array string (e.g., \code{"[id1, id2]"}), extracts OSM IDs,
and prints a pretty warning if any invalid IDs are found.
}
\details{
If no invalid IDs are found (i.e., input is \code{"[]"}), prints nothing.
}
\seealso{
Other support functions: 
\code{\link{exists_tiff}()},
\code{\link{fileurl_from_metadata}()},
\code{\link{get_gtfs_errors}()},
\code{\link{start_r5r_java}()},
\code{\link{stop_r5}()},
\code{\link{tempdir_unique}()},
\code{\link{travel_time_surface}}
}
\concept{support functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fare_structure.R
\name{assert_fare_structure}
\alias{assert_fare_structure}
\title{Assert fare structure}
\usage{
assert_fare_structure(fare_structure)
}
\arguments{
\item{fare_structure}{A fare structure object, following the convention
set in \code{\link[=setup_fare_structure]{setup_fare_structure()}}. This object describes how transit fares
should be calculated. Please see the fare structure vignette to
understand how this object is structured:
\code{vignette("fare_structure", package = "r5r")}.}
}
\value{
Throws and error upon failure and invisibly returns \code{TRUE} on
success.
}
\description{
Asserts whether the specified fare structure object complies with the
structure set in \code{\link[=setup_fare_structure]{setup_fare_structure()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assign.R
\name{assign_decay_function}
\alias{assign_decay_function}
\title{Assign decay function and parameter values}
\usage{
assign_decay_function(decay_function, decay_value)
}
\arguments{
\item{decay_function}{A string, the name of the decay function.}

\item{decay_value}{A number, the value of decay parameter.}
}
\value{
A \code{list} with the validated decay function and parameter value.
}
\description{
Checks and assigns decay function and values.
}
\seealso{
Other assigning functions: 
\code{\link{assign_departure}()},
\code{\link{assign_drop_geometry}()},
\code{\link{assign_max_street_time}()},
\code{\link{assign_max_trip_duration}()},
\code{\link{assign_mode}()},
\code{\link{assign_opportunities}()},
\code{\link{assign_osm_link_ids}()},
\code{\link{assign_points_input}()},
\code{\link{assign_shortest_path}()}
}
\concept{assigning functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r5r_sitrep.R
\name{r5r_sitrep}
\alias{r5r_sitrep}
\title{Generate an r5r situation report to help debug errors}
\usage{
r5r_sitrep()
}
\value{
A \code{list} with information of the versions of the r5r package, Java
and R5 Jar in use, the memory set to Java and user's Session Info.
}
\description{
The function reports a list with the following information:
\itemize{
\item The package version of \code{{r5r}} in use.
\item The installed version of \code{R5.jar}.
\item The Java version in use.
\item The amount of memory set to Java through the \code{java.parameters} option.
\item The user's Session Info.
}
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
r5r_sitrep()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isochrone.R
\name{isochrone}
\alias{isochrone}
\title{Estimate isochrones from a given location}
\usage{
isochrone(
  r5r_network,
  r5r_core = deprecated(),
  origins,
  mode = "transit",
  mode_egress = "walk",
  cutoffs = c(0, 15, 30),
  zoom = 10,
  departure_datetime = Sys.time(),
  polygon_output = TRUE,
  time_window = 10L,
  max_walk_time = Inf,
  max_bike_time = Inf,
  max_car_time = Inf,
  max_trip_duration = 120L,
  walk_speed = 3.6,
  bike_speed = 12,
  max_rides = 3,
  max_lts = 2,
  draws_per_minute = 5L,
  percentiles = NULL,
  n_threads = Inf,
  verbose = FALSE,
  progress = TRUE,
  sample_size = deprecated()
)
}
\arguments{
\item{r5r_network}{A routable transport network created with \code{\link[=build_network]{build_network()}}.}

\item{r5r_core}{The \code{r5r_core} argument is deprecated as of r5r v2.3.0.
Please use the \code{r5r_network} argument instead.}

\item{origins}{Either a \verb{POINT sf} object with WGS84 CRS, or a
\code{data.frame} containing the columns \code{id}, \code{lon} and \code{lat}.}

\item{mode}{A character vector. The transport modes allowed for access,
transfer and vehicle legs of the trips. Defaults to \code{WALK}. Please see
details for other options.}

\item{mode_egress}{A character vector. The transport mode used after egress
from the last public transport. It can be either \code{WALK}, \code{BICYCLE} or
\code{CAR}. Defaults to \code{WALK}. Ignored when public transport is not used.}

\item{cutoffs}{numeric vector. Number of minutes to define the time span of
each Isochrone. Defaults to \code{c(0, 15, 30)}.}

\item{zoom}{Resolution of the travel time surface used to create isochrones,
can be between \code{9} and \code{12.} The default is \code{10} (which uses cells of
153 meters at the Equator). More detailed isochrones will result from
larger numbers, at the expense of compute time. Specifically, a raster
grid of travel times in the Web Mercator projection at this zoom level
is created, and the isochrones are interpolated from this grid. For
more information on how the grid cells are defined, see
\href{https://docs.conveyal.com/analysis/methodology#zoom-levels}{the R5 documentation.}}

\item{departure_datetime}{A POSIXct object. Please note that the departure
time only influences public transport legs. When working with public
transport networks, please check the \code{calendar.txt} within your GTFS
feeds for valid dates. Please see details for further information on
how datetimes are parsed.}

\item{polygon_output}{A Logical. If \code{TRUE}, the function outputs
polygon-based isochrones (the default) based on travel times from each
origin to a sample of a random  sample nodes in the transport network
(see parameter \code{sample_size}). If \code{FALSE}, the function outputs
line-based isochrones based on travel times from each origin to the
centroids of all segments in the transport network.}

\item{time_window}{An integer. The time window in minutes for which \code{r5r}
will calculate multiple travel time matrices departing each minute.
Defaults to 10 minutes. The function returns the result based on
median travel times. Please read the time window vignette for more
details on its usage \code{vignette("time_window", package = "r5r")}}

\item{max_walk_time}{An integer. The maximum walking time (in minutes) to
access and egress the transit network, or to make transfers within the
network. Defaults to no restrictions, as long as \code{max_trip_duration}
is respected. The max time is considered separately for each leg (e.g.
if you set \code{max_walk_time} to 15, you could potentially walk up to 15
minutes to reach transit, and up to \emph{another} 15 minutes to reach the
destination after leaving transit). Defaults to \code{Inf}, no limit.}

\item{max_bike_time}{An integer. The maximum cycling time (in minutes) to
access and egress the transit network. Defaults to no restrictions, as
long as \code{max_trip_duration} is respected. The max time is considered
separately for each leg (e.g. if you set \code{max_bike_time} to 15 minutes,
you could potentially cycle up to 15 minutes to reach transit, and up
to \emph{another} 15 minutes to reach the destination after leaving
transit). Defaults to \code{Inf}, no limit.}

\item{max_car_time}{An integer. The maximum driving time (in minutes) to
access and egress the transit network. Defaults to no restrictions, as
long as \code{max_trip_duration} is respected. The max time is considered
separately for each leg (e.g. if you set \code{max_car_time} to 15 minutes,
you could potentially drive up to 15 minutes to reach transit, and up
to \emph{another} 15 minutes to reach the destination after leaving transit).
Defaults to \code{Inf}, no limit.}

\item{max_trip_duration}{An integer. The maximum trip duration in minutes.
Defaults to 120 minutes (2 hours).}

\item{walk_speed}{A numeric. Average walk speed in km/h. Defaults to 3.6 km/h.}

\item{bike_speed}{A numeric. Average cycling speed in km/h. Defaults to 12 km/h.}

\item{max_rides}{An integer. The maximum number of public transport rides
allowed in the same trip. Defaults to 3.}

\item{max_lts}{An integer between 1 and 4. The maximum level of traffic
stress that cyclists will tolerate. A value of 1 means cyclists will
only travel through the quietest streets, while a value of 4 indicates
cyclists can travel through any road. Defaults to 2. Please see
details for more information.}

\item{draws_per_minute}{An integer. The number of Monte Carlo draws to
perform per time window minute when calculating travel time matrices and
when estimating accessibility. Defaults to 5. This would mean 300 draws in
a 60-minute time window, for example. This parameter only affects the
results when the GTFS feeds contain a \code{frequencies.txt} table. If the GTFS
feed does not have a frequency table, r5r still allows for multiple runs
over the set \code{time_window} but in a deterministic way.}

\item{percentiles}{An integer vector (max length of 5). Specifies the
percentile to use when returning accessibility estimates within the given
time window. Please note that this parameter is applied to the travel time
estimates that generate the accessibility results, and not to the
accessibility distribution itself (i.e. if the 25th percentile is
specified, the accessibility is calculated from the 25th percentile travel
time, which may or may not be equal to the 25th percentile of the
accessibility distribution itself). Defaults to 50, returning the
accessibility calculated from the median travel time. If a vector with
length bigger than 1 is passed, the output contains an additional column
that specifies the percentile of each accessibility estimate. Due to
upstream restrictions, only 5 percentiles can be specified at a time. For
more details, please see \code{R5} documentation at \url{https://docs.conveyal.com/analysis/methodology#accounting-for-variability}.}

\item{n_threads}{An integer. The number of threads to use when running the
router in parallel. Defaults to use all available threads (\code{Inf}).}

\item{verbose}{A logical. Whether to show \code{R5} informative messages when
running the function. Defaults to \code{FALSE} (please note that in such case
\code{R5} error messages are still shown). Setting \code{verbose} to \code{TRUE} shows
detailed output, which can be useful for debugging issues not caught by
\code{r5r}.}

\item{progress}{A logical. Whether to show a progress counter when running
the router. Defaults to \code{FALSE}. Only works when \code{verbose} is set to
\code{FALSE}, so the progress counter does not interfere with \code{R5}'s output
messages. Setting \code{progress} to \code{TRUE} may impose a small penalty for
computation efficiency, because the progress counter must be
synchronized among all active threads.}

\item{sample_size}{deprecated, no longer has any effect.}
}
\value{
A \verb{"sf" "data.frame"} for each isochrone of each origin.
}
\description{
Fast computation of isochrones from a given location. The
function can return either polygon-based or line-based isochrones.
Polygon-based isochrones are generated as concave polygons based on the
travel times from the trip origin to all nodes in the transport network.
Meanwhile, line-based isochronesare based on travel times from each origin
to the centroids of all segments in the transport network.
}
\section{Transport modes}{


\code{R5} allows for multiple combinations of transport modes. The options
include:
\itemize{
\item \strong{Transit modes:} \code{TRAM}, \code{SUBWAY}, \code{RAIL}, \code{BUS}, \code{FERRY}, \code{CABLE_CAR},
\code{GONDOLA}, \code{FUNICULAR}. The option \code{TRANSIT} automatically considers all
public transport modes available.
\item \strong{Non transit modes:} \code{WALK}, \code{BICYCLE}, \code{CAR}, \code{BICYCLE_RENT},
\code{CAR_PARK}.
}
}

\section{Level of Traffic Stress (LTS)}{


When cycling is enabled in \code{R5} (by passing the value \code{BIKE} to either
\code{mode} or \code{mode_egress}), setting \code{max_lts} will allow cycling only on
streets with a given level of danger/stress. Setting \code{max_lts} to 1, for
example, will allow cycling only on separated bicycle infrastructure or
low-traffic streets and routing will revert to walking when traversing any
links with LTS exceeding 1. Setting \code{max_lts} to 3 will allow cycling on
links with LTS 1, 2 or 3. Routing also reverts to walking if the street
segment is tagged as non-bikable in OSM (e.g. a staircase), independently of
the specified max LTS.

The default methodology for assigning LTS values to network edges is based
on commonly tagged attributes of OSM ways. See more info about LTS in the
original documentation of R5 from Conveyal at
\url{https://docs.conveyal.com/learn-more/traffic-stress}. In summary:
\itemize{
\item \strong{LTS 1}: Tolerable for children. This includes low-speed, low-volume
streets, as well as those with separated bicycle facilities (such as
parking-protected lanes or cycle tracks).
\item \strong{LTS 2}: Tolerable for the mainstream adult population. This includes
streets where cyclists have dedicated lanes and only have to interact with
traffic at formal crossing.
\item \strong{LTS 3}: Tolerable for "enthused and confident" cyclists. This includes
streets which may involve close proximity to moderate- or high-speed
vehicular traffic.
\item \strong{LTS 4}: Tolerable only for "strong and fearless" cyclists. This
includes streets where cyclists are required to mix with moderate- to
high-speed vehicular traffic.
}

For advanced users, you can provide custom LTS values by adding a tag \verb{<key = "lts">} to the \code{osm.pbf} file.
}

\section{Datetime parsing}{


\code{r5r} ignores the timezone attribute of datetime objects when parsing dates
and times, using the study area's timezone instead. For example, let's say
you are running some calculations using Rio de Janeiro, Brazil, as your study
area. The datetime \code{as.POSIXct("13-05-2019 14:00:00", format = "\%d-\%m-\%Y \%H:\%M:\%S")} will be parsed as May 13th, 2019, 14:00h in
Rio's local time, as expected. But \code{as.POSIXct("13-05-2019 14:00:00", format = "\%d-\%m-\%Y \%H:\%M:\%S", tz = "Europe/Paris")} will also be parsed as
the exact same date and time in Rio's local time, perhaps surprisingly,
ignoring the timezone attribute.
}

\section{Routing algorithm}{


The \code{\link[=travel_time_matrix]{travel_time_matrix()}}, \code{\link[=expanded_travel_time_matrix]{expanded_travel_time_matrix()}},
\code{\link[=arrival_travel_time_matrix]{arrival_travel_time_matrix()}}  and \code{\link[=accessibility]{accessibility()}} functions use an
\code{R5}-specific extension to the RAPTOR routing algorithm (see Conway et al.,
2017). This RAPTOR extension uses a systematic sample of one departure per
minute over the time window set by the user in the 'time_window' parameter.
A detailed description of base RAPTOR can be found in Delling et al (2015).
However, whenever the user includes transit fares inputs to these functions,
they automatically switch to use an \code{R5}-specific extension to the McRAPTOR
routing algorithm.
\itemize{
\item Conway, M. W., Byrd, A., & van der Linden, M. (2017). Evidence-based
transit and land use sketch planning using interactive accessibility methods
on combined schedule and headway-based networks. Transportation Research
Record, 2653(1), 45-53. \doi{10.3141/2653-06}
\item Delling, D., Pajor, T., & Werneck, R. F. (2015). Round-based public
transit routing. Transportation Science, 49(3), 591-604.
\doi{10.1287/trsc.2014.0534}
}
}

\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
options(java.parameters = "-Xmx2G")
library(r5r)
library(ggplot2)

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path = data_path)

# load origin/point of interest
points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))
origin <- points[2,]

departure_datetime <- as.POSIXct(
 "13-05-2019 14:00:00",
 format = "\%d-\%m-\%Y \%H:\%M:\%S"
)

# estimate polygon-based isochrone from origin
iso_poly <- isochrone(
  r5r_network,
  origins = origin,
  mode = "walk",
  polygon_output = TRUE,
  departure_datetime = departure_datetime,
  cutoffs = seq(0, 120, 30)
  )

head(iso_poly)


# estimate line-based isochrone from origin
iso_lines <- isochrone(
  r5r_network,
  origins = origin,
  mode = "walk",
  polygon_output = FALSE,
  departure_datetime = departure_datetime,
  cutoffs = seq(0, 100, 25)
  )

head(iso_lines)


# plot colors
colors <- c('#ffe0a5','#ffcb69','#ffa600','#ff7c43','#f95d6a',
            '#d45087','#a05195','#665191','#2f4b7c','#003f5c')

# polygons
ggplot() +
  geom_sf(data=iso_poly, aes(fill=factor(isochrone))) +
  scale_fill_manual(values = colors) +
  theme_minimal()

# lines
ggplot() +
  geom_sf(data=iso_lines, aes(color=factor(isochrone))) +
  scale_color_manual(values = colors) +
  theme_minimal()

stop_r5(r5r_network)
\dontshow{\}) # examplesIf}
}
\concept{Isochrone}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_transit_availability.R
\name{check_transit_availability}
\alias{check_transit_availability}
\title{Check transit service availability by date}
\usage{
check_transit_availability(
  r5r_network,
  r5r_core = deprecated(),
  dates = NULL,
  start_date = NULL,
  end_date = NULL
)
}
\arguments{
\item{r5r_network}{A routable transport network created with \code{build_network()}.}

\item{r5r_core}{The \code{r5r_core} argument is deprecated as of r5r v2.3.0. Please use the \code{r5r_network} argument instead.}

\item{dates}{A vector of specific dates to be checked. Can be character strings
in #'   "YYYY-MM-DD" format, or objects of class \code{Date}. This argument
cannot be used with \code{start_date} or \code{end_date}.}

\item{start_date}{The start date for a continuous date range. Must be a single
character string in "YYYY-MM-DD" format or a \code{Date} object. Must be used
with \code{end_date}.}

\item{end_date}{The end date for a continuous date range. Must be a single
character string in "YYYY-MM-DD" format or a \code{Date} object. Must be used
with \code{start_date}.}
}
\value{
A \code{data.table} with four columns: \code{date}, \code{total_services},
\code{active_services}, and \code{pct_active} (the proportion of active services).
}
\description{
This function checks the number and proportion of public transport services
from the GTFS feeds in a \code{r5r_network} that are active on specified dates. This
is useful to verify that the selected departure dates for routing analysis are
valid and have adequate service levels. When routing with public transport, it
is crucial to use a departure date where services are operational, as indicated
in the GTFS \code{calendar.txt} file.
}
\details{
You can specify the dates to check in two ways:
\itemize{
\item Using the \code{dates} argument to provide a vector of specific dates.
\item Using the \code{start_date} and \code{end_date} arguments to provide a continuous date range.
}
You must use one of these two methods, but not both in the same function call.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)
data_path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path)

# Example 1: Check a vector of specific dates
# Let's check a regular weekday and a Sunday, where service may differ.
dates_to_check <- c("2019-05-13", "2019-05-19")
availability1 <- check_transit_availability(r5r_network, dates = dates_to_check)
availability1

# Example 2: Check a continuous date range using start_date and end_date
availability2 <- check_transit_availability(
  r5r_network,
  start_date = "2019-01-01",
  end_date = "2019-12-31"
)
availability2[121:124,]

# plot availability over the year
library(ggplot2)
ggplot(availability2, aes(x = date, y = pct_active)) +
  geom_line() +
  geom_point() +
  theme_classic(base_size = 16)

stop_r5(r5r_network)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arrival_travel_time_matrix.R
\name{arrival_travel_time_matrix}
\alias{arrival_travel_time_matrix}
\title{Calculate travel time matrix between origin destination pairs considering a
time of arrival}
\usage{
arrival_travel_time_matrix(
  r5r_network,
  r5r_core = deprecated(),
  origins,
  destinations,
  mode = "WALK",
  mode_egress = "WALK",
  arrival_datetime = Sys.time(),
  breakdown = FALSE,
  max_walk_time = Inf,
  max_bike_time = Inf,
  max_car_time = Inf,
  max_trip_duration = 120L,
  walk_speed = 3.6,
  bike_speed = 12,
  max_rides = 3,
  max_lts = 2,
  new_carspeeds = NULL,
  carspeed_scale = 1,
  new_lts = NULL,
  draws_per_minute = 5L,
  n_threads = Inf,
  verbose = FALSE,
  progress = FALSE,
  output_dir = NULL
)
}
\arguments{
\item{r5r_network}{A routable transport network created with \code{\link[=build_network]{build_network()}}.}

\item{r5r_core}{The \code{r5r_core} argument is deprecated as of r5r v2.3.0.
Please use the \code{r5r_network} argument instead.}

\item{origins, destinations}{Either a \verb{POINT sf} object with WGS84 CRS, or a
\code{data.frame} containing the columns \code{id}, \code{lon} and \code{lat}.}

\item{mode}{A character vector. The transport modes allowed for access,
transfer and vehicle legs of the trips. Defaults to \code{WALK}. Please see
details for other options.}

\item{mode_egress}{A character vector. The transport mode used after egress
from the last public transport. It can be either \code{WALK}, \code{BICYCLE} or
\code{CAR}. Defaults to \code{WALK}. Ignored when public transport is not used.}

\item{arrival_datetime}{A POSIXct object.}

\item{breakdown}{A logical. Whether to include detailed information about
each trip in the output. If \code{FALSE} (the default), the output lists the
total time between each origin-destination pair and the routes used to
complete the trip for each minute of the specified time window. If \code{TRUE},
the output includes the total access, waiting, in-vehicle and transfer
time of each trip. Please note that setting this parameter to \code{TRUE} makes
the function significantly slower.}

\item{max_walk_time}{An integer. The maximum walking time (in minutes) to
access and egress the transit network, to make transfers within the network
or to complete walk-only trips. Defaults to no restrictions (numeric value
of \code{Inf}), as long as \code{max_trip_duration} is respected. When routing
transit trips, the max time is considered separately for each leg (e.g. if
you set \code{max_walk_time} to 15, you could get trips with an up to 15 minutes
walk leg to reach transit and another up to 15 minutes walk leg to reach
the destination after leaving transit. In walk-only trips, whenever
\code{max_walk_time} differs from \code{max_trip_duration}, the lowest value is
considered.}

\item{max_bike_time}{An integer. The maximum cycling time (in minutes) to
access and egress the transit network, to make transfers within the network
or to complete bicycle-only trips. Defaults to no restrictions (numeric
value of \code{Inf}), as long as \code{max_trip_duration} is respected. When routing
transit trips, the max time is considered separately for each leg (e.g. if
you set \code{max_bike_time} to 15, you could get trips with an up to 15 minutes
cycle leg to reach transit and another up to 15 minutes cycle leg to reach
the destination after leaving transit. In bicycle-only trips, whenever
\code{max_bike_time} differs from \code{max_trip_duration}, the lowest value is
considered.}

\item{max_car_time}{An integer. The maximum driving time (in minutes) to
access and egress the transit network. Defaults to no restrictions, as long
as \code{max_trip_duration} is respected. The max time is considered separately
for each leg (e.g. if you set \code{max_car_time} to 15 minutes, you could
potentially drive up to 15 minutes to reach transit, and up to \emph{another} 15
minutes to reach the destination after leaving transit). Defaults to \code{Inf},
no limit.}

\item{max_trip_duration}{An integer. The maximum trip duration in minutes.
Defaults to 120 minutes (2 hours).}

\item{walk_speed}{A numeric. Average walk speed in km/h. Defaults to 3.6
km/h.}

\item{bike_speed}{A numeric. Average cycling speed in km/h. Defaults to 12
km/h.}

\item{max_rides}{An integer. The maximum number of public transport rides
allowed in the same trip. Defaults to 3.}

\item{max_lts}{An integer between 1 and 4. The maximum level of traffic
stress that cyclists will tolerate. A value of 1 means cyclists will only
travel through the quietest streets, while a value of 4 indicates cyclists
can travel through any road. Defaults to 2. Please see details for more
information.}

\item{new_carspeeds}{A \code{data.frame} specifying the new car speed for each OSM
edge id. This table must contain columns \code{osm_id}, \code{max_speed}
and \code{speed_type}. The \code{"speed_type"} column is of class character
and it indicates whether the values in \code{"max_speed"} should be
interpreted as percentages of original speeds (\code{"scale"}) or as absolute
speeds (\code{"km/h"}). Alternatively, the \code{new_carspeeds} parameter can
receive an \verb{sf data.frame} with POLYGON geometry that indicates the new
car speed for all the roads that fall within each polygon. In this case,
the table must contain the columns \code{poly_id} with a unique id for
each polygon, \code{scale} with the new speed scaling factors and
\code{priority}, which is a number ranking which polygon should be
considered in case of overlapping polygons. See more into in the
\verb{link to congestion vignette}.}

\item{carspeed_scale}{Numeric. The default car speed to use for road segments
not specified in \code{new_carspeeds}. By default, it is \code{NULL} and the speeds
of the unlisted roads are kept unchanged.}

\item{new_lts}{A \code{data.frame} specifying the new LTS levels for each OSM edge
id. The table must contain columns \code{osm_id} and \code{lts}.
Alternatively, the \code{new_lts} parameter can receive an \verb{sf data.frame}
with LINESTRING geometry. R5 will then find the nearest road for each
LINESTRING and update its LTS value accordingly.}

\item{draws_per_minute}{An integer. The number of Monte Carlo draws to
perform per time window minute when calculating travel time matrices and
when estimating accessibility. Defaults to 5. This would mean 300 draws in
a 60-minute time window, for example. This parameter only affects the
results when the GTFS feeds contain a \code{frequencies.txt} table. If the GTFS
feed does not have a frequency table, r5r still allows for multiple runs
over the set \code{time_window} but in a deterministic way.}

\item{n_threads}{An integer. The number of threads to use when running the
router in parallel. Defaults to use all available threads (Inf).}

\item{verbose}{A logical. Whether to show \code{R5} informative messages when
running the function. Defaults to \code{FALSE} (please note that in such case
\code{R5} error messages are still shown). Setting \code{verbose} to \code{TRUE} shows
detailed output, which can be useful for debugging issues not caught by
\code{r5r}.}

\item{progress}{A logical. Whether to show a progress counter when running
the router. Defaults to \code{FALSE}. Only works when \code{verbose} is set to
\code{FALSE}, so the progress counter does not interfere with \code{R5}'s output
messages. Setting \code{progress} to \code{TRUE} may impose a small penalty for
computation efficiency, because the progress counter must be synchronized
among all active threads.}

\item{output_dir}{Either \code{NULL} or a path to an existing directory. When not
\code{NULL} (the default), the function will write one \code{.csv} file with the
results for each origin in the specified directory. In such case, the
function returns the path specified in this parameter. This parameter is
particularly useful when running on memory-constrained settings because
writing the results directly to disk prevents \code{r5r} from loading them to
RAM memory.}
}
\value{
A \code{data.table} with travel time estimates (in minutes) and the
routes used in each trip between origin and destination pairs, for each
minute of the specified time window. Each set of origin, destination and
departure minute can appear up to N times, where N is the number of Monte
Carlo draws specified in the function arguments (please note that this
only applies when the GTFS feeds that describe the transit network include
a frequencies table, otherwise only a single draw is performed). A pair is
completely absent from the final output if no trips could be completed in
any of the minutes of the time window. If for a single pair trips could be
completed in some of the minutes of the time window, but not for all of
them, the minutes in which trips couldn't be completed will have \code{NA}
travel time and routes used. If \code{output_dir} is not \code{NULL}, the function
returns the path specified in that parameter, in which the \code{.csv} files
containing the results are saved.
}
\description{
Computation of travel time estimates between one or multiple origin
destination pairs considering a time of arrival. This function considers a
time of arrival set by the user. The function returns the travel time of the
trip with the latest departure time that arrives before the arrival time set
by the user. If you want to calculate travel times considering a departure
time, have a' look at the \code{\link[=travel_time_matrix]{travel_time_matrix()}} function. This function is a
wrapper around \code{\link[=expanded_travel_time_matrix]{expanded_travel_time_matrix()}}. On one hand, this means this
the output of this function has more columns (more info) compared the output
of \code{\link[=travel_time_matrix]{travel_time_matrix()}}. On the other hand, this function can be very memory
intensive if the user allows for really long max trip duration.
}
\section{Transport modes}{


\code{R5} allows for multiple combinations of transport modes. The options
include:
\itemize{
\item \strong{Transit modes:} \code{TRAM}, \code{SUBWAY}, \code{RAIL}, \code{BUS}, \code{FERRY}, \code{CABLE_CAR},
\code{GONDOLA}, \code{FUNICULAR}. The option \code{TRANSIT} automatically considers all
public transport modes available.
\item \strong{Non transit modes:} \code{WALK}, \code{BICYCLE}, \code{CAR}, \code{BICYCLE_RENT},
\code{CAR_PARK}.
}
}

\section{Level of Traffic Stress (LTS)}{


When cycling is enabled in \code{R5} (by passing the value \code{BIKE} to either
\code{mode} or \code{mode_egress}), setting \code{max_lts} will allow cycling only on
streets with a given level of danger/stress. Setting \code{max_lts} to 1, for
example, will allow cycling only on separated bicycle infrastructure or
low-traffic streets and routing will revert to walking when traversing any
links with LTS exceeding 1. Setting \code{max_lts} to 3 will allow cycling on
links with LTS 1, 2 or 3. Routing also reverts to walking if the street
segment is tagged as non-bikable in OSM (e.g. a staircase), independently of
the specified max LTS.

The default methodology for assigning LTS values to network edges is based
on commonly tagged attributes of OSM ways. See more info about LTS in the
original documentation of R5 from Conveyal at
\url{https://docs.conveyal.com/learn-more/traffic-stress}. In summary:
\itemize{
\item \strong{LTS 1}: Tolerable for children. This includes low-speed, low-volume
streets, as well as those with separated bicycle facilities (such as
parking-protected lanes or cycle tracks).
\item \strong{LTS 2}: Tolerable for the mainstream adult population. This includes
streets where cyclists have dedicated lanes and only have to interact with
traffic at formal crossing.
\item \strong{LTS 3}: Tolerable for "enthused and confident" cyclists. This includes
streets which may involve close proximity to moderate- or high-speed
vehicular traffic.
\item \strong{LTS 4}: Tolerable only for "strong and fearless" cyclists. This
includes streets where cyclists are required to mix with moderate- to
high-speed vehicular traffic.
}

For advanced users, you can provide custom LTS values by adding a tag \verb{<key = "lts">} to the \code{osm.pbf} file.
}

\section{Datetime parsing}{


\code{r5r} ignores the timezone attribute of datetime objects when parsing dates
and times, using the study area's timezone instead. For example, let's say
you are running some calculations using Rio de Janeiro, Brazil, as your study
area. The datetime \code{as.POSIXct("13-05-2019 14:00:00", format = "\%d-\%m-\%Y \%H:\%M:\%S")} will be parsed as May 13th, 2019, 14:00h in
Rio's local time, as expected. But \code{as.POSIXct("13-05-2019 14:00:00", format = "\%d-\%m-\%Y \%H:\%M:\%S", tz = "Europe/Paris")} will also be parsed as
the exact same date and time in Rio's local time, perhaps surprisingly,
ignoring the timezone attribute.
}

\section{Routing algorithm}{


The \code{\link[=travel_time_matrix]{travel_time_matrix()}}, \code{\link[=expanded_travel_time_matrix]{expanded_travel_time_matrix()}},
\code{\link[=arrival_travel_time_matrix]{arrival_travel_time_matrix()}}  and \code{\link[=accessibility]{accessibility()}} functions use an
\code{R5}-specific extension to the RAPTOR routing algorithm (see Conway et al.,
2017). This RAPTOR extension uses a systematic sample of one departure per
minute over the time window set by the user in the 'time_window' parameter.
A detailed description of base RAPTOR can be found in Delling et al (2015).
However, whenever the user includes transit fares inputs to these functions,
they automatically switch to use an \code{R5}-specific extension to the McRAPTOR
routing algorithm.
\itemize{
\item Conway, M. W., Byrd, A., & van der Linden, M. (2017). Evidence-based
transit and land use sketch planning using interactive accessibility methods
on combined schedule and headway-based networks. Transportation Research
Record, 2653(1), 45-53. \doi{10.3141/2653-06}
\item Delling, D., Pajor, T., & Werneck, R. F. (2015). Round-based public
transit routing. Transportation Science, 49(3), 591-604.
\doi{10.1287/trsc.2014.0534}
}
}

\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)

# build transport network
data_path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path )

# load origin/destination points
points <- read.csv(file.path(data_path, "poa_points_of_interest.csv"))

arrival_datetime <- as.POSIXct(
  "13-05-2019 14:00:00",
  format = "\%d-\%m-\%Y \%H:\%M:\%S"
)

# by default only returns the total time between each pair in each minute of
# the specified time window
arrival_ttm <- arrival_travel_time_matrix(
  r5r_network,
  origins = points,
  destinations = points,
  mode = c("WALK", "TRANSIT"),
  arrival_datetime = arrival_datetime,
  max_trip_duration = 60
)

head(arrival_ttm)

# when breakdown = TRUE the output contains much more information
arrival_ttm2 <- arrival_travel_time_matrix(
  r5r_network,
  origins = points,
  destinations = points,
  mode = c("WALK", "TRANSIT"),
  arrival_datetime = arrival_datetime,
  max_trip_duration = 60,
  breakdown = TRUE
)

head(arrival_ttm2)

stop_r5(r5r_network)
\dontshow{\}) # examplesIf}
}
\seealso{
Other routing: 
\code{\link{detailed_itineraries}()},
\code{\link{expanded_travel_time_matrix}()},
\code{\link{pareto_frontier}()},
\code{\link{travel_time_matrix}()}
}
\concept{routing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_snap.R
\name{find_snap}
\alias{find_snap}
\title{Find snapped locations of input points on street network}
\usage{
find_snap(
  r5r_network,
  r5r_core = deprecated(),
  points,
  radius = 1600,
  mode = "WALK"
)
}
\arguments{
\item{r5r_network}{A routable transport network created with \code{\link[=build_network]{build_network()}}.}

\item{r5r_core}{The \code{r5r_core} argument is deprecated as of r5r v2.3.0.
Please use the \code{r5r_network} argument instead.}

\item{points}{Either a \verb{POINT sf} object with WGS84 CRS, or a \code{data.frame}
containing the columns \code{id}, \code{lon} and \code{lat}.}

\item{radius}{Numeric. The maximum radius in meters within which to snap.
Defaults to 1600m.}

\item{mode}{A string. Which mode to consider when trying to snap the points
to the network. Defaults to \code{WALK}, also allows \code{BICYCLE} and \code{CAR}.}
}
\value{
A \code{data.table} with the original points, their respective
snapped coordinates on the street network and the Euclidean distance (in
meters) between the original points and their snapped location. Points that
could not be snapped show \code{NA} coordinates and \code{found = FALSE}.
}
\description{
Finds the snapped location of points on \code{R5} network. Snapping is an
important step of the routing process, which is when the origins and
destinations specified by the user are actually positioned on the network
created by \code{R5}. The snapping process in \code{R5} is composed of two rounds.
First, it tries to snap the points within a radius of 300 meters from
themselves. If the first round is unsuccessful, then \code{R5} expands the search
to the radius specified (by default 1.6km). If yet again it is unsuccessful,
then the unsnapped points won't be used during the routing process. The
snapped location of each point depends on the transport mode set by the user,
because some network edges are not available to specific modes (e.g. a
pedestrian-only street cannot be used to snap car trips).
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(r5r)

path <- system.file("extdata/poa", package = "r5r")
r5r_network <- build_network(data_path = path)
points <- read.csv(file.path(path, "poa_hexgrid.csv"))

snap_df <- find_snap(
  r5r_network,
  points = points,
  radius = 2000,
  mode = "WALK"
  )

stop_r5(r5r_network)
\dontshow{\}) # examplesIf}
}
\seealso{
Other network functions: 
\code{\link{street_network_bbox}()},
\code{\link{street_network_to_sf}()},
\code{\link{transit_network_to_sf}()}
}
\concept{network functions}
